@@ -43,7 +43,6 @@
 #include "SessionMgr.h"
 #include "md5.h"
 #include "ServiceAcceptor.h"
-#include "LookupService.h"
 #include "FileSettingsReader.h"
 #include "DBSettingsReader.h"
 #include "StreamPipe.h"
@@ -57,8 +56,12 @@
 #include "PipeThrottler.h"
 #include "mt19937ar.h"
 #include "Query.h"
-
-
+#ifdef _WIN32
+#include "SChannelPipe.h"
+#endif
+#ifdef WITH_OPENSSL
+#include "OpenSSLPipe.h"
+#endif
 
 #ifdef _WIN32
 #include <condition_variable>
@@ -91,6 +94,7 @@
 #include <mach/mach.h>
 #endif
 
+
 const size_t SEND_BLOCKSIZE=8192;
 const size_t MAX_THREAD_ID=std::string::npos;
 
@@ -189,13 +193,16 @@ CServer::CServer()
 #ifdef _WIN32
 	initialize_GetTickCount64();
 	log_rotation_size = 20*1024*1024; //20MB
+
+	send_window_size = -1;
+	recv_window_size = -1;
 #endif
 }
 
 void CServer::setup(void)
 {
 	sessmgr=new CSessionMgr();
-	threadpool=new CThreadPool();
+	threadpool=new CThreadPool(std::string::npos, 2, "idle pool thread");
 
 #ifndef NO_SQLITE
 	CDatabase::initMutex();
@@ -207,6 +214,17 @@ void CServer::setup(void)
 #ifdef MODE_WIN
 	File::init_mutex();
 #endif
+
+#ifdef _WIN32
+	SChannelPipe::init();
+#endif
+#ifdef WITH_OPENSSL
+	OpenSSLPipe::init();
+#endif
+
+#ifdef ENABLE_C_ARES
+	LookupInit();
+#endif
 }
 
 void CServer::destroyAllDatabases(void)
@@ -989,10 +1007,7 @@ ISessionMgr *CServer::getSessionMgr(void)
 std::string CServer::GenerateHexMD5(const std::string &input)
 {
 	MD5 md((unsigned char*)input.c_str() );
-	char *p=md.hex_digest();
-	std::string ret=p;
-	delete []p;
-	return ret;
+	return md.hex_digest();
 }
 
 std::string CServer::GenerateBinaryMD5(const std::string &input)
@@ -1013,21 +1028,74 @@ void CServer::StartCustomStreamService(IService *pService, std::string pServiceN
 
 IPipe* CServer::ConnectStream(std::string pServer, unsigned short pPort, unsigned int pTimeoutms)
 {
-	sockaddr_in server;
-	memset(&server, 0, sizeof(server));
-	LookupBlocking(pServer, &server.sin_addr);
-	server.sin_port = htons(pPort);
-	server.sin_family = AF_INET;
+	std::vector<SLookupBlockingResult> lookup_result;
+#ifdef ENABLE_C_ARES
+	lookup_result = LookupWithTimeout(pServer, pTimeoutms, pTimeoutms/2);
+#else
+	lookup_result = LookupBlocking(pServer);
+#endif
+	if (lookup_result.empty())
+	{
+#ifdef _WIN32
+		Server->Log("No result when looking up hostname \"" + pServer + "\". Err: " + convert(WSAGetLastError()), LL_DEBUG);
+#else
+		Server->Log("No result when looking up hostname \"" + pServer + "\". Err: " + convert(errno), LL_DEBUG);
+#endif
+		return NULL;
+	}
+
+	for (size_t i = 0; i < lookup_result.size(); ++i)
+	{
+		IPipe* ret = ConnectStream(lookup_result[i], pPort, pTimeoutms);
+		if (ret != NULL)
+			return ret;
+	}
+
+	return NULL;
+}
+
+
+IPipe* CServer::ConnectStream(const SLookupBlockingResult& lookup_result, unsigned short pPort, unsigned int pTimeoutms)
+{
+	union
+	{
+		sockaddr_in addr_v4;
+		sockaddr_in6 addr_v6;
+	} addr = {};
+
+	if (lookup_result.is_ipv6)
+	{
+		memcpy(&addr.addr_v6.sin6_addr, lookup_result.addr_v6, sizeof(addr.addr_v6.sin6_addr));
+		addr.addr_v6.sin6_port = htons(pPort);
+		addr.addr_v6.sin6_family = AF_INET6;
+		addr.addr_v6.sin6_scope_id = lookup_result.zone;
+	}
+	else
+	{
+		addr.addr_v4.sin_addr.s_addr = lookup_result.addr_v4;
+		addr.addr_v4.sin_port = htons(pPort);
+		addr.addr_v4.sin_family = AF_INET;
+	}
 
 	int type = SOCK_STREAM;
 #if !defined(_WIN32) && defined(SOCK_CLOEXEC)
 	type |= SOCK_CLOEXEC;
 #endif
 
-	SOCKET s = socket(AF_INET, type, 0);
+	SOCKET s = socket(lookup_result.is_ipv6 ? AF_INET6 : AF_INET, type, 0);
 	if (s == SOCKET_ERROR)
 	{
-		return NULL;
+#if !defined(_WIN32) && defined(SOCK_CLOEXEC)
+		if (errno == EINVAL)
+		{
+			type &= ~SOCK_CLOEXEC;
+			s = socket(lookup_result.is_ipv6 ? AF_INET6 : AF_INET, type, 0);
+		}
+#endif
+		if (s == SOCKET_ERROR)
+		{
+			return NULL;
+		}
 	}
 
 #if !defined(_WIN32) && !defined(SOCK_CLOEXEC)
@@ -1064,7 +1132,10 @@ IPipe* CServer::ConnectStream(std::string pServer, unsigned short pPort, unsigne
 	setsockopt(s, SOL_SOCKET, SO_NOSIGPIPE, (void*)&val, sizeof(val));
 #endif
 
-	int rc=connect(s, (sockaddr*)&server, sizeof(sockaddr_in) );
+	sockaddr* addrptr = lookup_result.is_ipv6 ? 
+		reinterpret_cast<sockaddr*>(&addr.addr_v6) : reinterpret_cast<sockaddr*>(&addr.addr_v4);
+
+	int rc=connect(s, addrptr, lookup_result.is_ipv6 ? sizeof(addr.addr_v6) : sizeof(addr.addr_v4) );
 #ifndef _WIN32
 	if(rc==SOCKET_ERROR)
 	{
@@ -1126,9 +1197,10 @@ IPipe* CServer::ConnectStream(std::string pServer, unsigned short pPort, unsigne
 		else
 		{
 #ifdef _WIN32
-			int window_size=512*1024;
-			setsockopt(s, SOL_SOCKET, SO_SNDBUF, (char *) &window_size, sizeof(window_size));
-			setsockopt(s, SOL_SOCKET, SO_RCVBUF, (char *) &window_size, sizeof(window_size));
+			if(send_window_size>0)
+				setsockopt(s, SOL_SOCKET, SO_SNDBUF, (char *) &send_window_size, sizeof(send_window_size));
+			if(recv_window_size>0)
+				setsockopt(s, SOL_SOCKET, SO_RCVBUF, (char *) &recv_window_size, sizeof(recv_window_size));
 #endif
 		    return new CStreamPipe(s);
 		}
@@ -1140,6 +1212,37 @@ IPipe* CServer::ConnectStream(std::string pServer, unsigned short pPort, unsigne
 	}
 }
 
+IPipe * CServer::ConnectSslStream(const std::string & pServer, unsigned short pPort, unsigned int pTimeoutms)
+{
+#if defined(_WIN32) || defined(WITH_OPENSSL)
+
+	int64 starttime = Server->getTimeMS();
+	CStreamPipe* bpipe = static_cast<CStreamPipe*>(ConnectStream(pServer, pPort, pTimeoutms));
+
+	if (bpipe == NULL)
+		return NULL;
+
+	int64 remaining_time = pTimeoutms - (Server->getTimeMS() - starttime);
+	if (remaining_time < 0) remaining_time = 0;
+
+#ifdef WITH_OPENSSL
+	OpenSSLPipe* ssl_pipe = new OpenSSLPipe(bpipe);
+#else
+	SChannelPipe* ssl_pipe = new SChannelPipe(bpipe);
+#endif
+
+	if (!ssl_pipe->ssl_connect(pServer, static_cast<int>(remaining_time)))
+	{
+		delete ssl_pipe;
+		return NULL;
+	}
+
+	return ssl_pipe;
+#else
+	return NULL;
+#endif
+}
+
 IPipe *CServer::PipeFromSocket(SOCKET pSocket)
 {
 	return new CStreamPipe(pSocket);
@@ -1300,18 +1403,52 @@ IPipe *CServer::createMemoryPipe(void)
 }
 
 #ifdef _WIN32
-void thread_helper_f(IThread *t, const std::string& name)
+struct SThreadInfo
 {
+	std::string name;
+	IThread* t;
+};
+
+void thread_helper_f2(IThread* t)
+{
+#ifndef _DEBUG
+	try
+	{
+#endif
+		(*t)();
+		Server->destroyDatabases(Server->getThreadID());
+#ifndef _DEBUG
+	}
+	catch (std::exception& e)
+	{
+		Server->Log(std::string("Thread exit with unhandled std::exception ") + e.what(), LL_ERROR);
+		throw;
+	}
+	catch (...)
+	{
+		Server->Log(std::string("Thread exit with unhandled C++ exception "), LL_ERROR);
+		throw;
+	}
+#endif
+}
+
+DWORD WINAPI thread_helper_f(LPVOID param)
+{
+	IThread* t;
 #if defined(_WIN32) && defined(_DEBUG)
-	SetThreadName(-1, name.c_str());
+	SThreadInfo* thread_info = reinterpret_cast<SThreadInfo*>(param);
+	SetThreadName(-1, thread_info->name.c_str());
+	t = thread_info->t;
+	delete thread_info;
+#else
+	t = reinterpret_cast<IThread*>(param);
 #endif
 
 #ifndef _DEBUG
 	__try
 	{
 #endif
-		(*t)();
-		Server->destroyDatabases(Server->getThreadID());
+		thread_helper_f2(t);
 #ifndef _DEBUG
 	}
 	__except(CServer::WriteDump(GetExceptionInformation()))
@@ -1319,6 +1456,7 @@ void thread_helper_f(IThread *t, const std::string& name)
 		throw;
 	}
 #endif
+	return 0;
 }
 #else //_WIN32
 void os_reset_priority();
@@ -1329,20 +1467,74 @@ void* thread_helper_f(void * t)
 	os_reset_priority();
 #endif
 	IThread *tmp=(IThread*)t;
-	(*tmp)();
+#ifndef _DEBUG
+	try
+	{
+#endif
+		(*tmp)();
+#ifndef _DEBUG
+	}
+	catch (std::exception& e)
+	{
+		Server->Log(std::string("Thread exit with unhandled std::exception ") + e.what(), LL_ERROR);
+		throw;
+	}
+	catch (...)
+	{
+		Server->Log(std::string("Thread exit with unhandled C++ exception "), LL_ERROR);
+		throw;
+	}
+#endif
 	Server->destroyDatabases(Server->getThreadID());
 	return NULL;
 }
 #endif //_WIN32
 
-void CServer::createThread(IThread *thread, const std::string& name)
+bool CServer::createThread(IThread *thread, const std::string& name, CreateThreadFlags flags)
 {
+	const size_t thread_large_stack_size = 64 * 1024 * 1024;
+
 #ifdef _WIN32
-	std::thread tr(thread_helper_f, thread, name);
-	tr.detach();
+	SIZE_T stack_size = 0;
+	if (flags & IServer::CreateThreadFlags_LargeStackSize)
+	{
+		stack_size = thread_large_stack_size;
+	}
+
+	void* param;
+#if defined(_WIN32) && defined(_DEBUG)
+	SThreadInfo* thread_info = new SThreadInfo;
+	thread_info->name = name;
+	thread_info->t = thread;
+	param = thread_info;
+#else
+	param = thread;
+#endif
+
+	HANDLE hThread = CreateThread(NULL, stack_size, &thread_helper_f, param, 0, NULL);
+	if (hThread == NULL
+		&& flags & IServer::CreateThreadFlags_LargeStackSize)
+	{
+		Server->Log("Error creating thread with large stack size. Trying to create thread without LargeStackSize", LL_WARNING);
+		hThread = CreateThread(NULL, 0, &thread_helper_f, param, 0, NULL);
+	}
+	if (hThread == NULL)
+	{
+		Server->Log("Error creating thread. Errno: " + convert((int)GetLastError()), LL_ERROR);
+		return false;
+	}
+	else
+	{
+		CloseHandle(hThread);
+		return true;
+	}
 #else
 	pthread_attr_t attr;
-	pthread_attr_init(&attr);
+	if (pthread_attr_init(&attr) != 0)
+	{
+		Server->Log("Error initializing pthread attrs", LL_ERROR);
+		return false;
+	}
 #if !defined(URB_THREAD_STACKSIZE64) || !defined(URB_THREAD_STACKSIZE32)
 
 #ifndef _LP64
@@ -1360,9 +1552,33 @@ void CServer::createThread(IThread *thread, const std::string& name)
 
 #endif
 
+	if (flags & IServer::CreateThreadFlags_LargeStackSize)
+	{
+		pthread_attr_setstacksize(&attr, thread_large_stack_size);
+	}
+
+	if (pthread_attr_setdetachstate(&attr, PTHREAD_CREATE_DETACHED) != 0)
+	{
+		Server->Log("Error setting thread to detached", LL_ERROR);
+		pthread_attr_destroy(&attr);
+		return false;
+	}
+
 	pthread_t t;
-	pthread_create(&t, &attr, &thread_helper_f,  (void*)thread);
-	pthread_detach(t);
+	if (pthread_create(&t, &attr, &thread_helper_f, (void*)thread) != 0)
+	{
+		if (errno == ENOMEM
+			&& flags & IServer::CreateThreadFlags_LargeStackSize)
+		{
+			pthread_attr_destroy(&attr);
+			Server->Log("Error creating pthread (ENOMEM). Trying to create thread without LargeStackSize", LL_WARNING);
+			return createThread(thread, name, 
+				static_cast<CreateThreadFlags>(flags & ~(IServer::CreateThreadFlags_LargeStackSize)));
+		}
+		Server->Log("Error creating pthread. Errno: " + convert(errno), LL_ERROR);
+		pthread_attr_destroy(&attr);
+		return false;
+	}
 
 #ifdef HAVE_PTHREAD_SETNAME_NP
 	if (!name.empty())
@@ -1382,6 +1598,7 @@ void CServer::createThread(IThread *thread, const std::string& name)
 #endif //HAVE_PTHREAD_SETNAME_NP
 
 	pthread_attr_destroy(&attr);
+	return true;
 #endif
 }
 
@@ -1519,7 +1736,7 @@ bool CServer::fileExists(std::string pFilename)
 #ifndef WIN32
 	return ::FileExists(pFilename);
 #else
-	fstream in(ConvertToWchar(pFilename).c_str(), ios::in);
+	std::fstream in(ConvertToWchar(pFilename).c_str(), std::ios::in);
 	if( in.is_open()==false )
 		return false;
 
@@ -1778,6 +1995,11 @@ IPipeThrottler* CServer::createPipeThrottler(
 	return new PipeThrottler(bps, percent_max, updater);
 }
 
+IThreadPool * CServer::createThreadPool(size_t max_threads, size_t max_waiting_threads, const std::string & idle_name)
+{
+	return new CThreadPool(max_threads, max_waiting_threads, idle_name);
+}
+
 void CServer::shutdown(void)
 {
 	run=false;
@@ -2029,3 +2251,19 @@ void CServer::setLogConsoleTime(bool b)
 {
 	log_console_time = b;
 }
+
+#ifdef _WIN32
+void CServer::setSocketWindowSizes(int p_send_window_size, int p_recv_window_size)
+{
+	send_window_size = p_send_window_size;
+	recv_window_size = p_recv_window_size;
+}
+int CServer::getSendWindowSize()
+{
+	return send_window_size;
+}
+int CServer::getRecvWindowSize()
+{
+	return recv_window_size;
+}
+#endif
