@@ -68,12 +68,20 @@
 
 #include "FileMetadataPipe.h"
 
+#ifdef __linux__
 #ifndef SEEK_DATA
 #define SEEK_DATA 3
 #endif
 #ifndef SEEK_HOLE
 #define SEEK_HOLE 4
 #endif
+#endif
+
+#if defined(__FreeBSD__)
+//2017-05-12: Make illegal. Performance problems with ZFS on FreeBSD
+#undef SEEK_HOLE
+#define SEEK_HOLE 400
+#endif
 
 #define CLIENT_TIMEOUT	120
 #define CHECK_BASE_PATH
@@ -81,7 +89,8 @@
 
 
 CClientThread::CClientThread(SOCKET pSocket, CTCPFileServ* pParent)
-	: extra_buffer(NULL), waiting_for_chunk(false)
+	: extra_buffer(NULL), waiting_for_chunk(false),
+	backup_semantics(true)
 {
 	int_socket=pSocket;
 
@@ -100,10 +109,23 @@ CClientThread::CClientThread(SOCKET pSocket, CTCPFileServ* pParent)
 	hFile=INVALID_HANDLE_VALUE;
 
 #ifdef _WIN32
-	int window_size;
-	int window_size_len=sizeof(window_size);
-	getsockopt(pSocket, SOL_SOCKET, SO_SNDBUF,(char *) &window_size, &window_size_len );
-	Log("Info: Window size="+convert(window_size));
+	int window_size = Server->getSendWindowSize();
+	if (window_size > 0)
+	{
+		int err = setsockopt(pSocket, SOL_SOCKET, SO_SNDBUF, (char *)&window_size, sizeof(window_size));
+
+		if (err == SOCKET_ERROR)
+			Log("Error: Can't modify SO_SNDBUF", LL_DEBUG);
+	}
+
+	window_size = Server->getRecvWindowSize();
+	if (window_size > 0)
+	{
+		int err = setsockopt(pSocket, SOL_SOCKET, SO_RCVBUF, (char *)&window_size, sizeof(window_size));
+
+		if (err == SOCKET_ERROR)
+			Log("Error: Can't modify SO_RCVBUF", LL_DEBUG);
+	}
 #endif
 
 	close_the_socket=true;
@@ -166,6 +188,8 @@ void CClientThread::operator()(void)
 	}
 #endif
 
+	backup_semantics = FileServFactory::backupSemanticsEnabled();
+
 	while( RecvMessage() && !stopped )
 	{
 	}
@@ -188,7 +212,7 @@ void CClientThread::operator()(void)
 				{
 					if (next_chunks.front().pipe_file_user == NULL)
 					{
-						FileServ::decrShareActive(next_chunks.front().s_filename);
+						FileServ::decrShareActive(next_chunks.front().s_filename, next_chunks.front().share_active_gen);
 						Server->destroy(next_chunks.front().update_file);
 					}
 					delete next_chunks.front().pipe_file_user;
@@ -430,7 +454,11 @@ bool CClientThread::ProcessPacket(CRData *data)
                 {
                     Log("Sending meta data of "+o_filename +" metadata_id=" + convert(metadata_id) , LL_DEBUG);
 
-					assert(metadata_id!=0);
+					if (metadata_id == 0)
+					{
+						Log("Trying to send metadata bug metadata_id is zero", LL_ERROR);
+						return false;
+					}
                 }
 				
 
@@ -606,8 +634,10 @@ bool CClientThread::ProcessPacket(CRData *data)
 #ifndef BACKUP_SEM
 				hFile=CreateFileW(Server->ConvertToWchar(filename).c_str(), FILE_READ_DATA, FILE_SHARE_READ, NULL, OPEN_EXISTING, FILE_FLAG_OVERLAPPED|FILE_FLAG_SEQUENTIAL_SCAN, NULL);
 #else
+				if (backup_semantics)
+					extra_flags |= FILE_FLAG_BACKUP_SEMANTICS;
 				hFile=CreateFileW(Server->ConvertToWchar(filename).c_str(), FILE_READ_DATA, FILE_SHARE_READ, NULL, OPEN_EXISTING,
-					FILE_FLAG_OVERLAPPED|FILE_FLAG_BACKUP_SEMANTICS|FILE_FLAG_SEQUENTIAL_SCAN|extra_flags, NULL);
+					FILE_FLAG_OVERLAPPED|FILE_FLAG_SEQUENTIAL_SCAN|extra_flags, NULL);
 #endif
 
 				if(hFile == INVALID_HANDLE_VALUE)
@@ -954,7 +984,8 @@ bool CClientThread::ProcessPacket(CRData *data)
 				flags |= O_CLOEXEC;
 #endif
 #if defined(O_NOATIME)
-				flags |= O_NOATIME;
+				if(backup_semantics)
+					flags |= O_NOATIME;
 #endif
 				hFile=open64(filename.c_str(), flags);
 				
@@ -1224,7 +1255,13 @@ bool CClientThread::ProcessPacket(CRData *data)
 						if(with_hashes && foffset==next_checkpoint)
 						{
 							hash_func.finalize();
-							SendInt((char*)hash_func.raw_digest_int(), 16);
+							rc = SendInt((char*)hash_func.raw_digest_int(), 16);
+							if (rc == SOCKET_ERROR)
+							{
+								Log("Error: Sending hash failed");
+								CloseHandle(hFile);
+								return false;
+							}
 							next_checkpoint+=c_checkpoint_dist;
 							if(next_checkpoint>curr_filesize)
 								next_checkpoint=curr_filesize;
@@ -1263,13 +1300,13 @@ bool CClientThread::ProcessPacket(CRData *data)
 					Handle_ID_BLOCK_REQUEST(data);
 				}
 			}break;
-		case ID_GET_FILE_HASH_AND_METADATA:
+		/*case ID_GET_FILE_HASH_AND_METADATA:
 			{
 				if(!GetFileHashAndMetadata(data))
 				{
 					return false;
 				}
-			}break;
+			}break;*/
 
 		case ID_INFORM_METADATA_STREAM_END:
 			{
@@ -1285,6 +1322,13 @@ bool CClientThread::ProcessPacket(CRData *data)
 					return false;
 				}
 			} break;
+		case ID_STOP_PHASH:
+			{
+				if (!StopPhash(data))
+				{
+					return false;
+				}
+			} break;
 		case ID_FREE_SERVER_FILE:
 			{
 				if (chunk_send_thread_ticket != ILLEGAL_THREADPOOL_TICKET)
@@ -1495,24 +1539,10 @@ int CClientThread::SendData()
 					ts=ldata->bsize;
 
 				_i32 rc=SendInt(&ldata->buffer[sent], ts);
+				bool has_error = false;
 				if( rc==SOCKET_ERROR )
 				{
-					int err;
-	#ifdef _WIN32
-					err=WSAGetLastError();
-	#else
-					err=errno;
-	#endif
-					Log("SOCKET_ERROR in SendData(). BSize: "+convert(ldata->bsize)+" WSAGetLastError: "+convert(err), LL_DEBUG);
-				
-					if( ldata->delbuf )
-					{
-						bufmgr->releaseBuffer(ldata->delbufptr);
-						ldata->delbuf=false;
-					}
-					t_send.erase( t_send.begin() );
-					delete ldata;
-					return -1;
+					has_error = true;
 				}
 				else if(with_hashes)
 				{
@@ -1520,18 +1550,46 @@ int CClientThread::SendData()
 				}
 				sent+=ts;
 				sent_bytes+=ts;
-				if(with_hashes)
+				if(with_hashes
+					&& !has_error)
 				{
 					if(next_checkpoint-sent_bytes==0)
 					{
 						hash_func.finalize();
-						SendInt((char*)hash_func.raw_digest_int(), 16);
-						next_checkpoint+=c_checkpoint_dist;
-						if(next_checkpoint>curr_filesize)
-							next_checkpoint=curr_filesize;
-						hash_func.init();
+						rc = SendInt((char*)hash_func.raw_digest_int(), 16);
+						if (rc == SOCKET_ERROR)
+						{
+							has_error = true;
+						}
+						else
+						{
+							next_checkpoint += c_checkpoint_dist;
+							if (next_checkpoint > curr_filesize)
+								next_checkpoint = curr_filesize;
+							hash_func.init();
+						}
 					}
 				}
+
+				if (has_error)
+				{
+					int err;
+#ifdef _WIN32
+					err = WSAGetLastError();
+#else
+					err = errno;
+#endif
+					Log("SOCKET_ERROR in SendData(). BSize: " + convert(ldata->bsize) + " WSAGetLastError: " + convert(err), LL_DEBUG);
+
+					if (ldata->delbuf)
+					{
+						bufmgr->releaseBuffer(ldata->delbufptr);
+						ldata->delbuf = false;
+					}
+					t_send.erase(t_send.begin());
+					delete ldata;
+					return -1;
+				}
 			}
 		}
 		else
@@ -1826,7 +1884,10 @@ bool CClientThread::GetFileBlockdiff(CRData *data, bool with_metadata)
 #ifndef BACKUP_SEM
 		hFile=CreateFileW(Server->ConvertToWchar(filename).c_str(), FILE_READ_DATA, FILE_SHARE_READ, NULL, OPEN_EXISTING, FILE_FLAG_SEQUENTIAL_SCAN, NULL);
 #else
-		hFile=CreateFileW(Server->ConvertToWchar(filename).c_str(), FILE_READ_DATA, FILE_SHARE_READ, NULL, OPEN_EXISTING, FILE_FLAG_BACKUP_SEMANTICS|FILE_FLAG_SEQUENTIAL_SCAN, NULL);
+		DWORD flags = FILE_FLAG_SEQUENTIAL_SCAN;
+		if (backup_semantics)
+			flags |= FILE_FLAG_BACKUP_SEMANTICS;
+		hFile=CreateFileW(Server->ConvertToWchar(filename).c_str(), FILE_READ_DATA, FILE_SHARE_READ, NULL, OPEN_EXISTING, flags, NULL);
 #endif
 #else //_WIN32
 		int flags = O_RDONLY | O_LARGEFILE;
@@ -1834,7 +1895,8 @@ bool CClientThread::GetFileBlockdiff(CRData *data, bool with_metadata)
 		flags |= O_CLOEXEC;
 #endif
 #if defined(O_NOATIME)
-		flags |= O_NOATIME;
+		if(backup_semantics)
+			flags |= O_NOATIME;
 #endif
 		hFile=open64(filename.c_str(), flags);
 #endif //_WIN32
@@ -1941,7 +2003,7 @@ bool CClientThread::GetFileBlockdiff(CRData *data, bool with_metadata)
 
 	hFile=INVALID_HANDLE_VALUE;
 
-	scoped_share_active.release();
+	chunk.share_active_gen = scoped_share_active.release();
 
 	queueChunk(chunk);
 
@@ -2083,7 +2145,10 @@ bool CClientThread::GetFileHashAndMetadata( CRData* data )
 #ifndef BACKUP_SEM
 	hFile=CreateFileW(Server->ConvertToWchar(filename).c_str(), FILE_READ_DATA, FILE_SHARE_READ, NULL, OPEN_EXISTING, FILE_FLAG_SEQUENTIAL_SCAN, NULL);
 #else
-	hFile=CreateFileW(Server->ConvertToWchar(filename).c_str(), FILE_READ_DATA, FILE_SHARE_READ, NULL, OPEN_EXISTING, FILE_FLAG_BACKUP_SEMANTICS|FILE_FLAG_SEQUENTIAL_SCAN, NULL);
+	DWORD flags = FILE_FLAG_SEQUENTIAL_SCAN;
+	if (backup_semantics)
+		flags |= FILE_FLAG_BACKUP_SEMANTICS;
+	hFile=CreateFileW(Server->ConvertToWchar(filename).c_str(), FILE_READ_DATA, FILE_SHARE_READ, NULL, OPEN_EXISTING, flags, NULL);
 #endif
 #else //_WIN32
 	int flags = O_RDONLY | O_LARGEFILE;
@@ -2091,7 +2156,8 @@ bool CClientThread::GetFileHashAndMetadata( CRData* data )
 	flags |= O_CLOEXEC;
 #endif
 #if defined(O_NOATIME)
-	flags |= O_NOATIME;
+	if(backup_semantics)
+		flags |= O_NOATIME;
 #endif
 	hFile=open64(filename.c_str(), flags);
 #endif //_WIN32
@@ -2274,7 +2340,7 @@ bool CClientThread::sendFullFile(IFile* file, _i64 start_offset, bool with_hashe
 
 		bool curr_flush = rc<count;
 
-		if (last_flush_foffset - foffset>1 * 1024 * 1024)
+		if (foffset - last_flush_foffset>1 * 1024 * 1024)
 		{
 			curr_flush = true;
 			last_flush_foffset = foffset;
@@ -2296,7 +2362,12 @@ bool CClientThread::sendFullFile(IFile* file, _i64 start_offset, bool with_hashe
 		if(with_hashes && foffset==next_checkpoint)
 		{
 			hash_func.finalize();
-			SendInt((char*)hash_func.raw_digest_int(), 16, curr_flush);
+			rc=SendInt((char*)hash_func.raw_digest_int(), 16, curr_flush);
+			if (rc == SOCKET_ERROR)
+			{
+				Log("Error: Sending hash failed");
+				return false;
+			}
 			next_checkpoint+=c_checkpoint_dist;
 			if(next_checkpoint>curr_filesize && curr_filesize>0)
 				next_checkpoint=curr_filesize;
@@ -2357,6 +2428,42 @@ bool CClientThread::InformMetadataStreamEnd( CRData * data )
 	return true;
 }
 
+bool CClientThread::StopPhash(CRData * data)
+{
+#ifdef CHECK_IDENT
+	std::string ident;
+	data->getStr(&ident);
+	if (!FileServ::checkIdentity(ident))
+	{
+		Log("Identity check failed -hash", LL_DEBUG);
+		return false;
+	}
+#endif
+
+	std::string token;
+	data->getStr(&token);
+
+	std::string fn;
+	data->getStr(&fn);
+
+	PipeSessions::phashEnd(token, fn);
+
+	char ch = ID_PONG;
+	int rc = SendInt(&ch, 1);
+	if (rc == SOCKET_ERROR)
+	{
+		Log("Error: Sending data failed (InformMetadataStreamEnd)");
+		return false;
+	}
+
+	if (!clientpipe->Flush(CLIENT_TIMEOUT * 1000))
+	{
+		Server->Log("Error flushing output socket (2)", LL_INFO);
+	}
+
+	return true;
+}
+
 bool CClientThread::FinishScript( CRData * data )
 {
 #ifdef CHECK_IDENT
@@ -2625,7 +2732,7 @@ int64 CClientThread::getFileExtents(int64 fsize, int64& n_sparse_extents, std::v
 			return -1;
 		}
 	}
-#else
+#elif SEEK_HOLE
 	off64_t last_sparse_pos = 0;
 	has_file_extents = false;
 	real_fsize = fsize;
@@ -2689,6 +2796,9 @@ int64 CClientThread::getFileExtents(int64 fsize, int64& n_sparse_extents, std::v
 
 	n_sparse_extents = file_extents.size();
 
+#else
+	has_file_extents = false;
+	real_fsize = fsize;
 #endif
 	return real_fsize;
 }
