@@ -25,15 +25,19 @@
 #include "../common/data.h"
 #include "PipeFile.h"
 #include "PipeFileTar.h"
+#include "PipeFileExt.h"
 #include <string>
 #include <stdlib.h>
+#include "map_buffer.h"
 
 volatile bool PipeSessions::do_stop = false;
 IMutex* PipeSessions::mutex = NULL;
+IMutex* PipeSessions::active_shares_mutex = NULL;
 std::map<std::string, SPipeSession> PipeSessions::pipe_files;
 std::map<std::string, SExitInformation> PipeSessions::exit_information;
 std::map<std::pair<std::string, std::string>, IFileServ::IMetadataCallback*> PipeSessions::metadata_callbacks;
-std::map<std::string, size_t> PipeSessions::active_shares;
+std::map<std::pair<std::string, size_t>, size_t> PipeSessions::active_shares;
+size_t PipeSessions::active_shares_gen = 0;
 
 const int64 pipe_file_timeout = 1*60*60*1000;
 const int64 pipe_file_read_timeout = 30 * 60 * 1000;
@@ -108,7 +112,8 @@ IFile* PipeSessions::getFile(const std::string& cmd, ScopedPipeFileUser& pipe_fi
 			script_cmd.erase(script_cmd.size()-output_filename.size(), output_filename.size());
 
 			bool l_tar_file = false;
-			std::string script_filename = FileServ::mapScriptOutputNameToScript(output_filename, l_tar_file);
+			IPipeFile* pipe_file = NULL;
+			std::string script_filename = FileServ::mapScriptOutputNameToScript(output_filename, l_tar_file, pipe_file);
 
 			script_cmd = "\"" + script_cmd + script_filename +
 				"\" \"" + output_filename + "\" "+convert(backupnum);
@@ -119,7 +124,11 @@ IFile* PipeSessions::getFile(const std::string& cmd, ScopedPipeFileUser& pipe_fi
 			}
 
 			IPipeFile* nf;
-			if (!l_tar_file)
+			if (pipe_file != NULL)
+			{
+				nf = pipe_file;
+			}
+			else if (!l_tar_file)
 			{
 				nf = new PipeFile(script_cmd);
 			}
@@ -166,6 +175,7 @@ void PipeSessions::injectPipeSession(const std::string & session_key, int backup
 void PipeSessions::init()
 {
 	mutex = Server->createMutex();
+	active_shares_mutex = Server->createMutex();
 
 	Server->getThreadPool()->execute(new PipeSessions, "PipeSession: timeout");
 }
@@ -173,6 +183,7 @@ void PipeSessions::init()
 void PipeSessions::destroy()
 {
 	delete mutex;
+	delete active_shares_mutex;
 
 	do_stop=true;
 }
@@ -398,21 +409,26 @@ IFileServ::IMetadataCallback* PipeSessions::transmitFileMetadata( const std::str
 
 	IScopedLock lock(mutex);
 
-	std::map<std::pair<std::string, std::string>, IFileServ::IMetadataCallback*>::iterator iter_cb = 
-		metadata_callbacks.find(std::make_pair(sharename, identity));
-
 	IFileServ::IMetadataCallback* ret = NULL;
-	if(iter_cb!=metadata_callbacks.end())
-	{
-		data.addVoidPtr(iter_cb->second);
-		ret = iter_cb->second;
-	}
 
 	std::map<std::string, SPipeSession>::iterator it = pipe_files.find("urbackup/FILE_METADATA|"+server_token);
 
 	if(it!=pipe_files.end() && it->second.input_pipe!=NULL)
 	{
-		++active_shares[sharename + "|" + server_token];
+		{
+			IScopedLock alock(active_shares_mutex);
+			++active_shares[std::make_pair(sharename + "|" + server_token, active_shares_gen)];
+			data.addUInt64(active_shares_gen);
+		}
+
+		std::map<std::pair<std::string, std::string>, IFileServ::IMetadataCallback*>::iterator iter_cb =
+			metadata_callbacks.find(std::make_pair(sharename, identity));
+
+		if(iter_cb!=metadata_callbacks.end())
+		{
+			data.addVoidPtr(iter_cb->second);
+			ret = iter_cb->second;
+		}
 
 		it->second.input_pipe->Write(data.getDataPtr(), data.getDataSize());
 	}
@@ -449,7 +465,11 @@ void PipeSessions::transmitFileMetadata(const std::string & public_fn, const std
 
 	if (it != pipe_files.end() && it->second.input_pipe != NULL)
 	{
-		++active_shares[sharename + "|" + server_token];
+		{
+			IScopedLock alock(active_shares_mutex);
+			++active_shares[std::make_pair(sharename + "|" + server_token, active_shares_gen)];
+			data.addUInt64(active_shares_gen);
+		}
 
 		it->second.input_pipe->Write(data.getDataPtr(), data.getDataSize());
 	}
@@ -492,12 +512,15 @@ void PipeSessions::transmitFileMetadataAndFiledataWait(const std::string & publi
 
 	if (it != pipe_files.end() && it->second.input_pipe != NULL)
 	{
-		++active_shares[sharename + "|" + server_token];
+		{
+			IScopedLock alock(active_shares_mutex);
+			active_shares[std::make_pair(sharename + "|" + server_token, active_shares_gen)]+=2;
+			metadatamsg.addUInt(active_shares_gen);
+			datamsg.addUInt(active_shares_gen);
+		}	
 
 		it->second.input_pipe->Write(datamsg.getDataPtr(), datamsg.getDataSize());
 
-		++active_shares[sharename + "|" + server_token];
-
 		it->second.input_pipe->Write(metadatamsg.getDataPtr(), metadatamsg.getDataSize());
 
 		lock.relock(NULL);
@@ -512,7 +535,7 @@ void PipeSessions::transmitFileMetadataAndFiledataWait(const std::string & publi
 
 
 
-void PipeSessions::fileMetadataDone(const std::string & public_fn, const std::string& server_token)
+void PipeSessions::fileMetadataDone(const std::string & public_fn, const std::string& server_token, size_t active_gen)
 {
 	std::string sharename = getuntil("/", public_fn);
 	if (sharename.empty())
@@ -520,9 +543,9 @@ void PipeSessions::fileMetadataDone(const std::string & public_fn, const std::st
 		sharename = public_fn;
 	}
 
-	IScopedLock lock(mutex);
+	IScopedLock lock(active_shares_mutex);
 
-	std::map<std::string, size_t>::iterator it = active_shares.find(sharename + "|" + server_token);
+	std::map<std::pair<std::string, size_t>, size_t>::iterator it = active_shares.find(std::make_pair(sharename + "|" + server_token, active_gen));
 
 	if (it != active_shares.end())
 	{
@@ -536,11 +559,37 @@ void PipeSessions::fileMetadataDone(const std::string & public_fn, const std::st
 
 bool PipeSessions::isShareActive(const std::string & sharename, const std::string& server_token)
 {
-	IScopedLock lock(mutex);
+	IScopedLock lock(active_shares_mutex);
 
-	std::map<std::string, size_t>::iterator it = active_shares.find(sharename + "|" + server_token);
+	for (std::map<std::pair<std::string, size_t>, size_t>::iterator it = active_shares.begin();
+		it != active_shares.end(); ++it)
+	{
+		if (it->first.first == sharename + "|" + server_token)
+			return true;
+	}
 
-	return it != active_shares.end();
+	return false;
+}
+
+bool PipeSessions::isShareActiveGen(const std::string& sharename, const std::string& server_token, size_t gen)
+{
+	IScopedLock lock(active_shares_mutex);
+
+	for (std::map<std::pair<std::string, size_t>, size_t>::iterator it = active_shares.begin();
+		it != active_shares.end(); ++it)
+	{
+		if (it->first.first == sharename + "|" + server_token &&
+			it->first.second<=gen)
+			return true;
+	}
+
+	return false;
+}
+
+void PipeSessions::setActiveSharesGen(size_t gen)
+{
+	IScopedLock lock(active_shares_mutex);
+	active_shares_gen = gen;
 }
 
 void PipeSessions::metadataStreamEnd( const std::string& server_token )
@@ -558,6 +607,26 @@ void PipeSessions::metadataStreamEnd( const std::string& server_token )
 	}
 }
 
+void PipeSessions::phashEnd(const std::string & server_token, const std::string & phash_fn)
+{
+	IScopedLock lock(mutex);
+
+	bool allow_exec;
+	std::string fn = map_file("phash_{9c28ff72-5a74-487b-b5e1-8f1c96cd0cf4}/phash_" + phash_fn, std::string(), allow_exec, NULL);
+
+	if (fn.empty())
+		return;
+
+	std::map<std::string, SPipeSession>::iterator it = pipe_files.find(fn+"|0|0|" + server_token);
+
+	if (it != pipe_files.end())
+	{
+		PipeFileExt* pext = dynamic_cast<PipeFileExt*>(it->second.file);
+		if(pext!=NULL)
+			pext->forceExit();
+	}
+}
+
 void PipeSessions::registerMetadataCallback( const std::string &name, const std::string& identity, IFileServ::IMetadataCallback* callback )
 {
 	IScopedLock lock(mutex);
@@ -595,7 +664,7 @@ std::string PipeSessions::getKey( const std::string& cmd, int& backupnum, int64&
 	}
 
 	std::vector<std::string> cmd_toks;
-	TokenizeMail(cmd, cmd_toks, "|");
+	Tokenize(cmd, cmd_toks, "|");
 	std::string script_cmd = cmd_toks[0];
 
 	if(script_cmd=="urbackup/FILE_METADATA" && cmd_toks.size()>1)
