@@ -58,6 +58,9 @@ extern IServer* Server;
 #include "cowfile.h"
 #endif
 #include "fs/ntfs.h"
+#ifdef _WIN32
+#include "fs/ntfs_win.h"
+#endif
 
 #include "pluginmgr.h"
 
@@ -378,13 +381,13 @@ namespace
 
 				for(int64 block=0;block<numblocks;++block)
 				{
-					char* buf = input_fs[i]->readBlock(block);
+					IFilesystem::IFsBuffer* buf = input_fs[i]->readBlock(block);
 					if(buf!=NULL)
 					{
 						fs_buffer fsb(input_fs[i], buf);
 
 						vhdout.Seek(out_pos);
-						if(vhdout.Write(buf, static_cast<_u32>(blocksize))!=static_cast<_u32>(blocksize))
+						if(vhdout.Write(buf->getBuf(), static_cast<_u32>(blocksize))!=static_cast<_u32>(blocksize))
 						{
 							Server->Log("Error writing to VHD output file", LL_ERROR);
 							return false;
@@ -704,7 +707,7 @@ DLLEXPORT void LoadActions(IServer* pServer)
 
 		if(!selected_via_gui)
 		{
-			TokenizeMail(assemble, input_files, ";");
+			Tokenize(assemble, input_files, ";");
 			output_file = Server->getServerParameter("output_file");
 		}
 
@@ -743,18 +746,16 @@ DLLEXPORT void LoadActions(IServer* pServer)
 	if(!vhdcopy_in.empty())
 	{
 		Server->Log("VHDCopy.");
-		VHDFile in(vhdcopy_in, true,0);
-		if(in.isOpen()==false)
+		IVHDFile* in = open_device_file(vhdcopy_in);
+		if(in==NULL || in->isOpen()==false)
 		{
-			Server->Log("Error opening VHD-File \""+vhdcopy_in+"\"", LL_ERROR);
+			Server->Log("Error opening image file \""+vhdcopy_in+"\"", LL_ERROR);
 			exit(4);
 		}
 
-		uint64 vhdsize=in.getSize();
-		float vhdsize_gb=(vhdsize/1024)/1024.f/1024.f;
-		uint64 vhdsize_mb=vhdsize/1024/1024;
-		Server->Log("VHD Info: Size: "+convert(vhdsize_gb)+" GB "+convert(vhdsize_mb)+" MB",LL_INFO);
-		unsigned int vhd_blocksize=in.getBlocksize();
+		uint64 vhdsize=in->getSize();
+		Server->Log("Image info: Size: "+PrettyPrintBytes(vhdsize),LL_INFO);
+		unsigned int vhd_blocksize=in->getBlocksize();
 		
 		std::string vhdcopy_out=Server->getServerParameter("vhdcopy_out");
 		if(vhdcopy_out.empty())
@@ -778,9 +779,13 @@ DLLEXPORT void LoadActions(IServer* pServer)
 				{
 					skip=atoi(skip_s.c_str());
 				}
+				else if (is_disk_mbr(vhdcopy_in + ".mbr"))
+				{
+					skip = 0;
+				}
 
 				Server->Log("Skipping "+convert(skip)+" bytes...", LL_INFO);
-				in.Seek(skip);
+				in->Seek(skip);
 				char buffer[4096];
 				size_t read;
 				int last_pc=0;
@@ -791,7 +796,7 @@ DLLEXPORT void LoadActions(IServer* pServer)
 				out->Seek(0);
 				while(currpos%vhd_blocksize!=0)
 				{
-					is_ok=in.Read(buffer, 512, read);
+					is_ok=in->Read(buffer, 512, read);
 					if(read>0)
 					{
 						_u32 rc=out->Write(buffer, (_u32)read);
@@ -811,9 +816,9 @@ DLLEXPORT void LoadActions(IServer* pServer)
 
 				do
 				{
-					if(in.has_sector())
+					if(in->has_sector())
 					{
-						is_ok=in.Read(buffer, 4096, read);
+						is_ok=in->Read(buffer, 4096, read);
 						if(read>0)
 						{
 							_u32 rc=out->Write(buffer, (_u32)read);
@@ -829,7 +834,7 @@ DLLEXPORT void LoadActions(IServer* pServer)
 					{
 						read=4096;
 						currpos+=read;
-						in.Seek(currpos);
+						in->Seek(currpos);
 						out->Seek(currpos-skip);
 					}
 					
@@ -960,6 +965,12 @@ DLLEXPORT void LoadActions(IServer* pServer)
 		
 		const int64 vhd_blocksize=(1024*1024)/2;
 		int skip=1024*512;
+		std::string s_verify_skip = Server->getServerParameter("verify_skip");
+		if (!s_verify_skip.empty())
+		{
+			skip = watoi(s_verify_skip);
+		}
+
 		in->Seek(skip);
 		uint64 currpos=skip;
 		uint64 size=in->getSize();
@@ -1327,6 +1338,172 @@ DLLEXPORT void LoadActions(IServer* pServer)
 	}
 #endif
 
+#ifdef _WIN32
+	std::string disk_scrub = Server->getServerParameter("disk_scrub");
+	if (!disk_scrub.empty())
+	{
+		if (disk_scrub == "SelectViaGUI")
+		{
+			std::string disk_path = select_dir_via_dialog("Please select the disk to scrub");
+			if (disk_path.empty())
+			{
+				Server->Log("No disk selected", LL_ERROR);
+				exit(1);
+			}
+			disk_scrub = disk_path;
+		}
+
+		if (disk_scrub.size() > 2
+			&& isletter(disk_scrub[0])
+			&& disk_scrub[1] == ':'
+			&& disk_scrub[2] == '\\')
+		{
+			disk_scrub = disk_scrub.substr(0, 1);
+		}
+		else if (disk_scrub.size() == 2
+			&& isletter(disk_scrub[0])
+			&& disk_scrub[1] == ':')
+		{
+			disk_scrub = disk_scrub.substr(0, 1);
+		}
+		else if (!(disk_scrub.size() == 1
+			&& isletter(disk_scrub[0])))
+		{
+			Server->Log("Disk not specified correctly: " + disk_scrub, LL_ERROR);
+			exit(2);
+		}
+
+		FSNTFSWIN fs("\\\\?\\" + disk_scrub + ":", IFSImageFactory::EReadaheadMode_Overlapped,
+			false, NULL);
+
+		if (fs.hasError())
+		{
+			Server->Log("Error opening volume " + disk_scrub, LL_ERROR);
+			exit(5);
+		}
+
+		Server->Log("Scrubbing volume " + disk_scrub + " size " + PrettyPrintBytes(fs.getSize())
+			+" used "+PrettyPrintBytes(fs.calculateUsedSpace()));
+
+		int64 drivesize = fs.getSize();
+		int64 blocksize = fs.getBlocksize();
+		int pc = -1;
+		std::vector<int64> error_blocks;
+		int64 n_blocks = drivesize / blocksize;
+		for (int64 i = 0, blocks = n_blocks; i < blocks; ++i)
+		{
+			bool has_error = false;
+			IFilesystem::IFsBuffer* buf = fs.readBlock(i, &has_error);
+
+			if (has_error)
+			{
+				Server->Log("Block " + convert(i) + " has read error", LL_ERROR);
+				error_blocks.push_back(i);
+			}
+
+			if (buf != NULL)
+				fs.releaseBuffer(buf);
+
+			int curr_pc = static_cast<int>((i * 100) / n_blocks);
+			if (curr_pc != pc)
+			{
+				Server->Log("Reading volume. " + convert(curr_pc) + "% finished");
+				pc = curr_pc;
+			}
+		}
+
+		fs.readaheadSetMaxNBuffers(1);
+
+		Server->Log("Re-reading blocks with errors with smaller block size (4096 bytes)...");
+
+		size_t err_num = 0;
+		size_t total_errs = error_blocks.size();
+		for (size_t i = 0; i < error_blocks.size();)
+		{
+			int64 pos = error_blocks[i];
+			Server->Log("Error block position " +convert(pos)+" error "+ convert(err_num) + "/" + convert(total_errs));
+			
+			bool has_error = false;
+			IFilesystem::IFsBuffer* buf = fs.readBlock(pos, &has_error);
+
+			if (has_error)
+			{
+				Server->Log("Block " + convert(pos) + " has read error", LL_ERROR);
+				++i;
+			}
+			else
+			{
+				Server->Log("Block " + convert(pos) + " is ok", LL_ERROR);
+				error_blocks.erase(error_blocks.begin() + i);
+			}
+
+			if (buf != NULL)
+				fs.releaseBuffer(buf);
+		}
+
+		if (!error_blocks.empty())
+		{
+			Server->Log("Volume has "+convert(error_blocks.size())+" read errors", LL_INFO);
+
+			std::cout << "Search through all file extents to find out which files are damaged? Y/n" << std::endl;
+			char ch;
+			std::cin >> ch;
+
+			if (ch != 'n'
+				&& ch != 'N')
+			{
+				fs.logFileBlocks("\\\\?\\" + disk_scrub + ":\\", error_blocks);
+			}
+
+			std::cout << "Overwrite damaged blocks with zeros? y/N" << std::endl;
+			ch = 0;
+			std::cin >> ch;
+
+			if (ch == 'Y'
+				|| ch == 'y')
+			{
+				HANDLE hDev = CreateFileW(Server->ConvertToWchar("\\\\?\\" + disk_scrub+":").c_str(), 
+					GENERIC_READ|GENERIC_WRITE, FILE_SHARE_READ | FILE_SHARE_WRITE, NULL, OPEN_EXISTING, 
+					FILE_ATTRIBUTE_NORMAL | FILE_FLAG_NO_BUFFERING, NULL);
+
+				if (hDev == INVALID_HANDLE_VALUE)
+				{
+					Server->Log("Error opening device. Last error: "+convert((int64)GetLastError()), LL_ERROR);
+					exit(3);
+				}
+
+				size_t buf_size = fs.getBlocksize();
+				char* buf = reinterpret_cast<char*>(_aligned_malloc(buf_size, buf_size));
+				memset(buf, 0, buf_size);
+
+				IFsFile* vol = Server->openFileFromHandle(hDev, disk_scrub);
+
+				for (size_t i = 0; i < error_blocks.size(); ++i)
+				{
+					Server->Log("Zeroing sector of volume " + disk_scrub + " at position " + convert(error_blocks[i] * fs.getBlocksize())+"...");
+					_u32 rc = vol->Write(error_blocks[i] * fs.getBlocksize(),
+						buf, static_cast<_u32>(buf_size));
+
+					if (rc != buf_size)
+					{
+						Server->Log("Error writing to volume " + disk_scrub + " at position " + convert(error_blocks[i] * fs.getBlocksize()) +
+							". Last error: " + convert((int64)GetLastError()), LL_ERROR);
+					}
+				}
+
+				_aligned_free(buf);
+				Server->Log("Zeroing complete.");
+			}
+		}
+		else
+		{
+			Server->Log("Volume scrub completed. No errors detected.");
+		}
+
+		exit(0);
+	}
+#endif //_WIN32
+
 	imagepluginmgr=new CImagePluginMgr;
 
 	Server->RegisterPluginThreadsafeModel( imagepluginmgr, "fsimageplugin");
