@@ -21,8 +21,9 @@ extern char **environ;
 
 const int mode_btrfs=0;
 const int mode_zfs=1;
+const int mode_zfs_file=2;
 
-CServer *Server;
+IServer *Server;
 
 #ifdef _WIN32
 #include <Windows.h>
@@ -70,6 +71,10 @@ std::string getBackupfolderPath(int mode)
 	{
 		fn_name = "dataset";
 	}
+	else if(mode==mode_zfs_file)
+	{
+		fn_name = "dataset_file";
+	}
 	else
 	{
 		return std::string();
@@ -203,6 +208,8 @@ int exec_wait(const std::string& path, std::string& stdout, ...)
 			return -1;
 		}
 		
+		close(pipefd[1]);
+		
 		int rc = execvp(path.c_str(), args.data());
 		exit(rc);
 	}
@@ -288,41 +295,39 @@ std::string find_zfs_cmd()
 {
 	static std::string zfs_cmd;
 	
-	if(!zfs_cmd.empty())
+	const char* const zfs_locations[] = { "zfs",
+								   "/sbin/zfs", 
+								   "/bin/zfs", 
+								   "/usr/sbin/zfs",
+								   "/usr/bin/zfs" };
+								   
+	for(size_t i=0;i<sizeof(zfs_locations)/sizeof(zfs_locations[0]);++i)
 	{
-		return zfs_cmd;
+		const char* location = zfs_locations[i];
+		if(exec_wait(location, false, "version", NULL)==0)
+		{
+			zfs_cmd=location;
+			return zfs_cmd;
+		}
+		if(exec_wait(location, false, "--version", NULL)==2)
+		{
+			zfs_cmd=location;
+			return zfs_cmd;
+		}
 	}
 	
-	if(exec_wait("zfs", false, "--version", NULL)==2)
+	zfs_cmd="zfs";
+	return zfs_cmd;
+}
+
+void zfs_elevate()
+{
+#ifdef __linux__
+	if(setuid(0)!=0)
 	{
-		zfs_cmd="zfs";
-		return zfs_cmd;
-	}
-	else if(exec_wait("/sbin/zfs", false, "--version", NULL)==2)
-	{
-		zfs_cmd="/sbin/zfs";
-		return zfs_cmd;
-	}
-	else if(exec_wait("/bin/zfs", false, "--version", NULL)==2)
-	{
-		zfs_cmd="/bin/zfs";
-		return zfs_cmd;
-	}
-	else if(exec_wait("/usr/sbin/zfs", false, "--version", NULL)==2)
-	{
-		zfs_cmd="/usr/sbin/zfs";
-		return zfs_cmd;
-	}
-	else if(exec_wait("/usr/bin/zfs", false, "--version", NULL)==2)
-	{
-		zfs_cmd="/usr/bin/zfs";
-		return zfs_cmd;
-	}
-	else
-	{
-		zfs_cmd="zfs";
-		return zfs_cmd;
+		std::cout << "Cannot become root user (2)" << std::endl;
 	}
+#endif
 }
 #endif
 
@@ -339,6 +344,7 @@ bool create_subvolume(int mode, std::string subvolume_folder)
 	}
 	else if(mode==mode_zfs)
 	{
+		zfs_elevate();
 		int rc=exec_wait(find_zfs_cmd(), true, "create", "-p", subvolume_folder.c_str(), NULL);
 		chown_dir(subvolume_folder);
 		return rc==0;
@@ -360,6 +366,7 @@ bool get_mountpoint(int mode, std::string subvolume_folder)
 	}
 	else if(mode==mode_zfs)
 	{
+		zfs_elevate();
 		int rc=exec_wait(find_zfs_cmd(), true, "get", "-H", "-o", "value", "mountpoint", subvolume_folder.c_str(), NULL);
 		return rc==0;
 	}
@@ -380,6 +387,7 @@ bool create_snapshot(int mode, std::string snapshot_src, std::string snapshot_ds
 	}
 	else if(mode==mode_zfs)
 	{
+		zfs_elevate();
 		int rc=exec_wait(find_zfs_cmd(), true, "clone", (snapshot_src+"@ro").c_str(), snapshot_dst.c_str(), NULL);
 		chown_dir(snapshot_dst);
 		return rc==0;
@@ -400,6 +408,7 @@ bool is_subvolume(int mode, std::string subvolume_folder)
 	}
 	else if(mode==mode_zfs)
 	{
+		zfs_elevate();
 		int rc=exec_wait(find_zfs_cmd(), false, "list", subvolume_folder.c_str(), NULL);
 		return rc==0;
 	}
@@ -409,6 +418,8 @@ bool is_subvolume(int mode, std::string subvolume_folder)
 
 bool promote_dependencies(const std::string& snapshot, std::vector<std::string>& dependencies)
 {
+	zfs_elevate();
+	
 	std::cout << "Searching for origin " << snapshot << std::endl;
 	
 	std::string snap_data;
@@ -417,7 +428,7 @@ bool promote_dependencies(const std::string& snapshot, std::vector<std::string>&
 		return false;
 	
 	std::vector<std::string> snaps;
-	TokenizeMail(snap_data, snaps, "\n");
+	Tokenize(snap_data, snaps, "\n");
 		
 	std::string snap_folder = ExtractFilePath(snapshot);
 	for(size_t i=0;i<snaps.size();++i)
@@ -482,6 +493,8 @@ bool remove_subvolume(int mode, std::string subvolume_folder, bool quiet=false)
 	}
 	else if(mode==mode_zfs)
 	{
+		zfs_elevate();
+		
 		exec_wait(find_zfs_cmd(), false, "destroy", (subvolume_folder+"@ro").c_str(), NULL);
 		int rc = exec_wait(find_zfs_cmd(), false, "destroy", subvolume_folder.c_str(), NULL);
 		if(rc!=0)
@@ -536,6 +549,7 @@ bool make_readonly(int mode, std::string subvolume_folder)
 	}
 	else if(mode==mode_zfs)
 	{
+		zfs_elevate();
 		int rc=exec_wait(find_zfs_cmd(), true, "snapshot", (subvolume_folder+"@ro").c_str(), NULL);
 		return rc==0;
 	}
@@ -559,6 +573,19 @@ int zfs_test()
 		&& remove_subvolume(mode_zfs, clientdir) )
 	{
 		std::cout << "ZFS TEST OK" << std::endl;
+		
+		clientdir=getBackupfolderPath(mode_zfs_file)+os_file_sep()+"testA54hj5luZtlorr494";
+		
+		if(getBackupfolderPath(mode_zfs_file).empty())
+		{
+			return 10 + mode_zfs;
+		}
+		else if(create_subvolume(mode_zfs, clientdir)
+			&& remove_subvolume(mode_zfs, clientdir))
+		{
+			return 10 + mode_zfs_file;
+		}
+	
 		return 10 + mode_zfs;
 	}
 	else
@@ -591,7 +618,7 @@ int main(int argc, char *argv[])
 	else
 	{
 		cmd=argv[1];
-	}	
+	}
 
 	std::string backupfolder=getBackupfolderPath(mode);
 	
@@ -603,7 +630,11 @@ int main(int argc, char *argv[])
 		}
 		else if(mode==mode_zfs)
 		{
-			std::cout << "Dataset not set" << std::endl;
+			std::cout << "ZFS image dataset not set" << std::endl;
+		}
+		else if(mode==mode_zfs_file)
+		{
+			std::cout << "ZFS file dataset not set" << std::endl;
 		}
 		else
 		{
@@ -612,6 +643,11 @@ int main(int argc, char *argv[])
 		return 1;
 	}
 	
+	if(cmd!="test" && mode==mode_zfs_file)
+	{
+		mode=mode_zfs;
+	}
+	
 #ifndef _WIN32
 	if(seteuid(0)!=0)
 	{
