@@ -40,6 +40,12 @@
 #include <errno.h>
 #endif
 
+#ifdef __ANDROID__
+extern "C" {
+#include "ifaddrs.c"
+}
+#endif
+
 namespace
 {
 	const std::string str_tmpdir="C:\\Windows\\Temp";
@@ -51,13 +57,7 @@ namespace
 
 	const size_t maxQueuedFiles = 3000;
 	const size_t queuedFilesLow = 100;
-
-	std::string ipToString(sockaddr_in sa)
-	{
-		char str[INET_ADDRSTRLEN];
-		inet_ntop(AF_INET, &(sa.sin_addr), str, INET_ADDRSTRLEN);
-		return str;
-	}
+	const char* multicast_group = "ff12::f894:d:dd00:ef91";
 }
 
 void Log(std::string str)
@@ -92,7 +92,7 @@ bool setSockP(SOCKET sock)
 FileClient::FileClient(bool enable_find_servers, std::string identity, int protocol_version, bool internet_connection,
 	FileClient::ReconnectionCallback *reconnection_callback, FileClient::NoFreeSpaceCallback *nofreespace_callback)
 	: tcpsock(NULL), starttime(0), connect_starttime(0), socket_open(false), connected(false),
-	serveraddr(), local_ip(),
+	local_ip(),
 	max_version(), server_addr(), connection_id(), 
 	protocol_version(protocol_version), internet_connection(internet_connection),
 	transferred_bytes(0), reconnection_callback(reconnection_callback),
@@ -125,111 +125,179 @@ void FileClient::bindToNewInterfaces()
 {
 	std::string s_broadcast_source_port = Server->getServerParameter("broadcast_source_port");
 	unsigned short broadcast_source_port = UDP_SOURCE_PORT;
-	if(!s_broadcast_source_port.empty())
+	if (!s_broadcast_source_port.empty())
 	{
 		broadcast_source_port = static_cast<unsigned short>(atoi(s_broadcast_source_port.c_str()));
 	}
 
-	#ifndef _WIN32
-	std::string bcast_interfaces=Server->getServerParameter("broadcast_interfaces", "");
+#ifndef _WIN32
+	std::string bcast_interfaces = Server->getServerParameter("broadcast_interfaces", "");
 
 	std::vector<std::string> bcast_filter;
-	if(!bcast_interfaces.empty())
+	if (!bcast_interfaces.empty())
 	{
 		Tokenize(bcast_interfaces, bcast_filter, ";,");
 	}
 
 	ifaddrs *start_ifap;
-	int rc=getifaddrs(&start_ifap);
-	if(rc==0)
+	int rc = getifaddrs(&start_ifap);
+	if (rc == 0)
 	{
-		ifaddrs* ifap=start_ifap;
-		for(;ifap!=NULL;ifap=ifap->ifa_next)
+		ifaddrs* ifap = start_ifap;
+		for (; ifap != NULL; ifap = ifap->ifa_next)
 		{
-			bool found_name = bcast_filter.empty() || std::find(bcast_filter.begin(), bcast_filter.end(), ifap->ifa_name)!=bcast_filter.end();
+			bool found_name = bcast_filter.empty() || std::find(bcast_filter.begin(), bcast_filter.end(), ifap->ifa_name) != bcast_filter.end();
 
-			if(found_name &&
-				!(ifap->ifa_flags & IFF_LOOPBACK) 
-				&& !(ifap->ifa_flags & IFF_POINTOPOINT) 
-				&&  (ifap->ifa_flags & IFF_BROADCAST)
-				&&  ifap->ifa_addr->sa_family == AF_INET )
-			{			
-				sockaddr_in source_addr;
-				memset(&source_addr, 0, sizeof(source_addr));
-				source_addr.sin_addr=((struct sockaddr_in *)ifap->ifa_addr)->sin_addr;
-				source_addr.sin_family = AF_INET;
-				source_addr.sin_port = htons(broadcast_source_port);
+			if (found_name &&
+				!(ifap->ifa_flags & IFF_LOOPBACK)
+				&& !(ifap->ifa_flags & IFF_POINTOPOINT))
+			{
+				if (ifap->ifa_addr!=NULL
+					&& ifap->ifa_addr->sa_family == AF_INET
+					&& (ifap->ifa_flags & IFF_BROADCAST))
+				{
+					sockaddr_in source_addr;
+					memset(&source_addr, 0, sizeof(source_addr));
+					source_addr.sin_addr = ((struct sockaddr_in *)ifap->ifa_addr)->sin_addr;
+					source_addr.sin_family = AF_INET;
+					source_addr.sin_port = htons(broadcast_source_port);
 
-				if(std::find(broadcast_iface_addrs.begin(), broadcast_iface_addrs.end(), source_addr.sin_addr.s_addr)!=broadcast_iface_addrs.end())
-					continue;
 
-				int type = SOCK_DGRAM;
+					if (alreadyHasAddrv4(source_addr))
+						continue;
+
+					int type = SOCK_DGRAM;
 #if defined(SOCK_CLOEXEC)
-				type |= SOCK_CLOEXEC;
+					type |= SOCK_CLOEXEC;
 #endif
-				SOCKET udpsock=socket(AF_INET, type,0);
-				if(udpsock==-1)
-				{
-					Server->Log(std::string("Error creating socket for interface ")+std::string(ifap->ifa_name), LL_ERROR);
-					continue;
-				}
+					SOCKET udpsock = socket(AF_INET, type, 0);
+					if (udpsock == -1)
+					{
+						Server->Log(std::string("Error creating socket for interface ") + std::string(ifap->ifa_name), LL_ERROR);
+						continue;
+					}
 #if !defined(SOCK_CLOEXEC)
-				fcntl(udpsock, F_SETFD, fcntl(udpsock, F_GETFD, 0) | FD_CLOEXEC);
+					fcntl(udpsock, F_SETFD, fcntl(udpsock, F_GETFD, 0) | FD_CLOEXEC);
 #endif
 
-				BOOL val=TRUE;
-				int rc = setsockopt(udpsock, SOL_SOCKET, SO_REUSEADDR, &val, sizeof(BOOL));
-				if(rc<0)
-				{
-					Server->Log(std::string("Setting SO_REUSEADDR failed for interface ")+std::string(ifap->ifa_name), LL_ERROR);
+					BOOL val = TRUE;
+					int rc = setsockopt(udpsock, SOL_SOCKET, SO_REUSEADDR, &val, sizeof(BOOL));
+					if (rc < 0)
+					{
+						Server->Log(std::string("Setting SO_REUSEADDR failed for interface ") + std::string(ifap->ifa_name), LL_ERROR);
+					}
+
+					Server->Log(std::string("Binding to interface ") + std::string(ifap->ifa_name) + " (ipv4) for broadcasting...", LL_DEBUG);
+
+					rc = bind(udpsock, (struct sockaddr *)&source_addr, sizeof(source_addr));
+					if (rc < 0)
+					{
+						Server->Log(std::string("Binding UDP socket failed for interface ") + std::string(ifap->ifa_name), LL_ERROR);
+					}
+
+					rc = setsockopt(udpsock, SOL_SOCKET, SO_BROADCAST, (char*)&val, sizeof(BOOL));
+					if (rc < 0)
+					{
+						Server->Log(std::string("Enabling SO_BROADCAST for UDP socket failed for interface ") + std::string(ifap->ifa_name), LL_ERROR);
+						closesocket(udpsock);
+						continue;
+					}
+
+#if defined(__FreeBSD__)
+					int optval = 1;
+					if (setsockopt(udpsock, IPPROTO_IP, IP_ONESBCAST, &optval, sizeof(int)) == -1)
+					{
+						Server->Log(std::string("Error setting IP_ONESBCAST for interface ") + std::string(ifap->ifa_name), LL_ERROR);
+					}
+#endif
+
+
+					SSocket new_udpsock;
+
+					new_udpsock.is_ipv6 = false;
+					new_udpsock.addr_ipv4 = source_addr;
+					new_udpsock.udpsock = udpsock;
+					new_udpsock.broadcast_addr = *((struct sockaddr_in *)ifap->ifa_broadaddr);
+
+					udpsocks.push_back(new_udpsock);
+
+					Server->Log("Broadcasting on ipv4 interface "+std::string(ifap->ifa_name) + " addr " + ipToString(new_udpsock));
 				}
-
-				Server->Log(std::string("Binding to interface ")+std::string(ifap->ifa_name)+" for broadcasting...", LL_DEBUG);
-
-				rc = bind(udpsock, (struct sockaddr *)&source_addr, sizeof(source_addr));
-				if(rc<0)
+				else if (ifap->ifa_addr != NULL 
+					&& ifap->ifa_addr->sa_family == AF_INET6
+					&& Server->getServerParameter("disable_ipv6").empty())
 				{
-					Server->Log(std::string("Binding UDP socket failed for interface ")+std::string(ifap->ifa_name), LL_ERROR);
-				}
-				
-				rc = setsockopt(udpsock, SOL_SOCKET, SO_BROADCAST, (char*)&val, sizeof(BOOL) );
-				if(rc<0)
-				{
-					Server->Log(std::string("Enabling SO_BROADCAST for UDP socket failed for interface ")+std::string(ifap->ifa_name), LL_ERROR);
-					closesocket(udpsock);
-					continue;
-				}
+					sockaddr_in6 source_addr;
+					memset(&source_addr, 0, sizeof(source_addr));
+					source_addr.sin6_addr = ((struct sockaddr_in6 *)ifap->ifa_addr)->sin6_addr;
+					source_addr.sin6_family = AF_INET6;
+					source_addr.sin6_port = htons(broadcast_source_port);
+					source_addr.sin6_scope_id = ((struct sockaddr_in6 *)ifap->ifa_addr)->sin6_scope_id;
 
-				#if defined(__FreeBSD__)
-				int optval=1;
-				if(setsockopt(udpsock, IPPROTO_IP, IP_ONESBCAST, &optval, sizeof(int))==-1)
-				{
-					Server->Log(std::string("Error setting IP_ONESBCAST for interface " )+std::string(ifap->ifa_name), LL_ERROR);
+
+					if (alreadyHasAddrv6(source_addr))
+						continue;
+
+					int type = SOCK_DGRAM;
+#if defined(SOCK_CLOEXEC)
+					type |= SOCK_CLOEXEC;
+#endif
+					SOCKET udpsock = socket(AF_INET6, type, 0);
+					if (udpsock == -1)
+					{
+						Server->Log(std::string("Error creating ipv6 socket for interface ") + std::string(ifap->ifa_name), LL_ERROR);
+						continue;
+					}
+#if !defined(SOCK_CLOEXEC)
+					fcntl(udpsock, F_SETFD, fcntl(udpsock, F_GETFD, 0) | FD_CLOEXEC);
+#endif
+
+					BOOL val = TRUE;
+					int rc = setsockopt(udpsock, SOL_SOCKET, SO_REUSEADDR, &val, sizeof(BOOL));
+					if (rc < 0)
+					{
+						Server->Log(std::string("Setting SO_REUSEADDR failed for interface ") + std::string(ifap->ifa_name), LL_ERROR);
+					}
+
+					int optval = 1;
+					setsockopt(udpsock, IPPROTO_IPV6, IPV6_V6ONLY, (char*)&optval, sizeof(optval));
+
+					Server->Log(std::string("Binding to interface ") + std::string(ifap->ifa_name) + " (ipv6) for broadcasting...", LL_DEBUG);
+
+					SSocket new_udpsock;
+
+					new_udpsock.is_ipv6 = true;
+					new_udpsock.addr_ipv6 = source_addr;
+					new_udpsock.udpsock = udpsock;
+
+					rc = bind(udpsock, (struct sockaddr *)&source_addr, sizeof(source_addr));
+					if (rc < 0)
+					{
+						Server->Log(std::string("Binding UDP socket failed for interface ") + std::string(ifap->ifa_name)+" errno="+convert(errno)+" interface ipv6 "+ipToString(new_udpsock), LL_ERROR);
+					}
+
+
+					udpsocks.push_back(new_udpsock);
+
+					Server->Log("Broadcasting on ipv6 interface "+ std::string(ifap->ifa_name) +" addr " + ipToString(new_udpsock));
 				}
-				#endif
-
-				broadcast_iface_addrs.push_back(source_addr.sin_addr.s_addr);
-				broadcast_addrs.push_back(*((struct sockaddr_in *)ifap->ifa_broadaddr));
-				udpsocks.push_back(udpsock);
-
-				Server->Log("Broadcasting on interface IP "+ ipToString(source_addr));
 			}
 		}
 		freeifaddrs(start_ifap);
 	}
 	else
 	{
-		retryBindToNewInterfaces=false;
+		retryBindToNewInterfaces = false;
 
-		Server->Log("Getting interface ips failed. errno="+convert(errno)+
+		Server->Log("Getting interface ips failed. errno=" + convert(errno) +
 			". Server may not listen properly on all network devices when discovering clients.", LL_ERROR);
 
 		int type = SOCK_DGRAM;
 #if defined(SOCK_CLOEXEC)
 		type |= SOCK_CLOEXEC;
 #endif
-		SOCKET udpsock=socket(AF_INET, type,0);
-		if(udpsock==-1)
+		SOCKET udpsock = socket(AF_INET, type, 0);
+		if (udpsock == -1)
 		{
 			Server->Log("Error creating socket", LL_ERROR);
 		}
@@ -238,9 +306,9 @@ void FileClient::bindToNewInterfaces()
 #if !defined(SOCK_CLOEXEC)
 			fcntl(udpsock, F_SETFD, fcntl(udpsock, F_GETFD, 0) | FD_CLOEXEC);
 #endif
-			BOOL val=TRUE;
+			BOOL val = TRUE;
 			int rc = setsockopt(udpsock, SOL_SOCKET, SO_REUSEADDR, &val, sizeof(BOOL));
-			if(rc<0)
+			if (rc < 0)
 			{
 				Server->Log("Setting SO_REUSEADDR failed", LL_ERROR);
 			}
@@ -253,26 +321,72 @@ void FileClient::bindToNewInterfaces()
 
 			Server->Log("Binding to no interface for broadcasting. Entering IP on restore CD won't work.", LL_WARNING);
 
-			rc = setsockopt(udpsock, SOL_SOCKET, SO_BROADCAST, (char*)&val, sizeof(BOOL) );
-			if(rc<0)
+			rc = setsockopt(udpsock, SOL_SOCKET, SO_BROADCAST, (char*)&val, sizeof(BOOL));
+			if (rc < 0)
 			{
 				Server->Log("Enabling SO_BROADCAST for UDP socket failed", LL_ERROR);
 				closesocket(udpsock);
 			}
 			else
 			{
-				udpsocks.push_back(udpsock);
-				source_addr.sin_addr.s_addr = INADDR_BROADCAST;
-				broadcast_addrs.push_back(source_addr);
-				broadcast_iface_addrs.push_back(source_addr.sin_addr.s_addr);
+				SSocket new_udpsock;
 
-				Server->Log("Broadcasting on interface IP "+ ipToString(source_addr));
+				source_addr.sin_addr.s_addr = INADDR_BROADCAST;
+
+				new_udpsock.is_ipv6 = false;
+				new_udpsock.addr_ipv4 = source_addr;
+				new_udpsock.udpsock = udpsock;
+				new_udpsock.broadcast_addr = source_addr;
+
+				udpsocks.push_back(new_udpsock);
+
+				Server->Log("Broadcasting on interface IPv4 " + ipToString(new_udpsock));
+			}
+		}
+
+		if (Server->getServerParameter("disable_ipv6").empty())
+		{
+			SOCKET udpsock = socket(AF_INET6, type, 0);
+			if (udpsock == -1)
+			{
+				Server->Log("Error creating ipv6 socket", LL_ERROR);
+			}
+			else
+			{
+#if !defined(SOCK_CLOEXEC)
+				fcntl(udpsock, F_SETFD, fcntl(udpsock, F_GETFD, 0) | FD_CLOEXEC);
+#endif
+				BOOL val = TRUE;
+				int rc = setsockopt(udpsock, SOL_SOCKET, SO_REUSEADDR, &val, sizeof(BOOL));
+				if (rc < 0)
+				{
+					Server->Log("Setting SO_REUSEADDR failed", LL_ERROR);
+				}
+
+				int optval = 1;
+				setsockopt(udpsock, IPPROTO_IPV6, IPV6_V6ONLY, (char*)&optval, sizeof(optval));
+
+				sockaddr_in6 source_addr;
+				memset(&source_addr, 0, sizeof(source_addr));
+				source_addr.sin6_addr = in6addr_any;
+				source_addr.sin6_family = AF_INET6;
+				source_addr.sin6_port = htons(broadcast_source_port);
+
+				Server->Log("Binding to no interface for broadcasting. Entering IP on restore CD won't work.", LL_WARNING);
+
+				SSocket new_udpsock;
+				new_udpsock.is_ipv6 = true;
+				new_udpsock.addr_ipv6 = source_addr;
+				new_udpsock.udpsock = udpsock;
+
+				udpsocks.push_back(new_udpsock);
+
+				Server->Log("Broadcasting on interface IPv6 " + ipToString(new_udpsock));
 			}
 		}
 	}
 #else
 	char hostname[MAX_PATH];
-    struct    hostent* h;
 
     _i32 rc=gethostname(hostname, MAX_PATH);
     if(rc==SOCKET_ERROR)
@@ -283,24 +397,35 @@ void FileClient::bindToNewInterfaces()
 	}
 	std::vector<_u32> addresses;
 
-    if(NULL != (h = gethostbyname(hostname)))
+	ADDRINFOA hints = {};
+	hints.ai_flags = AI_PASSIVE;
+	hints.ai_family = AF_INET;
+
+	PADDRINFOA h = NULL;
+	rc = getaddrinfo(hostname, NULL, &hints, &h);
+	PADDRINFOA orig_h = h;
+
+    if(rc==0 && h!=NULL)
     {
-		if(h->h_addrtype!=AF_INET)
+		if(h->ai_family!=AF_INET)
 		{
 			Server->Log("Hostname hostent is not AF_INET (ipv4)", LL_ERROR);
 		}
 		else
 		{
-			for(_u32 x = 0; (h->h_addr_list[x]); x++)
+			while(h!=NULL)
 			{
 				sockaddr_in source_addr;
 				memset(&source_addr, 0, sizeof(source_addr));
 				source_addr.sin_family = AF_INET;
-				source_addr.sin_addr.s_addr = *((_u32*)h->h_addr_list[x]);
+				source_addr.sin_addr = reinterpret_cast<sockaddr_in*>(h->ai_addr)->sin_addr;
 				source_addr.sin_port = htons(broadcast_source_port);
 
-				if(std::find(broadcast_iface_addrs.begin(), broadcast_iface_addrs.end(), source_addr.sin_addr.s_addr)!=broadcast_iface_addrs.end())
+				if (alreadyHasAddrv4(source_addr))
+				{
+					h = h->ai_next;
 					continue;
+				}
 
 				SOCKET udpsock=socket(AF_INET,SOCK_DGRAM,0);	
 
@@ -326,13 +451,93 @@ void FileClient::bindToNewInterfaces()
 					Server->Log("Failed setting SO_BROADCAST in FileClient", LL_ERROR);
 				}
 
-				udpsocks.push_back(udpsock);
-				broadcast_iface_addrs.push_back(source_addr.sin_addr.s_addr);
+				SSocket new_udpsock;
 
-				Server->Log("Broadcasting on interface IP "+ ipToString(source_addr));
+				new_udpsock.is_ipv6 = false;
+				new_udpsock.addr_ipv4 = source_addr;
+				new_udpsock.udpsock = udpsock;
+
+				udpsocks.push_back(new_udpsock);
+
+				Server->Log("Broadcasting on interface IPv4 "+ ipToString(new_udpsock));
+
+				h = h->ai_next;
 			}
 		}				
     }
+
+	if (orig_h != NULL)
+		freeaddrinfo(orig_h);
+
+
+	if (Server->getServerParameter("disable_ipv6").empty())
+	{
+		hints.ai_family = AF_INET6;
+		PADDRINFOA h = NULL;
+		rc = getaddrinfo(hostname, NULL, &hints, &h);
+		PADDRINFOA orig_h = h;
+
+		if (rc == 0 && h != NULL)
+		{
+			if (h->ai_family != AF_INET6)
+			{
+				Server->Log("Hostname hostent is not AF_INET6 (ipv6)", LL_ERROR);
+			}
+			else
+			{
+				while (h != NULL)
+				{
+					sockaddr_in6 source_addr;
+					memset(&source_addr, 0, sizeof(source_addr));
+					source_addr.sin6_family = AF_INET6;
+					source_addr.sin6_addr = reinterpret_cast<sockaddr_in6*>(h->ai_addr)->sin6_addr;
+					source_addr.sin6_port = htons(broadcast_source_port);
+
+					if (alreadyHasAddrv6(source_addr))
+					{
+						h = h->ai_next;
+						continue;
+					}
+
+					SOCKET udpsock = socket(AF_INET6, SOCK_DGRAM, 0);
+
+					int optval = 1;
+					int rc = setsockopt(udpsock, SOL_SOCKET, SO_REUSEADDR, (char*)&optval, sizeof(int));
+					if (rc == SOCKET_ERROR)
+					{
+						Server->Log("Failed setting SO_REUSEADDR in FileClient (2)", LL_ERROR);
+					}
+
+					optval = 1;
+					setsockopt(udpsock, IPPROTO_IPV6, IPV6_V6ONLY, (char*)&optval, sizeof(optval));
+
+					rc = bind(udpsock, (struct sockaddr *)&source_addr, sizeof(source_addr));
+					if (rc<0)
+					{
+						Server->Log("Binding UDP socket failed (2)", LL_ERROR);
+					}
+
+					setSockP(udpsock);
+
+					SSocket new_udpsock;
+
+					new_udpsock.is_ipv6 = true;
+					new_udpsock.addr_ipv6 = source_addr;
+					new_udpsock.udpsock = udpsock;
+
+					udpsocks.push_back(new_udpsock);
+
+					Server->Log("Broadcasting on interface IPv6 " + ipToString(new_udpsock));
+
+					h = h->ai_next;
+				}
+			}
+		}
+
+		if (orig_h != NULL)
+			freeaddrinfo(orig_h);
+	}
+
 #endif
 }
 
@@ -344,7 +549,7 @@ FileClient::~FileClient(void)
 	}
 	for(size_t i=0;i<udpsocks.size();++i)
 	{
-		closesocket(udpsocks[i]);
+		closesocket(udpsocks[i].udpsock);
 	}
 
 	Server->destroy(mutex);
@@ -357,7 +562,7 @@ FileClient::~FileClient(void)
 	}
 }
 
-std::vector<sockaddr_in> FileClient::getServers(void)
+std::vector<FileClient::SAddrHint> FileClient::getServers(void)
 {
         return servers;
 }
@@ -367,7 +572,7 @@ std::vector<std::string> FileClient::getServerNames(void)
         return servernames;
 }
 
-std::vector<sockaddr_in> FileClient::getWrongVersionServers(void)
+std::vector<FileClient::SAddrHint> FileClient::getWrongVersionServers(void)
 {
         return wvservers;
 }
@@ -377,9 +582,9 @@ _u32 FileClient::getLocalIP(void)
         return local_ip;
 }
 
-_u32 FileClient::GetServers(bool start, const std::vector<in_addr> &addr_hints)
+_u32 FileClient::GetServers(bool start, const std::vector<SAddrHint> &addr_hints)
 {
-        if(start==true)
+        if(start)
         {
 			if(retryBindToNewInterfaces)
 			{
@@ -389,21 +594,41 @@ _u32 FileClient::GetServers(bool start, const std::vector<in_addr> &addr_hints)
 			max_version=0;
 			for(size_t i=0;i<udpsocks.size();++i)
 			{
-				sockaddr_in addr_udp;
-				addr_udp.sin_family=AF_INET;
-				addr_udp.sin_port=htons(UDP_PORT);
-#ifdef __FreeBSD__
-				addr_udp.sin_addr.s_addr=broadcast_addrs[i].sin_addr.s_addr;
-#else
-				addr_udp.sin_addr.s_addr=INADDR_BROADCAST;
-#endif
-				memset(addr_udp.sin_zero,0, sizeof(addr_udp.sin_zero));
-
-				char ch=ID_PING;
-				int rc=sendto(udpsocks[i], &ch, 1, 0, (sockaddr*)&addr_udp, sizeof(sockaddr_in));
-				if(rc==-1)
+				if (!udpsocks[i].is_ipv6)
 				{
-					Server->Log("Sending broadcast failed!", LL_ERROR);
+					sockaddr_in addr_udp;
+					addr_udp.sin_family = AF_INET;
+					addr_udp.sin_port = htons(UDP_PORT);
+#ifdef __FreeBSD__
+					addr_udp.sin_addr.s_addr = udpsocks[i].broadcast_addr.sin_addr.s_addr;
+#else
+					addr_udp.sin_addr.s_addr = INADDR_BROADCAST;
+#endif
+					memset(addr_udp.sin_zero, 0, sizeof(addr_udp.sin_zero));
+
+					char ch = ID_PING;
+					int rc = sendto(udpsocks[i].udpsock, &ch, 1, 0, (sockaddr*)&addr_udp, sizeof(addr_udp));
+					if (rc == -1)
+					{
+						Server->Log("Sending broadcast failed!", LL_ERROR);
+					}
+				}
+				else
+				{
+					sockaddr_in6 addr_udp = {};
+					addr_udp.sin6_family = AF_INET6;
+					addr_udp.sin6_port = htons(UDP_PORT);
+					if (inet_pton(AF_INET6, multicast_group, &addr_udp.sin6_addr) != 1)
+					{
+						Server->Log("inet_pton failed", LL_ERROR);
+					}
+
+					char ch = ID_PING;
+					int rc = sendto(udpsocks[i].udpsock, &ch, 1, 0, (sockaddr*)&addr_udp, sizeof(addr_udp));
+					if (rc == -1)
+					{
+						Server->Log("Sending broadcast failed! (ipv6)", LL_ERROR);
+					}
 				}
 			}
 
@@ -418,46 +643,65 @@ _u32 FileClient::GetServers(bool start, const std::vector<in_addr> &addr_hints)
 #define SETSOCK_CAST
 #endif
 
-					if(setsockopt(udpsocks[i], SOL_SOCKET, SO_BROADCAST, SETSOCK_CAST &broadcast, sizeof(int))==-1)
+					if(!udpsocks[i].is_ipv6)
 					{
-						Server->Log("Error setting socket to not broadcast", LL_ERROR);
-					}
+						if(setsockopt(udpsocks[i].udpsock, SOL_SOCKET, SO_BROADCAST, SETSOCK_CAST &broadcast, sizeof(int))==-1)
+						{
+							Server->Log("Error setting socket to not broadcast", LL_ERROR);
+						}
+
+						#if defined(__FreeBSD__)
+						int optval=0;
+						if(setsockopt(udpsocks[i].udpsock, IPPROTO_IP, IP_ONESBCAST, &optval, sizeof(int))==-1)
+						{
+							Server->Log(std::string("Error setting IP_ONESBCAST" ), LL_ERROR);
+						}
+						#endif
 
-					#if defined(__FreeBSD__)
-					int optval=0;
-					if(setsockopt(udpsocks[i], IPPROTO_IP, IP_ONESBCAST, &optval, sizeof(int))==-1)
-					{
-						Server->Log(std::string("Error setting IP_ONESBCAST" ), LL_ERROR);
 					}
-					#endif
 
 					for(size_t j=0;j<addr_hints.size();++j)
 					{
 						char ch=ID_PING;
-						sockaddr_in addr_udp;
-						addr_udp.sin_family=AF_INET;
-						addr_udp.sin_port=htons(UDP_PORT);
-						addr_udp.sin_addr.s_addr=addr_hints[j].s_addr;
-						memset(addr_udp.sin_zero,0, sizeof(addr_udp.sin_zero));
 
-						sendto(udpsocks[i], &ch, 1, 0, (sockaddr*)&addr_udp, sizeof(sockaddr_in) );
+						if (!addr_hints[j].is_ipv6)
+						{
+							sockaddr_in addr_udp = {};
+							addr_udp.sin_family = AF_INET;
+							addr_udp.sin_port = htons(UDP_PORT);
+							addr_udp.sin_addr.s_addr = addr_hints[j].addr_ipv4;
+
+							sendto(udpsocks[i].udpsock, &ch, 1, 0, (sockaddr*)&addr_udp, sizeof(sockaddr_in));
+						}
+						else
+						{
+							sockaddr_in6 addr_udp = {};
+							addr_udp.sin6_family = AF_INET6;
+							addr_udp.sin6_port = htons(UDP_PORT);
+							memcpy(&addr_udp.sin6_addr, addr_hints[j].addr_ipv6, sizeof(addr_hints[j].addr_ipv6));
+
+							sendto(udpsocks[i].udpsock, &ch, 1, 0, (sockaddr*)&addr_udp, sizeof(sockaddr_in));
+						}
 					}
 
-					broadcast=1;
-					if(setsockopt(udpsocks[i], SOL_SOCKET, SO_BROADCAST, SETSOCK_CAST &broadcast, sizeof(int))==-1)
+					if (!udpsocks[i].is_ipv6)
 					{
-						Server->Log("Error setting socket to broadcast", LL_ERROR);
-					}
+						broadcast = 1;
+						if (setsockopt(udpsocks[i].udpsock, SOL_SOCKET, SO_BROADCAST, SETSOCK_CAST &broadcast, sizeof(int)) == -1)
+						{
+							Server->Log("Error setting socket to broadcast", LL_ERROR);
+						}
 
 #undef SETSOCK_CAST
 
-					#if defined(__FreeBSD__)
-					optval=1;
-					if(setsockopt(udpsocks[i], IPPROTO_IP, IP_ONESBCAST, &optval, sizeof(int))==-1)
-					{
-						Server->Log(std::string("Error setting IP_ONESBCAST" ), LL_ERROR);
+#if defined(__FreeBSD__)
+						int optval = 1;
+						if (setsockopt(udpsocks[i].udpsock, IPPROTO_IP, IP_ONESBCAST, &optval, sizeof(int)) == -1)
+						{
+							Server->Log(std::string("Error setting IP_ONESBCAST"), LL_ERROR);
+						}
+#endif
 					}
-					#endif
 				}
 			}
 
@@ -480,13 +724,13 @@ _u32 FileClient::GetServers(bool start, const std::vector<in_addr> &addr_hints)
 
 			if(!udpsocks.empty())
 			{
-				max_socket=udpsocks[0];
+				max_socket=udpsocks[0].udpsock;
 			}
 
 			for(size_t i=0;i<udpsocks.size();++i)
 			{
-				FD_SET(udpsocks[i], &fdset);
-				max_socket=(std::max)(max_socket, udpsocks[i]);
+				FD_SET(udpsocks[i].udpsock, &fdset);
+				max_socket=(std::max)(max_socket, udpsocks[i].udpsock);
 			}
 
 			timeval lon;
@@ -498,7 +742,7 @@ _u32 FileClient::GetServers(bool start, const std::vector<in_addr> &addr_hints)
 			conn.resize(udpsocks.size());
 			for(size_t i=0;i<udpsocks.size();++i)
 			{
-				conn[i].fd=udpsocks[i];
+				conn[i].fd=udpsocks[i].udpsock;
 				conn[i].events=POLLIN;
 				conn[i].revents=0;
 			}
@@ -510,14 +754,26 @@ _u32 FileClient::GetServers(bool start, const std::vector<in_addr> &addr_hints)
 				for(size_t i=0;i<udpsocks.size();++i)
 				{
 #ifdef _WIN32
-					if(FD_ISSET(udpsocks[i], &fdset))
+					if(FD_ISSET(udpsocks[i].udpsock, &fdset))
 #else
 					if(conn[i].revents!=0)
 #endif
 					{
-						socklen_t addrsize=sizeof(sockaddr_in);
-						sockaddr_in sender;
-						_i32 err = recvfrom(udpsocks[i], buffer, BUFFERSIZE_UDP, 0, (sockaddr*)&sender, &addrsize);
+						socklen_t addrsize;
+						sockaddr* sender;
+						sockaddr_in6 sender_v6;
+						sockaddr_in sender_v4;
+						if (udpsocks[i].is_ipv6)
+						{
+							sender = (sockaddr*)&sender_v6;
+							addrsize = sizeof(sender_v6);
+						}
+						else
+						{
+							sender = (sockaddr*)&sender_v4;
+							addrsize = sizeof(sender_v4);
+						}
+						_i32 err = recvfrom(udpsocks[i].udpsock, buffer, BUFFERSIZE_UDP, 0, sender, &addrsize);
 						if(err==SOCKET_ERROR)
 						{
 							continue;
@@ -525,7 +781,7 @@ _u32 FileClient::GetServers(bool start, const std::vector<in_addr> &addr_hints)
 						if(err>2&&buffer[0]==ID_PONG)
 						{
 							int version=(unsigned char)buffer[1];
-							if(version==VERSION)
+							if(version==FILECLIENT_VERSION)
 							{
 								std::string sn;
 
@@ -534,13 +790,41 @@ _u32 FileClient::GetServers(bool start, const std::vector<in_addr> &addr_hints)
 
 								if (!sn.empty())
 								{
-									servers.push_back(sender);
+									if (udpsocks[i].is_ipv6)
+									{
+										SAddrHint res;
+										res.is_ipv6 = true;
+										memcpy(res.addr_ipv6, &sender_v6.sin6_addr, sizeof(sender_v6.sin6_addr));
+										res.zone = sender_v6.sin6_scope_id;
+										servers.push_back(res);
+									}
+									else
+									{
+										SAddrHint res;
+										res.is_ipv6 = false;
+										res.addr_ipv4 = sender_v4.sin_addr.s_addr;
+										servers.push_back(res);
+									}									
 									servernames.push_back(sn);
 								}
 							}
 							else
 							{
-								wvservers.push_back(sender);
+								if (udpsocks[i].is_ipv6)
+								{
+									SAddrHint res;
+									res.is_ipv6 = true;
+									memcpy(res.addr_ipv6, &sender_v6.sin6_addr, sizeof(sender_v6.sin6_addr));
+									res.zone = sender_v6.sin6_scope_id;
+									wvservers.push_back(res);
+								}
+								else
+								{
+									SAddrHint res;
+									res.is_ipv6 = false;
+									res.addr_ipv4 = sender_v4.sin_addr.s_addr;
+									wvservers.push_back(res);
+								}
 							}
                                 
 							if( version>max_version )
@@ -570,14 +854,14 @@ int FileClient::getMaxVersion(void)
         return max_version;        
 }
 
-_u32 FileClient::Connect(sockaddr_in *addr)
+_u32 FileClient::Connect(const SAddrHint& addr)
 {
 	if( socket_open==true )
     {
             Server->destroy(tcpsock);
     }
 
-	tcpsock=Server->ConnectStream(inet_ntoa(addr->sin_addr), TCP_PORT, 10000);
+	tcpsock=Server->ConnectStream(addr.toString(), TCP_PORT, 10000);
 
 	if(tcpsock!=NULL)
 	{
@@ -589,7 +873,7 @@ _u32 FileClient::Connect(sockaddr_in *addr)
 		}
 	}
 
-	server_addr=*addr;
+	server_addr=addr;
 
     if(tcpsock==NULL)
 		return ERR_ERROR;
@@ -671,7 +955,7 @@ bool FileClient::Reconnect(void)
 		IPipe* new_tcpsock;
 		if(reconnection_callback==NULL)
 		{
-			new_tcpsock=Server->ConnectStream(inet_ntoa(server_addr.sin_addr), TCP_PORT, 10000);
+			new_tcpsock=Server->ConnectStream(server_addr.toString(), TCP_PORT, 10000);
 		}
 		else
 		{
@@ -708,7 +992,11 @@ bool FileClient::Reconnect(void)
 
 	resetSparseExtentsFile();
 
-	setReconnectTries(50);
+	{
+		IScopedLock lock(mutex);
+		if (reconnect_tries < 50)
+			reconnect_tries = 50;
+	}
 
 	if(!hashed && protocol_version>1)
 	{
@@ -774,6 +1062,7 @@ bool FileClient::Reconnect(void)
 	last_progress_log=0;
 
     starttime=Server->getTimeMS();
+	int64 last_timeouttime = 0;
 
 
 	enum EReceiveState
@@ -792,11 +1081,13 @@ bool FileClient::Reconnect(void)
 
 	char* buf = dl_buf;
 
+	_u32 queue_err = ERR_SUCCESS;
+
 	while(true)
 	{        
 		if(!is_script)
 		{
-			fillQueue();
+			queue_err = fillQueue();
 		}
 		else
 		{
@@ -808,7 +1099,12 @@ bool FileClient::Reconnect(void)
 		}
 
 		size_t rc;
-		if(tcpsock->isReadable() || dl_off==0 ||
+		if (queue_err != ERR_SUCCESS)
+		{
+			Server->Log("Queue err " + convert(queue_err), LL_DEBUG);
+			rc = 0;
+		}
+		else if(tcpsock->isReadable() || dl_off==0 ||
 			(firstpacket 
 				&& ( (dl_buf[0]==ID_FILESIZE && dl_off<1+sizeof(_u64))
 					|| (dl_buf[0] == ID_FILESIZE_AND_EXTENTS && dl_off<1 + 2*sizeof(_u64)) )
@@ -830,12 +1126,27 @@ bool FileClient::Reconnect(void)
         if( rc==0 )
         {
 			Server->Log("Server timeout (2) in FileClient", LL_DEBUG);
+
+			if (last_timeouttime == 0
+				|| Server->getTimeMS() - last_timeouttime>30*60*1000)
+			{
+				IScopedLock lock(mutex);
+				if (reconnect_tries < 50)
+					reconnect_tries = 50;
+			}
+
+			last_timeouttime = Server->getTimeMS();
+
 			bool b = false;
 			int tries = getReconnectTriesDecr();
 			if (tries > 0)
 			{
 				b = Reconnect();
 			}
+			else
+			{
+				Server->Log("FileClient: Reconnect tries exhausted (1)", LL_INFO);
+			}
 			if(!b )
 			{
 				Server->Log("FileClient: ERR_TIMEOUT", LL_INFO);
@@ -885,12 +1196,26 @@ bool FileClient::Reconnect(void)
 				{
 					Server->Log("Timeout during file request (4). Reconnecting...", LL_DEBUG);
 
+					if (last_timeouttime == 0
+						|| Server->getTimeMS() - last_timeouttime>30 * 60 * 1000)
+					{
+						IScopedLock lock(mutex);
+						if (reconnect_tries < 50)
+							reconnect_tries = 50;
+					}
+
+					last_timeouttime = Server->getTimeMS();
+
 					bool b = false;
 					int tries = getReconnectTriesDecr();
 					if (tries > 0)
 					{
 						b = Reconnect();
 					}
+					else
+					{
+						Server->Log("FileClient: Reconnect tries exhausted (2)", LL_INFO);
+					}
 					if (!b)
 					{
 						Server->Log("Timeout during file request (4)", LL_ERROR);
@@ -1047,6 +1372,7 @@ bool FileClient::Reconnect(void)
 						memmove(dl_buf, dl_buf+1, rc-1);
 						dl_off = rc-1;
 					}
+					Server->Log("Unknown packet id " + convert(PID), LL_ERROR);
 					return ERR_ERROR;
 				}
             }
@@ -1098,6 +1424,9 @@ bool FileClient::Reconnect(void)
 						return ERR_ERROR;
 					}
 
+					if(received>0)
+						file->Seek(received);
+
 					{
 						IScopedLock lock(mutex);
 						sparse_bytes += sparse_file->getSparseSize();
@@ -1269,12 +1598,19 @@ bool FileClient::Reconnect(void)
 	    if( Server->getTimeMS()-starttime > SERVER_TIMEOUT )
 		{
 			Server->Log("Server timeout in FileClient. Trying to reconnect...", LL_INFO);
+
+			last_timeouttime = Server->getTimeMS();
+
 			bool b = false;
 			int tries = getReconnectTriesDecr();
 			if (tries > 0)
 			{
 				b = Reconnect();
 			}
+			else
+			{
+				Server->Log("FileClient: Reconnect tries exhausted (3)", LL_INFO);
+			}
 			if(!b)
 			{
 				Server->Log("FileClient: ERR_TIMEOUT", LL_INFO);
@@ -1421,17 +1757,17 @@ void FileClient::setQueueCallback( QueueCallback* cb )
 	queue_callback = cb;
 }
 
-void FileClient::fillQueue()
+_u32 FileClient::fillQueue()
 {
 	if(queue_callback==NULL)
 	{
 		if(needs_flush)
 		{
 			needs_flush=false;
-			Flush();
+			return Flush();
 		}
 
-		return;
+		return ERR_SUCCESS;
 	}
 
 	if(queued.size()>queuedFilesLow)
@@ -1439,10 +1775,10 @@ void FileClient::fillQueue()
 		if (needs_flush)
 		{
 			needs_flush = false;
-			Flush();
+			return Flush();
 		}
 
-		return;
+		return ERR_SUCCESS;
 	}
 
 	bool needs_send_flush=false;
@@ -1458,10 +1794,10 @@ void FileClient::fillQueue()
 			if (needs_flush)
 			{
 				needs_flush = false;
-				Flush();
+				return Flush();
 			}
 
-			return;
+			return ERR_SUCCESS;
 		}
 
 		MetadataQueue metadata_queue = MetadataQueue_Data;
@@ -1476,10 +1812,10 @@ void FileClient::fillQueue()
 			{
 				needs_flush=false;
 				needs_send_flush=false;
-				Flush();
+				return Flush();
 			}
 
-			return;
+			return ERR_SUCCESS;
 		}
 
 		CWData data;
@@ -1523,15 +1859,15 @@ void FileClient::fillQueue()
 		if(stack.Send( tcpsock, data.getDataPtr(), data.getDataSize(), c_default_timeout, false)!=data.getDataSize())
 		{
 			Server->Log("Queueing file failed", LL_DEBUG);
-			queue_callback->unqueueFileFull(queue_fn, finish_script);
 
-			if (needs_flush)
+			for (size_t i = 0; i<queued_files.size(); ++i)
 			{
-				needs_flush = false;
-				Flush();
+				queue_callback->unqueueFileFull(queued_files[i].fn, queued_files[i].finish_script);
 			}
 
-			return;
+			queue_callback->unqueueFileFull(queue_fn, finish_script);
+
+			return ERR_TIMEOUT;
 		}
 
 		queued.push_back(SQueueItem(queue_fn, finish_script));
@@ -1542,7 +1878,7 @@ void FileClient::fillQueue()
 	if (needs_flush)
 	{
 		needs_flush = false;
-		Flush();
+		return Flush();
 	}
 	else if(needs_send_flush)
 	{
@@ -1553,8 +1889,11 @@ void FileClient::fillQueue()
 			{
 				queue_callback->unqueueFileFull(queued_files[i].fn, queued_files[i].finish_script);
 			}
+			return ERR_TIMEOUT;
 		}
 	}
+
+	return ERR_SUCCESS;
 }
 
 void FileClient::logProgress(const std::string& remotefn, _u64 filesize, _u64 received)
@@ -1582,6 +1921,44 @@ void FileClient::logProgress(const std::string& remotefn, _u64 filesize, _u64 re
 	}	
 }
 
+bool FileClient::alreadyHasAddrv4(sockaddr_in addr)
+{
+	for (size_t i = 0; i < udpsocks.size(); ++i)
+	{
+		if (!udpsocks[i].is_ipv6
+			&& addr.sin_addr.s_addr == udpsocks[i].addr_ipv4.sin_addr.s_addr)
+			return true;
+	}
+	return false;
+}
+
+bool FileClient::alreadyHasAddrv6(sockaddr_in6 addr)
+{
+	for (size_t i = 0; i < udpsocks.size(); ++i)
+	{
+		if (udpsocks[i].is_ipv6
+			&& memcmp(&addr.sin6_addr, &udpsocks[i].addr_ipv6.sin6_addr, sizeof(addr.sin6_addr))==0)
+			return true;
+	}
+	return false;
+}
+
+std::string FileClient::ipToString(SSocket& s)
+{
+	if (!s.is_ipv6)
+	{
+		char str[INET_ADDRSTRLEN];
+		inet_ntop(AF_INET, &(s.addr_ipv4.sin_addr), str, INET_ADDRSTRLEN);
+		return str;
+	}
+	else
+	{
+		char str[INET6_ADDRSTRLEN];
+		inet_ntop(AF_INET6, &(s.addr_ipv6.sin6_addr), str, INET6_ADDRSTRLEN);
+		return str;
+	}
+}
+
 IFile * FileClient::temporaryFileRetry()
 {
 	while (true)
@@ -1643,7 +2020,7 @@ void FileClient::setNoFreeSpaceCallback(FileClient::NoFreeSpaceCallback * cb)
 	nofreespace_callback = cb;
 }
 
-_u32 FileClient::GetFileHashAndMetadata( std::string remotefn, std::string& hash, std::string& permissions, int64& filesize, int64& created, int64& modified )
+/*_u32 FileClient::GetFileHashAndMetadata( std::string remotefn, std::string& hash, std::string& permissions, int64& filesize, int64& created, int64& modified )
 {
 	if (tcpsock == NULL)
 		return ERR_ERROR;
@@ -1874,7 +2251,7 @@ _u32 FileClient::GetFileHashAndMetadata( std::string remotefn, std::string& hash
 			}
 		}
 	}
-}
+}*/
 
 _u32 FileClient::InformMetadataStreamEnd( const std::string& server_token, int tries)
 {
@@ -1941,6 +2318,72 @@ _u32 FileClient::InformMetadataStreamEnd( const std::string& server_token, int t
 	}
 }
 
+_u32 FileClient::StopPhashLoad(const std::string & server_token, const std::string & phash_fn, int tries)
+{
+	if (tcpsock == NULL)
+		return ERR_ERROR;
+
+	setReconnectTries(tries);
+
+	CWData data;
+	data.addUChar(ID_STOP_PHASH);
+	data.addString(identity);
+	data.addString(server_token);
+	data.addString(phash_fn);
+
+	if (stack.Send(tcpsock, data.getDataPtr(), data.getDataSize()) != data.getDataSize())
+	{
+		Server->Log("Timeout during sending phash stream end (1)", LL_ERROR);
+		return ERR_TIMEOUT;
+	}
+
+	while (true)
+	{
+		size_t rc = tcpsock->Read(dl_buf, 1, 120000);
+
+		if (rc == 0)
+		{
+			Server->Log("Server timeout (2) in FileClient sending phash stream end", LL_DEBUG);
+			int tries = getReconnectTriesDecr();
+			bool b = false;
+			if (tries > 0)
+			{
+				b = Reconnect();
+			}
+			if (!b)
+			{
+				Server->Log("FileClient: ERR_TIMEOUT (phash stream)", LL_INFO);
+				return ERR_TIMEOUT;
+			}
+			else
+			{
+				CWData data;
+				data.addUChar(ID_INFORM_METADATA_STREAM_END);
+				data.addString(identity);
+				data.addString(server_token);
+
+				rc = stack.Send(tcpsock, data.getDataPtr(), data.getDataSize());
+				if (rc == 0)
+				{
+					Server->Log("FileClient: Error sending phash stream end", LL_INFO);
+				}
+				starttime = Server->getTimeMS();
+			}
+		}
+		else
+		{
+			if (*dl_buf == ID_PONG)
+			{
+				return ERR_SUCCESS;
+			}
+			else
+			{
+				return ERR_ERROR;
+			}
+		}
+	}
+}
+
 _u32 FileClient::FinishScript(std::string remotefn)
 {
 	if (tcpsock == NULL)
@@ -2077,3 +2520,27 @@ void FileClient::resetSparseExtentsFile()
 	}
 }
 
+std::string FileClient::SAddrHint::toString() const
+{
+	if (!is_ipv6)
+	{
+		char str[INET_ADDRSTRLEN];
+		inet_ntop(AF_INET, (void*)&addr_ipv4, str, INET_ADDRSTRLEN);
+		return str;
+	}
+	else
+	{
+		char str[INET6_ADDRSTRLEN];
+		sockaddr_in6 saddr = {};
+		saddr.sin6_family = AF_INET6;
+		saddr.sin6_port = htons(UDP_SOURCE_PORT);
+		memcpy(&saddr.sin6_addr, addr_ipv6, sizeof(addr_ipv6));
+		saddr.sin6_scope_id = zone;
+		if (getnameinfo(reinterpret_cast<sockaddr*>(&saddr), sizeof(saddr), 
+			str, sizeof(str), NULL, 0, NI_NUMERICHOST) != 0)
+		{
+			return std::string();
+		}
+		return str;
+	}
+}
