@@ -1,5 +1,6 @@
 enum InternetPipeCapabilities
 {
 	IPC_ENCRYPTED=1,
-	IPC_COMPRESSED=2
+	IPC_COMPRESSED=2,
+	IPC_COMPRESSED_ZSTD = 4,
 };
\ No newline at end of file
