@@ -42,11 +42,14 @@
 #include <errno.h>
 #include <time.h>
 #include <sys/time.h>
+#include <limits.h>
 #include "../config.h"
 #ifdef __linux__
 #include <sys/resource.h>
 #include <sys/syscall.h>
+#include <linux/fs.h>
 #endif
+#include <stack>
 
 #if defined(__FreeBSD__) || defined(__APPLE__)
 #define lstat64 lstat
@@ -55,6 +58,16 @@
 #define open64 open
 #define readdir64 readdir
 #define dirent64 dirent
+#define fsblkcnt64_t fsblkcnt_t
+#endif
+
+#if defined(__APPLE__)
+#include <pthread.h>
+#endif
+
+#if defined(__ANDROID__)
+#define fsblkcnt64_t fsblkcnt_t
+#include "android_popen.h"
 #endif
 
 
@@ -114,6 +127,30 @@ std::vector<SFile> getFiles(const std::string &path, bool *has_error, bool ignor
 		if(f.name=="." || f.name==".." )
 			continue;		
 
+#ifdef __APPLE__
+//        Cannot stat certain locations due to macOS SIP restrictions
+        if(upath+dirp->d_name=="/Library/Caches/com.apple.aned" ||
+            upath+dirp->d_name=="/private/var/db/appinstalld" ||
+            upath+dirp->d_name=="/private/var/db/ConfigurationProfiles/Store" ||
+            upath+dirp->d_name=="/private/var/db/CoreDuet/Knowledge" ||
+            upath+dirp->d_name=="/private/var/db/DifferentialPrivacy" ||
+            upath+dirp->d_name=="/private/var/db/DumpPanic" ||
+            upath+dirp->d_name=="/private/var/db/fpsd/dvp" ||
+            upath+dirp->d_name=="/private/var/db/KernelExtensionManagement/Staging" ||
+            upath+dirp->d_name=="/private/var/db/lockdown" ||
+            upath+dirp->d_name=="/private/var/db/MobileIdentityService" ||
+            upath+dirp->d_name=="/private/var/db/oah" ||
+            upath+dirp->d_name=="/private/var/db/searchparty" ||
+            upath+dirp->d_name=="/private/var/db/sysdiagnose/com.apple.sysdiagnose" ||
+            upath+dirp->d_name=="/private/var/networkd/db" ||
+            upath+dirp->d_name=="/private/var/protected/trustd/private" ||
+            upath+dirp->d_name=="/System/Library/Templates/Data/private/var/db/oah")
+        {
+            Log("[macOS] Skipping \""+upath+dirp->d_name+"\" due to macOS SIP restriction", LL_INFO);
+            continue;
+        }
+#endif
+    
 		f.isdir=(dirp->d_type==DT_DIR);
 		
 		struct stat64 f_info;
@@ -170,6 +207,7 @@ std::vector<SFile> getFiles(const std::string &path, bool *has_error, bool ignor
 			{
 				*has_error=true;
 			}
+			errno=0;
 			continue;
 		}
 		tmp.push_back(f);
@@ -260,6 +298,12 @@ int os_get_file_type(const std::string &path)
 			ret |= EFileType_Symlink;
 		}
 		
+		if(!S_ISDIR(f_info.st_mode)
+			&& !S_ISREG(f_info.st_mode) )
+		{
+			ret |= EFileType_Special;
+		}
+		
 		if(rc1!=0)
 		{
 			ret |= EFileType_File;
@@ -350,8 +394,17 @@ int64 os_free_space(const std::string &path)
     int rc=statvfs64((path).c_str(), &buf);
 	if(rc==0)
 	{
-		int64 blocksize = buf.f_frsize ? buf.f_frsize : buf.f_bsize;
-		return blocksize*buf.f_bavail;
+#if defined(__FreeBSD__) || defined(__APPLE__)
+		int64 free = (int64)buf.f_frsize*buf.f_bavail;
+#else
+		fsblkcnt64_t blocksize = buf.f_frsize ? buf.f_frsize : buf.f_bsize;
+		fsblkcnt64_t free = blocksize*buf.f_bavail;
+#endif
+		if(free>LLONG_MAX)
+		{
+			return LLONG_MAX;
+		}
+		return free;
 	}
 	else
 	{
@@ -373,8 +426,17 @@ int64 os_total_space(const std::string &path)
     int rc=statvfs64((path).c_str(), &buf);
 	if(rc==0)
 	{
-		fsblkcnt_t used=buf.f_blocks-buf.f_bfree;
-		return buf.f_bsize*(used+buf.f_bavail);
+		fsblkcnt64_t used=buf.f_blocks-buf.f_bfree;
+#if defined(__FreeBSD__) || defined(__APPLE__)
+		int64 total = (int64)(used+buf.f_bavail)*buf.f_frsize;
+#else
+		fsblkcnt64_t total = (used+buf.f_bavail)*buf.f_bsize;
+#endif
+		if(total>LLONG_MAX)
+		{
+			return LLONG_MAX;
+		}
+		return total;
 	}
 	else
 		return -1;
@@ -389,134 +451,147 @@ bool os_directory_exists(const std::string &path)
 	return isDirectory(path);
 }
 
-bool os_remove_nonempty_dir(const std::string &path, os_symlink_callback_t symlink_callback, void* userdata, bool delete_root)
+//TODO: fix returning error and check users
+bool os_remove_nonempty_dir(const std::string &root_path, os_symlink_callback_t symlink_callback, void* userdata, bool delete_root)
 {
-    std::string upath=(path);
-
 	if(delete_root)
 	{
 		struct stat64 f_info;
-		int rc = lstat64(upath.c_str(), &f_info);
+		int rc = lstat64(root_path.c_str(), &f_info);
 		if(rc==0 && S_ISLNK(f_info.st_mode))
 		{
-			if(unlink(upath.c_str())!=0)
+			if(unlink(root_path.c_str())!=0)
 			{
-				Log("Error deleting symlink \""+upath+"\"", LL_ERROR);
+				Log("Error deleting symlink \""+root_path+"\" (root)", LL_ERROR);
 			}
 			return true;
 		}
 	}
 
-	std::vector<SFile> tmp;
-	DIR *dp;
-    struct dirent *dirp;
-    if((dp  = opendir(upath.c_str())) == NULL)
-	{
-        Log("No permission to access \""+upath+"\"", LL_ERROR);
-        return false;
-    }
+	std::stack<std::pair<std::string, bool> > paths;
 
-	
+	paths.push(std::make_pair(root_path, false));
 	bool ok=true;
-	std::vector<std::string> subdirs;
-	while ((dirp = readdir(dp)) != NULL)
+
+	while(!paths.empty())
 	{
-		if( (std::string)dirp->d_name!="." && (std::string)dirp->d_name!=".." )
+		std::string upath=paths.top().first;
+		std::string path = upath;
+
+		if(paths.top().second)
 		{
-#ifndef sun
-			if(dirp->d_type==DT_UNKNOWN)
+			if(paths.size()>1
+				|| delete_root)
 			{
-#endif
-				struct stat64 f_info;
-				int rc=lstat64((upath+"/"+(std::string)dirp->d_name).c_str(), &f_info);
-				if(rc==0)
+				if(rmdir(upath.c_str())!=0)
 				{
-					if(S_ISLNK(f_info.st_mode))
+						Log("Error deleting directory \""+upath+"\"", LL_ERROR);
+				}
+			}
+			paths.pop();
+			continue;
+		}
+		
+		paths.top().second=true;
+
+		DIR *dp;
+		struct dirent *dirp;
+		if((dp  = opendir(upath.c_str())) == NULL)
+		{
+			Log("No permission to access \""+upath+"\"", LL_ERROR);
+			paths.pop();
+			ok=false;
+			continue;
+		}
+	
+		bool ok=true;
+		while ((dirp = readdir(dp)) != NULL)
+		{
+			if( (std::string)dirp->d_name!="." && (std::string)dirp->d_name!=".." )
+			{
+	#ifndef sun
+				if(dirp->d_type==DT_UNKNOWN)
+				{
+	#endif
+					struct stat64 f_info;
+					int rc=lstat64((upath+"/"+(std::string)dirp->d_name).c_str(), &f_info);
+					if(rc==0)
 					{
-						if(symlink_callback!=NULL)
+						if(S_ISLNK(f_info.st_mode))
 						{
-                            symlink_callback((upath+"/"+(std::string)dirp->d_name), NULL, userdata);
+							if(symlink_callback!=NULL)
+							{
+								symlink_callback((upath+"/"+(std::string)dirp->d_name), NULL, userdata);
+							}
+							else
+							{
+								if(unlink((upath+"/"+(std::string)dirp->d_name).c_str())!=0)
+								{
+									Log("Error deleting symlink \""+upath+"/"+(std::string)dirp->d_name+"\"", LL_ERROR);
+								}
+							}
+						}
+						else if(S_ISDIR(f_info.st_mode) )
+						{
+							paths.push(std::make_pair(upath+"/"+std::string(dirp->d_name), false));
 						}
 						else
 						{
 							if(unlink((upath+"/"+(std::string)dirp->d_name).c_str())!=0)
 							{
-                                Log("Error deleting symlink \""+upath+"/"+(std::string)dirp->d_name+"\"", LL_ERROR);
+								Log("Error deleting file \""+upath+"/"+(std::string)dirp->d_name+"\"", LL_ERROR);
 							}
 						}
 					}
-					else if(S_ISDIR(f_info.st_mode) )
+					else
 					{
-                        subdirs.push_back((dirp->d_name));
+						std::string e=convert(errno);
+						switch(errno)
+						{
+							case EACCES: e="EACCES"; break;
+							case EBADF: e="EBADF"; break;
+							case EFAULT: e="EFAULT"; break;
+							case ELOOP: e="ELOOP"; break;
+							case ENAMETOOLONG: e="ENAMETOOLONG"; break;
+							case ENOENT: e="ENOENT"; break;
+							case ENOMEM: e="ENOMEM"; break;
+							case ENOTDIR: e="ENOTDIR"; break;
+						}
+						Log("No permission to stat \""+upath+"/"+dirp->d_name+"\" error: "+e, LL_ERROR);
+					}
+	#ifndef sun
+				}
+				else if(dirp->d_type==DT_DIR )
+				{
+					paths.push(std::make_pair(upath+"/"+std::string(dirp->d_name), false));
+				}
+				else if(dirp->d_type==DT_LNK )
+				{
+					if(symlink_callback!=NULL)
+					{
+						symlink_callback((upath+"/"+(std::string)dirp->d_name), NULL, userdata);
 					}
 					else
 					{
 						if(unlink((upath+"/"+(std::string)dirp->d_name).c_str())!=0)
 						{
-                            Log("Error deleting file \""+upath+"/"+(std::string)dirp->d_name+"\"", LL_ERROR);
+							Log("Error deleting symlink \""+upath+"/"+(std::string)dirp->d_name+"\"", LL_ERROR);
 						}
 					}
 				}
 				else
-				{
-					std::string e=convert(errno);
-					switch(errno)
-					{
-					    case EACCES: e="EACCES"; break;
-					    case EBADF: e="EBADF"; break;
-					    case EFAULT: e="EFAULT"; break;
-					    case ELOOP: e="ELOOP"; break;
-					    case ENAMETOOLONG: e="ENAMETOOLONG"; break;
-					    case ENOENT: e="ENOENT"; break;
-					    case ENOMEM: e="ENOMEM"; break;
-					    case ENOTDIR: e="ENOTDIR"; break;
-					}
-                    Log("No permission to stat \""+upath+"/"+dirp->d_name+"\" error: "+e, LL_ERROR);
-				}
-#ifndef sun
-			}
-			else if(dirp->d_type==DT_DIR )
-			{
-                subdirs.push_back((dirp->d_name));
-			}
-			else if(dirp->d_type==DT_LNK )
-			{
-				if(symlink_callback!=NULL)
-				{
-                    symlink_callback((upath+"/"+(std::string)dirp->d_name), NULL, userdata);
-				}
-				else
 				{
 					if(unlink((upath+"/"+(std::string)dirp->d_name).c_str())!=0)
 					{
-                        Log("Error deleting symlink \""+upath+"/"+(std::string)dirp->d_name+"\"", LL_ERROR);
+						Log("Error deleting file \""+upath+"/"+(std::string)dirp->d_name+"\"", LL_ERROR);
 					}
 				}
+	#endif
 			}
-			else
-			{
-				if(unlink((upath+"/"+(std::string)dirp->d_name).c_str())!=0)
-				{
-                    Log("Error deleting file \""+upath+"/"+(std::string)dirp->d_name+"\"", LL_ERROR);
-				}
-			}
-#endif
-		}
-    }
-    closedir(dp);
-    for(size_t i=0;i<subdirs.size();++i)
-    {
-		bool b=os_remove_nonempty_dir(path+"/"+subdirs[i], symlink_callback, userdata);
-		if(!b)
-		    ok=false;
-    }
-	if(delete_root)
-	{
-		if(rmdir(upath.c_str())!=0)
-		{
-        		Log("Error deleting directory \""+upath+"\"", LL_ERROR);
 		}
+		closedir(dp);
 	}
+
 	return ok;
 }
 
@@ -530,6 +605,7 @@ bool os_link_symbolic(const std::string &target, const std::string &lname, void*
     return symlink(target.c_str(), lname.c_str())==0;
 }
 
+#ifndef OS_FUNC_NO_NET
 bool os_lookuphostname(std::string pServer, unsigned int *dest)
 {
 	const char* host=pServer.c_str();
@@ -577,6 +653,85 @@ bool os_lookuphostname(std::string pServer, unsigned int *dest)
 	return true;
 }
 
+bool os_lookuphostname(std::string pServer, SLookupResult & dest)
+{
+	const char* host = pServer.c_str();
+	unsigned int addr = inet_addr(host);
+	if (addr != INADDR_NONE)
+	{
+		dest.is_ipv6 = false;
+		dest.addr_v4 = addr;
+		return true;
+	}
+	int rc = inet_pton(AF_INET6, host, dest.addr_v6);
+	if (rc == 1)
+	{
+		dest.is_ipv6 = true;
+		return true;
+	}
+
+	addrinfo hints;
+	memset(&hints, 0, sizeof(hints));
+	hints.ai_family = AF_UNSPEC;
+	hints.ai_socktype = SOCK_STREAM;
+	hints.ai_protocol = IPPROTO_TCP;
+
+	addrinfo* h;
+	if (getaddrinfo(pServer.c_str(), NULL, &hints, &h) == 0)
+	{
+		addrinfo* orig_h = h;
+		while (h != NULL)
+		{
+			if (h->ai_family != AF_INET6
+				&& h->ai_family != AF_INET)
+			{
+				h = h->ai_next;
+				continue;
+			}
+
+			if (h->ai_family == AF_INET6)
+			{
+				if (h->ai_addrlen >= sizeof(sockaddr_in6))
+				{
+					dest.is_ipv6 = true;
+					memcpy(dest.addr_v6, &reinterpret_cast<sockaddr_in6*>(h->ai_addr)->sin6_addr, 16);
+					freeaddrinfo(orig_h);
+					return true;
+				}
+				else
+				{
+					freeaddrinfo(orig_h);
+					return false;
+				}
+			}
+			else
+			{
+				if (h->ai_addrlen >= sizeof(sockaddr_in))
+				{
+					dest.is_ipv6 = false;
+					dest.addr_v4 = reinterpret_cast<sockaddr_in*>(h->ai_addr)->sin_addr.s_addr;
+					freeaddrinfo(orig_h);
+					return true;
+				}
+				else
+				{
+					freeaddrinfo(orig_h);
+					return false;
+				}
+			}
+
+			h = h->ai_next;
+		}
+		freeaddrinfo(orig_h);
+		return false;
+	}
+	else
+	{
+		return false;
+	}
+}
+#endif //OS_FUNC_NO_NET
+
 std::string os_file_prefix(std::string path)
 {
 	return path;
@@ -870,6 +1025,17 @@ SFile getFileMetadata( const std::string &path )
 			ret.isdir = true;
 		}
 
+		if (S_ISLNK(f_info.st_mode))
+		{
+			ret.issym = true;
+			ret.isspecialf = true;
+		}
+
+		if (!ret.isdir && !S_ISREG(f_info.st_mode))
+		{
+			ret.isspecialf = true;
+		}
+
 		ret.size = f_info.st_size;
 		ret.last_modified = f_info.st_mtime;
 		ret.created = f_info.st_ctime;
@@ -911,6 +1077,10 @@ int os_popen(const std::string& cmd, std::string& ret)
 {
 	ret.clear();
 
+#ifdef __ANDROID__
+    POFILE* pin = NULL;
+#endif
+
 	FILE* in = NULL;
 
 #ifndef _WIN32
@@ -918,7 +1088,15 @@ int os_popen(const std::string& cmd, std::string& ret)
 #define _pclose pclose
 #endif
 
+#ifdef __ANDROID__
+    pin = and_popen(cmd.c_str(), "r");
+    if(pin!=NULL) in=pin->fp;
+#elif __linux__
+	in = _popen(cmd.c_str(), "re");
+	if(!in) in = _popen(cmd.c_str(), "r");
+#else
 	in = _popen(cmd.c_str(), "r");
+#endif
 
 	if(in==NULL)
 	{
@@ -937,7 +1115,11 @@ int os_popen(const std::string& cmd, std::string& ret)
 	}
 	while(read==sizeof(buf));
 
-	return _pclose(in);
+#ifdef __ANDROID__
+    return and_pclose(pin);
+#else
+    return _pclose(in);
+#endif
 }
 
 std::string os_last_error_str()
@@ -1069,7 +1251,7 @@ bool os_disable_background_priority(SPrioInfo& prio_info)
 	return true;
 }
 
-bool os_enable_prioritize(SPrioInfo& prio_info)
+bool os_enable_prioritize(SPrioInfo& prio_info, EPrio prio)
 {	
 	if(getpid() == syscall(SYS_gettid))
 	{
@@ -1082,12 +1264,23 @@ bool os_enable_prioritize(SPrioInfo& prio_info)
 	
 	int ioprio = 0;
 	int ioprio_class = IOPRIO_CLASS_BE;
+	int cpuprio = -10;
+	
+	if(prio==Prio_SlightPrioritize)
+	{
+		ioprio=2;
+		cpuprio=-3;
+	}
+	else if(prio==Prio_SlightBackground)
+	{
+		ioprio=7;
+		cpuprio=5;
+	}
 	
 	if(ioprio_set(IOPRIO_WHO_PROCESS, 0, ioprio | ioprio_class << IOPRIO_CLASS_SHIFT)==-1)
 	{
 		return false;
-	}
-	int cpuprio = -10;
+	}	
 	if(setpriority(PRIO_PROCESS, 0, cpuprio)==-1)
 	{
 		os_disable_prioritize(prio_info);
@@ -1118,6 +1311,128 @@ void os_reset_priority()
 	setpriority(PRIO_PROCESS, 0, 0);
 }
 
+#elif defined (__APPLE__)
+
+struct SPrioInfoInt
+{
+	int io_prio;
+	int cpu_prio;
+};
+
+SPrioInfo::SPrioInfo()
+ : prio_info(new SPrioInfoInt)
+{
+}
+
+SPrioInfo::~SPrioInfo()
+{
+	delete prio_info;
+}
+
+bool os_enable_background_priority(SPrioInfo& prio_info)
+{
+	if(prio_info.prio_info==NULL)
+	{
+		return false;
+	}
+
+	uint64 thread_id;
+	if(pthread_threadid_np(NULL, &thread_id) == 0
+		&& (uint64)getpid() == thread_id )
+	{
+		//This would set it for the whole process
+		return false;
+	}
+	
+	prio_info.prio_info->io_prio = getiopolicy_np(IOPOL_TYPE_DISK, IOPOL_SCOPE_THREAD);
+	prio_info.prio_info->cpu_prio = getpriority(PRIO_DARWIN_THREAD, 0);
+
+	if(setiopolicy_np(IOPOL_TYPE_DISK, IOPOL_SCOPE_THREAD, IOPOL_THROTTLE)==-1)
+	{
+		return false;
+	}
+	int cpuprio = 19;
+	if(setpriority(PRIO_DARWIN_THREAD, 0, cpuprio)==-1)
+	{
+		os_disable_background_priority(prio_info);
+		return false;
+	}
+	
+	return true;
+}
+
+bool os_disable_background_priority(SPrioInfo& prio_info)
+{
+	if(prio_info.prio_info==NULL)
+	{
+		return false;
+	}
+		
+	bool success = (setiopolicy_np(IOPOL_TYPE_DISK, IOPOL_SCOPE_THREAD, prio_info.prio_info->io_prio)==0);
+	success &= (setpriority(PRIO_DARWIN_THREAD, 0, prio_info.prio_info->cpu_prio)==0);
+	return success;
+}
+
+bool os_enable_prioritize(SPrioInfo& prio_info, EPrio prio)
+{
+	if(prio_info.prio_info==NULL)
+	{
+		return false;
+	}
+
+	uint64 thread_id;
+	if(pthread_threadid_np(NULL, &thread_id) == 0
+		&& (uint64)getpid() == thread_id )
+	{
+		//This would set it for the whole process
+		return false;
+	}
+
+	prio_info.prio_info->io_prio = getiopolicy_np(IOPOL_TYPE_DISK, IOPOL_SCOPE_THREAD);
+	prio_info.prio_info->cpu_prio = getpriority(PRIO_DARWIN_THREAD, 0);
+	
+	int ioprio = IOPOL_STANDARD;
+	int cpuprio = -10;
+	
+	if(prio==Prio_SlightPrioritize)
+	{
+		ioprio=IOPOL_IMPORTANT;
+		cpuprio=-3;
+	}
+	else if(prio==Prio_SlightBackground)
+	{
+		ioprio=IOPOL_UTILITY;
+		cpuprio=5;
+	}
+	
+	if(setiopolicy_np(IOPOL_TYPE_DISK, IOPOL_SCOPE_THREAD, ioprio)==-1)
+	{
+		return false;
+	}	
+	if(setpriority(PRIO_DARWIN_THREAD, 0, cpuprio)==-1)
+	{
+		os_disable_prioritize(prio_info);
+		return false;
+	}
+	
+	return true;
+}
+
+bool os_disable_prioritize(SPrioInfo& prio_info)
+{
+	return os_disable_background_priority(prio_info);
+}
+
+void assert_process_priority()
+{
+}
+
+void os_reset_priority()
+{
+	setiopolicy_np(IOPOL_TYPE_DISK, IOPOL_SCOPE_PROCESS, IOPOL_STANDARD);
+	setpriority(PRIO_PROCESS, 0, 0);
+}
+
 #else //__NR_ioprio_set
 
 SPrioInfo::SPrioInfo()
@@ -1138,7 +1453,7 @@ bool os_disable_background_priority(SPrioInfo& prio_info)
 	return false;
 }
 
-bool os_enable_prioritize(SPrioInfo& prio_info)
+bool os_enable_prioritize(SPrioInfo& prio_info, EPrio prio)
 {
 	return false;
 }
@@ -1158,24 +1473,52 @@ void os_reset_priority()
 
 #endif //__NR_ioprio_set
 
+#define BTRFS_IOC_SYNC _IO(BTRFS_IOCTL_MAGIC, 8)
+
 bool os_sync(const std::string & path)
 {
-#if defined(__linux__) && defined(HAVE_SYNCFS)
+#if defined(__linux__)
 	int fd = open(path.c_str(), O_RDONLY|O_CLOEXEC);
 	
 	if(fd!=-1)
 	{
-		if(syncfs(fd)!=0)
+		if(ioctl(fd, BTRFS_IOC_SYNC, NULL)==-1)
 		{
-			sync();
-			close(fd);
-			return true;
+			if(errno!=ENOTTY && errno!=ENOSYS 
+				&& errno!=EINVAL)
+			{
+				close(fd);
+				return false;
+			}
 		}
 		else
 		{
 			close(fd);
 			return true;
 		}
+
+#if defined(HAVE_SYNCFS)
+		if(syncfs(fd)!=0)
+		{
+			if(errno==ENOSYS)
+			{
+				close(fd);
+				sync();
+				return true;
+			}
+			close(fd);
+			return false;
+		}
+		else
+		{
+			close(fd);
+			return true;
+		}
+#else
+		close(fd);
+		sync();
+		return true;
+#endif
 	}
 	else
 	{
@@ -1188,3 +1531,16 @@ bool os_sync(const std::string & path)
 #endif
 }
 
+size_t os_get_num_cpus()
+{
+	return sysconf(_SC_NPROCESSORS_ONLN);
+}
+
+int os_system(const std::string& cmd)
+{
+#ifdef __ANDROID__
+	return and_system(cmd.c_str());
+#else
+	return system(cmd.c_str());
+#endif
+}
