@@ -39,6 +39,7 @@
 #include "server.h"
 #include "../urbackupcommon/TreeHash.h"
 #include "../common/data.h"
+#include "PhashLoad.h"
 
 #ifndef NAME_MAX
 #define NAME_MAX _POSIX_NAME_MAX
@@ -118,7 +119,7 @@ bool FileBackup::request_filelist_construct(bool full, bool resume, int group,
 	CTCPStack tcpstack(client_main->isOnInternetConnection());
 
 	ServerLogger::Log(logid, clientname+": Connecting for filelist...", LL_DEBUG);
-	IPipe *cc=client_main->getClientCommandConnection(10000);
+	IPipe *cc=client_main->getClientCommandConnection(server_settings.get(), 10000);
 	if(cc==NULL)
 	{
 		ServerLogger::Log(logid, "Connecting to ClientService of \""+clientname+"\" failed - CONNECT error during filelist construction", LL_ERROR);
@@ -181,6 +182,14 @@ bool FileBackup::request_filelist_construct(bool full, bool resume, int group,
 		start_backup_cmd += "&status_id=" + convert(status_id);
 	}
 
+	bool phash = false;
+	if (client_main->getProtocolVersions().phash_version > 0
+		&& server_settings->getSettings()->internet_parallel_file_hashing)
+	{
+		start_backup_cmd += "&phash=1";
+		phash = true;
+	}
+
 	bool async_index = false;
 	if (client_main->getProtocolVersions().async_index_version > 0)
 	{
@@ -223,7 +232,7 @@ bool FileBackup::request_filelist_construct(bool full, bool resume, int group,
 				&& Server->getTimeMS() - starttime <= timeout_time)
 			{
 				ServerLogger::Log(logid, clientname + ": Connecting for filelist (async)...", LL_DEBUG);
-				cc = client_main->getClientCommandConnection(10000);
+				cc = client_main->getClientCommandConnection(server_settings.get(), 10000);
 
 				if (ServerStatus::getProcess(clientname, status_id).stop)
 				{
@@ -252,6 +261,7 @@ bool FileBackup::request_filelist_construct(bool full, bool resume, int group,
 			}
 
 			tcpstack.reset();
+			tcpstack.setAddChecksum(client_main->isOnInternetConnection());
 
 			std::string cmd = identity + "WAIT FOR INDEX async_id=" + async_id;
 
@@ -297,12 +307,13 @@ bool FileBackup::request_filelist_construct(bool full, bool resume, int group,
 		tcpstack.AddData((char*)ret.c_str(), ret.size());
 
 		std::string ret;
-		if(tcpstack.getPacket(ret))
+		bool has_error = false;
+		while(tcpstack.getPacket(ret))
 		{
 			if(ret!="DONE")
 			{
 				if (async_id.empty()
-					&& next(ret, 0, "ASYNC-"))
+					&& next(ret, 0, "ASYNC-") )
 				{
 					str_map params;
 					ParseParamStrHttp(ret.substr(6), &params);
@@ -311,11 +322,26 @@ bool FileBackup::request_filelist_construct(bool full, bool resume, int group,
 					Server->destroy(cc);
 					cc = NULL;
 					starttime = Server->getTimeMS();
+
+					
 				}
 				else if(ret=="BUSY")
 				{
 					starttime=Server->getTimeMS();
 				}
+				else if (ret == "PHASH")
+				{
+					starttime = Server->getTimeMS();
+					if (phash)
+					{
+						if (!startPhashDownloadThread(async_id))
+						{
+							ServerLogger::Log(logid, "Error starting parallel hash load", LL_ERROR);
+							has_error = true;
+							break;
+						}
+					}
+				}
 				else if(ret!="no backup dirs")
 				{
 					if (ret == "ERR")
@@ -324,12 +350,20 @@ bool FileBackup::request_filelist_construct(bool full, bool resume, int group,
 					}
 					logVssLogdata(Server->getTimeSeconds()-total_starttime_s);
 					ServerLogger::Log(logid, "Constructing of filelist of \""+clientname+"\" failed: "+ret, LL_ERROR);
+					if (ret.find("Async indexing process not found") != std::string::npos)
+					{
+						ServerLogger::Log(logid, "Hint: The most likely explanation for this error is that the client was restarted/rebooted during indexing", LL_ERROR);
+						should_backoff = false;
+						has_timeout_error = true;
+					}
+					has_error = true;
 					break;
 				}
 				else
 				{
 					ServerLogger::Log(logid, "Constructing of filelist of \""+clientname+"\" failed: "+ret+". Please add paths to backup on the client (via tray icon) or configure default paths to backup.", LL_ERROR);
 					no_backup_dirs=true;
+					has_error = true;
 					break;
 				}				
 			}
@@ -340,6 +374,11 @@ bool FileBackup::request_filelist_construct(bool full, bool resume, int group,
 				return true;
 			}
 		}
+
+		if (has_error)
+		{
+			break;
+		}
 	}
 
 	if (has_total_timeout)
@@ -354,6 +393,112 @@ bool FileBackup::request_filelist_construct(bool full, bool resume, int group,
 	return false;
 }
 
+bool FileBackup::wait_for_async(const std::string& async_id, int64 timeout_time)
+{
+	int64 starttime = Server->getTimeMS();
+	std::auto_ptr<IPipe> cc;
+	CTCPStack tcpstack(client_main->isOnInternetConnection());
+
+	while (Server->getTimeMS() - starttime <= timeout_time)
+	{
+		if (cc.get() == NULL)
+		{
+			while (cc.get() == NULL
+				&& Server->getTimeMS() - starttime <= timeout_time)
+			{
+				ServerLogger::Log(logid, clientname + ": Connecting for async...", LL_DEBUG);
+				cc.reset(client_main->getClientCommandConnection(server_settings.get(), 10000));
+
+				if (ServerStatus::getProcess(clientname, status_id).stop)
+				{
+					cc.reset();
+					ServerLogger::Log(logid, "Sever admin stopped backup)", LL_WARNING);
+					break;
+				}
+
+				if (cc.get() == NULL)
+				{
+					ServerLogger::Log(logid, clientname + ": Failed to connect to client. Retrying in 10s", LL_DEBUG);
+					Server->wait(10000);
+				}
+			}
+
+			if (cc.get() == NULL)
+			{
+				if (!ServerStatus::getProcess(clientname, status_id).stop)
+				{
+					ServerLogger::Log(logid, "Connecting to ClientService of \"" + clientname + "\" failed - CONNECT error during async (2)", LL_ERROR);
+					has_timeout_error = true;
+				}
+				return false;
+			}
+
+			starttime = Server->getTimeMS();
+			tcpstack.reset();
+			std::string cmd = client_main->getIdentity() + "WAIT FOR INDEX async_id=" + async_id + "#token=" + server_token;
+			tcpstack.Send(cc.get(), cmd);
+		}
+
+		std::string ret;
+		size_t rc = cc->Read(&ret, 60000);
+		if (rc == 0)
+		{
+			cc.reset();
+			continue;
+		}
+
+		tcpstack.AddData((char*)ret.c_str(), ret.size());
+
+		while (tcpstack.getPacket(ret))
+		{
+			if (ret == "DONE")
+			{
+				return true;
+			}
+			else if (ret == "BUSY")
+			{
+				starttime = Server->getTimeMS();
+			}
+			else
+			{
+				if (ret == "ERR")
+				{
+					client_main->forceReauthenticate();
+				}
+
+				ServerLogger::Log(logid, "Async cmd of \"" + clientname + "\" failed: " + ret, LL_ERROR);
+				return false;
+			}
+		}
+	}
+
+	ServerLogger::Log(logid, "Async cmd of \"" + clientname + "\" failed: Timeout", LL_ERROR);
+	has_timeout_error = true;
+	return false;
+}
+
+bool FileBackup::request_client_write_tokens()
+{
+	std::string ret = client_main->sendClientMessage("WRITE TOKENS ", "Error requesting client to write tokens", 10000, true, LL_WARNING);
+	if (ret == "OK")
+	{
+		return true;
+	}
+	else if (next(ret, 0, "ASYNC-"))
+	{
+		str_map params;
+		ParseParamStrHttp(ret.substr(6), &params);
+
+		return wait_for_async(params["async_id"]);
+	}
+	else
+	{
+		ServerLogger::Log(logid, "Error requesting client to write tokens: " + ret, LL_WARNING);
+	}
+
+	return false;
+}
+
 bool FileBackup::hasEarlyError()
 {
 	return has_early_error;
@@ -380,8 +525,16 @@ void FileBackup::logVssLogdata(int64 vss_duration_s)
 	}
 }
 
-bool FileBackup::getTokenFile(FileClient &fc, bool hashed_transfer )
+bool FileBackup::getTokenFile(FileClient &fc, bool hashed_transfer, bool request)
 {
+	if (request)
+	{
+		if (!request_client_write_tokens())
+		{
+			return false;
+		}
+	}
+
 	bool has_token_file=true;
 	
 	IFsFile *tokens_file=Server->openFile(os_file_prefix(backuppath_hashes+os_file_sep()+".urbackup_tokens.properties"), MODE_WRITE);
@@ -434,7 +587,7 @@ void FileBackup::createHashThreads(bool use_reflink, bool ignore_hash_mismatches
 	hashpipe=Server->createMemoryPipe();
 	hashpipe_prepare=Server->createMemoryPipe();
 
-	bsh=new BackupServerHash(hashpipe, clientid, use_snapshots, use_reflink, use_tmpfiles, logid, use_snapshots);
+	bsh=new BackupServerHash(hashpipe, clientid, use_snapshots, use_reflink, use_tmpfiles, logid, use_snapshots, max_file_id);
 	bsh_prepare=new BackupServerPrepareHash(hashpipe_prepare, hashpipe, clientid, logid, ignore_hash_mismatches);
 	bsh_ticket = Server->getThreadPool()->execute(bsh, "fbackup write");
 	bsh_prepare_ticket = Server->getThreadPool()->execute(bsh_prepare, "fbackup hash");
@@ -490,8 +643,10 @@ _i64 FileBackup::getIncrementalSize(IFile *f, const std::vector<size_t> &diffs,
 			{
 				if(cf.isdir==true)
 				{
-					if (indir_currdepth == 0
-						&& cf.name == "windows_components_config")
+					if (depth == 0
+						&& cf.name == "windows_components_config"
+						&& (client_main->getProtocolVersions().os_simple.empty()
+							|| client_main->getProtocolVersions().os_simple=="windows") )
 					{
 						backup_with_components = true;
 					}
@@ -621,7 +776,7 @@ void FileBackup::calculateEtaFileBackup( int64 &last_eta_update, int64& eta_set_
 
 bool FileBackup::doBackup()
 {
-	if(!client_main->handle_not_enough_space(""))
+	if(!client_main->handle_not_enough_space("", logid, *server_settings.get()))
 	{
 		return false;
 	}
@@ -646,9 +801,12 @@ bool FileBackup::doBackup()
 
 	if(!cdp_path)
 	{
-		if(!constructBackupPath(use_snapshots, !r_incremental))
+		std::string errmsg;
+		if(!constructBackupPath(use_snapshots, !r_incremental, errmsg))
 		{
-            ServerLogger::Log(logid, "Cannot create directory "+backuppath+" for backup (server error). "+os_last_error_str(), LL_ERROR);
+			errmsg = trim(errmsg);
+            ServerLogger::Log(logid, "Cannot create directory "+backuppath+" for backup (server error). "
+				+ (errmsg.empty() ? os_last_error_str() : errmsg), LL_ERROR);
 			return false;
 		}
 	}
@@ -664,7 +822,7 @@ bool FileBackup::doBackup()
 	pingthread =new ServerPingThread(client_main, clientname, status_id, client_main->getProtocolVersions().eta_version>0, server_token);
 	pingthread_ticket=Server->getThreadPool()->execute(pingthread, "client ping");
 
-	local_hash.reset(new BackupServerHash(NULL, clientid, use_snapshots, use_reflink, use_tmpfiles, logid, use_snapshots));
+	local_hash.reset(new BackupServerHash(NULL, clientid, use_snapshots, use_reflink, use_tmpfiles, logid, use_snapshots, max_file_id));
 	local_hash->setupDatabase();
 
 	createHashThreads(use_reflink, server_settings->getSettings()->ignore_disk_errors);
@@ -680,7 +838,8 @@ bool FileBackup::doBackup()
 	}	
 
 	local_hash->deinitDatabase();
-	
+
+	stopPhashDownloadThread(filelist_async_id);
 
 	if(disk_error)
 	{
@@ -713,7 +872,7 @@ bool FileBackup::hasChange(size_t line, const std::vector<size_t> &diffs)
 	return std::binary_search(diffs.begin(), diffs.end(), line);
 }
 
-std::string FileBackup::fixFilenameForOS(std::string fn, std::set<std::string>& samedir_filenames, const std::string& curr_path, bool log_warnings, logid_t logid, std::map<std::string, std::string>& filepath_corrections)
+std::string FileBackup::fixFilenameForOS(std::string fn, std::set<std::string>& samedir_filenames, const std::string& curr_path, bool log_warnings, logid_t logid, FilePathCorrections& filepath_corrections)
 {
 	std::string orig_fn = fn;
 
@@ -874,11 +1033,11 @@ std::string FileBackup::fixFilenameForOS(std::string fn, std::set<std::string>&
 	{
 		if(curr_path.empty())
 		{
-			filepath_corrections[orig_fn] = fn;
+			filepath_corrections.add(orig_fn, fn);
 		}
 		else
 		{
-			filepath_corrections[curr_path+"/"+ orig_fn] = fn;
+			filepath_corrections.add(curr_path + "/" + orig_fn, fn);
 		}
 	}
 
@@ -912,6 +1071,22 @@ void FileBackup::log_progress(const std::string & fn, int64 total, int64 downloa
 	}
 }
 
+bool FileBackup::create_hardlink(const std::string & linkname, const std::string & fname, bool use_ioref, bool * too_many_links, bool* copy)
+{
+	if (use_ioref && BackupServer::isReflinkCopy())
+	{
+		if (copy != NULL) *copy = true;
+		return copy_file(fname, linkname);
+	}
+
+	return os_create_hardlink(linkname, fname, use_ioref, too_many_links);
+}
+
+bool FileBackup::handle_not_enough_space(const std::string & path)
+{
+	return client_main->handle_not_enough_space(path, logid, *server_settings.get());
+}
+
 std::string FileBackup::convertToOSPathFromFileClient(std::string path)
 {
 	if(os_file_sep()!="/")
@@ -977,10 +1152,12 @@ void FileBackup::sendBackupOkay(bool b_okay)
 {
 	if(b_okay)
 	{
-		notifyClientBackupSuccessfull();
+		notifyClientBackupSuccessful();
 	}
 	else
 	{
+		notifyClientBackupFailed();
+
 		if(pingthread!=NULL)
 		{
 			pingthread->setStop(true);
@@ -990,7 +1167,7 @@ void FileBackup::sendBackupOkay(bool b_okay)
 	}
 }
 
-void FileBackup::notifyClientBackupSuccessfull(void)
+void FileBackup::notifyClientBackupSuccessful(void)
 {
 	if (client_main->getProtocolVersions().cmd_version == 0)
 	{
@@ -1010,6 +1187,22 @@ void FileBackup::notifyClientBackupSuccessfull(void)
 	}
 }
 
+void FileBackup::notifyClientBackupFailed()
+{
+	if (client_main->getProtocolVersions().cmd_version < 2)
+		return;
+
+	std::string params = "status_id=" + convert(status_id) + "&server_token=" + EscapeParamString(server_token)
+		+ "&group=" + convert(group);
+
+	if (!clientsubname.empty())
+	{
+		params += "&clientsubname=" + EscapeParamString(clientsubname);
+	}
+
+	client_main->sendClientMessageRetry("BACKUP FAILED " + params, "OK", "Sending status (BACKUP FAILED) to client failed", 10000, 2);
+}
+
 void FileBackup::waitForFileThreads(void)
 {
 	SStatus status=ServerStatus::getStatus(clientname);
@@ -1285,7 +1478,7 @@ bool FileBackup::hasDiskError()
 	return disk_error;
 }
 
-bool FileBackup::constructBackupPath(bool on_snapshot, bool create_fs)
+bool FileBackup::constructBackupPath(bool on_snapshot, bool create_fs, std::string& errmsg)
 {
 	if(!createDirectoryForClient())
 	{
@@ -1313,10 +1506,85 @@ bool FileBackup::constructBackupPath(bool on_snapshot, bool create_fs)
 	{
 		if(create_fs)
 		{
-			return SnapshotHelper::createEmptyFilesystem(clientname, backuppath_single)
-				&& (os_create_dir(os_file_prefix(backuppath_hashes))
-					|| (os_directory_exists(os_file_prefix(backuppath_hashes))
-						&& getFiles(os_file_prefix(backuppath_hashes)).empty()));
+			bool zfs_file = BackupServer::getSnapshotMethod(false) == BackupServer::ESnapshotMethod_ZfsFile;
+			if (zfs_file)
+			{
+				if (os_get_file_type(backuppath) != 0)
+				{
+					ServerLogger::Log(logid, "File/Directory " + backuppath + " already exists.", LL_ERROR);
+					allow_remove_backup_folder = false;
+					return false;
+				}
+
+				std::auto_ptr<IFile> touch_f(Server->openFile(backuppath, MODE_WRITE));
+				if (touch_f.get() == NULL)
+				{
+					ServerLogger::Log(logid, "Could not touch file " + backuppath + ". " + os_last_error_str(), LL_ERROR);
+					return false;
+				}
+				touch_f->Sync();
+			}
+
+			if (SnapshotHelper::isSubvolume(false, clientname, backuppath_single))
+			{
+				if (zfs_file)
+				{
+					Server->deleteFile(backuppath);
+				}
+
+				ServerLogger::Log(logid, "Subvolume " + clientname + "/"+backuppath_single + " already exists.", LL_ERROR);
+				allow_remove_backup_folder = false;
+				return false;
+			}
+
+			bool b = SnapshotHelper::createEmptyFilesystem(false, clientname, backuppath_single, errmsg);
+
+			if (!b)
+			{
+				if (zfs_file)
+				{
+					Server->deleteFile(backuppath);
+				}
+
+				ServerLogger::Log(logid, "Error creating empty file subvolume. "
+					+ (errmsg.empty() ? "" : ("\"" + errmsg + "\"")), LL_ERROR);
+
+				return false;
+			}
+
+			if (zfs_file)
+			{
+				std::string mountpoint = SnapshotHelper::getMountpoint(false, clientname, backuppath_single);
+				if (mountpoint.empty())
+				{
+					ServerLogger::Log(logid, "Could not find mountpoint of snapshot of client " + clientname + " path " + backuppath_single, LL_ERROR);
+					return false;
+				}
+
+				if (!os_link_symbolic(mountpoint, backuppath + "_new"))
+				{
+					ServerLogger::Log(logid, "Could create symlink to mountpoint at " + backuppath+ " to " + mountpoint + ". " + os_last_error_str(), LL_ERROR);
+					return false;
+				}
+
+				if (!os_rename_file(backuppath + "_new", backuppath))
+				{
+					ServerLogger::Log(logid, "Could rename symlink at " + backuppath + "_new to " + backuppath + ". " + os_last_error_str(), LL_ERROR);
+					return false;
+				}
+			}
+
+			b = os_create_dir(os_file_prefix(backuppath_hashes))
+				|| (os_directory_exists(os_file_prefix(backuppath_hashes))
+					&& getFiles(os_file_prefix(backuppath_hashes)).empty());
+
+			if (!b)
+			{
+				ServerLogger::Log(logid, "Error creating hash folder. " + os_last_error_str(), LL_ERROR);
+				return false;
+			}
+
+			return true;
 		}
 		else
 		{
@@ -1802,6 +2070,20 @@ void FileBackup::saveUsersOnClient()
 		}
 	}
 
+	s_uids = urbackup_tokens->getValue("ids", "");
+	uids.clear();
+	Tokenize(s_uids, uids, ",");
+
+	for (size_t i = 0; i < uids.size(); ++i)
+	{
+		std::string accountname = (base64_decode_dash(urbackup_tokens->getValue(uids[i] + ".accountname", std::string())));
+		if (accountname == "root")
+		{
+			backup_dao->addUserToken(accountname, clientid, urbackup_tokens->getValue(uids[i] + ".token", std::string()));
+			break;
+		}
+	}
+
 	std::vector<std::string> keys = urbackup_tokens->getKeys();
 	for(size_t i=0;i<keys.size();++i)
 	{
@@ -1816,13 +2098,22 @@ void FileBackup::deleteBackup()
 {
 	if(backupid==-1)
 	{
+		if (!allow_remove_backup_folder)
+		{
+			return;
+		}
+
 		if(use_snapshots)
 		{
-			if(!SnapshotHelper::removeFilesystem(clientname, backuppath_single) )
+			if(!SnapshotHelper::removeFilesystem(false, clientname, backuppath_single) )
 			{
 				ServerLinkDao link_dao(Server->getDatabase(Server->getThreadID(), URBACKUPDB_SERVER_LINKS));
 				remove_directory_link_dir(backuppath, link_dao, clientid);
 			}
+			else if (BackupServer::getSnapshotMethod(false) == BackupServer::ESnapshotMethod_ZfsFile)
+			{
+				Server->deleteFile(backuppath);
+			}
 		}
 		else
 		{
@@ -1832,6 +2123,13 @@ void FileBackup::deleteBackup()
 	}
 	else
 	{				
+		if (!allow_remove_backup_folder)
+		{
+			ServerCleanupDao cleanupdao(db);
+			cleanupdao.removeFileBackup(backupid);
+			return;
+		}
+
 		Server->getThreadPool()->executeWait(
 			new ServerCleanupThread(CleanupAction(server_settings->getSettings()->backupfolder, clientid, backupid, true, NULL) ),
 			"delete fbackup");
@@ -1841,7 +2139,7 @@ void FileBackup::deleteBackup()
 bool FileBackup::createSymlink(const std::string& name, size_t depth, const std::string& symlink_target, const std::string& dir_sep, bool isdir )
 {
 	std::vector<std::string> toks;
-	TokenizeMail(symlink_target, toks, dir_sep);
+	Tokenize(symlink_target, toks, dir_sep);
 
 	std::string target;
 
@@ -1889,12 +2187,12 @@ bool FileBackup::startFileMetadataDownloadThread()
 	{
 		std::string identity = client_main->getIdentity();
 		std::auto_ptr<FileClient> fc_metadata_stream(new FileClient(false, identity, client_main->getProtocolVersions().filesrv_protocol_version,
-			client_main->isOnInternetConnection(), client_main, use_tmpfiles?NULL:client_main));
+			client_main->isOnInternetConnection(), client_main, use_tmpfiles?NULL:this));
 
 		_u32 rc=client_main->getClientFilesrvConnection(fc_metadata_stream.get(), server_settings.get(), 10000);
 		if(rc!=ERR_CONNECTED)
 		{
-			ServerLogger::Log(logid, "Full Backup of "+clientname+" failed - CONNECT error (for metadata stream)", LL_ERROR);
+			ServerLogger::Log(logid, "Backup of "+clientname+" failed - CONNECT error (for metadata stream)", LL_ERROR);
 			has_early_error=true;
 			log_backup=false;
 			return false;
@@ -1924,6 +2222,13 @@ bool FileBackup::startFileMetadataDownloadThread()
 			stopFileMetadataDownloadThread(true, 0);
 			return false;
 		}
+
+		local_hash2.reset(new BackupServerHash(NULL, clientid, use_snapshots, use_reflink, use_tmpfiles, logid, use_snapshots, max_file_id));
+
+		metadata_apply_thread.reset(new server::FileMetadataDownloadThread::FileMetadataApplyThread(metadata_download_thread.get(),
+			backuppath_hashes, backuppath, client_main, local_hash2.get(), filepath_corrections, max_file_id));
+
+		metadata_apply_thread_ticket = Server->getThreadPool()->execute(metadata_apply_thread.get(), "fb meta apply");
 	}	
 
 	return true;
@@ -1935,6 +2240,8 @@ bool FileBackup::stopFileMetadataDownloadThread(bool stopped, size_t expected_em
 
 	if(metadata_download_thread.get()!=NULL)
 	{
+		metadata_download_thread->forceStart();
+
 		if (!Server->getThreadPool()->waitFor(metadata_download_thread_ticket, 1000))
 		{
 			ServerLogger::Log(logid, "Waiting for metadata download stream to finish", LL_INFO);
@@ -1947,7 +2254,7 @@ bool FileBackup::stopFileMetadataDownloadThread(bool stopped, size_t expected_em
 			{
 				std::string identity = client_main->getIdentity();
 				std::auto_ptr<FileClient> fc_metadata_stream_end(new FileClient(false, identity, client_main->getProtocolVersions().filesrv_protocol_version,
-					client_main->isOnInternetConnection(), client_main, use_tmpfiles ? NULL : client_main));
+					client_main->isOnInternetConnection(), client_main, use_tmpfiles ? NULL : this));
 
 				_u32 rc = client_main->getClientFilesrvConnection(fc_metadata_stream_end.get(), server_settings.get(), 10000);
 				if (rc == ERR_CONNECTED)
@@ -1998,29 +2305,27 @@ bool FileBackup::stopFileMetadataDownloadThread(bool stopped, size_t expected_em
 			++num_issues;
 		}
 
-		if(!stopped && !disk_error && !has_early_error && ( !metadata_download_thread->getHasError() || metadata_download_thread->getHasTimeoutError() ) )
-		{
-			size_t num_embedded_metadata_files = 0;
-			bool b = metadata_download_thread->applyMetadata(backuppath_hashes, backuppath, client_main, local_hash.get(),
-				filepath_corrections, !metadata_download_thread->getHasTimeoutError(), num_embedded_metadata_files);
+		Server->getThreadPool()->waitFor(metadata_apply_thread_ticket);
 
-			if (!b)
+		if(!stopped && !disk_error && !has_early_error && ( !metadata_download_thread->getHasError() || metadata_download_thread->getHasTimeoutError() ) )
+		{	
+			if (!metadata_apply_thread->hasSuccess())
 			{
 				++num_issues;
 			}
 
-			if (!b
+			if (!metadata_apply_thread->hasSuccess()
 				&& metadata_download_thread->getHasFatalError())
 			{
 				disk_error = true;
 			}
-			else if (num_embedded_metadata_files != expected_embedded_metadata_files)
+			else if (metadata_apply_thread->getNumEmbeddedFiles() != expected_embedded_metadata_files)
 			{
-				ServerLogger::Log(logid, "Wrong number of embedded meta-data files. Expected "+convert(expected_embedded_metadata_files)+" but got "+convert(num_embedded_metadata_files), LL_ERROR);
+				ServerLogger::Log(logid, "Wrong number of embedded meta-data files. Expected "+convert(expected_embedded_metadata_files)+" but got "+convert(metadata_apply_thread->getNumEmbeddedFiles()), LL_ERROR);
 				disk_error = true;
 			}
 
-			return b;
+			return metadata_apply_thread->hasSuccess();
 		}
 	}
 
@@ -2048,19 +2353,6 @@ void FileBackup::parseSnapshotFailed(const std::string & logline)
 	}
 }
 
-void FileBackup::addFilePathCorrections(const std::map<std::string, std::string>& c)
-{
-	for (std::map<std::string, std::string>::const_iterator it = c.begin();
-		it != c.end(); ++it)
-	{
-		std::map<std::string, std::string>::iterator curr_it = filepath_corrections.find(it->first);
-		if (curr_it == filepath_corrections.end())
-		{
-			filepath_corrections.insert(std::make_pair(it->first, it->second));
-		}
-	}
-}
-
 std::string FileBackup::permissionsAllowAll()
 {
 	CWData token_info;
@@ -2094,6 +2386,94 @@ bool FileBackup::loadWindowsBackupComponentConfigXml(FileClient &fc)
 	return true;
 }
 
+bool FileBackup::startPhashDownloadThread(const std::string& async_id)
+{
+	std::string identity = client_main->getIdentity();
+	std::auto_ptr<FileClient> fc_phash_stream(new FileClient(false, identity, client_main->getProtocolVersions().filesrv_protocol_version,
+		client_main->isOnInternetConnection(), client_main, use_tmpfiles ? NULL : this));
+
+	_u32 rc = client_main->getClientFilesrvConnection(fc_phash_stream.get(), server_settings.get(), 10000);
+	if (rc != ERR_CONNECTED)
+	{
+		ServerLogger::Log(logid, "Full Backup of " + clientname + " failed - CONNECT error (for metadata stream)", LL_ERROR);
+		has_early_error = true;
+		log_backup = false;
+		return false;
+	}
+
+	filelist_async_id = async_id;
+	fc_phash_stream->setProgressLogCallback(this);
+	phash_load.reset(new PhashLoad(fc_phash_stream.release(), logid, async_id));
+	phash_load_ticket = Server->getThreadPool()->execute(phash_load.get(), "phash load");
+
+	int64 starttime = Server->getTimeMS();
+
+	do
+	{
+		if (phash_load->isDownloading())
+		{
+			break;
+		}
+		Server->wait(100);
+	} while (Server->getTimeMS() - starttime<10000);
+
+	if (!phash_load->isDownloading())
+	{
+		stopPhashDownloadThread(async_id);
+		return false;
+	}
+
+	return true;
+}
+
+bool FileBackup::stopPhashDownloadThread(const std::string& async_id)
+{
+	if (phash_load.get() == NULL)
+	{
+		return true;
+	}
+
+	if (!Server->getThreadPool()->waitFor(phash_load_ticket, 1000))
+	{
+		ServerLogger::Log(logid, "Waiting for parallel hash load stream to finish", LL_INFO);
+
+		do
+		{
+			std::string identity = client_main->getIdentity();
+			std::auto_ptr<FileClient> fc_phash_stream_end(new FileClient(false, identity, client_main->getProtocolVersions().filesrv_protocol_version,
+				client_main->isOnInternetConnection(), client_main, use_tmpfiles ? NULL : this));
+
+			_u32 rc = client_main->getClientFilesrvConnection(fc_phash_stream_end.get(), server_settings.get(), 10000);
+			if (rc == ERR_CONNECTED)
+			{
+				fc_phash_stream_end->StopPhashLoad(server_token, async_id, 0);
+			}
+			else
+			{
+				ServerLogger::Log(logid, "Could not get filesrv connection when trying to end parallel hash data tranfer (" + clientname + ").", LL_DEBUG);
+			}
+
+			if (Server->getThreadPool()->waitFor(phash_load_ticket, 1000))
+			{
+				break;
+			}
+
+			phash_load->setProgressLogEnabled(true);
+			phash_load->shutdown();
+
+		} while (true);
+	}
+
+	if (phash_load->hasError() || phash_load->hasTimeoutError())
+	{
+		++num_issues;
+	}
+
+	phash_load.reset();
+
+	return true;
+}
+
 void FileBackup::save_debug_data(const std::string& rfn, const std::string& local_hash, const std::string& remote_hash)
 {
 	ServerLogger::Log(logid, "Local hash: "+local_hash+" remote hash: "+remote_hash, LL_INFO);
@@ -2101,7 +2481,7 @@ void FileBackup::save_debug_data(const std::string& rfn, const std::string& loca
 
 	std::string identity = client_main->getIdentity();
 	FileClient fc(false, identity, client_main->getProtocolVersions().filesrv_protocol_version,
-		client_main->isOnInternetConnection(), client_main, use_tmpfiles?NULL:client_main);
+		client_main->isOnInternetConnection(), client_main, use_tmpfiles?NULL:this);
 
 	_u32 rc=client_main->getClientFilesrvConnection(&fc, server_settings.get(), 10000);
 	if(rc!=ERR_CONNECTED)
