@@ -31,6 +31,8 @@
 #include "FileMetadataDownloadThread.h"
 #include "snapshot_helper.h"
 #include <stack>
+#include "PhashLoad.h"
+#include "server.h"
 
 extern std::string server_identity;
 
@@ -78,6 +80,11 @@ bool FullFileBackup::doFileBackup()
 	{
 		has_early_error=true;
 
+		if (no_backup_dirs)
+		{
+			backup_dao->updateClientNumIssues(ServerBackupDao::num_issues_no_backuppaths, clientid);
+		}
+
 		if(no_backup_dirs || connect_fail)
 		{
 			log_backup=false;
@@ -118,7 +125,7 @@ bool FullFileBackup::doFileBackup()
 	}
 	std::string identity = client_main->getIdentity();
 	FileClient fc(false, identity, client_main->getProtocolVersions().filesrv_protocol_version,
-		client_main->isOnInternetConnection(), client_main, use_tmpfiles?NULL:client_main);
+		client_main->isOnInternetConnection(), client_main, use_tmpfiles?NULL:this);
 	_u32 rc=client_main->getClientFilesrvConnection(&fc, server_settings.get(), 10000);
 	if(rc!=ERR_CONNECTED)
 	{
@@ -149,7 +156,7 @@ bool FullFileBackup::doFileBackup()
 		return false;
 	}
 
-	getTokenFile(fc, hashed_transfer);
+	getTokenFile(fc, hashed_transfer, false);
 
 	if (!backup_dao->newFileBackup(0, clientid, backuppath_single, 0, Server->getTimeMS() - indexing_start_time, group))
 	{
@@ -170,7 +177,7 @@ bool FullFileBackup::doFileBackup()
 
 	if(clientlist==NULL )
 	{
-		ServerLogger::Log(logid, "Error creating clientlist for client "+clientname, LL_ERROR);
+		ServerLogger::Log(logid, "Error creating clientlist for client "+clientname+" at "+Server->getServerWorkingDir()+ "/"+clientlistName(backupid)+". "+os_last_error_str(), LL_ERROR);
 		has_early_error=true;
 		return false;
 	}
@@ -222,7 +229,7 @@ bool FullFileBackup::doFileBackup()
 		use_tmpfiles, tmpfile_path, server_token, use_reflink,
 		backupid, false, hashpipe_prepare, client_main, client_main->getProtocolVersions().filesrv_protocol_version,
 		0, logid, with_hashes, shares_without_snapshot, with_sparse_hashing, metadata_download_thread.get(),
-		backup_with_components));
+		backup_with_components, filepath_corrections, max_file_id));
 
 	bool queue_downloads = client_main->getProtocolVersions().filesrv_protocol_version>2;
 
@@ -461,6 +468,11 @@ bool FullFileBackup::doFileBackup()
 				}
 				else
 				{
+					if (depth == 0)
+					{
+						server_download->addToQueueStartShadowcopy(cf.name);
+					}
+
 					if(!has_orig_path)
 					{
 						if (curr_orig_path != orig_sep)
@@ -527,10 +539,30 @@ bool FullFileBackup::doFileBackup()
 						hash_it = extra_params.find("thash");
 					}
 
-					if( !file_ok
-						&& hash_it!=extra_params.end())
+					if (!file_ok
+						&& hash_it != extra_params.end())
 					{
 						curr_sha2 = base64_decode_dash(hash_it->second);
+					}
+					else if (!file_ok
+						&& phash_load.get() != NULL
+						&& !script_dir
+						&& extra_params.find("no_hash") == extra_params.end())
+					{
+						if (!phash_load->getHash(line, curr_sha2))
+						{
+							ServerLogger::Log(logid, "Error getting parallel hash for file \"" + cf.name + "\" line " + convert(line), LL_ERROR);
+							r_offline = true;
+							break;
+						}
+						else
+						{
+							metadata.shahash = curr_sha2;
+						}
+					}
+
+					if(!curr_sha2.empty())
+					{						
 						if(cf.size>= link_file_min_size
 							&& link_file(cf.name, osspecific_name, curr_path, curr_os_path, curr_sha2, cf.size,
 							             true, metadata))
@@ -557,8 +589,14 @@ bool FullFileBackup::doFileBackup()
 						server_download->addToQueueFull(line, cf.name, osspecific_name, curr_path, curr_os_path, queue_downloads?cf.size:-1,
 							metadata, script_dir, false, 0, curr_sha2);
 					}
+
+					if (depth == 0)
+					{
+						server_download->addToQueueStopShadowcopy(cf.name);
+					}
 				}
 
+				max_file_id.setMaxPreProcessed(line);
 				++line;
 			}
 		}
@@ -573,6 +611,8 @@ bool FullFileBackup::doFileBackup()
 		disk_error = true;
 	}
 
+	stopPhashDownloadThread(filelist_async_id);
+
 
 	server_download->queueStop();
 
@@ -604,8 +644,6 @@ bool FullFileBackup::doFileBackup()
 		calculateDownloadSpeed(ctime, fc, NULL);
 	}
 
-	addFilePathCorrections(server_download->getFilePathCorrections());
-
 	ServerStatus::setProcessSpeed(clientname, status_id, 0);
 
 	if(server_download->isOffline() && !r_offline)
@@ -659,6 +697,12 @@ bool FullFileBackup::doFileBackup()
 
 	stopFileMetadataDownloadThread(false, server_download->getNumEmbeddedMetadataFiles());
 
+	if (!r_offline && !c_has_error && !disk_error
+		&& client_main->getProtocolVersions().wtokens_version > 0 )
+	{
+		getTokenFile(fc, hashed_transfer, true);
+	}
+
 	server_download->deleteTempFolder();
 
 	ServerLogger::Log(logid, "Writing new file list...", LL_INFO);
@@ -837,29 +881,46 @@ bool FullFileBackup::doFileBackup()
 		if ( !os_sync(backuppath)
 			|| !os_sync(backuppath_hashes) )
 		{
-			ServerLogger::Log(logid, "Syncing file system failed. Backup may not be completely on disk. " + os_last_error_str(), LL_DEBUG);
+			ServerLogger::Log(logid, "Syncing file system failed. Backup may not be completely on disk. " + os_last_error_str(), BackupServer::canSyncFs() ? LL_ERROR : LL_DEBUG);
+
+			if(BackupServer::canSyncFs())
+				c_has_error = true;
 		}
 
-		std::auto_ptr<IFile> sync_f(Server->openFile(os_file_prefix(backuppath_hashes + os_file_sep() + sync_fn), MODE_WRITE));
-
-		if (use_snapshots)
+		std::auto_ptr<IFile> sync_f;
+		if (!c_has_error)
 		{
-			if (!SnapshotHelper::makeReadonly(clientname, backuppath_single))
+			sync_f.reset(Server->openFile(os_file_prefix(backuppath_hashes + os_file_sep() + sync_fn), MODE_WRITE));
+
+			if (sync_f.get() != NULL
+				&& !sync_f->Sync())
+			{
+				ServerLogger::Log(logid, "Error syncing sync file to disk. " + os_last_error_str(), LL_ERROR);
+				c_has_error = true;
+			}
+		}
+
+		if (!c_has_error
+			&& use_snapshots)
+		{
+			if (!SnapshotHelper::makeReadonly(false, clientname, backuppath_single))
 			{
 				ServerLogger::Log(logid, "Making backup snapshot read only failed", LL_WARNING);
 			}
 		}
 
-		DBScopedSynchronous synchronous(db);
-		backup_dao->setFileBackupDone(backupid);
-
 		if (sync_f.get() != NULL)
 		{
+			DBScopedSynchronous synchronous(db);
+			DBScopedWriteTransaction trans(db);
+
+			backup_dao->setFileBackupDone(backupid);
 			backup_dao->setFileBackupSynced(backupid);
 		}
-		else
+		else if(!c_has_error)
 		{
-			ServerLogger::Log(logid, "Error creating sync file at "+ backuppath_hashes + os_file_sep() + sync_fn, LL_WARNING);
+			ServerLogger::Log(logid, "Error creating sync file at "+ backuppath_hashes + os_file_sep() + sync_fn+". Not setting backup to done.", LL_ERROR);
+			c_has_error = true;
 		}
 
 		Server->destroy(clientlist);
