@@ -342,15 +342,25 @@ namespace
 bool ImageBackup::doImage(const std::string &pLetter, const std::string &pParentvhd, int incremental, int incremental_ref,
 	bool transfer_checksum, std::string image_file_format, bool transfer_bitmap, bool transfer_prev_cbitmap)
 {
-	std::string sletter = pLetter;
+	std::string sletter = conv_filename(pLetter);
 	if (pLetter != "SYSVOL" && pLetter != "ESP")
 	{
-		sletter = pLetter[0];
+		if ((pLetter.size() <= 3
+			&& pLetter[1] == ':'
+			&& (pLetter[2] == '\\'
+				|| pLetter[2] == '/'))
+			|| (pLetter.size() <= 2
+				&& pLetter[1] == ':'))
+		{
+			sletter = sletter.substr(0,1);
+		}
 	}
 
 	std::string imagefn;
 	bool fatal_mbr_error;
-	std::string mbrd = getMBR(sletter, fatal_mbr_error);
+	std::string loadfn;
+	bool disk_backup = false;
+	std::string mbrd = getMBR(sletter, pLetter, pParentvhd.empty(), snapshot_id, fatal_mbr_error, loadfn);
 	if (mbrd.empty())
 	{
 		if (pLetter != "SYSVOL" && pLetter != "ESP")
@@ -373,10 +383,12 @@ bool ImageBackup::doImage(const std::string &pLetter, const std::string &pParent
 
 		if (imagefn.empty())
 		{
+			ServerLogger::Log(logid, "Error creating image backup destination.", LL_ERROR);
 			return false;
 		}
 
-		std::auto_ptr<IFile> mbr_file(Server->openFile(os_file_prefix(imagefn + ".mbr"), MODE_WRITE));
+		std::auto_ptr<IFsFile> mbr_file(Server->openFile(os_file_prefix(imagefn + ".mbr"), MODE_WRITE));
+
 		if (mbr_file.get() != NULL)
 		{
 			_u32 w = mbr_file->Write(mbrd);
@@ -385,13 +397,143 @@ bool ImageBackup::doImage(const std::string &pLetter, const std::string &pParent
 				ServerLogger::Log(logid, "Error writing mbr data. " + os_last_error_str(), LL_ERROR);
 				return false;
 			}
-			mbr_file.reset();
 		}
 		else
 		{
 			ServerLogger::Log(logid, "Error creating file for writing MBR data. " + os_last_error_str(), LL_ERROR);
 			return false;
 		}
+
+		if(!loadfn.empty())
+		{
+			disk_backup = true;
+
+			FileClient fc(false, client_main->getIdentity(), client_main->getProtocolVersions().filesrv_protocol_version,
+				client_main->isOnInternetConnection(), client_main);
+			_u32 rc = client_main->getClientFilesrvConnection(&fc, server_settings.get(), 10000);
+			if (rc != ERR_CONNECTED)
+			{
+				ServerLogger::Log(logid, "Error getting MBR zip data - CONNECT error", LL_ERROR);
+				return false;
+			}
+
+			ServerLogger::Log(logid, clientname + ": Loading MBR zip file...", LL_INFO);
+
+			int64 full_backup_starttime = Server->getTimeMS();
+
+			int64 starttime = Server->getTimeMS();
+
+			IFsFile* tmpf = Server->openTemporaryFile();
+			if (tmpf == NULL)
+			{
+				ServerLogger::Log(logid, "Error opening temporay file for MBR zip load", LL_ERROR);
+				return false;
+			}
+
+			ScopedDeleteFile tmpf_del(tmpf);
+
+			bool has_err_file = true;
+			{
+				_u32 rc2 = fc.GetFile(loadfn + ".err", tmpf, true, false, 0, false, 0);
+				if (rc2 == ERR_CANNOT_OPEN_FILE)
+				{
+					has_err_file = false;
+				}
+			}
+
+			tmpf->Resize(0, false);
+			tmpf->Seek(0);
+			rc = ERR_CANNOT_OPEN_FILE;
+			int retry = 0;
+			while (Server->getTimeMS() - starttime < 60 * 60 * 1000
+				&& rc== ERR_CANNOT_OPEN_FILE)
+			{
+				rc = fc.GetFile(loadfn, tmpf, true, false, 0, false, 0);
+
+				if (rc != ERR_TIMEOUT
+					&& rc!=ERR_SUCCESS)
+				{
+					if (retry < 3)
+					{
+						Server->wait(1000);
+					}
+					else
+					{
+						if (!has_err_file)
+						{
+							tmpf->Resize(0, false);
+							tmpf->Seek(0);
+							_u32 rc2 = fc.GetFile(loadfn + ".err", tmpf, true, false, 0, false, 0);
+
+							if (rc2 == ERR_SUCCESS)
+							{
+								if (tmpf->Size() < 100 * 1024)
+									ServerLogger::Log(logid, "Error loading ZIP file MBR (1): " + tmpf->Read(0LL, static_cast<_u32>(tmpf->Size())), LL_ERROR);
+								else
+									ServerLogger::Log(logid, "Error loading ZIP file MBR (1)(error too large)", LL_ERROR);
+								return false;
+							}
+						}
+
+						Server->wait(20000);
+					}
+					++retry;
+				}
+			}
+
+			if (rc != ERR_SUCCESS)
+			{
+				tmpf->Resize(0, false);
+				tmpf->Seek(0);
+				_u32 rc2 = fc.GetFile(loadfn+".err", tmpf, true, false, 0, false, 0);
+
+				if (rc2 == ERR_SUCCESS)
+				{
+					if(tmpf->Size()<100 * 1024)
+						ServerLogger::Log(logid, "Error loading ZIP file MBR: "+tmpf->Read(0LL, static_cast<_u32>(tmpf->Size()) ), LL_ERROR);
+					else
+						ServerLogger::Log(logid, "Error loading ZIP file MBR (error too large)", LL_ERROR);
+					return false;
+				}
+				else
+				{
+					ServerLogger::Log(logid, "Error loading ZIP file MBR: "+fc.getErrorString(rc), LL_ERROR);
+					return false;
+				}
+			}
+
+			char buf[1024];
+			tmpf->Seek(0);
+			bool has_read_error=false;
+			while ((rc = tmpf->Read(buf, sizeof(buf), &has_read_error)) > 0)
+			{
+				if (mbr_file->Write(buf, rc) != rc)
+				{
+					ServerLogger::Log(logid, "Error writing to "+mbr_file->getFilename()+". "+os_last_error_str(), LL_ERROR);
+					return false;
+				}
+			}
+
+			if (has_read_error)
+			{
+				ServerLogger::Log(logid, "Error reading from " + tmpf->getFilename() + ". " + os_last_error_str(), LL_ERROR);
+				return false;
+			}
+
+			ServerLogger::Log(logid, clientname + ": Loaded MBR zip file (" + PrettyPrintBytes(tmpf->Size()) + ")", LL_INFO);
+
+			if (snapshot_id == 0)
+			{
+				tmpf->Resize(0, false);
+				tmpf->Seek(0);
+				_u32 rc2 = fc.GetFile(loadfn + ".save_id", tmpf, true, false, 0, false, 0);
+				if (rc2 == ERR_SUCCESS
+					&& tmpf->Size() < 100)
+				{
+					snapshot_id = watoi(tmpf->Read(0LL, static_cast<_u32>(tmpf->Size())));
+				}
+			}
+		}
 	}
 
 	ScopedLockImageFromCleanup cleanup_lock(0);
@@ -404,7 +546,7 @@ bool ImageBackup::doImage(const std::string &pLetter, const std::string &pParent
 	}
 
 	CTCPStack tcpstack(client_main->isOnInternetConnection());
-	IPipe *cc=client_main->getClientCommandConnection(10000);
+	IPipe *cc=client_main->getClientCommandConnection(server_settings.get(), 10000);
 	if(cc==NULL)
 	{
 		ServerLogger::Log(logid, "Connecting to \""+clientname+"\" for image backup failed", LL_ERROR);
@@ -441,6 +583,8 @@ bool ImageBackup::doImage(const std::string &pLetter, const std::string &pParent
 		chksum_str += "&shadowid=" + convert(snapshot_id);
 	}
 
+	chksum_str += "&zero_skipped=1";
+
 	if(pParentvhd.empty())
 	{
 		tcpstack.Send(cc, identity+"FULL IMAGE letter="+pLetter+"&token="+server_token+chksum_str);
@@ -645,7 +789,7 @@ bool ImageBackup::doImage(const std::string &pLetter, const std::string &pParent
 					else
 					{
 						Server->Log(clientname +": Trying to reconnect in doImage", LL_DEBUG);
-						cc = client_main->getClientCommandConnection(10000);
+						cc = client_main->getClientCommandConnection(server_settings.get(), 10000);
 						if (cc == NULL)
 						{
 							Server->wait(60000);
@@ -695,6 +839,7 @@ bool ImageBackup::doImage(const std::string &pLetter, const std::string &pParent
 					{
 						cmd += "&clientsubname=" + EscapeParamString(clientsubname);
 					}
+					cmd += "&zero_skipped=1";
 					size_t sent = tcpstack.Send(cc, cmd);
 					if(sent==0)
 					{
@@ -722,6 +867,7 @@ bool ImageBackup::doImage(const std::string &pLetter, const std::string &pParent
 					{
 						ts += "&clientsubname=" + EscapeParamString(clientsubname);
 					}
+					ts += "&zero_skipped=1";
 					std::auto_ptr<IFile> prevbitmap;
 					if (transfer_prev_cbitmap)
 					{
@@ -971,15 +1117,18 @@ bool ImageBackup::doImage(const std::string &pLetter, const std::string &pParent
 						ServerStatus::setProcessTotalBytes(clientname, status_id, blockcnt*blocksize);
 					}
 
-					mbr_offset=writeMBR(vhdfile, drivesize);
-					if( mbr_offset==0 )
+					if (!disk_backup)
 					{
-						ServerLogger::Log(logid, "Error writing image MBR", LL_ERROR);
-						goto do_image_cleanup;
-					}
-					else
-					{
-						vhdfile->setMbrOffset(mbr_offset);
+						mbr_offset = writeMBR(vhdfile, drivesize);
+						if (mbr_offset == 0)
+						{
+							ServerLogger::Log(logid, "Error writing image MBR", LL_ERROR);
+							goto do_image_cleanup;
+						}
+						else
+						{
+							vhdfile->setMbrOffset(mbr_offset);
+						}
 					}
 				}
 				else
@@ -1396,25 +1545,41 @@ bool ImageBackup::doImage(const std::string &pLetter, const std::string &pParent
 							IFile *t_file=Server->openFile(os_file_prefix(imagefn), MODE_READ);
 							if(t_file!=NULL)
 							{
+								int64 image_size = t_file->RealSize();
+								Server->destroy(t_file);
+
 								std::auto_ptr<IFile> sync_f;
 								if (!vhdfile_err)
 								{
 									if (!os_sync(imagefn))
 									{
-										ServerLogger::Log(logid, "Syncing file system failed. Image backup may not be completely on disk. " + os_last_error_str(), LL_DEBUG);
+										ServerLogger::Log(logid, "Syncing file system failed. Image backup may not be completely on disk. " + os_last_error_str(), BackupServer::canSyncFs() ? LL_ERROR : LL_DEBUG);
+
+										if(BackupServer::canSyncFs())
+											vhdfile_err = true;
 									}
 
-									sync_f.reset(Server->openFile(os_file_prefix(imagefn + ".sync"), MODE_WRITE));
+									if(!vhdfile_err)
+									{
+										sync_f.reset(Server->openFile(os_file_prefix(imagefn + ".sync"), MODE_WRITE));
+
+										if (sync_f.get() != NULL
+											&& !sync_f->Sync())
+										{
+											ServerLogger::Log(logid, "Error syncing sync file to disk. " + os_last_error_str(), LL_ERROR);
+											vhdfile_err = true;
+										}
+									}									
 								}
 
-								int64 image_size = t_file->RealSize();
 								db->BeginWriteTransaction();
 								backup_dao->setImageSize(image_size, backupid);
 								backup_dao->addImageSizeToClient(clientid, image_size);
 								if(!vhdfile_err)
 								{
 									if ( dependencies.empty()
-										 && set_complete )
+										 && set_complete 
+										&& sync_f.get()!=NULL)
 									{
 										backup_dao->setImageBackupComplete(backupid);
 									}
@@ -1424,13 +1589,15 @@ bool ImageBackup::doImage(const std::string &pLetter, const std::string &pParent
 									}
 									else
 									{
-										ServerLogger::Log(logid, "Error creating sync file at " + imagefn + ".sync", LL_WARNING);
+										ServerLogger::Log(logid, "Error creating sync file at " + imagefn + ".sync", LL_ERROR);
+										vhdfile_err = true;
 									}
 									db->EndTransaction();
 
-									if (image_file_format == image_file_format_cowraw)
+									if (image_file_format == image_file_format_cowraw
+										&& BackupServer::getSnapshotMethod(true)!= BackupServer::ESnapshotMethod_None)
 									{
-										if(!SnapshotHelper::makeReadonly(clientname, backuppath_single))
+										if(!SnapshotHelper::makeReadonly(true, clientname, backuppath_single))
 										{
 											ServerLogger::Log(logid, "Making image backup snapshot read only failed", LL_WARNING);
 										}
@@ -1440,8 +1607,6 @@ bool ImageBackup::doImage(const std::string &pLetter, const std::string &pParent
 								{
 									db->EndTransaction();
 								}
-								
-								Server->destroy(t_file);
 							}
 
 							running_updater->stop();
@@ -1583,6 +1748,7 @@ bool ImageBackup::doImage(const std::string &pLetter, const std::string &pParent
 								nextblock = updateNextblock(nextblock, vhdblock+vhd_blocksize, &shactx, zeroblockdata.data(), has_parent,
 									hashfile, parenthashfile, blocksize, mbr_offset, vhd_blocksize, warned_about_parenthashfile_error,
 									vhdblock, vhdfile, 0);
+								off += sizeof(int64);
 							}
 							else
 							{
@@ -1590,6 +1756,22 @@ bool ImageBackup::doImage(const std::string &pLetter, const std::string &pParent
 							}
 							currblock=-1;
 						}
+						else if (currblock == -128) //Number of skipped blocks
+						{
+							if (r - off >= 2 * sizeof(int64))
+							{
+								int64 zero_skipped;
+								memcpy(&zero_skipped, &buffer[off + sizeof(int64)], sizeof(int64));
+								zero_skipped = little_endian(zero_skipped);
+								numblocks += zero_skipped;
+								off += 2*sizeof(int64);
+							}
+							else
+							{
+								accum = true;
+							}
+							currblock = -1;
+						}
 						else if(currblock<0
 							|| currblock>totalblocks)
 						{
@@ -1926,9 +2108,14 @@ std::string ImageBackup::constructImagePath(const std::string &letter, std::stri
 	{
 		imgpath+=".raw";
 		create_folder = false;
-		if (full_backup)
+		if (BackupServer::getSnapshotMethod(true) == BackupServer::ESnapshotMethod_None
+			&& BackupServer::canReflink())
 		{
-			if (BackupServer::getSnapshotMethod() == BackupServer::ESnapshotMethod_Zfs)
+			create_folder = true;
+		}
+		else if (full_backup)
+		{
+			if (BackupServer::getSnapshotMethod(true) == BackupServer::ESnapshotMethod_Zfs)
 			{
 				std::auto_ptr<IFile> touch_f(Server->openFile(image_folder, MODE_WRITE));
 				if (touch_f.get()==NULL)
@@ -1939,17 +2126,21 @@ std::string ImageBackup::constructImagePath(const std::string &letter, std::stri
 				touch_f->Sync();
 			}
 
-			if (!SnapshotHelper::createEmptyFilesystem(clientname, backuppath_single))
+			std::string errmsg;
+			if (!SnapshotHelper::createEmptyFilesystem(true, clientname, backuppath_single, errmsg))
 			{
-				if (BackupServer::getSnapshotMethod() == BackupServer::ESnapshotMethod_Zfs)
+				if (BackupServer::getSnapshotMethod(true) == BackupServer::ESnapshotMethod_Zfs)
 				{
 					Server->deleteFile(image_folder);
 				}
+				errmsg = trim(errmsg);
+				ServerLogger::Log(logid, "Error creating empty image subvolume. "
+					+ (errmsg.empty() ? "" : ("\"" + errmsg + "\"")), LL_ERROR);
 				return std::string();
 			}
-			else if (BackupServer::getSnapshotMethod() == BackupServer::ESnapshotMethod_Zfs)
+			else if (BackupServer::getSnapshotMethod(true) == BackupServer::ESnapshotMethod_Zfs)
 			{
-				std::string mountpoint = SnapshotHelper::getMountpoint(clientname, backuppath_single);
+				std::string mountpoint = SnapshotHelper::getMountpoint(true, clientname, backuppath_single);
 				if (mountpoint.empty())
 				{
 					ServerLogger::Log(logid, "Could not find mountpoint of snapshot of client " + clientname + " path " + backuppath_single, LL_ERROR);
@@ -1988,61 +2179,86 @@ std::string ImageBackup::constructImagePath(const std::string &letter, std::stri
 		std::string parent_backuppath_single = ExtractFileName(ExtractFilePath(pParentvhd));
 		std::string parent_fn = ExtractFileName(pParentvhd);
 
-		if (BackupServer::getSnapshotMethod() == BackupServer::ESnapshotMethod_Zfs)
+		if (BackupServer::getSnapshotMethod(true) == BackupServer::ESnapshotMethod_None
+			&& BackupServer::canReflink() )
 		{
-			std::auto_ptr<IFile> touch_f(Server->openFile(image_folder, MODE_WRITE));
-			if(touch_f.get()==NULL)
+			if (!os_create_hardlink(imgpath, pParentvhd,
+				true, NULL))
 			{
-				ServerLogger::Log(logid, "Could not touch file " + image_folder + ". "+os_last_error_str(), LL_ERROR);
+				ServerLogger::Log(logid, "Could not reflink \"" + pParentvhd + "\" to \""
+					+ imgpath + "\". "+os_last_error_str(), LL_ERROR);
 				return std::string();
 			}
-			touch_f->Sync();
-		}
 
-		ServerLogger::Log(logid, "Creating writable snapshot of previous image backup...", LL_INFO);
-		if (!SnapshotHelper::snapshotFileSystem(clientname, parent_backuppath_single, backuppath_single))
-		{
-			ServerLogger::Log(logid, "Could not create snapshot of previous image backup at " + parent_backuppath_single, LL_ERROR);
-			if (BackupServer::getSnapshotMethod() == BackupServer::ESnapshotMethod_Zfs)
+			if (!os_create_hardlink(imgpath+".bitmap", pParentvhd+".bitmap",
+				true, NULL))
 			{
-				Server->deleteFile(image_folder);
+				ServerLogger::Log(logid, "Could not reflink \"" + pParentvhd + ".bitmap\" to \""
+					+ imgpath + ".bitmap\". " + os_last_error_str(), LL_ERROR);
+				return std::string();
 			}
-			return std::string();
 		}
 		else
 		{
-			if (BackupServer::getSnapshotMethod() == BackupServer::ESnapshotMethod_Zfs)
+			if (BackupServer::getSnapshotMethod(true) == BackupServer::ESnapshotMethod_Zfs)
 			{
-				std::string mountpoint = SnapshotHelper::getMountpoint(clientname, backuppath_single);
-				if (mountpoint.empty())
+				std::auto_ptr<IFile> touch_f(Server->openFile(image_folder, MODE_WRITE));
+				if (touch_f.get() == NULL)
 				{
-					ServerLogger::Log(logid, "Could not find mountpoint of snapshot of client " + clientname+ " path "+ backuppath_single, LL_ERROR);
-					return std::string();
-				}
-
-				if (!os_link_symbolic(mountpoint, image_folder+"_new"))
-				{
-					ServerLogger::Log(logid, "Could create symlink to mountpoint at " + image_folder + " to " + mountpoint+". "+os_last_error_str(), LL_ERROR);
-					return std::string();
-				}
-
-				if (!os_rename_file(image_folder + "_new", image_folder))
-				{
-					ServerLogger::Log(logid, "Could rename symlink at " + image_folder + "_new to " + image_folder + ". " + os_last_error_str(), LL_ERROR);
+					ServerLogger::Log(logid, "Could not touch file " + image_folder + ". " + os_last_error_str(), LL_ERROR);
 					return std::string();
 				}
+				touch_f->Sync();
 			}
 
-			Server->deleteFile(image_folder + os_file_sep() + parent_fn + ".hash");
-			Server->deleteFile(image_folder + os_file_sep() + parent_fn + ".cbitmap");
-			Server->deleteFile(image_folder + os_file_sep() + parent_fn + ".mbr");
-			Server->deleteFile(image_folder + os_file_sep() + parent_fn + ".sync");
-			os_rename_file(image_folder + os_file_sep() + parent_fn + ".bitmap", imgpath+".bitmap");
-			if (!os_rename_file(image_folder + os_file_sep() + parent_fn, imgpath))
+			ServerLogger::Log(logid, "Creating writable snapshot of previous image backup...", LL_INFO);
+			std::string errmsg;
+			if (!SnapshotHelper::snapshotFileSystem(true, clientname, parent_backuppath_single, backuppath_single, errmsg))
 			{
-				ServerLogger::Log(logid, "Error renaming in snapshot (\"" + image_folder + os_file_sep() + parent_fn + "\" to \""+imgpath+"\")", LL_ERROR);
+				errmsg = trim(errmsg);
+				ServerLogger::Log(logid, "Could not create snapshot of previous image backup at " + parent_backuppath_single
+					+ (errmsg.empty() ? "" : (" \"" + errmsg + "\"")), LL_ERROR);
+				if (BackupServer::getSnapshotMethod(true) == BackupServer::ESnapshotMethod_Zfs)
+				{
+					Server->deleteFile(image_folder);
+				}
 				return std::string();
 			}
+			else
+			{
+				if (BackupServer::getSnapshotMethod(true) == BackupServer::ESnapshotMethod_Zfs)
+				{
+					std::string mountpoint = SnapshotHelper::getMountpoint(true, clientname, backuppath_single);
+					if (mountpoint.empty())
+					{
+						ServerLogger::Log(logid, "Could not find mountpoint of snapshot of client " + clientname + " path " + backuppath_single, LL_ERROR);
+						return std::string();
+					}
+
+					if (!os_link_symbolic(mountpoint, image_folder + "_new"))
+					{
+						ServerLogger::Log(logid, "Could create symlink to mountpoint at " + image_folder + " to " + mountpoint + ". " + os_last_error_str(), LL_ERROR);
+						return std::string();
+					}
+
+					if (!os_rename_file(image_folder + "_new", image_folder))
+					{
+						ServerLogger::Log(logid, "Could rename symlink at " + image_folder + "_new to " + image_folder + ". " + os_last_error_str(), LL_ERROR);
+						return std::string();
+					}
+				}
+
+				Server->deleteFile(image_folder + os_file_sep() + parent_fn + ".hash");
+				Server->deleteFile(image_folder + os_file_sep() + parent_fn + ".cbitmap");
+				Server->deleteFile(image_folder + os_file_sep() + parent_fn + ".mbr");
+				Server->deleteFile(image_folder + os_file_sep() + parent_fn + ".sync");
+				os_rename_file(image_folder + os_file_sep() + parent_fn + ".bitmap", imgpath + ".bitmap");
+				if (!os_rename_file(image_folder + os_file_sep() + parent_fn, imgpath))
+				{
+					ServerLogger::Log(logid, "Error renaming in snapshot (\"" + image_folder + os_file_sep() + parent_fn + "\" to \"" + imgpath + "\")", LL_ERROR);
+					return std::string();
+				}
+			}
 		}
 	}
 
@@ -2090,10 +2306,30 @@ SBackup ImageBackup::getLastImage(const std::string &letter, bool incr)
 	}
 }
 
-std::string ImageBackup::getMBR(const std::string &dl, bool& fatal_error)
+std::string ImageBackup::getMBR(const std::string &dl, const std::string& disk_path,
+	bool image_full, int64 snapshot_id, bool& fatal_error, std::string& loadfn)
 {
 	fatal_error = true;
-	std::string ret=client_main->sendClientMessage("MBR driveletter="+dl, "Getting MBR for drive "+dl+" failed", 10000);
+
+	std::string params = "driveletter=" + EscapeParamString(dl);
+
+	if (!clientsubname.empty())
+	{
+		params += "&clientsubname=" + EscapeParamString(clientsubname);
+	}
+
+	params += "&disk_path=" + EscapeParamString(disk_path);
+
+	if (snapshot_id != 0)
+	{
+		params += "&shadowid=" + convert(snapshot_id);
+	}
+
+	params += std::string("&image_full=") + (image_full ? "1" : "0");
+	params += "&running_jobs=" + convert(ServerStatus::numRunningJobs(clientname));
+	params += "&token=" + EscapeParamString(server_token);
+
+	std::string ret=client_main->sendClientMessage("MBR "+ params, "Getting MBR for drive "+dl+" failed", 10000);
 	CRData r(&ret);
 	char b;
 	if(r.getChar(&b) && b==1 )
@@ -2101,10 +2337,19 @@ std::string ImageBackup::getMBR(const std::string &dl, bool& fatal_error)
 		char ver;
 		if(r.getChar(&ver) )
 		{
-			if(ver!=0 && ver!=1)
+			std::string zipfn;
+			if(ver!=0 && ver!=1
+				&& ver!=100)
 			{
 				ServerLogger::Log(logid, "MBR version "+convert((int)ver)+" is not supported by this server", LL_ERROR);
 			}
+			else if (ver == 100
+				&& r.getStr2(&zipfn))
+			{
+				ServerLogger::Log(logid, "Loading ZIP metadata from "+zipfn, LL_INFO);
+				loadfn = zipfn;
+				return ret.substr(0,2);
+			}
 			else
 			{
 				CRData r2(&ret);
