@@ -1,5 +1,8 @@
 #include "ImageMount.h"
 #include "../Interface/Server.h"
+#include "../Interface/File.h"
+#include "../fsimageplugin/IFSImageFactory.h"
+#include "../fsimageplugin/IVHDFile.h"
 #include "dao/ServerBackupDao.h"
 #include "database.h"
 #include "../urbackupcommon/os_functions.h"
@@ -10,21 +13,25 @@
 #include <Windows.h>
 #include <Subauth.h>
 #include <imdisk.h>
-bool os_link_symbolic_junctions_raw(const std::string &target, const std::string &lname);
+bool os_link_symbolic_junctions_raw(const std::string &target, const std::string &lname, void* transaction);
+#else
+#include <errno.h>
 #endif
 
-std::map<int, size_t> ImageMount::mounted_images;
+std::map<ImageMount::SMountId, size_t> ImageMount::mounted_images;
 IMutex* ImageMount::mounted_images_mutex;
 std::set<int> ImageMount::locked_images;
 IMutex* ImageMount::mount_processes_mutex;
-std::map<int, THREADPOOL_TICKET> ImageMount::mount_processes;
+std::map<ImageMount::SMountId, THREADPOOL_TICKET> ImageMount::mount_processes;
+
+extern IFSImageFactory *image_fak;
 
 namespace
 {
 #ifdef _WIN32
 	const std::string alt_mount_path = "C:\\UrBackupMounts";
 
-	bool os_mount_image(const std::string& path, int backupid)
+	bool os_mount_image(const std::string& path, int backupid, int partnum, IFSImageFactory::SPartition partition, std::string& errmsg)
 	{
 		if (path.size() <= 1)
 		{
@@ -33,6 +40,14 @@ namespace
 
 		std::wstring filename = Server->ConvertToWchar(path);
 
+		std::string mpath = "contents";
+
+		if (partition.offset >= 0)
+		{
+			filename += Server->ConvertToWchar(":" + convert(partition.offset) + "-" + convert(partition.length));
+			mpath += convert(partnum);
+		}
+
 		std::vector<char> create_data_buf(sizeof(IMDISK_CREATE_DATA) + filename.size() * sizeof(wchar_t));
 		IMDISK_CREATE_DATA* create_data = reinterpret_cast<IMDISK_CREATE_DATA*>(create_data_buf.data());
 
@@ -52,7 +67,8 @@ namespace
 
 		if (hDriver == INVALID_HANDLE_VALUE)
 		{
-			Server->Log("Error communicating with ImDisk driver. " + os_last_error_str(), LL_ERROR);
+			errmsg = "Error communicating with ImDisk driver. " + os_last_error_str();
+			Server->Log(errmsg, LL_ERROR);
 			return false;
 		}
 
@@ -62,7 +78,8 @@ namespace
 			create_data, buffer_size, create_data,
 			buffer_size, &bytes_returned, NULL))
 		{
-			Server->Log("Error creating ImDisk device. " + os_last_error_str(), LL_ERROR);
+			errmsg = "Error creating ImDisk device. " + os_last_error_str();
+			Server->Log(errmsg, LL_ERROR);
 			return false;
 		}
 
@@ -71,9 +88,9 @@ namespace
 		std::string device_path = Server->ConvertFromWchar(IMDISK_DEVICE_BASE_NAME) +
 			convert(static_cast<int64>(create_data->DeviceNumber)) + "\\";
 
-		std::string mountpoint = ExtractFilePath(path) + os_file_sep() + "contents";
+		std::string mountpoint = ExtractFilePath(path) + os_file_sep() + mpath;
 
-		if (!os_link_symbolic_junctions_raw(device_path, mountpoint))
+		if (!os_link_symbolic_junctions_raw(device_path, mountpoint, NULL))
 		{
 			Server->Log("Error creating junction on ImDisk mountpoint at \"" + mountpoint + "\". " + os_last_error_str(), LL_WARNING);
 
@@ -82,7 +99,8 @@ namespace
 			if (!((os_directory_exists(alt_mount_path) || os_create_dir_recursive(alt_mount_path))
 				&& os_link_symbolic(device_path, mountpoint)))
 			{
-				Server->Log("Error creating junction on ImDisk mountpoint. " + os_last_error_str(), LL_ERROR);
+				errmsg = "Error creating junction on ImDisk mountpoint. " + os_last_error_str();
+				Server->Log(errmsg, LL_ERROR);
 				return false;
 			}
 		}
@@ -96,16 +114,25 @@ namespace
 			Server->wait(1000);
 		}
 
-		return true;
+		if (!getFiles(mountpoint, &has_error).empty()
+			|| !has_error)
+		{
+			return true;
+		}
+
+		errmsg = "Timeout while mounting volume image";
+
+		return false;
 	}
 
-	bool os_unmount_image(const std::string& mountpoint, const std::string& path, int backupid)
+	bool os_unmount_image(const std::string& mountpoint, const std::string& path, int backupid, int partnum, std::string& errmsg)
 	{
 		std::string target;
 		if (!os_get_symlink_target(mountpoint, target)
 			|| target.empty())
 		{
-			Server->Log("Error getting mountpoint target. " + os_last_error_str(), LL_ERROR);
+			errmsg = "Error getting mountpoint target. " + os_last_error_str();
+			Server->Log(errmsg, LL_ERROR);
 			return false;
 		}
 
@@ -124,7 +151,8 @@ namespace
 
 		if (hDevice == INVALID_HANDLE_VALUE)
 		{
-			Server->Log("Cannot open ImDisk device " + target + ". " + os_last_error_str(), LL_WARNING);
+			errmsg = "Cannot open ImDisk device " + target + ". " + os_last_error_str();
+			Server->Log(errmsg, LL_WARNING);
 			os_remove_symlink_dir(mountpoint);
 			return false;
 		}
@@ -160,7 +188,8 @@ namespace
 			if (hDriver == INVALID_HANDLE_VALUE)
 			{
 				CloseHandle(hDevice);
-				Server->Log("Error communicating with ImDisk driver (2). " + os_last_error_str(), LL_ERROR);
+				errmsg = "Error communicating with ImDisk driver (2). " + os_last_error_str();
+				Server->Log(errmsg, LL_ERROR);
 				os_remove_symlink_dir(mountpoint);
 				return false;
 			}
@@ -170,7 +199,8 @@ namespace
 				&deviceNumber, sizeof(deviceNumber), NULL,
 				0, &bytes_returned, NULL))
 			{
-				Server->Log("Error removing ImDisk device. " + os_last_error_str(), LL_ERROR);
+				errmsg = "Error removing ImDisk device. " + os_last_error_str();
+				Server->Log(errmsg, LL_ERROR);
 				CloseHandle(hDevice);
 				CloseHandle(hDriver);
 				os_remove_symlink_dir(mountpoint);
@@ -186,7 +216,7 @@ namespace
 		return os_remove_symlink_dir(mountpoint);
 	}
 #else
-	bool image_helper_action(const std::string& path, int backupid, const std::string& action)
+	bool image_helper_action(const std::string& path, int backupid, const std::string& action, int partnum, IFSImageFactory::SPartition partition, std::string& errmsg)
 	{
 		IDatabase* db = Server->getDatabase(Server->getThreadID(), URBACKUPDB_SERVER);
 		ServerBackupDao backup_dao(db);
@@ -206,17 +236,35 @@ namespace
 			mount_helper = "urbackup_mount_helper";
 		}
 
-		return system((mount_helper + " " + action + " \"" + clientname.value + "\" \"" + foldername + "\" \"" + imagename + "\"").c_str()) == 0;
+		std::string cmd = mount_helper + " " + action + " \"" + clientname.value 
+			+ "\" \"" + foldername + "\" \"" + imagename + "\"";
+
+		if (partnum >= 0)
+		{
+			cmd += " " + convert(partnum);
+		}
+
+		if (partition.offset >= 0)
+		{
+			cmd += " " + convert(partition.offset) + " " + convert(partition.length);
+		}
+		cmd += " 2>&1";
+		int rc = os_popen(cmd.c_str(), errmsg);
+		if (rc != 0)
+		{
+			Server->Log("Image mounting failed: "+errmsg, LL_ERROR);
+		}
+		return rc == 0;
 	}
 
-	bool os_mount_image(const std::string& path, int backupid)
+	bool os_mount_image(const std::string& path, int backupid, int partnum, IFSImageFactory::SPartition partition, std::string& errmsg)
 	{
-		return image_helper_action(path, backupid, "mount");
+		return image_helper_action(path, backupid, "mount", partnum, partition, errmsg);
 	}
 
-	bool os_unmount_image(const std::string& mountpoint, const std::string& path, int backupid)
+	bool os_unmount_image(const std::string& mountpoint, const std::string& path, int backupid, int partnum, std::string& errmsg)
 	{
-		return image_helper_action(path, backupid, "umount");
+		return image_helper_action(path, backupid, "umount", partnum, IFSImageFactory::SPartition(), errmsg);
 	}
 #endif
 
@@ -248,15 +296,22 @@ namespace
 		}
 	};
 
-	bool unmount_image(ServerBackupDao& backup_dao, int backupid)
+	bool unmount_image(ServerBackupDao& backup_dao, int backupid, int partition, std::string& errmsg)
 	{
-		ServerBackupDao::SMountedImage image_inf = backup_dao.getMountedImage(backupid);
+		ServerBackupDao::SMountedImage image_inf = backup_dao.getMountedImage(backupid, partition);
 		if (!image_inf.exists)
 		{
 			return false;
 		}
 
-		std::string mountpoint = ExtractFilePath(image_inf.path) + os_file_sep() + "contents";
+		std::string mpath = "contents";
+
+		if (partition >= 0)
+		{
+			mpath += convert(partition);
+		}
+
+		std::string mountpoint = ExtractFilePath(image_inf.path) + os_file_sep() + mpath;
 
 #ifdef _WIN32
 		if (!os_directory_exists(mountpoint))
@@ -269,19 +324,25 @@ namespace
 			return false;
 		}
 #else
-		if (!os_directory_exists(mountpoint))
+		if (!os_directory_exists(mountpoint) || errno == EACCES || errno == ENOTCONN)
 		{
 			mountpoint = ExtractFilePath(image_inf.path) + "_mnt";
+
+			if (partition >= 0)
+			{
+				mountpoint += convert(partition);
+			}
 		}
 
-		if (!os_directory_exists(mountpoint))
+		if (!os_directory_exists(mountpoint) && errno != EACCES && errno != ENOTCONN)
 		{
 			return false;
 		}
 #endif
 
-		return os_unmount_image(mountpoint, image_inf.path, backupid);
+		return os_unmount_image(mountpoint, image_inf.path, backupid, partition, errmsg);
 	}
+
 }
 
 void ImageMount::operator()()
@@ -310,30 +371,36 @@ void ImageMount::operator()()
 			IScopedLock lock(mounted_images_mutex);
 			for (size_t i = 0; i < old_mounted_images.size(); ++i)
 			{
-				if (mounted_images.find(old_mounted_images[i].id) == mounted_images.end()
-					&& locked_images.find(old_mounted_images[i].id) == locked_images.end())
+				if (mounted_images.find(SMountId(old_mounted_images[i].backupid, old_mounted_images[i].partition)) == mounted_images.end()
+					&& locked_images.find(old_mounted_images[i].backupid) == locked_images.end())
 				{
-					locked_images.insert(old_mounted_images[i].id);
+					locked_images.insert(old_mounted_images[i].backupid);
 					lock.relock(NULL);
 
 					int64 passed_time_s = Server->getTimeSeconds() - old_mounted_images[i].mounttime;
-					Server->Log("Unmounting mounted image backup id " + convert(old_mounted_images[i].id) +
-						" path \"" + old_mounted_images[i].path + "\" mounted " + PrettyPrintTime(passed_time_s * 1000) + " ago", LL_INFO);
-					unmount_image(backup_dao, old_mounted_images[i].id);
-					backup_dao.setImageUnmounted(old_mounted_images[i].id);
+					Server->Log("Unmounting mounted image backup id " + convert(old_mounted_images[i].backupid) +
+						" partition "+convert(old_mounted_images[i].partition)+" path \"" + old_mounted_images[i].path + "\" mounted " + PrettyPrintTime(passed_time_s * 1000) + " ago", LL_INFO);
+					std::string errmsg;
+					if (!unmount_image(backup_dao, old_mounted_images[i].backupid, old_mounted_images[i].partition, errmsg) )
+					{
+						Server->Log("Unmounting image backup id " + convert(old_mounted_images[i].backupid) +
+							" path \"" + old_mounted_images[i].path + "\" mounted " + PrettyPrintTime(passed_time_s * 1000)
+							+ " ago failed: " + errmsg, LL_ERROR);
+					}
+					backup_dao.delImageMounted(old_mounted_images[i].id);
 
 					lock.relock(mounted_images_mutex);
-					locked_images.erase(locked_images.find(old_mounted_images[i].id));
+					locked_images.erase(locked_images.find(old_mounted_images[i].backupid));
 				}
 			}
 		}
 
 		{
 			IScopedLock lock(mount_processes_mutex);
-			for (std::map<int, THREADPOOL_TICKET>::iterator it = mount_processes.begin();
+			for (std::map<SMountId, THREADPOOL_TICKET>::iterator it = mount_processes.begin();
 				it != mount_processes.end();)
 			{
-				std::map<int, THREADPOOL_TICKET>::iterator it_curr = it++;
+				std::map<SMountId, THREADPOOL_TICKET>::iterator it_curr = it++;
 				if (Server->getThreadPool()->waitFor(it_curr->second))
 				{
 					mount_processes.erase(it_curr);
@@ -343,7 +410,7 @@ void ImageMount::operator()()
 	}
 }
 
-bool ImageMount::mount_image(int backupid, ScopedMountedImage& mounted_image, int64 timeoutms, bool& has_timeout)
+bool ImageMount::mount_image(int backupid, int partition, ScopedMountedImage& mounted_image, int64 timeoutms, bool& has_timeout, std::string& errmsg)
 {
 	has_timeout = false;
 
@@ -353,7 +420,7 @@ bool ImageMount::mount_image(int backupid, ScopedMountedImage& mounted_image, in
 		has_timeout = true;
 		return false;
 	}
-	return mount_image_int(backupid, mounted_image, timeoutms, has_timeout);
+	return mount_image_int(backupid, partition, mounted_image, timeoutms, has_timeout, errmsg);
 }
 
 namespace
@@ -361,25 +428,28 @@ namespace
 	class MountImageThread : public IThread
 	{
 		int backupid;
+		std::string& errmsg;
+		int partition;
 	public:
-		MountImageThread(int backupid)
-			: backupid(backupid)
+		MountImageThread(int backupid, int partition, std::string& errmsg)
+			: backupid(backupid), partition(partition), errmsg(errmsg)
 		{
 
 		}
 
 		void operator()()
 		{
-			ImageMount::mount_image_thread(backupid);
+			ImageMount::mount_image_thread(backupid, partition, errmsg);
 			delete this;
 		}
 	};
 }
 
-bool ImageMount::mount_image_int(int backupid, ScopedMountedImage& mounted_image, int64 timeoutms, bool& has_timeout)
+bool ImageMount::mount_image_int(int backupid, int partition, ScopedMountedImage& mounted_image,
+	int64 timeoutms, bool& has_timeout, std::string& errmsg)
 {
 	IScopedLock lock(mount_processes_mutex);
-	std::map<int, THREADPOOL_TICKET>::iterator it = mount_processes.find(backupid);
+	std::map<SMountId, THREADPOOL_TICKET>::iterator it = mount_processes.find(SMountId(backupid, partition));
 	THREADPOOL_TICKET ticket;
 	if (it != mount_processes.end())
 	{
@@ -388,15 +458,15 @@ bool ImageMount::mount_image_int(int backupid, ScopedMountedImage& mounted_image
 	}
 	else
 	{
-		ticket = Server->getThreadPool()->execute(new MountImageThread(backupid), "mnt image");
-		mount_processes.insert(std::make_pair(backupid, ticket));
+		ticket = Server->getThreadPool()->execute(new MountImageThread(backupid, partition, errmsg), "mnt image");
+		mount_processes.insert(std::make_pair(SMountId(backupid, partition), ticket));
 		lock.relock(NULL);
 	}
 
 	if (Server->getThreadPool()->waitFor(ticket, static_cast<int>(timeoutms)))
 	{
 		IScopedLock lock(mount_processes_mutex);
-		std::map<int, THREADPOOL_TICKET>::iterator it = mount_processes.find(backupid);
+		std::map<SMountId, THREADPOOL_TICKET>::iterator it = mount_processes.find(SMountId(backupid, partition) );
 		if (it != mount_processes.end()
 			&& it->second == ticket)
 		{
@@ -405,7 +475,7 @@ bool ImageMount::mount_image_int(int backupid, ScopedMountedImage& mounted_image
 		lock.relock(NULL);
 		IDatabase* db = Server->getDatabase(Server->getThreadID(), URBACKUPDB_SERVER);
 		ServerBackupDao backup_dao(db);
-		ServerBackupDao::SMountedImage image_inf = backup_dao.getMountedImage(backupid);
+		ServerBackupDao::SMountedImage image_inf = backup_dao.getMountedImage(backupid, partition);
 		if (image_inf.mounttime != 0)
 		{
 			mounted_image.reset(backupid);
@@ -419,40 +489,44 @@ bool ImageMount::mount_image_int(int backupid, ScopedMountedImage& mounted_image
 	}
 }
 
-std::string ImageMount::get_mount_path(int backupid, bool do_mount, ScopedMountedImage& mounted_image, 
-	int64 timeoutms, bool& has_timeout)
+std::string ImageMount::get_mount_path(int backupid, int clientid, int partition, bool do_mount, ScopedMountedImage& mounted_image,
+	int64 timeoutms, bool& has_timeout, std::string& errmsg)
 {
 	has_timeout = false;
 
+	IDatabase* db = Server->getDatabase(Server->getThreadID(), URBACKUPDB_SERVER);
+	ServerBackupDao backup_dao(db);
+
+	ServerBackupDao::SMountedImage image_inf = backup_dao.getImageInfo(backupid);
+
+	if (!image_inf.exists
+		|| image_inf.clientid != clientid)
+	{
+		return std::string();
+	}
+
 	ScopedLockImage lock_image(backupid, timeoutms);
 	if (!lock_image.has_lock())
 	{
 		has_timeout = true;
 		return std::string();
-	}
+	}		
 
-	IDatabase* db = Server->getDatabase(Server->getThreadID(), URBACKUPDB_SERVER);
-	ServerBackupDao backup_dao(db);
-
-	ServerBackupDao::SMountedImage image_inf = backup_dao.getMountedImage(backupid);
-	if (!image_inf.exists)
-	{
-		return std::string();
-	}
+	image_inf = backup_dao.getMountedImage(backupid, partition);
 
 	bool has_mount_process = false;
 	{
 		IScopedLock lock(mount_processes_mutex);
-		std::map<int, THREADPOOL_TICKET>::iterator it = mount_processes.find(backupid);
+		std::map<SMountId, THREADPOOL_TICKET>::iterator it = mount_processes.find(SMountId(backupid, partition));
 		has_mount_process = it != mount_processes.end();
 	}
 
-	if (image_inf.mounttime == 0
+	if ( (!image_inf.exists || image_inf.mounttime == 0)
 		|| has_mount_process)
 	{
 		if (do_mount)
 		{
-			if (!mount_image_int(backupid, mounted_image, timeoutms, has_timeout))
+			if (!mount_image_int(backupid, partition, mounted_image, timeoutms, has_timeout, errmsg))
 			{
 				return std::string();
 			}
@@ -463,12 +537,28 @@ std::string ImageMount::get_mount_path(int backupid, bool do_mount, ScopedMounte
 		}
 	}
 
-	std::string ret = ExtractFilePath(image_inf.path) + os_file_sep() + "contents";
+	std::string mpath = "contents";
+
+	if (partition >= 0)
+	{
+		mpath += convert(partition);
+	}
+
+	ServerBackupDao::SMountedImage image_path;
+	if (image_inf.exists)
+		image_path = image_inf;
+	else
+		image_path = backup_dao.getImageInfo(backupid);
+
+	std::string ret = ExtractFilePath(image_path.path) + os_file_sep() + mpath;
 
 	if (os_directory_exists(ret))
 	{
 		mounted_image.reset(backupid);
-		backup_dao.setImageMounted(backupid);
+		if (image_inf.exists)
+			backup_dao.updateImageMounted(image_inf.id);
+		else
+			backup_dao.addImageMounted(backupid, partition);
 		return ret;
 	}
 
@@ -478,16 +568,27 @@ std::string ImageMount::get_mount_path(int backupid, bool do_mount, ScopedMounte
 	if (os_directory_exists(ret))
 	{
 		mounted_image.reset(backupid);
-		backup_dao.setImageMounted(backupid);
+		if (image_inf.exists)
+			backup_dao.updateImageMounted(image_inf.id);
+		else
+			backup_dao.addImageMounted(backupid, partition);
 		return ret;
 	}
 #else
 	ret = ExtractFilePath(image_inf.path) + "_mnt";
 
+	if (partition >= 0)
+	{
+		ret += convert(partition);
+	}
+
 	if (os_directory_exists(ret))
 	{
 		mounted_image.reset(backupid);
-		backup_dao.setImageMounted(backupid);
+		if (image_inf.exists)
+			backup_dao.updateImageMounted(image_inf.id);
+		else
+			backup_dao.addImageMounted(backupid, partition);
 		return ret;
 	}
 #endif
@@ -495,22 +596,76 @@ std::string ImageMount::get_mount_path(int backupid, bool do_mount, ScopedMounte
 	return std::string();
 }
 
-void ImageMount::incrImageMounted(int backupid)
+bool ImageMount::unmount_images(int backupid)
+{
+	IDatabase* db = Server->getDatabase(Server->getThreadID(), URBACKUPDB_SERVER);
+	ServerBackupDao backup_dao(db);
+
+	std::vector<ServerBackupDao::SMountedImage> old_mounted_images =
+		backup_dao.getOldMountedImages(0);
+
+	IScopedLock lock(mounted_images_mutex);
+	for (size_t i = 0; i < old_mounted_images.size(); ++i)
+	{
+		if (old_mounted_images[i].backupid != backupid)
+		{
+			continue;
+		}
+
+		if (mounted_images.find(SMountId(old_mounted_images[i].backupid, old_mounted_images[i].partition)) != mounted_images.end())
+		{
+			Server->Log("Image id " + convert(old_mounted_images[i].backupid) +
+				" partition " + convert(old_mounted_images[i].partition)
+				+ " currently mounted", LL_INFO);
+			return false;
+		}
+
+		if (locked_images.find(old_mounted_images[i].backupid) != locked_images.end())
+		{
+			Server->Log("Image id " + convert(old_mounted_images[i].backupid) +
+				" partition " + convert(old_mounted_images[i].partition)
+				+ " currently locked", LL_INFO);
+			return false;
+		}
+
+		locked_images.insert(old_mounted_images[i].backupid);
+		lock.relock(NULL);
+
+		int64 passed_time_s = Server->getTimeSeconds() - old_mounted_images[i].mounttime;
+		Server->Log("Unmounting mounted image backup id " + convert(old_mounted_images[i].backupid) +
+			" partition " + convert(old_mounted_images[i].partition) + " path \"" + old_mounted_images[i].path + "\" mounted " + PrettyPrintTime(passed_time_s * 1000) + " ago (unmount_images)", LL_INFO);
+		std::string errmsg;
+		if (!unmount_image(backup_dao, old_mounted_images[i].backupid, old_mounted_images[i].partition, errmsg))
+		{
+			Server->Log("Unmounting image backup id " + convert(old_mounted_images[i].backupid) +
+				" path \"" + old_mounted_images[i].path + "\" mounted " + PrettyPrintTime(passed_time_s * 1000)
+				+ " ago failed: " + errmsg + " (unmount_images)", LL_ERROR);
+		}
+		backup_dao.delImageMounted(old_mounted_images[i].id);
+
+		lock.relock(mounted_images_mutex);
+		locked_images.erase(locked_images.find(old_mounted_images[i].backupid));
+	}
+
+	return true;
+}
+
+void ImageMount::incrImageMounted(int backupid, int partition)
 {
 	IScopedLock lock(mounted_images_mutex);
-	std::map<int, size_t>::iterator it = mounted_images.find(backupid);
+	std::map<SMountId, size_t>::iterator it = mounted_images.find(SMountId(backupid, partition) );
 	if (it == mounted_images.end())
 	{
 		ServerCleanupThread::lockImageFromCleanup(backupid);
 	}
 
-	++mounted_images[backupid];
+	++mounted_images[SMountId(backupid, partition)];
 }
 
-void ImageMount::decrImageMounted(int backupid)
+void ImageMount::decrImageMounted(int backupid, int partition)
 {
 	IScopedLock lock(mounted_images_mutex);
-	std::map<int, size_t>::iterator it = mounted_images.find(backupid);
+	std::map<SMountId, size_t>::iterator it = mounted_images.find(SMountId(backupid, partition) );
 	assert(it != mounted_images.end()
 		&& it->second > 0);
 	if (it == mounted_images.end())
@@ -557,22 +712,54 @@ void ImageMount::unlockImage(int backupid)
 	locked_images.erase(locked_images.find(backupid));
 }
 
-void ImageMount::mount_image_thread(int backupid)
+void ImageMount::mount_image_thread(int backupid, int partition, std::string& errmsg)
 {	
 	IDatabase* db = Server->getDatabase(Server->getThreadID(), URBACKUPDB_SERVER);
 	ServerBackupDao backup_dao(db);
 
-	ServerBackupDao::SMountedImage image_inf = backup_dao.getMountedImage(backupid);
-	if (!image_inf.exists)
+	ServerBackupDao::SMountedImage image_inf = backup_dao.getImageInfo(backupid);
+	if (!image_inf.exists
+		|| backup_dao.getMountedImage(backupid, partition).exists)
 	{
 		return;
 	}
 
-	backup_dao.setImageMounted(backupid);
+	IFSImageFactory::SPartition sel_part;
 
-	if (!os_mount_image(image_inf.path, backupid))
+	if (partition >= 0)
 	{
-		backup_dao.setImageUnmounted(backupid);
+		std::string ext = findextension(image_inf.path);
+		std::auto_ptr<IVHDFile> vhdfile;
+		if (ext == "raw")
+		{
+			vhdfile.reset(image_fak->createVHDFile(image_inf.path, true, 0, 2 * 1024 * 1024, false, IFSImageFactory::ImageFormat_RawCowFile));
+		}
+		else
+		{
+			vhdfile.reset(image_fak->createVHDFile(image_inf.path, true, 0));
+		}
+
+		if (vhdfile.get() != NULL)
+		{
+			if (!vhdfile->isOpen())
+				return;
+
+			bool gpt_style;
+			std::vector<IFSImageFactory::SPartition> partitions = image_fak->readPartitions(vhdfile.get(), 0, gpt_style);
+
+			if (partition < partitions.size())
+			{
+				sel_part = partitions[partition];
+			}
+		}
+	}
+
+	backup_dao.addImageMounted(backupid, partition);
+	int64 insert_id = db->getLastInsertID();
+
+	if (!os_mount_image(image_inf.path, backupid, partition, sel_part, errmsg))
+	{
+		backup_dao.delImageMounted(insert_id);
 		return;
 	}
 }
