@@ -36,6 +36,7 @@
 #include "FileMetadataDownloadThread.h"
 #include "database.h"
 #include <algorithm>
+#include "PhashLoad.h"
 
 extern std::string server_identity;
 
@@ -49,7 +50,7 @@ namespace
 IncrFileBackup::IncrFileBackup( ClientMain* client_main, int clientid, std::string clientname, std::string clientsubname, LogAction log_action,
 	int group, bool use_tmpfiles, std::string tmpfile_path, bool use_reflink, bool use_snapshots, std::string server_token, std::string details, bool scheduled)
 	: FileBackup(client_main, clientid, clientname, clientsubname, log_action, true, group, use_tmpfiles, tmpfile_path, use_reflink, use_snapshots, server_token, details, scheduled), 
-	intra_file_diffs(intra_file_diffs), hash_existing_mutex(NULL), filesdao(NULL), link_dao(NULL), link_journal_dao(NULL)
+	hash_existing_mutex(NULL), filesdao(NULL), link_dao(NULL), link_journal_dao(NULL)
 {
 
 }
@@ -127,6 +128,11 @@ bool IncrFileBackup::doFileBackup()
 	{
 		has_early_error=true;
 
+		if (no_backup_dirs)
+		{
+			backup_dao->updateClientNumIssues(ServerBackupDao::num_issues_no_backuppaths, clientid);
+		}
+
 		if(no_backup_dirs || connect_fail)
 		{
 			log_backup=false;
@@ -163,11 +169,11 @@ bool IncrFileBackup::doFileBackup()
 
 	Server->Log(clientname+": Connecting to client...", LL_DEBUG);
 	std::string identity = client_main->getIdentity();
-	FileClient fc(false, identity, client_main->getProtocolVersions().filesrv_protocol_version, client_main->isOnInternetConnection(), client_main, use_tmpfiles?NULL:client_main);
+	FileClient fc(false, identity, client_main->getProtocolVersions().filesrv_protocol_version, client_main->isOnInternetConnection(), client_main, use_tmpfiles?NULL:this);
 	std::auto_ptr<FileClientChunked> fc_chunked;
 	if(intra_file_diffs)
 	{
-		if(client_main->getClientChunkedFilesrvConnection(fc_chunked, server_settings.get(), 10000))
+		if(client_main->getClientChunkedFilesrvConnection(fc_chunked, server_settings.get(), this, 10000))
 		{
 			fc_chunked->setProgressLogCallback(this);
 			fc_chunked->setDestroyPipe(true);
@@ -276,6 +282,7 @@ bool IncrFileBackup::doFileBackup()
 		if(!client_main->isOnInternetConnection())
 		{
 			ServerLogger::Log(logid, "Error while calculating tree diff. Doing full backup.", LL_ERROR);
+			deleteBackup();
 			return doFullBackup();
 		}
 		else
@@ -286,20 +293,94 @@ bool IncrFileBackup::doFileBackup()
 		}
 	}
 
+
+	bool make_subvolume_readonly = false;
+	bool crossvolume_links = false;
 	if(use_snapshots)
 	{
-		ServerLogger::Log(logid, clientname+": Creating snapshot...", LL_INFO);
-		if(!SnapshotHelper::snapshotFileSystem(clientname, last.path, backuppath_single)
-			|| !SnapshotHelper::isSubvolume(clientname, backuppath_single) )
-		{
-			ServerLogger::Log(logid, "Creating new snapshot failed (Server error)", LL_WARNING);
+		make_subvolume_readonly = true;
+		crossvolume_links = true;
 
-			if(!SnapshotHelper::createEmptyFilesystem(clientname, backuppath_single) )
+		bool zfs_file = BackupServer::getSnapshotMethod(false) == BackupServer::ESnapshotMethod_ZfsFile;
+		if (zfs_file)
+		{
+			if (os_get_file_type(backuppath) != 0)
 			{
-				ServerLogger::Log(logid, "Creating empty filesystem failed (Server error)", LL_ERROR);
+				ServerLogger::Log(logid, "File/Directory at " + backuppath + " already exists.", LL_ERROR);
+				has_early_error = true;
+				allow_remove_backup_folder = false;
+				return false;
+			}
+
+			std::auto_ptr<IFile> touch_f(Server->openFile(backuppath, MODE_WRITE));
+			if (touch_f.get() == NULL)
+			{
+				ServerLogger::Log(logid, "Could not touch file " + backuppath + ". " + os_last_error_str(), LL_ERROR);
+				has_early_error = true;
+				return false;
+			}
+			touch_f->Sync();
+		}
+
+		ServerLogger::Log(logid, clientname+": Creating snapshot...", LL_INFO);
+		std::string errmsg;
+		if(!SnapshotHelper::snapshotFileSystem(false, clientname, last.path, backuppath_single, errmsg)
+			|| !SnapshotHelper::isSubvolume(false, clientname, backuppath_single) )
+		{
+			errmsg = trim(errmsg);
+			ServerLogger::Log(logid, "Creating new snapshot failed (Server error) "
+				+(errmsg.empty()?os_last_error_str(): ("\""+errmsg+"\"")), LL_WARNING);
+
+			if (SnapshotHelper::isSubvolume(false, clientname, backuppath_single))
+			{
+				if (zfs_file)
+				{
+					Server->deleteFile(backuppath);
+				}
+
+				ServerLogger::Log(logid, "Subvolume already exists.", LL_ERROR);
+				has_early_error = true;
+				allow_remove_backup_folder = false;
+				return false;
+			}
+
+			errmsg.clear();
+			if(!SnapshotHelper::createEmptyFilesystem(false, clientname, backuppath_single, errmsg) )
+			{
+				if (zfs_file)
+				{
+					Server->deleteFile(backuppath);
+				}
+
+				ServerLogger::Log(logid, "Creating empty filesystem failed (Server error) "
+					+ (errmsg.empty() ? os_last_error_str() : ("\"" + errmsg + "\"")), LL_ERROR);
 				has_early_error=true;
 				return false;
 			}
+			else if (zfs_file)
+			{
+				std::string mountpoint = SnapshotHelper::getMountpoint(false, clientname, backuppath_single);
+				if (mountpoint.empty())
+				{
+					ServerLogger::Log(logid, "Could not find mountpoint of snapshot of client " + clientname + " path " + backuppath_single, LL_ERROR);
+					has_early_error = true;
+					return false;
+				}
+
+				if (!os_link_symbolic(mountpoint, backuppath + "_new"))
+				{
+					ServerLogger::Log(logid, "Could create symlink to mountpoint at " + backuppath + " to " + mountpoint + ". " + os_last_error_str(), LL_ERROR);
+					has_early_error = true;
+					return false;
+				}
+
+				if (!os_rename_file(backuppath + "_new", backuppath))
+				{
+					ServerLogger::Log(logid, "Could rename symlink at " + backuppath + "_new to " + backuppath + ". " + os_last_error_str(), LL_ERROR);
+					has_early_error = true;
+					return false;
+				}
+			}
 
 			if(!os_create_dir(os_file_prefix(backuppath_hashes)))
 			{
@@ -312,6 +393,31 @@ bool IncrFileBackup::doFileBackup()
 		}
 		else
 		{
+			if (zfs_file)
+			{
+				std::string mountpoint = SnapshotHelper::getMountpoint(false, clientname, backuppath_single);
+				if (mountpoint.empty())
+				{
+					ServerLogger::Log(logid, "Could not find mountpoint of snapshot of client " + clientname + " path " + backuppath_single, LL_ERROR);
+					has_early_error = true;
+					return false;
+				}
+
+				if (!os_link_symbolic(mountpoint, backuppath + "_new"))
+				{
+					ServerLogger::Log(logid, "Could create symlink to mountpoint at " + backuppath + " to " + mountpoint + ". " + os_last_error_str(), LL_ERROR);
+					has_early_error = true;
+					return false;
+				}
+
+				if (!os_rename_file(backuppath + "_new", backuppath))
+				{
+					ServerLogger::Log(logid, "Could rename symlink at " + backuppath + "_new to " + backuppath + ". " + os_last_error_str(), LL_ERROR);
+					has_early_error = true;
+					return false;
+				}
+			}
+
 			Server->deleteFile(os_file_prefix(backuppath_hashes + os_file_sep() + sync_fn));
 			os_sync(backuppath_hashes);
 
@@ -323,8 +429,13 @@ bool IncrFileBackup::doFileBackup()
 			}
 		}
 	}
+	else if (BackupServer::canReflink()
+		&& !BackupServer::canHardlink())
+	{
+		crossvolume_links = true;
+	}
 
-	getTokenFile(fc, hashed_transfer);
+	getTokenFile(fc, hashed_transfer, false);
 
 	if(use_snapshots)
 	{
@@ -396,7 +507,7 @@ bool IncrFileBackup::doFileBackup()
 		use_tmpfiles, tmpfile_path, server_token, use_reflink,
 		backupid, r_incremental, hashpipe_prepare, client_main, client_main->getProtocolVersions().filesrv_protocol_version,
 		incremental_num, logid, with_hashes, shares_without_snapshot, with_sparse_hashing, metadata_download_thread.get(),
-		backup_with_components));
+		backup_with_components, filepath_corrections, max_file_id));
 
 	bool queue_downloads = client_main->getProtocolVersions().filesrv_protocol_version>2;
 
@@ -454,6 +565,7 @@ bool IncrFileBackup::doFileBackup()
 	std::stack<bool> dir_diff_stack;
 	std::stack<int64> dir_ids;
 	std::map<int64, int64> dir_end_ids;
+	bool phash_load_offline = false;
 
 	bool has_read_error = false;
 	while( (read=tmp_filelist->Read(buffer, 4096, &has_read_error))>0 )
@@ -492,6 +604,10 @@ bool IncrFileBackup::doFileBackup()
 								folder_files.pop();
 								dir_ids.pop();
 							}
+							else
+							{
+								max_file_id.setMaxPreProcessed(line);
+							}
 						}
 						else
 						{
@@ -643,10 +759,10 @@ bool IncrFileBackup::doFileBackup()
 							std::string srcpath=last_backuppath+local_curr_os_path;
 							std::string src_hashpath = last_backuppath_hashes+local_curr_os_path;
 							if(link_directory_pool(clientid, backuppath+local_curr_os_path, srcpath, dir_pool_path,
-								 BackupServer::isFilesystemTransactionEnabled(), link_dao, link_journal_dao) )
+								 BackupServer::isFilesystemTransactionEnabled(), link_dao, link_journal_dao, depth) )
 							{
 								if(link_directory_pool(clientid, backuppath_hashes + local_curr_os_path, src_hashpath, dir_pool_path,
-									BackupServer::isFilesystemTransactionEnabled(), link_dao, link_journal_dao) )
+									BackupServer::isFilesystemTransactionEnabled(), link_dao, link_journal_dao, depth) )
 								{
 									skip_dir_completely = 1;
 									dir_linked = true;
@@ -846,11 +962,11 @@ bool IncrFileBackup::doFileBackup()
 							
 							if( !dir_diff && !indirchange && curr_path!="/urbackup_backup_scripts")
 							{
-								if(!os_create_hardlink(os_file_prefix(metadata_fn), os_file_prefix(metadata_srcpath), use_snapshots, NULL))
+								if(!create_hardlink(os_file_prefix(metadata_fn), os_file_prefix(metadata_srcpath), crossvolume_links, NULL, NULL))
 								{
 									if(!copy_file(metadata_srcpath, metadata_fn))
 									{
-										if(client_main->handle_not_enough_space(metadata_fn))
+										if(client_main->handle_not_enough_space(metadata_fn, logid, *server_settings.get()))
 										{
 											if(!copy_file(metadata_srcpath, metadata_fn))
 											{
@@ -942,6 +1058,11 @@ bool IncrFileBackup::doFileBackup()
 					std::string local_curr_os_path=convertToOSPathFromFileClient(curr_os_path+"/"+osspecific_name);
 					std::string srcpath=last_backuppath+local_curr_os_path;
 
+					if (depth == 0)
+					{
+						server_download->addToQueueStartShadowcopy(cf.name);
+					}
+
 					if(!has_orig_path)
 					{
 						if (curr_orig_path != orig_sep)
@@ -970,6 +1091,27 @@ bool IncrFileBackup::doFileBackup()
 						{
 							curr_sha2 = base64_decode_dash(hash_it->second);
 						}
+
+						if (curr_sha2.empty()
+							&& phash_load.get() != NULL
+							&& !script_dir
+							&& extra_params.find("sym_target")==extra_params.end()
+							&& extra_params.find("special") == extra_params.end()
+							&& !phash_load_offline
+							&& extra_params.find("no_hash")==extra_params.end())
+						{
+							if (!phash_load->getHash(line, curr_sha2))
+							{
+								ServerLogger::Log(logid, "Error getting parallel hash for file \"" + cf.name + "\" line " + convert(line)+" (2)", LL_ERROR);
+								r_offline = true;
+								server_download->queueSkip();
+								phash_load_offline = true;
+							}
+							else
+							{
+								metadata.shahash = curr_sha2;
+							}
+						}
 					}
 
                     bool download_metadata=false;
@@ -1058,11 +1200,11 @@ bool IncrFileBackup::doFileBackup()
 					else if(!use_snapshots) //is not changed
 					{						
 						bool too_many_hardlinks;
-						bool b=os_create_hardlink(os_file_prefix(backuppath+local_curr_os_path), os_file_prefix(srcpath), use_snapshots, &too_many_hardlinks);
+						bool b=create_hardlink(os_file_prefix(backuppath+local_curr_os_path), os_file_prefix(srcpath), crossvolume_links, &too_many_hardlinks, NULL);
 
 						if(b)
 						{
-							b = os_create_hardlink(os_file_prefix(backuppath_hashes+local_curr_os_path), os_file_prefix(last_backuppath_hashes+local_curr_os_path), use_snapshots, &too_many_hardlinks);
+							b = create_hardlink(os_file_prefix(backuppath_hashes+local_curr_os_path), os_file_prefix(last_backuppath_hashes+local_curr_os_path), crossvolume_links, &too_many_hardlinks, NULL);
 
 							if(!b)
 							{
@@ -1078,7 +1220,7 @@ bool IncrFileBackup::doFileBackup()
 						else if(!b && too_many_hardlinks)
 						{
 							ServerLogger::Log(logid, "Creating hardlink from \""+srcpath+"\" to \""+backuppath+local_curr_os_path+"\" failed. Hardlink limit was reached. Copying file...", LL_DEBUG);
-							copyFile(srcpath, backuppath+local_curr_os_path,
+							copyFile(line, srcpath, backuppath+local_curr_os_path,
 								last_backuppath_hashes+local_curr_os_path,
 								backuppath_hashes+local_curr_os_path,
 								metadata);
@@ -1175,7 +1317,14 @@ bool IncrFileBackup::doFileBackup()
                         server_download->addToQueueFull(line, cf.name, osspecific_name, curr_path, curr_os_path, queue_downloads?0:-1,
                             metadata, script_dir, true, 0, std::string(), false, 0, std::string(), write_file_metadata);
                     }
+
+					if (depth == 0)
+					{
+						server_download->addToQueueStopShadowcopy(cf.name);
+					}
 				}
+
+				max_file_id.setMaxPreProcessed(line);
 				++line;
 			}
 		}
@@ -1193,6 +1342,8 @@ bool IncrFileBackup::doFileBackup()
 		disk_error = true;
 	}
 
+	stopPhashDownloadThread(filelist_async_id);
+
 	server_download->queueStop();
 
 	ServerLogger::Log(logid, "Waiting for file transfers...", LL_INFO);
@@ -1224,8 +1375,6 @@ bool IncrFileBackup::doFileBackup()
 		calculateDownloadSpeed(ctime, fc, fc_chunked.get());
 	}
 
-	addFilePathCorrections(server_download->getFilePathCorrections());
-
 	ServerStatus::setProcessSpeed(clientname, status_id, 0);
 
 	if(server_download->isOffline() && !r_offline)
@@ -1275,6 +1424,12 @@ bool IncrFileBackup::doFileBackup()
 
 	stopFileMetadataDownloadThread(false, server_download->getNumEmbeddedMetadataFiles());
 
+	if (!r_offline && !c_has_error && !disk_error
+		&& client_main->getProtocolVersions().wtokens_version > 0)
+	{
+		getTokenFile(fc, hashed_transfer, true);
+	}
+
 	server_download->deleteTempFolder();
 
 	ServerLogger::Log(logid, "Writing new file list...", LL_INFO);
@@ -1285,7 +1440,7 @@ bool IncrFileBackup::doFileBackup()
 
 	if(clientlist==NULL)
 	{
-		ServerLogger::Log(logid, "Error creating client file list at "+clientlistName(backupid), LL_ERROR);
+		ServerLogger::Log(logid, "Error creating client file list for client "+clientname+" at "+Server->getServerWorkingDir() + "/" + clientlistName(backupid) + ". "+os_last_error_str(), LL_ERROR);
 		disk_error=true;
 	}
 	else
@@ -1429,7 +1584,7 @@ bool IncrFileBackup::doFileBackup()
 
 		if(num_copied_file_entries>0)
 		{
-			ServerLogger::Log(logid, "Number of copyied file entries from last backup is "+convert(num_copied_file_entries), LL_INFO);
+			ServerLogger::Log(logid, "Number of copied file entries from last backup is "+convert(num_copied_file_entries), LL_INFO);
 		}
 
 		if (copy_last_file_entries)
@@ -1470,32 +1625,50 @@ bool IncrFileBackup::doFileBackup()
 			if (!os_sync(backuppath)
 				|| !os_sync(backuppath_hashes))
 			{
-				ServerLogger::Log(logid, "Syncing file system failed. Backup may not be completely on disk. " + os_last_error_str(), LL_DEBUG);
+				ServerLogger::Log(logid, "Syncing file system failed. Backup may not be completely on disk. " + os_last_error_str(), BackupServer::canSyncFs() ? LL_ERROR : LL_DEBUG);
+
+				if(BackupServer::canSyncFs())
+					c_has_error = true;
 			}
 
-			std::auto_ptr<IFile> sync_f(Server->openFile(os_file_prefix(backuppath_hashes + os_file_sep() + sync_fn), MODE_WRITE));
-
-			if (use_snapshots)
+			std::auto_ptr<IFile> sync_f;
+			if (!c_has_error)
 			{
-				if (!SnapshotHelper::makeReadonly(clientname, backuppath_single))
+				sync_f.reset(Server->openFile(os_file_prefix(backuppath_hashes + os_file_sep() + sync_fn), MODE_WRITE));
+
+				if (sync_f.get() != NULL
+					&& !sync_f->Sync())
+				{
+					ServerLogger::Log(logid, "Error syncing sync file to disk. " + os_last_error_str(), LL_ERROR);
+					c_has_error = true;
+				}
+			}
+
+			if (!c_has_error
+				&& make_subvolume_readonly)
+			{
+				if (!SnapshotHelper::makeReadonly(false, clientname, backuppath_single))
 				{
 					ServerLogger::Log(logid, "Making backup snapshot read only failed", LL_WARNING);
 				}
 			}
 
-			DBScopedSynchronous synchronous(db);
-			backup_dao->setFileBackupDone(backupid);
-
 			if (sync_f.get() != NULL)
 			{
+				DBScopedSynchronous synchronous(db);
+				DBScopedWriteTransaction trans(db);
+
+				backup_dao->setFileBackupDone(backupid);
 				backup_dao->setFileBackupSynced(backupid);
 			}
-			else
+			else if(!c_has_error)
 			{
-				ServerLogger::Log(logid, "Error creating sync file at " + backuppath_hashes + os_file_sep() + sync_fn, LL_WARNING);
+				ServerLogger::Log(logid, "Error creating sync file at " + backuppath_hashes + os_file_sep() + sync_fn+". Not setting backup to done.", LL_ERROR);
+				c_has_error = true;
 			}
 
-			if (ServerCleanupThread::isClientlistDeletionAllowed())
+			if (!c_has_error
+				&& ServerCleanupThread::isClientlistDeletionAllowed())
 			{
 				Server->deleteFile(clientlist_name);
 			}
@@ -1559,38 +1732,68 @@ bool IncrFileBackup::doFileBackup()
 		if (!os_sync(backuppath)
 			|| !os_sync(backuppath_hashes))
 		{
-			ServerLogger::Log(logid, "Syncing file system failed. Backup may not be completely on disk. " + os_last_error_str(), LL_DEBUG);
+			ServerLogger::Log(logid, "Syncing file system failed. Backup may not be completely on disk. " + os_last_error_str(), BackupServer::canSyncFs() ? LL_ERROR : LL_DEBUG);
+
+			if (BackupServer::canSyncFs())
+				c_has_error = true;
 		}
 
-		std::auto_ptr<IFile> sync_f(Server->openFile(os_file_prefix(backuppath_hashes + os_file_sep() + sync_fn), MODE_WRITE));
-
-		if (use_snapshots)
+		std::auto_ptr<IFile> sync_f;
+		if (!c_has_error)
 		{
-			if (!SnapshotHelper::makeReadonly(clientname, backuppath_single))
+			sync_f.reset(Server->openFile(os_file_prefix(backuppath_hashes + os_file_sep() + sync_fn), MODE_WRITE));
+
+			if (sync_f.get() != NULL
+				&& !sync_f->Sync())
+			{
+				ServerLogger::Log(logid, "Error syncing sync file to disk. " + os_last_error_str(), LL_ERROR);
+				c_has_error = true;
+			}
+		}
+
+		if (!c_has_error
+			&& make_subvolume_readonly)
+		{
+			if (!SnapshotHelper::makeReadonly(false, clientname, backuppath_single))
 			{
 				ServerLogger::Log(logid, "Making backup snapshot read only failed", LL_WARNING);
 			}
 		}
 
-		DBScopedSynchronous synchronous(db);
-		backup_dao->setFileBackupDone(backupid);
-
 		if (sync_f.get() != NULL)
 		{
+			DBScopedSynchronous synchronous(db);
+			DBScopedWriteTransaction trans(db);
+
+			backup_dao->setFileBackupDone(backupid);
 			backup_dao->setFileBackupSynced(backupid);
 		}
-		else
+		else if(!c_has_error)
 		{
-			ServerLogger::Log(logid, "Error creating sync file at " + backuppath_hashes + os_file_sep() + sync_fn, LL_WARNING);
+			ServerLogger::Log(logid, "Error creating sync file at " + backuppath_hashes + os_file_sep() + sync_fn+". Not setting backup to done", LL_ERROR);
+			c_has_error = true;
 		}
 
-		if (ServerCleanupThread::isClientlistDeletionAllowed())
+		if (!c_has_error
+			&& ServerCleanupThread::isClientlistDeletionAllowed())
 		{
 			Server->deleteFile(clientlist_name);
 		}
 	}
 	else
 	{
+		if (phash_load.get() != NULL)
+		{
+			std::auto_ptr<IFile> tf(Server->openTemporaryFile());
+			if (tf.get() != NULL)
+			{
+				if(copy_file(tmp_filelist, tf.get()))
+				{
+					ServerLogger::Log(logid, "Copied file list to "+tf->getFilename()+" for debugging.", LL_ERROR);
+				}				
+			}
+		}
+
 		ServerLogger::Log(logid, "Fatal error during backup. Backup not completed", LL_ERROR);
 		ClientMain::sendMailToAdmins("Fatal error occurred during incremental file backup", ServerLogger::getWarningLevelTextLogdata(logid));
 	}
@@ -1718,7 +1921,9 @@ bool IncrFileBackup::deleteFilesInSnapshot(const std::string clientlist_fn, cons
 				}
 
 				if( hasChange(line, deleted_ids) 
-					&& (deleted_inplace_ids ==NULL || !hasChange(line, *deleted_inplace_ids) ) )
+					&& ( (deleted_inplace_ids ==NULL || !hasChange(line, *deleted_inplace_ids) ) 
+						|| (!curr_file.isdir && curr_dir_exists && curr_path == snapshot_path + os_file_sep() + "urbackup_backup_scripts" )
+						) )
 				{					
 					std::string curr_fn=convertToOSPathFromFileClient(curr_os_path+os_file_sep()+osspecific_name);
 					if(curr_file.isdir)
@@ -1738,7 +1943,8 @@ bool IncrFileBackup::deleteFilesInSnapshot(const std::string clientlist_fn, cons
 									}
 								}
 							}
-							else if(!os_remove_nonempty_dir(os_file_prefix(curr_fn)) )
+							else if(!os_remove_nonempty_dir(os_file_prefix(curr_fn))
+								|| os_directory_exists(os_file_prefix(curr_fn)) )
 							{
 								ServerLogger::Log(logid, "Could not remove directory \"" + curr_fn + "\" in ::deleteFilesInSnapshot - " + systemErrorInfo(), no_error ? LL_WARNING : LL_ERROR);
 
@@ -1758,12 +1964,20 @@ bool IncrFileBackup::deleteFilesInSnapshot(const std::string clientlist_fn, cons
 						if( curr_dir_exists )
 						{
 							int ftype = EFileType_File;
+							bool keep_inplace = false;
 							if (curr_path == snapshot_path + os_file_sep() + "urbackup_backup_scripts")
 							{
 								ftype = os_get_file_type(os_file_prefix(curr_fn));
+
+								if (ftype & EFileType_File
+									&& !hash_dir
+									&& (deleted_inplace_ids == NULL || !hasChange(line, *deleted_inplace_ids) ) )
+								{
+									keep_inplace = true;
+								}
 							}
 
-							if(ftype & EFileType_File
+							if(ftype & EFileType_File && !keep_inplace
 								&& !Server->deleteFile(os_file_prefix(curr_fn)) )
 							{
 								std::auto_ptr<IFile> tf(Server->openFile(os_file_prefix(curr_fn), MODE_READ));
@@ -1782,7 +1996,8 @@ bool IncrFileBackup::deleteFilesInSnapshot(const std::string clientlist_fn, cons
 								}
 							}
 							else if (ftype & EFileType_Directory
-								&& !os_remove_nonempty_dir(os_file_prefix(curr_fn)) )
+								&& (!os_remove_nonempty_dir(os_file_prefix(curr_fn))
+									|| os_directory_exists(os_file_prefix(curr_fn)) ) )
 							{
 								ServerLogger::Log(logid, "Could not remove directory \"" + curr_fn + "\" in ::deleteFilesInSnapshot (2) - " + systemErrorInfo(), no_error ? LL_WARNING : LL_ERROR);
 
@@ -1907,12 +2122,15 @@ void IncrFileBackup::addSparseFileEntry( std::string curr_path, SFile &cf, int c
 	}
 }
 
-void IncrFileBackup::copyFile(const std::string& source, const std::string& dest,
+void IncrFileBackup::copyFile(size_t fileid, const std::string& source, const std::string& dest,
 	const std::string& hash_src, const std::string& hash_dest,
 	const FileMetadata& metadata)
 {
+	max_file_id.setMinDownloaded(fileid);
+
 	CWData data;
 	data.addInt(BackupServerHash::EAction_Copy);
+	data.addVarInt(fileid);
 	data.addString((source));
 	data.addString((dest));
 	data.addString((hash_src));
