@@ -17,6 +17,7 @@
 **************************************************************************/
 
 #include "InternetServiceConnector.h"
+#include "server.h"
 #include "../Interface/Server.h"
 #include "../Interface/Mutex.h"
 #include "../Interface/Condition.h"
@@ -25,11 +26,13 @@
 #include "../urbackupcommon/InternetServiceIDs.h"
 #include "../urbackupcommon/InternetServicePipe.h"
 #include "../urbackupcommon/CompressedPipe2.h"
+#include "../urbackupcommon/CompressedPipeZstd.h"
 #include "../urbackupcommon/CompressedPipe.h"
 #include "server_settings.h"
 #include "database.h"
 #include "../stringtools.h"
 #include "../cryptoplugin/ICryptoFactory.h"
+#include "serverinterface/login.h"
 
 #include <memory.h>
 #include <algorithm>
@@ -41,6 +44,7 @@ const unsigned int ping_timeout=30000;
 const unsigned int offline_timeout=ping_interval+10000;
 const unsigned int establish_timeout=60000;
 const int64 max_ecdh_key_age = 6 * 60 * 60 * 1000; //6h
+const std::string restore_prefix = "##restore##";
 
 std::map<std::string, SClientData> InternetServiceConnector::client_data;
 IMutex *InternetServiceConnector::mutex=NULL;
@@ -49,14 +53,20 @@ std::map<unsigned int, SOnetimeToken> InternetServiceConnector::onetime_tokens;
 unsigned int InternetServiceConnector::onetime_token_id=0;
 int64 InternetServiceConnector::last_token_remove=0;
 std::vector<std::pair<IECDHKeyExchange*, int64> > InternetServiceConnector::ecdh_key_exchange_buffer;
+std::set<std::string> InternetServiceConnector::internet_expect_endpoint;
 
 
 extern ICryptoFactory *crypto_fak;
 const size_t pbkdf2_iterations=20000;
 
+InternetService::InternetService(BackupServer * backup_server)
+	:backup_server(backup_server)
+{
+}
+
 ICustomClient* InternetService::createClient()
 {
-	return new InternetServiceConnector;
+	return new InternetServiceConnector(backup_server);
 }
 
 void InternetService::destroyClient( ICustomClient * pClient)
@@ -64,7 +74,8 @@ void InternetService::destroyClient( ICustomClient * pClient)
 	delete ((InternetServiceConnector*)pClient);
 }
 
-InternetServiceConnector::InternetServiceConnector(void)
+InternetServiceConnector::InternetServiceConnector(BackupServer * backup_server)
+	: backup_server(backup_server)
 {
 	local_mutex=Server->createMutex();
 	ecdh_key_exchange=NULL;
@@ -104,13 +115,22 @@ void InternetServiceConnector::Init(THREAD_ID pTID, IPipe *pPipe, const std::str
 	is_connected=false;
 	pinging=false;
 	free_connection=false;
-	state=ISS_AUTH;
+	if (internet_expect_endpoint.find(pEndpointName)!=internet_expect_endpoint.end())
+	{
+		state = ISS_RECEIVE_ENDPOINT;
+		endpoint_remaining = 50;
+	}
+	else
+	{
+		endpoint_name = pEndpointName;
+		state = ISS_AUTH;
+	}
 	has_timeout=false;
-	endpoint_name=pEndpointName;
 	connection_done_cond=NULL;
 	tcpstack.reset();
 	tcpstack.setAddChecksum(true);
 	tcpstack.setMaxPacketSize(32768);
+	client_ping_interval = ping_interval;
 	challenge=ServerSettings::generateRandomBinaryKey();
 	{
 		CWData data;
@@ -121,6 +141,9 @@ void InternetServiceConnector::Init(THREAD_ID pTID, IPipe *pPipe, const std::str
 		SSettings *settings=server_settings.getSettings();
 		capa|=IPC_ENCRYPTED;
 		capa|=IPC_COMPRESSED;
+#ifndef NO_ZSTD_COMPRESSION
+		capa |= IPC_COMPRESSED_ZSTD;
+#endif
 
 		compression_level=settings->internet_compression_level;
 		data.addUInt(capa);
@@ -157,7 +180,8 @@ void InternetServiceConnector::cleanup_pipes(bool remove_connection)
 	delete comp_pipe;
 	comp_pipe=NULL;
 
-	if(remove_connection)
+	if(remove_connection
+		&& client_data.find(clientname)!=client_data.end())
 	{
 		std::vector<InternetServiceConnector*>& spare_connections = client_data[clientname].spare_connections;
 		std::vector<InternetServiceConnector*>::iterator it=std::find(spare_connections.begin(), spare_connections.end(), this);
@@ -196,6 +220,22 @@ bool InternetServiceConnector::Run(IRunOtherCallback* run_other)
 		return false;
 	}
 
+	if (state == ISS_RECEIVE_ENDPOINT)
+	{
+		if (Server->getTimeMS() - lastpingtime > ping_timeout)
+		{
+			Server->Log("ISS_RECEIVE_ENDPOINT timeout in InternetServiceConnector::Run", LL_DEBUG);
+			IScopedLock lock(mutex);
+			if (!connect_start)
+			{
+				has_timeout = true;
+				cleanup_pipes(true);
+				return false;
+			}
+		}
+		return true;
+	}
+
 	if(do_connect && !pinging && state==ISS_AUTHED )
 	{
 		CWData data;
@@ -210,7 +250,7 @@ bool InternetServiceConnector::Run(IRunOtherCallback* run_other)
 		starttime=Server->getTimeMS();
 	}
 
-	if(state==ISS_AUTHED && Server->getTimeMS()-lastpingtime>ping_interval && !pinging)
+	if(state==ISS_AUTHED && Server->getTimeMS()-lastpingtime>client_ping_interval && !pinging)
 	{
 		lastpingtime=Server->getTimeMS();
 		pinging=true;
@@ -242,11 +282,37 @@ void InternetServiceConnector::ReceivePackets(IRunOtherCallback* run_other)
 		return;
 	}
 
+	if (state == ISS_RECEIVE_ENDPOINT)
+	{
+		char buf[51];
+		size_t rc = comm_pipe->Read(buf, endpoint_remaining);
+		if (rc == 0)
+		{
+			IScopedLock lock(mutex);
+			if (!connect_start)
+			{
+				has_timeout = true;
+				cleanup_pipes(true);
+			}
+			return;
+		}
+		buf[rc] = 0;
+		endpoint_remaining -= rc;
+		endpoint_name += buf;
+
+		if (endpoint_remaining == 0)
+		{
+			state = ISS_AUTH;
+		}
+		return;
+	}
+
 	std::string ret;
 	size_t rc;
 	rc=comm_pipe->Read(&ret);
 
-	if(rc==0)
+	if(rc==0 
+		|| !tcpstack.AddData((char*)ret.c_str(), ret.size()))
 	{
 		if( state!=ISS_CONNECTING && state!=ISS_USED )
 		{
@@ -260,8 +326,6 @@ void InternetServiceConnector::ReceivePackets(IRunOtherCallback* run_other)
 		return;
 	}
 
-	tcpstack.AddData((char*)ret.c_str(), ret.size());
-
 	char *buf;
 	size_t packetsize;
 	while((buf=tcpstack.getPacket(&packetsize))!=NULL)
@@ -419,6 +483,7 @@ void InternetServiceConnector::ReceivePackets(IRunOtherCallback* run_other)
 						CWData data;
 						if(!errmsg.empty())
 						{
+							logFailedAuth(clientname, endpoint_name);
 							Server->Log("Authentication failed in InternetServiceConnector::ReceivePackets: "+errmsg, LL_INFO);
 							data.addChar(ID_ISC_AUTH_FAILED);
 							data.addString(errmsg);
@@ -452,7 +517,28 @@ void InternetServiceConnector::ReceivePackets(IRunOtherCallback* run_other)
 								comm_pipe=is_pipe;
 								capa_debug_str += std::string("encrypted-") + (conn_version==2 ? "v2" : "v1");
 							}	
-							if(capa & IPC_COMPRESSED )
+							if (capa & IPC_COMPRESSED_ZSTD)
+							{
+#ifndef NO_ZSTD_COMPRESSION
+								if (conn_version == 2)
+								{
+									comp_pipe = new CompressedPipeZstd(comm_pipe, compression_level, -1);
+								}
+								else
+								{
+									Server->Log("Unknown connection version " + convert((int)conn_version) + " in state ISS_CAPA -1", LL_ERROR);
+									assert(false);
+								}
+
+								comm_pipe = comp_pipe;
+
+								if (!capa_debug_str.empty()) capa_debug_str += ", ";
+								capa_debug_str += "compressed-zstd";
+#else
+								Server->Log("Server does not support zstd compression in ISS_CAPA", LL_ERROR);
+#endif
+							}
+							else if(capa & IPC_COMPRESSED )
 							{
 								if(conn_version==1)
 								{
@@ -477,15 +563,23 @@ void InternetServiceConnector::ReceivePackets(IRunOtherCallback* run_other)
 
 							size_t spare_connections_num;
 
+							bool wakeup_new_client = false;
 							{
 								IScopedLock lock(mutex);
 								SClientData& curr_client_data = client_data[clientname];
+								if (curr_client_data.last_seen == -1
+									&& backup_server!=NULL)
+								{
+									wakeup_new_client = true;
+								}
 								curr_client_data.spare_connections.push_back(this);
 								curr_client_data.last_seen=Server->getTimeMS();
 								curr_client_data.endpoint_name = endpoint_name;
 
 								spare_connections_num = curr_client_data.spare_connections.size();
 							}
+							if (wakeup_new_client)
+								backup_server->wakeupNewClient();
 
 							if (token_auth)
 							{
@@ -539,6 +633,21 @@ void InternetServiceConnector::init_mutex(void)
 {
 	mutex=Server->createMutex();
 	onetime_token_mutex=Server->createMutex();
+
+	IDatabase* db = Server->getDatabase(Server->getThreadID(), URBACKUPDB_SERVER);
+	if (db != NULL)
+	{
+		db_results res = db->Read("SELECT value FROM settings_db.settings WHERE key='internet_expect_endpoint' AND clientid=0");
+		if (!res.empty())
+		{
+			std::vector<std::string> toks;
+			Tokenize(res[0]["value"], toks, ",;|");
+			for (size_t i = 0; i < toks.size(); ++i)
+			{
+				internet_expect_endpoint.insert(toks[i]);
+			}
+		}
+	}
 }
 
 void InternetServiceConnector::destroy_mutex(void)
@@ -596,9 +705,22 @@ IPipe *InternetServiceConnector::getConnection(const std::string &clientname, ch
 				IPipe *ret=isc->getISPipe();
 				isc->freeConnection(); //deletes ics
 
-				CompressedPipe *comp_pipe=dynamic_cast<CompressedPipe*>(ret);
-				CompressedPipe2 *comp_pipe2=dynamic_cast<CompressedPipe2*>(ret);
-				if(comp_pipe2!=NULL)
+				CompressedPipe *comp_pipe;
+				CompressedPipe2 *comp_pipe2;
+#ifndef NO_ZSTD_COMPRESSION
+				CompressedPipeZstd* comp_zstd;
+				if ((comp_zstd = dynamic_cast<CompressedPipeZstd*>(ret)) != NULL)
+				{
+					InternetServicePipe2 *isc_pipe2 = dynamic_cast<InternetServicePipe2*>(comp_zstd->getRealPipe());
+					if (isc_pipe2 != NULL)
+					{
+						isc_pipe2->destroyBackendPipeOnDelete(true);
+					}
+					comp_zstd->destroyBackendPipeOnDelete(true);
+				}
+				else
+#endif
+					if((comp_pipe2 = dynamic_cast<CompressedPipe2*>(ret))!=NULL)
 				{
 					InternetServicePipe2 *isc_pipe2=dynamic_cast<InternetServicePipe2*>(comp_pipe2->getRealPipe());
 					if(isc_pipe2!=NULL)
@@ -607,7 +729,7 @@ IPipe *InternetServiceConnector::getConnection(const std::string &clientname, ch
 					}
 					comp_pipe2->destroyBackendPipeOnDelete(true);
 				}
-				else if(comp_pipe!=NULL)
+				else if((comp_pipe = dynamic_cast<CompressedPipe*>(ret) )!=NULL)
 				{
 					InternetServicePipe *isc_pipe=dynamic_cast<InternetServicePipe*>(comp_pipe->getRealPipe());
 					if(isc_pipe!=NULL)
@@ -739,6 +861,10 @@ std::vector<std::pair<std::string, std::string> > InternetServiceConnector::getO
 			{
 				todel.push_back(it->first);
 			}
+			else if (ct - it->second.last_seen < 10000)
+			{
+				ret.push_back(std::make_pair(it->first, it->second.endpoint_name));
+			}
 		}
 	}
 
@@ -794,6 +920,27 @@ std::string InternetServiceConnector::getOnetimeToken(unsigned int id, std::stri
 std::string InternetServiceConnector::getAuthkeyFromDB(const std::string &clientname, bool &db_timeout)
 {
 	IDatabase *db=Server->getDatabase(tid, URBACKUPDB_SERVER);
+
+	if (next(clientname, 0, restore_prefix))
+	{
+		IQuery *q = db->Prepare("SELECT value FROM settings_db.settings WHERE key='restore_authkey' AND clientid=0", false);
+		int timeoutms = 1000;
+		db_results res = q->Read(&timeoutms);
+		db->destroyQuery(q);
+		if (!res.empty())
+		{
+			client_ping_interval = 30*1000;
+			db_timeout = false;
+			return res[0]["value"];
+		}
+		else if (timeoutms == 1)
+		{
+			db_timeout = true;
+		}
+
+		return std::string();
+	}
+
 	IQuery *q=db->Prepare("SELECT value FROM settings_db.settings WHERE key='internet_authkey' AND clientid=(SELECT id FROM clients WHERE name=?)", false);
 	q->Bind(clientname);
 	int timeoutms=1000;
@@ -850,4 +997,4 @@ void InternetServiceConnector::removeOldTokens(void)
 			onetime_tokens.erase(iter);
 		}
 	}
-}
\ No newline at end of file
+}
