@@ -44,7 +44,8 @@ namespace
 
 ServerDownloadThread::ServerDownloadThread( FileClient& fc, FileClientChunked* fc_chunked, const std::string& backuppath, const std::string& backuppath_hashes, const std::string& last_backuppath, const std::string& last_backuppath_complete, bool hashed_transfer, bool save_incomplete_file, int clientid,
 	const std::string& clientname, const std::string& clientsubname, bool use_tmpfiles, const std::string& tmpfile_path, const std::string& server_token, bool use_reflink, int backupid, bool r_incremental, IPipe* hashpipe_prepare, ClientMain* client_main,
-	int filesrv_protocol_version, int incremental_num, logid_t logid, bool with_hashes, const std::vector<std::string>& shares_without_snapshot, bool with_sparse_hashing, server::FileMetadataDownloadThread* file_metadata_download, bool sc_failure_fatal)
+	int filesrv_protocol_version, int incremental_num, logid_t logid, bool with_hashes, const std::vector<std::string>& shares_without_snapshot, bool with_sparse_hashing, server::FileMetadataDownloadThread* file_metadata_download, bool sc_failure_fatal,
+	FilePathCorrections& filepath_corrections, MaxFileId& max_file_id)
 	: fc(fc), fc_chunked(fc_chunked), backuppath(backuppath), backuppath_hashes(backuppath_hashes), 
 	last_backuppath(last_backuppath), last_backuppath_complete(last_backuppath_complete), hashed_transfer(hashed_transfer), save_incomplete_file(save_incomplete_file), clientid(clientid),
 	clientname(clientname), clientsubname(clientsubname),
@@ -52,7 +53,7 @@ ServerDownloadThread::ServerDownloadThread( FileClient& fc, FileClientChunked* f
 	is_offline(false), client_main(client_main), filesrv_protocol_version(filesrv_protocol_version), skipping(false), queue_size(0),
 	all_downloads_ok(true), incremental_num(incremental_num), logid(logid), has_timeout(false), with_hashes(with_hashes), with_metadata(client_main->getProtocolVersions().file_meta>0), shares_without_snapshot(shares_without_snapshot),
 	with_sparse_hashing(with_sparse_hashing), exp_backoff(false), num_embedded_metadata_files(0), file_metadata_download(file_metadata_download), num_issues(0), last_snap_num_issues(0), has_disk_error(false), sc_failure_fatal(sc_failure_fatal),
-	tmpfile_num(0)
+	tmpfile_num(0), filepath_corrections(filepath_corrections), max_file_id(max_file_id)
 {
 	mutex = Server->createMutex();
 	cond = Server->createCondition();
@@ -272,6 +273,12 @@ void ServerDownloadThread::addToQueueFull(size_t id, const std::string &fn, cons
 	ni.sha_dig = sha_dig;
 	ni.write_metadata = write_metadata;
 
+	if (id != 0
+		&& !at_front_postpone_quitstop)
+	{
+		max_file_id.setMinDownloaded(id);
+	}
+
 	if(is_script)
 	{
 		if (p_script_random != 0)
@@ -324,6 +331,11 @@ void ServerDownloadThread::addToQueueChunked(size_t id, const std::string &fn, c
     ni.metadata_only = false;
 	ni.sha_dig=sha_dig;
 
+	if (id != 0)
+	{
+		max_file_id.setMinDownloaded(id);
+	}
+
 	if(is_script)
 	{
 		if (p_script_random != 0)
@@ -398,6 +410,7 @@ size_t ServerDownloadThread::insertFullQueueEarliest(const SQueueItem& ni, bool
 	size_t idx=0;
 	size_t earlies_other_idx;
 	bool no_queued=true;
+	bool needs_postpone = false;
 	std::deque<SQueueItem>::iterator earliest_other=dl_queue.end();
 	for(std::deque<SQueueItem>::iterator it=dl_queue.begin();it!=dl_queue.end();++it)
 	{
@@ -408,6 +421,9 @@ size_t ServerDownloadThread::insertFullQueueEarliest(const SQueueItem& ni, bool
 			{
 				if (!after_switched || !it->switched)
 				{
+					if (needs_postpone)
+						max_file_id.postponeDownloaded(ni.id);
+
 					if (earliest_other != dl_queue.end())
 					{
 						dl_queue.insert(earliest_other, ni);
@@ -427,12 +443,25 @@ size_t ServerDownloadThread::insertFullQueueEarliest(const SQueueItem& ni, bool
 			}
 			else
 			{
+				if (it->id != std::string::npos
+					&& it->id>ni.id)
+				{
+					needs_postpone = true;
+				}
+
 				no_queued=false;
 				earliest_other=dl_queue.end();
 			}
 		}
 		else
 		{
+			if (it->id != std::string::npos
+				&& it->queued
+				&& it->id>ni.id)
+			{
+				needs_postpone = true;
+			}
+
 			if(earliest_other==dl_queue.end())
 			{
 				earliest_other = it;
@@ -441,6 +470,10 @@ size_t ServerDownloadThread::insertFullQueueEarliest(const SQueueItem& ni, bool
 		}
 		++idx;
 	}
+
+	if (needs_postpone)
+		max_file_id.postponeDownloaded(ni.id);
+
 	if(no_queued)
 	{
 		dl_queue.push_front(ni);
@@ -550,6 +583,7 @@ bool ServerDownloadThread::load_file(SQueueItem todl)
 
 			if (rc == ERR_BASE_DIR_LOST)
 			{
+				base_dir_lost_hint();
 				exp_backoff = true;
 			}
 		}
@@ -585,6 +619,7 @@ bool ServerDownloadThread::load_file(SQueueItem todl)
 					{
 						write_file_metadata(hashpath + os_file_sep() + metadata_dir_fn, client_main, todl.metadata, false);
 					}
+					ClientMain::destroyTemporaryFile(fd);
 				}
 			}
 		}
@@ -625,7 +660,7 @@ bool ServerDownloadThread::load_file(SQueueItem todl)
 			Server->destroy(file_old);
 		}
 
-		hashFile(dstpath, hashpath, fd, NULL, filepath_old, fd->Size(), todl.metadata, todl.is_script, todl.sha_dig, fc.releaseSparseExtendsFile(),
+		hashFile(todl.id, dstpath, hashpath, fd, NULL, filepath_old, fd->Size(), todl.metadata, todl.is_script, todl.sha_dig, fc.releaseSparseExtendsFile(),
 			todl.is_script ? HASH_FUNC_SHA512_NO_SPARSE : default_hashing_method, fileHasSnapshot(todl));
 	}
 	else
@@ -668,8 +703,8 @@ bool ServerDownloadThread::link_or_copy_file(const SQueueItem& todl)
 	
 	
 	if( ( os_get_file_type(os_file_prefix(dstpath))!=0
-			|| os_create_hardlink(os_file_prefix(dstpath), dlfiles.orig_file->getFilename(), use_reflink, NULL) )
-	    && os_create_hardlink(os_file_prefix(dsthashpath), dlfiles.chunkhashes->getFilename(), use_reflink, NULL) )
+			|| FileBackup::create_hardlink(os_file_prefix(dstpath), dlfiles.orig_file->getFilename(), use_reflink, NULL, NULL) )
+	    && FileBackup::create_hardlink(os_file_prefix(dsthashpath), dlfiles.chunkhashes->getFilename(), use_reflink, NULL, NULL) )
 	{
 		return true;
 	}
@@ -693,8 +728,8 @@ bool ServerDownloadThread::link_or_copy_file(const SQueueItem& todl)
 		    && (dlfiles.hashoutput=Server->openFile(hashoutput_fn, MODE_RW))!=NULL )
 		{
 			pfd_destroy.release();
-			hashFile(dstpath, dlfiles.hashpath, dlfiles.patchfile, dlfiles.hashoutput,
-			    (dlfiles.filepath_old), orig_filesize, todl.metadata, todl.is_script, todl.sha_dig, NULL,
+			hashFile(todl.id, dstpath, dlfiles.hashpath, dlfiles.patchfile, dlfiles.hashoutput,
+			    (dlfiles.filepath_old), orig_filesize, todl.metadata, todl.is_script, std::string(), NULL,
 				todl.is_script ? HASH_FUNC_SHA512_NO_SPARSE : default_hashing_method, fileHasSnapshot(todl));
 			return true;
 		}
@@ -830,6 +865,7 @@ bool ServerDownloadThread::load_file_patch(SQueueItem todl)
 
 			if (rc == ERR_BASE_DIR_LOST)
 			{
+				base_dir_lost_hint();
 				exp_backoff = true;
 			}
 		}
@@ -928,7 +964,7 @@ bool ServerDownloadThread::load_file_patch(SQueueItem todl)
 		pfd_destroy.release();
 		hash_tmp_destroy.release();
 		sparse_extents_f_delete.release();
-		hashFile(dstpath, dlfiles.hashpath, dlfiles.patchfile, dlfiles.hashoutput,
+		hashFile(todl.id, dstpath, dlfiles.hashpath, dlfiles.patchfile, dlfiles.hashoutput,
 			dlfiles.filepath_old, download_filesize, todl.metadata, todl.is_script, todl.sha_dig, sparse_extents_f,
 			todl.is_script ? HASH_FUNC_SHA512_NO_SPARSE : default_hashing_method, fileHasSnapshot(todl));
 	}
@@ -946,13 +982,14 @@ bool ServerDownloadThread::load_file_patch(SQueueItem todl)
 		return true;
 }
 
-void ServerDownloadThread::hashFile(std::string dstpath, std::string hashpath, IFile *fd, IFile *hashoutput, std::string old_file,
+void ServerDownloadThread::hashFile(int64 fileid, std::string dstpath, std::string hashpath, IFile *fd, IFile *hashoutput, std::string old_file,
 	int64 t_filesize, const FileMetadata& metadata, bool is_script, std::string sha_dig, IFile* sparse_extents_f, char hashing_method,
 	bool has_snapshot)
 {
 	int l_backup_id=backupid;
 
 	CWData data;
+	data.addVarInt(fileid);
 	data.addString(fd->getFilename());
 	data.addInt(l_backup_id);
 	data.addInt(r_incremental?1:0);
@@ -1279,6 +1316,12 @@ SPatchDownloadFiles ServerDownloadThread::preparePatchDownloadFiles( const SQueu
 	std::string cfn_short=todl.os_path+"/"+todl.short_fn;
 	if(cfn_short[0]=='/')
 		cfn_short.erase(0,1);
+#ifdef _WIN32
+	if(!cfn_short.empty()
+		&& cfn_short[0]==os_file_sep()[0])
+		cfn_short.erase(0, 1);
+#endif
+
 
 	std::string dstpath=backuppath+os_file_sep()+FileBackup::convertToOSPathFromFileClient(cfn_short);
 	std::string hashpath=backuppath_hashes+os_file_sep()+FileBackup::convertToOSPathFromFileClient(cfn_short);
@@ -1484,11 +1527,6 @@ bool ServerDownloadThread::sleepQueue()
 	return false;
 }
 
-std::map<std::string, std::string>& ServerDownloadThread::getFilePathCorrections()
-{
-	return filepath_corrections;
-}
-
 size_t ServerDownloadThread::getNumEmbeddedMetadataFiles()
 {
 	return num_embedded_metadata_files;
@@ -1930,7 +1968,7 @@ bool ServerDownloadThread::fileHasSnapshot(const SQueueItem& todl)
 std::string ServerDownloadThread::tarFnToOsPath(const std::string & tar_path)
 {
 	std::vector<std::string> toks;
-	TokenizeMail(tar_path, toks, "/");
+	Tokenize(tar_path, toks, "/");
 
 	std::string ret;
 	for (size_t i = 0; i < toks.size(); ++i)
@@ -1965,10 +2003,10 @@ std::string ServerDownloadThread::tarFnToOsPath(const std::string & tar_path)
 			}
 			std::string tmp_ret = ret + toks[i];
 
-			std::map<std::string, std::string>::iterator it = filepath_corrections.find(tmp_ret);
-			if (it != filepath_corrections.end())
+			std::string corr;
+			if (filepath_corrections.get(tmp_ret, corr))
 			{
-				ret += it->second;
+				ret += corr;
 			}
 			else
 			{
@@ -1998,4 +2036,11 @@ void ServerDownloadThread::logVssLogdata()
 			ServerLogger::Log(logid, entries[i].data, entries[i].loglevel);
 		}
 	}
-}
\ No newline at end of file
+}
+
+void ServerDownloadThread::base_dir_lost_hint()
+{
+	ServerLogger::Log(logid, "Hint: Current root backup path does not exist anymore. Usually this is because the current file "
+		"system snapshot it was backing up was deleted because it ran out of snapshot storage space. "
+		"See https://www.urbackup.org/faq.html#base_dir_lost for details and for how to fix this issue", LL_INFO);
+}
