@@ -21,6 +21,8 @@
 #include <assert.h>
 #include <string.h>
 
+int ServerBackupDao::num_issues_no_backuppaths = -1;
+
 /**
 * @-SQLGenTempSetup
 * @sql
@@ -333,6 +335,172 @@ std::vector<int> ServerBackupDao::getClientIds(void)
 	return ret;
 }
 
+/**
+* @-SQLGenAccess
+* @func vector<int> ServerBackupDao::getClientsByUid
+* @return int id
+* @sql
+*      SELECT id FROM clients WHERE uid=:uid(string)
+*/
+std::vector<int> ServerBackupDao::getClientsByUid(const std::string& uid)
+{
+	if(q_getClientsByUid==NULL)
+	{
+		q_getClientsByUid=db->Prepare("SELECT id FROM clients WHERE uid=?", false);
+	}
+	q_getClientsByUid->Bind(uid);
+	db_results res=q_getClientsByUid->Read();
+	q_getClientsByUid->Reset();
+	std::vector<int> ret;
+	ret.resize(res.size());
+	for(size_t i=0;i<res.size();++i)
+	{
+		ret[i]=watoi(res[i]["id"]);
+	}
+	return ret;
+}
+
+/**
+* @-SQLGenAccess
+* @func void ServerBackupDao::updateClientUid
+* @sql
+*      UPDATE clients SET uid=:uid(string) WHERE id=:clientid(int)
+*/
+void ServerBackupDao::updateClientUid(const std::string& uid, int clientid)
+{
+	if(q_updateClientUid==NULL)
+	{
+		q_updateClientUid=db->Prepare("UPDATE clients SET uid=? WHERE id=?", false);
+	}
+	q_updateClientUid->Bind(uid);
+	q_updateClientUid->Bind(clientid);
+	q_updateClientUid->Write();
+	q_updateClientUid->Reset();
+}
+
+/**
+* @-SQLGenAccess
+* @func void ServerBackupDao::deleteClient
+* @sql
+*      DELETE FROM clients WHERE id=:clientid(int)
+*/
+void ServerBackupDao::deleteClient(int clientid)
+{
+	if(q_deleteClient==NULL)
+	{
+		q_deleteClient=db->Prepare("DELETE FROM clients WHERE id=?", false);
+	}
+	q_deleteClient->Bind(clientid);
+	q_deleteClient->Write();
+	q_deleteClient->Reset();
+}
+
+/**
+* @-SQLGenAccess
+* @func void ServerBackupDao::changeClientName
+* @sql
+*      UPDATE clients SET name=:name(string) WHERE id=:id(int)
+*/
+void ServerBackupDao::changeClientName(const std::string& name, int id)
+{
+	if(q_changeClientName==NULL)
+	{
+		q_changeClientName=db->Prepare("UPDATE clients SET name=? WHERE id=?", false);
+	}
+	q_changeClientName->Bind(name);
+	q_changeClientName->Bind(id);
+	q_changeClientName->Write();
+	q_changeClientName->Reset();
+}
+
+/**
+* @-SQLGenAccess
+* @func void ServerBackupDao::changeClientNameWithVirtualmain
+* @sql
+*      UPDATE clients SET name=:name(string), virtualmain=:virtualmain(string) WHERE id=:id(int)
+*/
+void ServerBackupDao::changeClientNameWithVirtualmain(const std::string& name, const std::string& virtualmain, int id)
+{
+	if(q_changeClientNameWithVirtualmain==NULL)
+	{
+		q_changeClientNameWithVirtualmain=db->Prepare("UPDATE clients SET name=?, virtualmain=? WHERE id=?", false);
+	}
+	q_changeClientNameWithVirtualmain->Bind(name);
+	q_changeClientNameWithVirtualmain->Bind(virtualmain);
+	q_changeClientNameWithVirtualmain->Bind(id);
+	q_changeClientNameWithVirtualmain->Write();
+	q_changeClientNameWithVirtualmain->Reset();
+}
+
+/**
+* @-SQLGenAccess
+* @func void ServerBackupDao::addClientMoved
+* @sql
+*      INSERT INTO moved_clients (from_name, to_name) VALUES (:from_name(string), :to_name(string))
+*/
+void ServerBackupDao::addClientMoved(const std::string& from_name, const std::string& to_name)
+{
+	if(q_addClientMoved==NULL)
+	{
+		q_addClientMoved=db->Prepare("INSERT INTO moved_clients (from_name, to_name) VALUES (?, ?)", false);
+	}
+	q_addClientMoved->Bind(from_name);
+	q_addClientMoved->Bind(to_name);
+	q_addClientMoved->Write();
+	q_addClientMoved->Reset();
+}
+
+/**
+* @-SQLGenAccess
+* @func vector<string> ServerBackupDao::getClientMovedLimit5
+* @return string from_name
+* @sql
+*	   SELECT from_name FROM moved_clients WHERE to_name=:to_name(string) LIMIT 5
+*/
+std::vector<std::string> ServerBackupDao::getClientMovedLimit5(const std::string& to_name)
+{
+	if(q_getClientMovedLimit5==NULL)
+	{
+		q_getClientMovedLimit5=db->Prepare("SELECT from_name FROM moved_clients WHERE to_name=? LIMIT 5", false);
+	}
+	q_getClientMovedLimit5->Bind(to_name);
+	db_results res=q_getClientMovedLimit5->Read();
+	q_getClientMovedLimit5->Reset();
+	std::vector<std::string> ret;
+	ret.resize(res.size());
+	for(size_t i=0;i<res.size();++i)
+	{
+		ret[i]=res[i]["from_name"];
+	}
+	return ret;
+}
+
+/**
+* @-SQLGenAccess
+* @func vector<string> ServerBackupDao::getClientMovedFrom
+* @return string to_name
+* @sql
+*	   SELECT to_name FROM moved_clients WHERE from_name=:from_name(string)
+*/
+std::vector<std::string> ServerBackupDao::getClientMovedFrom(const std::string& from_name)
+{
+	if(q_getClientMovedFrom==NULL)
+	{
+		q_getClientMovedFrom=db->Prepare("SELECT to_name FROM moved_clients WHERE from_name=?", false);
+	}
+	q_getClientMovedFrom->Bind(from_name);
+	db_results res=q_getClientMovedFrom->Read();
+	q_getClientMovedFrom->Reset();
+	std::vector<std::string> ret;
+	ret.resize(res.size());
+	for(size_t i=0;i<res.size();++i)
+	{
+		ret[i]=res[i]["to_name"];
+	}
+	return ret;
+}
+
+
 /**
 * @-SQLGenAccess
 * @func string ServerBackupDao::getSetting
@@ -359,6 +527,26 @@ ServerBackupDao::CondString ServerBackupDao::getSetting(int clientid, const std:
 	return ret;
 }
 
+/**
+* @-SQLGenAccess
+* @func int ServerBackupDao::hasFileBackups
+* @return int_raw c
+* @sql
+*      SELECT COUNT(*) AS c FROM backups WHERE clientid=:clientid(int) AND done=1 LIMIT 1
+*/
+int ServerBackupDao::hasFileBackups(int clientid)
+{
+	if(q_hasFileBackups==NULL)
+	{
+		q_hasFileBackups=db->Prepare("SELECT COUNT(*) AS c FROM backups WHERE clientid=? AND done=1 LIMIT 1", false);
+	}
+	q_hasFileBackups->Bind(clientid);
+	db_results res=q_hasFileBackups->Read();
+	q_hasFileBackups->Reset();
+	assert(!res.empty());
+	return watoi(res[0]["c"]);
+}
+
 /**
 * @-SQLGenAccess
 * @func void ServerBackupDao::insertSetting
@@ -988,13 +1176,13 @@ void ServerBackupDao::saveImageAssociation(int img_id, int assoc_id)
 * @-SQLGenAccess
 * @func void ServerBackupDao::updateClientLastImageBackup
 * @sql
-*       UPDATE clients SET lastbackup_image=(SELECT b.backuptime FROM backup_images b WHERE b.id=:backupid(int)) WHERE id=:clientid(int)
+*       UPDATE clients SET lastbackup_image=(SELECT b.backuptime FROM backup_images b WHERE b.id=:backupid(int)), alerts_next_check=NULL WHERE id=:clientid(int)
 */
 void ServerBackupDao::updateClientLastImageBackup(int backupid, int clientid)
 {
 	if(q_updateClientLastImageBackup==NULL)
 	{
-		q_updateClientLastImageBackup=db->Prepare("UPDATE clients SET lastbackup_image=(SELECT b.backuptime FROM backup_images b WHERE b.id=?) WHERE id=?", false);
+		q_updateClientLastImageBackup=db->Prepare("UPDATE clients SET lastbackup_image=(SELECT b.backuptime FROM backup_images b WHERE b.id=?), alerts_next_check=NULL WHERE id=?", false);
 	}
 	q_updateClientLastImageBackup->Bind(backupid);
 	q_updateClientLastImageBackup->Bind(clientid);
@@ -1002,17 +1190,35 @@ void ServerBackupDao::updateClientLastImageBackup(int backupid, int clientid)
 	q_updateClientLastImageBackup->Reset();
 }
 
+/**
+* @-SQLGenAccess
+* @func void ServerBackupDao::updateClientNumIssues
+* @sql
+*       UPDATE clients SET last_filebackup_issues=:last_filebackup_issues(int) WHERE id=:clientid(int)
+*/
+void ServerBackupDao::updateClientNumIssues(int last_filebackup_issues, int clientid)
+{
+	if(q_updateClientNumIssues==NULL)
+	{
+		q_updateClientNumIssues=db->Prepare("UPDATE clients SET last_filebackup_issues=? WHERE id=?", false);
+	}
+	q_updateClientNumIssues->Bind(last_filebackup_issues);
+	q_updateClientNumIssues->Bind(clientid);
+	q_updateClientNumIssues->Write();
+	q_updateClientNumIssues->Reset();
+}
+
 /**
 * @-SQLGenAccess
 * @func void ServerBackupDao::updateClientLastFileBackup
 * @sql
-*       UPDATE clients SET lastbackup=(SELECT b.backuptime FROM backups b WHERE b.id=:backupid(int)), last_filebackup_issues=:last_filebackup_issues(int) WHERE id=:clientid(int)
+*       UPDATE clients SET lastbackup=(SELECT b.backuptime FROM backups b WHERE b.id=:backupid(int)), last_filebackup_issues=:last_filebackup_issues(int), alerts_next_check=NULL WHERE id=:clientid(int)
 */
 void ServerBackupDao::updateClientLastFileBackup(int backupid, int last_filebackup_issues, int clientid)
 {
 	if(q_updateClientLastFileBackup==NULL)
 	{
-		q_updateClientLastFileBackup=db->Prepare("UPDATE clients SET lastbackup=(SELECT b.backuptime FROM backups b WHERE b.id=?), last_filebackup_issues=? WHERE id=?", false);
+		q_updateClientLastFileBackup=db->Prepare("UPDATE clients SET lastbackup=(SELECT b.backuptime FROM backups b WHERE b.id=?), last_filebackup_issues=?, alerts_next_check=NULL WHERE id=?", false);
 	}
 	q_updateClientLastFileBackup->Bind(backupid);
 	q_updateClientLastFileBackup->Bind(last_filebackup_issues);
@@ -1571,61 +1777,116 @@ ServerBackupDao::CondInt ServerBackupDao::getImageMounttime(int backupid)
 
 /**
 * @-SQLGenAccess
-* @func void ServerBackupDao::setImageMounted
+* @func void ServerBackupDao::updateImageMounted
 * @sql
-*       UPDATE backup_images SET mounttime=strftime('%s','now') WHERE id=:backupid(int)
+*		UPDATE mounted_backup_images SET mounttime=strftime('%s','now') WHERE id=:id(int64)
 */
-void ServerBackupDao::setImageMounted(int backupid)
+void ServerBackupDao::updateImageMounted(int64 id)
 {
-	if(q_setImageMounted==NULL)
+	if(q_updateImageMounted==NULL)
 	{
-		q_setImageMounted=db->Prepare("UPDATE backup_images SET mounttime=strftime('%s','now') WHERE id=?", false);
+		q_updateImageMounted=db->Prepare("UPDATE mounted_backup_images SET mounttime=strftime('%s','now') WHERE id=?", false);
 	}
-	q_setImageMounted->Bind(backupid);
-	q_setImageMounted->Write();
-	q_setImageMounted->Reset();
+	q_updateImageMounted->Bind(id);
+	q_updateImageMounted->Write();
+	q_updateImageMounted->Reset();
 }
 
 /**
 * @-SQLGenAccess
-* @func void ServerBackupDao::setImageUnmounted
+* @func void ServerBackupDao::addImageMounted
 * @sql
-*       UPDATE backup_images SET mounttime=0 WHERE id=:backupid(int)
+*       INSERT INTO mounted_backup_images (backupid, partition, mounttime)
+*		VALUES (:backupid(int), :partition(int), strftime('%s','now'))
 */
-void ServerBackupDao::setImageUnmounted(int backupid)
+void ServerBackupDao::addImageMounted(int backupid, int partition)
 {
-	if(q_setImageUnmounted==NULL)
+	if(q_addImageMounted==NULL)
 	{
-		q_setImageUnmounted=db->Prepare("UPDATE backup_images SET mounttime=0 WHERE id=?", false);
+		q_addImageMounted=db->Prepare("INSERT INTO mounted_backup_images (backupid, partition, mounttime) VALUES (?, ?, strftime('%s','now'))", false);
 	}
-	q_setImageUnmounted->Bind(backupid);
-	q_setImageUnmounted->Write();
-	q_setImageUnmounted->Reset();
+	q_addImageMounted->Bind(backupid);
+	q_addImageMounted->Bind(partition);
+	q_addImageMounted->Write();
+	q_addImageMounted->Reset();
+}
+
+/**
+* @-SQLGenAccess
+* @func void ServerBackupDao::delImageMounted
+* @sql
+*		DELETE FROM mounted_backup_images WHERE id=:id(int64)
+*/
+void ServerBackupDao::delImageMounted(int64 id)
+{
+	if(q_delImageMounted==NULL)
+	{
+		q_delImageMounted=db->Prepare("DELETE FROM mounted_backup_images WHERE id=?", false);
+	}
+	q_delImageMounted->Bind(id);
+	q_delImageMounted->Write();
+	q_delImageMounted->Reset();
 }
 
 /**
 * @-SQLGenAccess
 * @func SMountedImage ServerBackupDao::getMountedImage
-* @return int id, string path, int64 mounttime
+* @return int id, int backupid, string path, int64 mounttime, int partition, int clientid
 * @sql
-*       SELECT id, path, mounttime FROM backup_images WHERE id=:backupid(int)
+*       SELECT b.id AS backupid, m.id AS id, path, m.mounttime AS mounttime, partition, clientid
+*		 FROM (mounted_backup_images m INNER JOIN backup_images b ON m.backupid=b.id) 
+*			WHERE b.id=:backupid(int) AND m.partition=:partition(int)
 */
-ServerBackupDao::SMountedImage ServerBackupDao::getMountedImage(int backupid)
+ServerBackupDao::SMountedImage ServerBackupDao::getMountedImage(int backupid, int partition)
 {
 	if(q_getMountedImage==NULL)
 	{
-		q_getMountedImage=db->Prepare("SELECT id, path, mounttime FROM backup_images WHERE id=?", false);
+		q_getMountedImage=db->Prepare("SELECT b.id AS backupid, m.id AS id, path, m.mounttime AS mounttime, partition, clientid FROM (mounted_backup_images m INNER JOIN backup_images b ON m.backupid=b.id)  WHERE b.id=? AND m.partition=?", false);
 	}
 	q_getMountedImage->Bind(backupid);
+	q_getMountedImage->Bind(partition);
 	db_results res=q_getMountedImage->Read();
 	q_getMountedImage->Reset();
-	SMountedImage ret = { false, 0, "", 0 };
+	SMountedImage ret = { false, 0, 0, "", 0, 0, 0 };
 	if(!res.empty())
 	{
 		ret.exists=true;
 		ret.id=watoi(res[0]["id"]);
+		ret.backupid=watoi(res[0]["backupid"]);
 		ret.path=res[0]["path"];
 		ret.mounttime=watoi64(res[0]["mounttime"]);
+		ret.partition=watoi(res[0]["partition"]);
+		ret.clientid=watoi(res[0]["clientid"]);
+	}
+	return ret;
+}
+
+/**
+* @-SQLGenAccess
+* @func SMountedImage ServerBackupDao::getImageInfo
+* @return int id, int backupid, string path, int clientid
+* @sql
+*       SELECT 0 AS id, id AS backupid, path, clientid
+*		 FROM backup_images
+*			WHERE id=:backupid(int)
+*/
+ServerBackupDao::SMountedImage ServerBackupDao::getImageInfo(int backupid)
+{
+	if(q_getImageInfo==NULL)
+	{
+		q_getImageInfo=db->Prepare("SELECT 0 AS id, id AS backupid, path, clientid FROM backup_images WHERE id=?", false);
+	}
+	q_getImageInfo->Bind(backupid);
+	db_results res=q_getImageInfo->Read();
+	q_getImageInfo->Reset();
+	SMountedImage ret = { false, 0, 0, "", 0 };
+	if(!res.empty())
+	{
+		ret.exists=true;
+		ret.id=watoi(res[0]["id"]);
+		ret.backupid=watoi(res[0]["backupid"]);
+		ret.path=res[0]["path"];
+		ret.clientid=watoi(res[0]["clientid"]);
 	}
 	return ret;
 }
@@ -1633,15 +1894,16 @@ ServerBackupDao::SMountedImage ServerBackupDao::getMountedImage(int backupid)
 /**
 * @-SQLGenAccess
 * @func vector<SMountedImage> ServerBackupDao::getOldMountedImages
-* @return int id, string path, int64 mounttime
+* @return int id, int backupid, string path, int64 mounttime, int partition
 * @sql
-*       SELECT id, path, mounttime FROM backup_images WHERE mounttime!=0 AND mounttime<(strftime('%s','now')-:times(int64))
+*       SELECT b.id AS backupid, m.id AS id, path, m.mounttime AS mounttime, partition FROM (mounted_backup_images m INNER JOIN backup_images b ON m.backupid=b.id) 
+*			WHERE m.mounttime!=0 AND m.mounttime<(strftime('%s','now')-:times(int64))
 */
 std::vector<ServerBackupDao::SMountedImage> ServerBackupDao::getOldMountedImages(int64 times)
 {
 	if(q_getOldMountedImages==NULL)
 	{
-		q_getOldMountedImages=db->Prepare("SELECT id, path, mounttime FROM backup_images WHERE mounttime!=0 AND mounttime<(strftime('%s','now')-?)", false);
+		q_getOldMountedImages=db->Prepare("SELECT b.id AS backupid, m.id AS id, path, m.mounttime AS mounttime, partition FROM (mounted_backup_images m INNER JOIN backup_images b ON m.backupid=b.id)  WHERE m.mounttime!=0 AND m.mounttime<(strftime('%s','now')-?)", false);
 	}
 	q_getOldMountedImages->Bind(times);
 	db_results res=q_getOldMountedImages->Read();
@@ -1652,8 +1914,10 @@ std::vector<ServerBackupDao::SMountedImage> ServerBackupDao::getOldMountedImages
 	{
 		ret[i].exists=true;
 		ret[i].id=watoi(res[i]["id"]);
+		ret[i].backupid=watoi(res[i]["backupid"]);
 		ret[i].path=res[i]["path"];
 		ret[i].mounttime=watoi64(res[i]["mounttime"]);
+		ret[i].partition=watoi(res[i]["partition"]);
 	}
 	return ret;
 }
@@ -1673,7 +1937,16 @@ void ServerBackupDao::prepareQueries( void )
 	q_getLastFullDurations=NULL;
 	q_getClientSetting=NULL;
 	q_getClientIds=NULL;
+	q_getClientsByUid=NULL;
+	q_updateClientUid=NULL;
+	q_deleteClient=NULL;
+	q_changeClientName=NULL;
+	q_changeClientNameWithVirtualmain=NULL;
+	q_addClientMoved=NULL;
+	q_getClientMovedLimit5=NULL;
+	q_getClientMovedFrom=NULL;
 	q_getSetting=NULL;
+	q_hasFileBackups=NULL;
 	q_insertSetting=NULL;
 	q_updateSetting=NULL;
 	q_getMiscValue=NULL;
@@ -1704,6 +1977,7 @@ void ServerBackupDao::prepareQueries( void )
 	q_updateImageBackupRunning=NULL;
 	q_saveImageAssociation=NULL;
 	q_updateClientLastImageBackup=NULL;
+	q_updateClientNumIssues=NULL;
 	q_updateClientLastFileBackup=NULL;
 	q_updateClientOsAndClientVersion=NULL;
 	q_deleteAllUsersOnClient=NULL;
@@ -1728,9 +2002,11 @@ void ServerBackupDao::prepareQueries( void )
 	q_getClientnameByImageid=NULL;
 	q_getClientidByImageid=NULL;
 	q_getImageMounttime=NULL;
-	q_setImageMounted=NULL;
-	q_setImageUnmounted=NULL;
+	q_updateImageMounted=NULL;
+	q_addImageMounted=NULL;
+	q_delImageMounted=NULL;
 	q_getMountedImage=NULL;
+	q_getImageInfo=NULL;
 	q_getOldMountedImages=NULL;
 }
 
@@ -1749,7 +2025,16 @@ void ServerBackupDao::destroyQueries( void )
 	db->destroyQuery(q_getLastFullDurations);
 	db->destroyQuery(q_getClientSetting);
 	db->destroyQuery(q_getClientIds);
+	db->destroyQuery(q_getClientsByUid);
+	db->destroyQuery(q_updateClientUid);
+	db->destroyQuery(q_deleteClient);
+	db->destroyQuery(q_changeClientName);
+	db->destroyQuery(q_changeClientNameWithVirtualmain);
+	db->destroyQuery(q_addClientMoved);
+	db->destroyQuery(q_getClientMovedLimit5);
+	db->destroyQuery(q_getClientMovedFrom);
 	db->destroyQuery(q_getSetting);
+	db->destroyQuery(q_hasFileBackups);
 	db->destroyQuery(q_insertSetting);
 	db->destroyQuery(q_updateSetting);
 	db->destroyQuery(q_getMiscValue);
@@ -1780,6 +2065,7 @@ void ServerBackupDao::destroyQueries( void )
 	db->destroyQuery(q_updateImageBackupRunning);
 	db->destroyQuery(q_saveImageAssociation);
 	db->destroyQuery(q_updateClientLastImageBackup);
+	db->destroyQuery(q_updateClientNumIssues);
 	db->destroyQuery(q_updateClientLastFileBackup);
 	db->destroyQuery(q_updateClientOsAndClientVersion);
 	db->destroyQuery(q_deleteAllUsersOnClient);
@@ -1804,9 +2090,11 @@ void ServerBackupDao::destroyQueries( void )
 	db->destroyQuery(q_getClientnameByImageid);
 	db->destroyQuery(q_getClientidByImageid);
 	db->destroyQuery(q_getImageMounttime);
-	db->destroyQuery(q_setImageMounted);
-	db->destroyQuery(q_setImageUnmounted);
+	db->destroyQuery(q_updateImageMounted);
+	db->destroyQuery(q_addImageMounted);
+	db->destroyQuery(q_delImageMounted);
 	db->destroyQuery(q_getMountedImage);
+	db->destroyQuery(q_getImageInfo);
 	db->destroyQuery(q_getOldMountedImages);
 }
 
