@@ -31,6 +31,11 @@ ServerCleanupDao::~ServerCleanupDao(void)
 	destroyQueries();
 }
 
+IDatabase * ServerCleanupDao::getDatabase()
+{
+	return db;
+}
+
 /**
 * @-SQLGenAccess
 * @func std::vector<SIncompleteImages> ServerCleanupDao::getIncompleteImages
@@ -60,6 +65,35 @@ std::vector<ServerCleanupDao::SIncompleteImages> ServerCleanupDao::getIncomplete
 	return ret;
 }
 
+/**
+* @-SQLGenAccess
+* @func int ServerCleanupDao::getIncompleteImage
+* @return int id
+* @sql
+*   SELECT id
+*   FROM backup_images
+*   WHERE
+*     complete=0 AND (archived & 1)=0 AND running<datetime('now','-300 seconds')
+*	  AND id=:id(int)
+*/
+ServerCleanupDao::CondInt ServerCleanupDao::getIncompleteImage(int id)
+{
+	if(q_getIncompleteImage==NULL)
+	{
+		q_getIncompleteImage=db->Prepare("SELECT id FROM backup_images WHERE complete=0 AND (archived & 1)=0 AND running<datetime('now','-300 seconds') AND id=?", false);
+	}
+	q_getIncompleteImage->Bind(id);
+	db_results res=q_getIncompleteImage->Read();
+	q_getIncompleteImage->Reset();
+	CondInt ret = { false, 0 };
+	if(!res.empty())
+	{
+		ret.exists=true;
+		ret.value=watoi(res[0]["id"]);
+	}
+	return ret;
+}
+
 /**
 * @-SQLGenAccess
 * @func std::vector<SIncompleteImages> ServerCleanupDao::getDeletePendingImages
@@ -69,12 +103,13 @@ std::vector<ServerCleanupDao::SIncompleteImages> ServerCleanupDao::getIncomplete
 *   FROM backup_images b, clients c
 *   WHERE
 *     b.delete_pending=1 AND b.clientid=c.id
+*   ORDER BY backuptime DESC
 */
 std::vector<ServerCleanupDao::SIncompleteImages> ServerCleanupDao::getDeletePendingImages(void)
 {
 	if(q_getDeletePendingImages==NULL)
 	{
-		q_getDeletePendingImages=db->Prepare("SELECT b.id AS id, b.path AS path, c.name AS clientname FROM backup_images b, clients c WHERE b.delete_pending=1 AND b.clientid=c.id", false);
+		q_getDeletePendingImages=db->Prepare("SELECT b.id AS id, b.path AS path, c.name AS clientname FROM backup_images b, clients c WHERE b.delete_pending=1 AND b.clientid=c.id ORDER BY backuptime DESC", false);
 	}
 	db_results res=q_getDeletePendingImages->Read();
 	std::vector<ServerCleanupDao::SIncompleteImages> ret;
@@ -136,7 +171,7 @@ std::vector<int> ServerCleanupDao::getClientsSortFilebackups(void)
 * @return int id
 * @sql
 *   SELECT DISTINCT c.id AS id FROM clients c 
-*		INNER JOIN (SELECT * FROM backup_images WHERE length(letter)<=2) b
+*		INNER JOIN (SELECT * FROM backup_images WHERE letter!='SYSVOL' AND letter!='ESP') b
 *				ON c.id=b.clientid
 *	ORDER BY b.backuptime ASC
 */
@@ -144,7 +179,7 @@ std::vector<int> ServerCleanupDao::getClientsSortImagebackups(void)
 {
 	if(q_getClientsSortImagebackups==NULL)
 	{
-		q_getClientsSortImagebackups=db->Prepare("SELECT DISTINCT c.id AS id FROM clients c  INNER JOIN (SELECT * FROM backup_images WHERE length(letter)<=2) b ON c.id=b.clientid ORDER BY b.backuptime ASC", false);
+		q_getClientsSortImagebackups=db->Prepare("SELECT DISTINCT c.id AS id FROM clients c  INNER JOIN (SELECT * FROM backup_images WHERE letter!='SYSVOL' AND letter!='ESP') b ON c.id=b.clientid ORDER BY b.backuptime ASC", false);
 	}
 	db_results res=q_getClientsSortImagebackups->Read();
 	std::vector<int> ret;
@@ -162,14 +197,14 @@ std::vector<int> ServerCleanupDao::getClientsSortImagebackups(void)
 * @return int id, string letter
 * @sql
 *   SELECT id, letter FROM backup_images 
-*	WHERE clientid=:clientid(int) AND incremental=0 AND complete=1 AND length(letter)<=2 AND archived=0
+*	WHERE clientid=:clientid(int) AND incremental=0 AND complete=1 AND letter!='SYSVOL' AND letter!='ESP' AND archived=0
 *	ORDER BY backuptime ASC
 */
 std::vector<ServerCleanupDao::SImageLetter> ServerCleanupDao::getFullNumImages(int clientid)
 {
 	if(q_getFullNumImages==NULL)
 	{
-		q_getFullNumImages=db->Prepare("SELECT id, letter FROM backup_images  WHERE clientid=? AND incremental=0 AND complete=1 AND length(letter)<=2 AND archived=0 ORDER BY backuptime ASC", false);
+		q_getFullNumImages=db->Prepare("SELECT id, letter FROM backup_images  WHERE clientid=? AND incremental=0 AND complete=1 AND letter!='SYSVOL' AND letter!='ESP' AND archived=0 ORDER BY backuptime ASC", false);
 	}
 	q_getFullNumImages->Bind(clientid);
 	db_results res=q_getFullNumImages->Read();
@@ -347,14 +382,14 @@ ServerCleanupDao::CondString ServerCleanupDao::getImagePath(int id)
 * @return int id, string letter
 * @sql
 *	SELECT id,letter FROM backup_images
-*	WHERE clientid=:clientid(int) AND incremental<>0 AND complete=1 AND length(letter)<=2 AND archived=0
+*	WHERE clientid=:clientid(int) AND incremental<>0 AND complete=1 AND letter!='SYSVOL' AND letter!='ESP' AND archived=0
 *	ORDER BY backuptime ASC
 */
 std::vector<ServerCleanupDao::SImageLetter> ServerCleanupDao::getIncrNumImages(int clientid)
 {
 	if(q_getIncrNumImages==NULL)
 	{
-		q_getIncrNumImages=db->Prepare("SELECT id,letter FROM backup_images WHERE clientid=? AND incremental<>0 AND complete=1 AND length(letter)<=2 AND archived=0 ORDER BY backuptime ASC", false);
+		q_getIncrNumImages=db->Prepare("SELECT id,letter FROM backup_images WHERE clientid=? AND incremental<>0 AND complete=1 AND letter!='SYSVOL' AND letter!='ESP' AND archived=0 ORDER BY backuptime ASC", false);
 	}
 	q_getIncrNumImages->Bind(clientid);
 	db_results res=q_getIncrNumImages->Read();
@@ -513,6 +548,24 @@ void ServerCleanupDao::removeFileBackup(int backupid)
 	q_removeFileBackup->Reset();
 }
 
+/**
+* @-SQLGenAccess
+* @func void ServerCleanupDao::changeImagePath
+* @sql
+*	UPDATE backup_images SET path=:path(string) WHERE id=:backupid(int)
+*/
+void ServerCleanupDao::changeImagePath(const std::string& path, int backupid)
+{
+	if(q_changeImagePath==NULL)
+	{
+		q_changeImagePath=db->Prepare("UPDATE backup_images SET path=? WHERE id=?", false);
+	}
+	q_changeImagePath->Bind(path);
+	q_changeImagePath->Bind(backupid);
+	q_changeImagePath->Write();
+	q_changeImagePath->Reset();
+}
+
 /**
 * @-SQLGenAccess
 * @func SFileBackupInfo ServerCleanupDao::getFileBackupInfo
@@ -1234,6 +1287,7 @@ ServerCleanupDao::CondInt ServerCleanupDao::hasMoreRecentFileBackup(int backupid
 void ServerCleanupDao::createQueries(void)
 {
 	q_getIncompleteImages=NULL;
+	q_getIncompleteImage=NULL;
 	q_getDeletePendingImages=NULL;
 	q_removeImage=NULL;
 	q_getClientsSortFilebackups=NULL;
@@ -1252,6 +1306,7 @@ void ServerCleanupDao::createQueries(void)
 	q_getClientName=NULL;
 	q_getFileBackupPath=NULL;
 	q_removeFileBackup=NULL;
+	q_changeImagePath=NULL;
 	q_getFileBackupInfo=NULL;
 	q_getImageBackupInfo=NULL;
 	q_removeImageSize=NULL;
@@ -1286,6 +1341,7 @@ void ServerCleanupDao::createQueries(void)
 void ServerCleanupDao::destroyQueries(void)
 {
 	db->destroyQuery(q_getIncompleteImages);
+	db->destroyQuery(q_getIncompleteImage);
 	db->destroyQuery(q_getDeletePendingImages);
 	db->destroyQuery(q_removeImage);
 	db->destroyQuery(q_getClientsSortFilebackups);
@@ -1304,6 +1360,7 @@ void ServerCleanupDao::destroyQueries(void)
 	db->destroyQuery(q_getClientName);
 	db->destroyQuery(q_getFileBackupPath);
 	db->destroyQuery(q_removeFileBackup);
+	db->destroyQuery(q_changeImagePath);
 	db->destroyQuery(q_getFileBackupInfo);
 	db->destroyQuery(q_getImageBackupInfo);
 	db->destroyQuery(q_removeImageSize);
