@@ -53,6 +53,7 @@ extern IServer* Server;
 #include "../fsimageplugin/IFSImageFactory.h"
 #include "../cryptoplugin/ICryptoFactory.h"
 #include "../urlplugin/IUrlFactory.h"
+#include "../luaplugin/ILuaInterpreter.h"
 
 #include "database.h"
 #include "actions.h"
@@ -83,6 +84,8 @@ SStartupStatus startup_status;
 #include "server_dir_links.h"
 #include "server_channel.h"
 #include "DataplanDb.h"
+#include "Alerts.h"
+#include "Mailer.h"
 
 #include <stdlib.h>
 #include "../Interface/DatabaseCursor.h"
@@ -94,6 +97,7 @@ SStartupStatus startup_status;
 #include "../urbackupcommon/WalCheckpointThread.h"
 #include "FileMetadataDownloadThread.h"
 #include "../urbackupcommon/chunk_hasher.h"
+#include "LogReport.h"
 
 #define MINIZ_NO_ZLIB_COMPATIBLE_NAMES
 #include "../common/miniz.h"
@@ -122,6 +126,7 @@ IFSImageFactory *image_fak;
 ICryptoFactory *crypto_fak;
 IUrlFactory *url_fak=NULL;
 IFileServ* fileserv=NULL;
+ILuaInterpreter* lua_interpreter = NULL;
 
 std::string server_identity;
 std::string server_token;
@@ -145,6 +150,7 @@ bool test_amatch(void);
 bool verify_hashes(std::string arg);
 void updateRights(int t_userid, std::string s_rights, IDatabase *db);
 int md5sum_check();
+int blockalign();
 
 std::string lang="en";
 std::string time_format_str="%Y-%m-%d %H:%M";
@@ -394,7 +400,7 @@ DLLEXPORT void LoadActions(IServer* pServer)
 		Server->Log("Amatch test failed! Stopping.", LL_ERROR);
 		return;
 	}*/
-	
+
 	std::string rmtest=Server->getServerParameter("rmtest");
 	if(!rmtest.empty())
 	{
@@ -402,6 +408,11 @@ DLLEXPORT void LoadActions(IServer* pServer)
 		return;
 	}
 
+	if (Server->getServerParameter("run-test") == "true")
+	{
+		exit(0);
+	}
+
 	std::string download_file=Server->getServerParameter("download_file");
 	if(!download_file.empty())
 	{
@@ -507,10 +518,14 @@ DLLEXPORT void LoadActions(IServer* pServer)
 				Server->Log("Cannot open file to hash (hash_file) \"" + Server->getServerParameter("hash_file") + "\" for reading. " + os_last_error_str(), LL_ERROR);
 			}
 		}
+		else if (app == "blockalign")
+		{
+			rc = blockalign();
+		}
 		else
 		{
 			rc=100;
-			Server->Log("App not found. Available apps: cleanup, remove_unknown, cleanup_database, repair_database, defrag_database, export_auth_log, check_fileindex, skiphash_copy, md5sum_check, hash");
+			Server->Log("App not found. Available apps: cleanup, remove_unknown, cleanup_database, repair_database, defrag_database, export_auth_log, check_fileindex, skiphash_copy, md5sum_check, hash, blockalign");
 		}
 		exit(rc);
 	}
@@ -608,7 +623,20 @@ DLLEXPORT void LoadActions(IServer* pServer)
 		}
 		else
 		{
-			fileserv = fileserv_fak->createFileServNoBind();
+			if (!fileserv_fak->optainBackupPrivileges())
+			{
+				Server->Log("Error optaining backup SYSTEM privileges. Restoring symlinks may not work.", LL_ERROR);
+			}
+			fileserv = fileserv_fak->createFileServNoBind(std::string(), false, false, true);
+		}
+	}
+
+	{
+		str_map params;
+		lua_interpreter = dynamic_cast<ILuaInterpreter*>(Server->getPlugin(Server->getThreadID(), Server->StartPlugin("lua", params)));
+		if (lua_interpreter == NULL)
+		{
+			Server->Log("Lua plugin missing. Alerts won't work.", LL_WARNING);
 		}
 	}
 
@@ -620,6 +648,7 @@ DLLEXPORT void LoadActions(IServer* pServer)
 	ServerSettings::init_mutex();
 	ClientMain::init_mutex();
 	DataplanDb::init();
+	init_log_report();
 
 	open_settings_database();
 	
@@ -633,7 +662,7 @@ DLLEXPORT void LoadActions(IServer* pServer)
 		}
 		else
 		{
-			Server->Log("Backup verification successfull.", LL_INFO);
+			Server->Log("Backup verification successful.", LL_INFO);
 			Server->deleteFile("verification_result.txt");
 			exit(0);
 		}
@@ -664,7 +693,11 @@ DLLEXPORT void LoadActions(IServer* pServer)
 		db->Write("PRAGMA journal_mode=WAL");
 	}
 		
-	if( FileExists("urbackup/backupfolder") )
+	if( FileExists("urbackup/backupfolder")
+#ifndef _WIN32
+		|| FileExists("/etc/urbackup/backupfolder")
+#endif
+		)
 	{
 		IDatabase *db=Server->getDatabase(Server->getThreadID(), URBACKUPDB_SERVER);
 		db_results res=db->Read("SELECT value FROM settings_db.settings WHERE key='backupfolder' AND clientid=0");
@@ -672,6 +705,12 @@ DLLEXPORT void LoadActions(IServer* pServer)
 		{
 			IQuery *q=db->Prepare("INSERT INTO settings_db.settings (key, value, clientid) VALUES ('backupfolder', ?, 0)", false);
 			std::string bf=getFile("urbackup/backupfolder");
+#ifndef _WIN32
+			if (bf.empty())
+			{
+				bf = getFile("/etc/urbackup/backupfolder");
+			}
+#endif
 			if(linecount(bf)>0)
 				bf=getline(0, bf);
 			q->Bind(trim(bf));
@@ -778,6 +817,8 @@ DLLEXPORT void LoadActions(IServer* pServer)
 	ADD_ACTION(start_backup);
 	ADD_ACTION(add_client);
 	ADD_ACTION(restore_prepare_wait);
+	ADD_ACTION(scripts);
+	ADD_ACTION(status_check);
 
 	if(Server->getServerParameter("allow_shutdown")=="true")
 	{
@@ -827,7 +868,7 @@ DLLEXPORT void LoadActions(IServer* pServer)
 			{
 				port=settings.getSettings()->internet_server_port;
 			}
-			Server->StartCustomStreamService(new InternetService, "InternetService", port);
+			Server->StartCustomStreamService(new InternetService(backup_server), "InternetService", port);
 		}
 	}
 
@@ -835,6 +876,8 @@ DLLEXPORT void LoadActions(IServer* pServer)
 	ServerCleanupThread::initMutex();
 	ServerAutomaticArchive::initMutex();
 	ServerCleanupThread *server_cleanup=new ServerCleanupThread(CleanupAction());
+	Mailer::init();
+	Server->createThread(new Alerts, "alerts");
 
 	is_leak_check=(Server->getServerParameter("leak_check")=="true");
 
@@ -1906,6 +1949,137 @@ bool upgrade52_53()
 	return b;
 }
 
+bool upgrade53_54()
+{
+	IDatabase *db = Server->getDatabase(Server->getThreadID(), URBACKUPDB_SERVER);
+
+	bool b = true;
+
+	b &= db->Write("INSERT INTO settings_db.settings (key, value, clientid) VALUES ('restore_authkey','"+ServerSettings::generateRandomAuthKey()+"', 0)");
+
+	return b;
+}
+
+bool upgrade54_55()
+{
+	IDatabase *db = Server->getDatabase(Server->getThreadID(), URBACKUPDB_SERVER);
+
+	bool b = true;
+
+	b &= db->Write("ALTER TABLE clients ADD file_ok INTEGER");
+	b &= db->Write("ALTER TABLE clients ADD image_ok INTEGER");
+	b &= db->Write("ALTER TABLE clients ADD alerts_state BLOB");
+	b &= db->Write("ALTER TABLE clients ADD alerts_next_check INTEGER");
+	b &= db->Write("ALTER TABLE clients ADD created INTEGER");
+	b &= db->Write("UPDATE clients SET created=0");
+
+	b &= db->Write("UPDATE clients SET file_ok=0, image_ok=0");
+
+	b &= db->Write("CREATE TABLE alert_scripts (id INTEGER PRIMARY KEY, "
+		"name TEXT, script TEXT)");
+	b &= db->Write("CREATE TABLE alert_script_params(id INTEGER PRIMARY KEY, "
+		"script_id INTEGER REFERENCES alert_scripts(id) ON DELETE CASCADE, "
+		"idx INTEGER, name TEXT, label TEXT, default_value TEXT, has_translation INTEGER DEFAULT 0, type TEXT)");
+
+	IQuery *q = db->Prepare("INSERT INTO alert_scripts (id, name, script) VALUES (?, ?, '')");
+	if (q != NULL)
+	{
+		q->Bind(1);
+		q->Bind("Default");
+		b &= q->Write();
+	}
+	else
+	{
+		b = false;
+	}
+
+	b &= db->Write("INSERT INTO alert_script_params (script_id, idx, name, label, default_value, has_translation, type) VALUES (1, 0, 'alert_file_mult', 'alert_file_mult', '3', 1, 'num')");
+	b &= db->Write("INSERT INTO alert_script_params (script_id, idx, name, label, default_value, has_translation, type) VALUES (1, 1, 'alert_image_mult', 'alert_image_mult', '3', 1, 'num')");
+	b &= db->Write("INSERT INTO alert_script_params (script_id, idx, name, label, default_value, has_translation, type) VALUES (1, 2, 'alert_emails', 'alert_emails', '', 1, 'str')");
+	b &= db->Write("INSERT INTO alert_script_params (script_id, idx, name, label, default_value, has_translation, type) VALUES (1, 3, 'alert_important', 'alert_important', '0', 1, 'bool')");
+
+	b &= db->Write("CREATE TABLE mail_queue (id INTEGER PRIMARY KEY, send_to TEXT, subject TEXT, message TEXT, next_try INTEGER, retry_count INTEGER DEFAULT 0)");
+
+	return b;
+}
+
+bool upgrade55_56()
+{
+	IDatabase *db = Server->getDatabase(Server->getThreadID(), URBACKUPDB_SERVER);
+
+	bool b = true;
+
+	b &= db->Write("INSERT INTO alert_script_params (script_id, idx, name, label, default_value, has_translation, type) VALUES (1, 4, 'alert_mail_ok', 'alert_mail_ok', '1', 1, 'bool')");
+
+	return b;
+}
+
+bool upgrade56_57()
+{
+	IDatabase *db = Server->getDatabase(Server->getThreadID(), URBACKUPDB_SERVER);
+
+	bool b = true;
+
+	b &= db->Write("CREATE TABLE moved_clients (id INTEGER PRIMARY KEY, "
+		"from_name TEXT, "
+		"to_name TEXT)");
+
+	b &= db->Write("CREATE INDEX moved_clients_from_idx ON moved_clients(from_name)");
+
+	b &= db->Write("ALTER TABLE clients ADD uid TEXT");
+	
+	b &= db->Write("CREATE TABLE mounted_backup_images (id INTEGER PRIMARY KEY, "
+		"backupid INTEGER REFERENCES backup_images(id) ON DELETE CASCADE, "
+		"partition INTEGER,"
+		"mounttime INTEGER)");
+
+	return b;
+}
+
+const int64 script_id_default = 1;
+const int64 script_id_pulseway = 100000;
+
+bool upgrade57_58()
+{
+	IDatabase *db = Server->getDatabase(Server->getThreadID(), URBACKUPDB_SERVER);
+
+	bool b = true;
+
+	b &= db->Write("ALTER TABLE alert_scripts ADD global_state BLOB");
+
+	int64 scriptid = script_id_pulseway;
+	b &= db->Write("INSERT INTO alert_scripts (id, name, script) VALUES (" + convert(scriptid) + ", 'Pulseway API', '')");
+	b &= db->Write("INSERT INTO alert_script_params (script_id, idx, name, label, default_value, has_translation, type) VALUES (" + convert(scriptid) + ", 0, 'username', 'pulseway_username', '', 1, 'str')");
+	b &= db->Write("INSERT INTO alert_script_params (script_id, idx, name, label, default_value, has_translation, type) VALUES (" + convert(scriptid) + ", 1, 'password', 'pulseway_password', '', 1, 'str')");
+	b &= db->Write("INSERT INTO alert_script_params (script_id, idx, name, label, default_value, has_translation, type) VALUES (" + convert(scriptid) + ", 2, 'instance_id', 'pulseway_instance_id', '', 1, 'str')");
+	b &= db->Write("INSERT INTO alert_script_params (script_id, idx, name, label, default_value, has_translation, type) VALUES (" + convert(scriptid) + ", 3, 'instance_name', 'pulseway_instance_name', '', 1, 'str')");
+	b &= db->Write("INSERT INTO alert_script_params (script_id, idx, name, label, default_value, has_translation, type) VALUES (" + convert(scriptid) + ", 4, 'instance_group', 'pulseway_instance_group', 'Default', 1, 'str')");
+	b &= db->Write("INSERT INTO alert_script_params (script_id, idx, name, label, default_value, has_translation, type) VALUES (" + convert(scriptid) + ", 5, 'instance_description', 'pulseway_instance_description', '', 1, 'str')");
+	b &= db->Write("INSERT INTO alert_script_params (script_id, idx, name, label, default_value, has_translation, type) VALUES (" + convert(scriptid) + ", 6, 'refresh_interval_minutes', 'pulseway_refresh_interval_minutes', '15', 1, 'num')");
+	b &= db->Write("INSERT INTO alert_script_params (script_id, idx, name, label, default_value, has_translation, type) VALUES (" + convert(scriptid) + ", 7, 'instance_notify_when_offline', 'pulseway_instance_notify_when_offline', '1', 1, 'bool')");
+	b &= db->Write("INSERT INTO alert_script_params (script_id, idx, name, label, default_value, has_translation, type) VALUES (" + convert(scriptid) + ", 8, 'api_url', 'pulseway_api_url', 'https://api.pulseway.com/v2', 1, 'str')");
+	b &= db->Write("INSERT INTO alert_script_params (script_id, idx, name, label, default_value, has_translation, type) VALUES (" + convert(scriptid) + ", 9, 'alert_file_mult', 'alert_file_mult', '3', 1, 'num')");
+	b &= db->Write("INSERT INTO alert_script_params (script_id, idx, name, label, default_value, has_translation, type) VALUES (" + convert(scriptid) + ", 10, 'alert_image_mult', 'alert_image_mult', '3', 1, 'num')");
+	b &= db->Write("INSERT INTO alert_script_params (script_id, idx, name, label, default_value, has_translation, type) VALUES (" + convert(scriptid) + ", 11, 'alert_emails', 'alert_emails', '', 1, 'str')");
+	b &= db->Write("INSERT INTO alert_script_params (script_id, idx, name, label, default_value, has_translation, type) VALUES (" + convert(scriptid) + ", 12, 'alert_important', 'alert_important', '0', 1, 'bool')");
+	b &= db->Write("INSERT INTO alert_script_params (script_id, idx, name, label, default_value, has_translation, type) VALUES (" + convert(scriptid) + ", 13, 'priority', 'pulseway_priority', 'normal|low|elevated|critical', 1, 'choice')");
+	b &= db->Write("INSERT INTO alert_script_params (script_id, idx, name, label, default_value, has_translation, type) VALUES (" + convert(scriptid) + ", 14, 'alert_ok', 'pulseway_alert_ok', '1', 1, 'bool')");
+
+	return b;
+}
+
+bool upgrade58_59()
+{
+	IDatabase *db = Server->getDatabase(Server->getThreadID(), URBACKUPDB_SERVER);
+
+	bool b = true;
+
+	b &= db->Write("INSERT INTO alert_script_params (script_id, idx, name, label, default_value, has_translation, type) VALUES (" + convert(script_id_pulseway) + ", 15, 'alert_nag_interval', 'pulseway_alert_nag_interval', '0|h', 1, 'interval')");
+	b &= db->Write("INSERT INTO alert_script_params (script_id, idx, name, label, default_value, has_translation, type) VALUES (" + convert(script_id_default) + ", 5, 'alert_nag_interval', 'alert_nag_interval', '0|h', 1, 'interval')");
+
+	return b;
+}
+
 void upgrade(void)
 {
 	Server->destroyAllDatabases();
@@ -1927,7 +2101,7 @@ void upgrade(void)
 	
 	int ver=watoi(res_v[0]["tvalue"]);
 	int old_v;
-	int max_v=53;
+	int max_v=59;
 	{
 		IScopedLock lock(startup_status.mutex);
 		startup_status.target_db_version=max_v;
@@ -2235,6 +2409,48 @@ void upgrade(void)
 				}
 				++ver;
 				break;
+			case 53:
+				if (!upgrade53_54())
+				{
+					has_error = true;
+				}
+				++ver;
+				break;
+			case 54:
+				if (!upgrade54_55())
+				{
+					has_error = true;
+				}
+				++ver;
+				break;
+			case 55:
+				if (!upgrade55_56())
+				{
+					has_error = true;
+				}
+				++ver;
+				break;
+			case 56:
+				if (!upgrade56_57())
+				{
+					has_error = true;
+				}
+				++ver;
+				break;
+			case 57:
+				if (!upgrade57_58())
+				{
+					has_error = true;
+				}
+				++ver;
+				break;
+			case 58:
+				if (!upgrade58_59())
+				{
+					has_error = true;
+				}
+				++ver;
+				break;
 			default:
 				break;
 		}
