@@ -89,7 +89,7 @@ namespace
 			std::string metadata_path = basedir;
 
 			std::vector<std::string> path_segments;
-			TokenizeMail(path.substr(1), path_segments, "/");
+			Tokenize(path.substr(1), path_segments, "/");
 
 			bool isdir = path[0]=='d';
 
@@ -229,20 +229,21 @@ namespace
 			const std::string& folder_log_name, int64 restore_id, size_t status_id, logid_t log_id, const std::string& restore_token, const std::string& identity,
 			const std::vector<std::pair<std::string, std::string> >& map_paths,
 			bool clean_other, bool ignore_other_fs, const std::string& share_path,
-			bool follow_symlinks, int64 restore_flags)
+			bool follow_symlinks, int64 restore_flags, const std::vector<std::string>& tokens, backupaccess::STokens access_tokens)
 			: curr_clientname(curr_clientname), curr_clientid(curr_clientid), restore_clientid(restore_clientid),
 			filelist_f(filelist_f),
 			skip_special_root(skip_special_root),
 			restore_token(restore_token), identity(identity), restore_id(restore_id), status_id(status_id), log_id(log_id),
 			single_file(false), map_paths(map_paths), clean_other(clean_other), ignore_other_fs(ignore_other_fs),
-			curr_restore_folder_idx(0), follow_symlinks(follow_symlinks), restore_flags(restore_flags)
+			curr_restore_folder_idx(0), follow_symlinks(follow_symlinks), restore_flags(restore_flags),
+			tokens(tokens), access_tokens(access_tokens)
 		{
 			SRestoreFolder restore_folder;
 			restore_folder.foldername = foldername;
 			restore_folder.hashfoldername = hashfoldername;
 			restore_folder.share_path = share_path;
 
-			TokenizeMail(filter, restore_folder.filter_fns, "/");
+			Tokenize(filter, restore_folder.filter_fns, "/");
 
 			if (restore_folder.filter_fns.size() == 1)
 			{
@@ -462,6 +463,34 @@ namespace
 					extra+="&single_item=1";
 				}
 
+				if (!tokens.empty())
+				{
+					if (!backupaccess::checkFileToken(access_tokens.tokens, tokens, metadata))
+					{
+						extra += "&skip=1";
+						recurse_dir = false;
+					}
+				}
+
+				if (depth == 0)
+				{
+					std::string tids;
+					for (size_t j = 0; j < access_tokens.tokens.size(); ++j)
+					{
+						if (access_tokens.tokens[j].token.empty())
+							continue;
+
+						for (size_t k = 0; k < tokens.size(); ++k)
+						{
+							if (access_tokens.tokens[j].token == tokens[k])
+							{
+								if (!tids.empty()) tids += ",";
+								tids += convert(access_tokens.tokens[j].id);
+							}
+						}
+					}
+					extra += "&tids=" + tids;
+				}
 
 				writeFileItem(filelist_f, file, extra);
 
@@ -488,7 +517,7 @@ namespace
 			bool& is_dir)
 		{
 			std::vector<std::string> symlink_path_toks;
-			TokenizeMail(symlink_target, symlink_path_toks, os_file_sep());
+			Tokenize(symlink_target, symlink_path_toks, os_file_sep());
 
 			SRestoreFolder restore_folder;
 			restore_folder.foldername = file_path;
@@ -593,6 +622,8 @@ namespace
 		size_t curr_restore_folder_idx;
 		bool follow_symlinks;
 		int64 restore_flags;
+		std::vector<std::string> tokens;
+		backupaccess::STokens access_tokens;
 	};
 }
 
@@ -600,7 +631,7 @@ bool create_clientdl_thread(const std::string& curr_clientname, int curr_clienti
 	const std::string& filter, bool skip_hashes,
 	const std::string& folder_log_name, int64& restore_id, size_t& status_id, logid_t& log_id, const std::string& restore_token,
 	const std::vector<std::pair<std::string, std::string> >& map_paths, bool clean_other, bool ignore_other_fs, const std::string& share_path,
-	bool follow_symlinks, int64 restore_flags, THREADPOOL_TICKET& ticket)
+	bool follow_symlinks, int64 restore_flags, THREADPOOL_TICKET& ticket, const std::vector<std::string>& tokens, const backupaccess::STokens& access_tokens)
 {
 	IFile* filelist_f = Server->openTemporaryFile();
 
@@ -644,7 +675,8 @@ bool create_clientdl_thread(const std::string& curr_clientname, int curr_clienti
 
 	ticket = Server->getThreadPool()->execute(new ClientDownloadThread(curr_clientname, curr_clientid, restore_clientid,
 		filelist_f, foldername, hashfoldername, filter, skip_hashes, folder_log_name, restore_id,
-		status_id, log_id, restore_token, identity, map_paths, clean_other, ignore_other_fs, share_path, follow_symlinks, restore_flags), "frestore preparation");
+		status_id, log_id, restore_token, identity, map_paths, clean_other, ignore_other_fs, share_path, follow_symlinks, 
+		restore_flags, tokens, access_tokens), "frestore preparation");
 
 	return true;
 }
@@ -652,7 +684,8 @@ bool create_clientdl_thread(const std::string& curr_clientname, int curr_clienti
 bool create_clientdl_thread( int backupid, const std::string& curr_clientname, int curr_clientid,
 	int64& restore_id, size_t& status_id, logid_t& log_id, const std::string& restore_token,
 	const std::vector<std::pair<std::string, std::string> >& map_paths,
-	bool clean_other, bool ignore_other_fs, bool follow_symlinks, int64 restore_flags)
+	bool clean_other, bool ignore_other_fs, bool follow_symlinks, int64 restore_flags,
+	const std::vector<std::string>& tokens)
 {
 	IDatabase* db = Server->getDatabase(Server->getThreadID(), URBACKUPDB_SERVER);
 	ServerBackupDao backup_dao(db);
@@ -682,12 +715,15 @@ bool create_clientdl_thread( int backupid, const std::string& curr_clientname, i
 		return false;
 	}
 
+	backupaccess::STokens access_tokens = backupaccess::readTokens(backupfolder.value, client_name.value, file_backup_info.path);
+
 	std::string curr_path=backupfolder.value+os_file_sep()+client_name.value+os_file_sep()+file_backup_info.path;
 	std::string curr_metadata_path=backupfolder.value+os_file_sep()+client_name.value+os_file_sep()+file_backup_info.path+os_file_sep()+".hashes";
 	std::string share_path = greplace(os_file_sep(), "/", file_backup_info.path);
 	THREADPOOL_TICKET ticket;
 
 	return create_clientdl_thread(curr_clientname, curr_clientid, file_backup_info.clientid, curr_path, curr_metadata_path, std::string(),
-		true, "", restore_id, status_id, log_id, restore_token, map_paths, clean_other, ignore_other_fs, share_path, follow_symlinks, restore_flags, ticket);
+		true, "", restore_id, status_id, log_id, restore_token, map_paths, clean_other, ignore_other_fs, share_path, follow_symlinks, restore_flags,
+		ticket, tokens, access_tokens);
 }
 
