@@ -45,6 +45,7 @@ IPipeThrottler *BackupServer::global_local_throttler=NULL;
 IMutex *BackupServer::throttle_mutex=NULL;
 bool BackupServer::file_snapshots_enabled=false;
 bool BackupServer::image_snapshots_enabled = false;
+bool BackupServer::reflink_is_copy = false;
 BackupServer::ESnapshotMethod BackupServer::snapshot_method = BackupServer::ESnapshotMethod_None;
 bool BackupServer::filesystem_transactions_enabled = false;
 bool BackupServer::use_tree_hashing = false;
@@ -54,6 +55,10 @@ std::vector<std::string> BackupServer::force_offline_clients;
 std::map<std::string, std::vector<std::string> >  BackupServer::virtual_clients;
 IMutex* BackupServer::virtual_clients_mutex=NULL;
 bool BackupServer::can_mount_images = false;
+bool BackupServer::can_reflink = false;
+bool BackupServer::can_hardlink = false;
+IMutex* BackupServer::fs_test_mutex = NULL;
+bool BackupServer::can_syncfs = true;
 
 extern IFSImageFactory *image_fak;
 
@@ -64,6 +69,7 @@ BackupServer::BackupServer(IPipe *pExitpipe)
 	exitpipe=pExitpipe;
 	force_offline_mutex=Server->createMutex();
 	virtual_clients_mutex=Server->createMutex();
+	fs_test_mutex = Server->createMutex();
 
 	if(Server->getServerParameter("internet_only_mode")=="true")
 		internet_only_mode=true;
@@ -76,6 +82,7 @@ BackupServer::~BackupServer()
 	Server->destroy(throttle_mutex);
 	Server->destroy(force_offline_mutex);
 	Server->destroy(virtual_clients_mutex);
+	Server->destroy(fs_test_mutex);
 }
 
 void BackupServer::operator()(void)
@@ -146,8 +153,6 @@ void BackupServer::operator()(void)
 #endif
 	}
 
-	testSnapshotAvailability(db);
-	testFilesystemTransactionAvailabiliy(db);
 	if (image_fak != NULL)
 	{
 		can_mount_images = image_fak->initializeImageMounting();
@@ -169,6 +174,7 @@ void BackupServer::operator()(void)
 #endif
 		}
 	}
+	testFilesystem(db);
 	runServerRecovery(db);
 
 	q_get_extra_hostnames=db->Prepare("SELECT id,hostname FROM settings_db.extra_clients");
@@ -214,7 +220,7 @@ void BackupServer::operator()(void)
 
 void BackupServer::findClients(FileClient &fc)
 {
-	std::vector<in_addr> addr_hints;
+	std::vector<FileClient::SAddrHint> addr_hints;
 	if(q_get_extra_hostnames!=NULL)
 	{
 		db_results res=q_get_extra_hostnames->Read();
@@ -222,18 +228,20 @@ void BackupServer::findClients(FileClient &fc)
 
 		for(size_t i=0;i<res.size();++i)
 		{
-			unsigned int dest;
-			bool b=os_lookuphostname((res[i]["hostname"]), &dest);
-			if(b)
+			std::string hostname = res[i]["hostname"];
+			SLookupResult lookup_res;
+			if (os_lookuphostname(hostname, lookup_res))
 			{
-				q_update_extra_ip->Bind((_i64)dest);
+				if(lookup_res.is_ipv6)
+					q_update_extra_ip->Bind(lookup_res.addr_v6, sizeof(lookup_res.addr_v6));
+				else
+					q_update_extra_ip->Bind((_i64)lookup_res.addr_v4);
 				q_update_extra_ip->Bind(res[i]["id"]);
 				q_update_extra_ip->Write();
 				q_update_extra_ip->Reset();
 
-				in_addr tmp;
-				tmp.s_addr=dest;
-				addr_hints.push_back(tmp);
+				FileClient::SAddrHint addr_hint(lookup_res);
+				addr_hints.push_back(addr_hint);
 			}
 		}
 	}
@@ -281,7 +289,7 @@ namespace
 		std::string name;
 		std::string subname;
 		std::string mainname;
-		sockaddr_in addr;
+		FileClient::SAddrHint addr;
 		std::string endpoint_name;
 		bool internetclient;
 		bool delete_pending;
@@ -296,7 +304,7 @@ void BackupServer::startClients(FileClient &fc)
 	if(!internet_only_mode)
 	{
 		std::vector<std::string> names=fc.getServerNames();
-		std::vector<sockaddr_in> servers=fc.getServers();
+		std::vector<FileClient::SAddrHint> servers=fc.getServers();
 
 		client_info.resize(servers.size());
 
@@ -325,6 +333,7 @@ void BackupServer::startClients(FileClient &fc)
 
 		SClientInfo new_client;
 		new_client.name = new_name;
+		new_client.addr.is_ipv6 = false;
 		memset(&(new_client.addr), 0, sizeof(sockaddr_in));
 		new_client.endpoint_name = anames[i].second;
 		new_client.internetclient = true;
@@ -418,11 +427,24 @@ void BackupServer::startClients(FileClient &fc)
 
 			if(c.internet_connection)
 			{
-				ServerStatus::setIP(curr_info.name, inet_addr(curr_info.endpoint_name.c_str()));
+				const char* host = curr_info.endpoint_name.c_str();
+				char dest[16];
+				int rc = inet_pton(AF_INET6, host, dest);
+				if (rc != 1)
+				{
+					ServerStatus::setIP(curr_info.name, inet_addr(curr_info.endpoint_name.c_str()));
+				}
+				else
+				{
+					ServerStatus::setIPv6(curr_info.name, dest);
+				}
 			}
 			else
 			{
-				ServerStatus::setIP(curr_info.name, c.addr.sin_addr.s_addr);
+				if(c.addr.is_ipv6)
+					ServerStatus::setIPv6(curr_info.name, c.addr.addr_ipv6);
+				else
+					ServerStatus::setIP(curr_info.name, c.addr.addr_ipv4);
 			}
 
 			clients.insert(std::pair<std::string, SClient>(curr_info.name, c) );
@@ -437,7 +459,7 @@ void BackupServer::startClients(FileClient &fc)
 				found_lan=true;
 			}
 
-			if(it->second.addr.sin_addr.s_addr==curr_info.addr.sin_addr.s_addr && !found_lan)
+			if(it->second.addr.same(curr_info.addr) && !found_lan)
 			{
 				it->second.offlinecount = 0;
 				it->second.changecount = 0;
@@ -448,7 +470,7 @@ void BackupServer::startClients(FileClient &fc)
 				for(size_t j=0;j<client_info.size();++j)
 				{
 					if(i!=j && client_info[j].name==curr_info.name 
-						&& it->second.addr.sin_addr.s_addr==client_info[j].addr.sin_addr.s_addr)
+						&& it->second.addr.same(client_info[j].addr))
 					{
 						none_fits=false;
 						break;
@@ -461,17 +483,18 @@ void BackupServer::startClients(FileClient &fc)
 					it->second.addr=curr_info.addr;
 					it->second.internet_connection=curr_info.internetclient;
 					std::string msg;
-					msg.resize(7+sizeof(sockaddr_in)+1);
-					msg[0]='a'; msg[1]='d'; msg[2]='d'; msg[3]='r'; msg[4]='e'; msg[5]='s'; msg[6]='s';
-					memcpy(&msg[7], &it->second.addr, sizeof(sockaddr_in));
-					msg[7+sizeof(sockaddr_in)]=(curr_info.internetclient?1:0);
+					msg.resize(7 + sizeof(FileClient::SAddrHint) + 1);
+					memcpy(&msg[0], "address", 7);
+					memcpy(&msg[7], &it->second.addr, sizeof(FileClient::SAddrHint));
+					msg[7+sizeof(FileClient::SAddrHint)]=(curr_info.internetclient?1:0);
 					it->second.pipe->Write(msg);
 
-					char *ip=(char*)&it->second.addr.sin_addr.s_addr;
+					Server->Log("New client address: "+it->second.addr.toString(), LL_INFO);
 
-					Server->Log("New client address: "+convert((unsigned char)ip[0])+"."+convert((unsigned char)ip[1])+"."+convert((unsigned char)ip[2])+"."+convert((unsigned char)ip[3]), LL_INFO);
-
-					ServerStatus::setIP(curr_info.name, it->second.addr.sin_addr.s_addr);
+					if(it->second.addr.is_ipv6)
+						ServerStatus::setIPv6(curr_info.name, it->second.addr.addr_ipv6);
+					else
+						ServerStatus::setIP(curr_info.name, it->second.addr.addr_ipv4);
 
 					it->second.offlinecount=0;
 				}
@@ -702,8 +725,18 @@ bool BackupServer::isImageSnapshotsEnabled()
 	return image_snapshots_enabled;
 }
 
-BackupServer::ESnapshotMethod BackupServer::getSnapshotMethod()
+bool BackupServer::isReflinkCopy()
 {
+	return reflink_is_copy;
+}
+
+BackupServer::ESnapshotMethod BackupServer::getSnapshotMethod(bool image)
+{
+	if (image 
+		&& snapshot_method == ESnapshotMethod_ZfsFile)
+	{
+		return ESnapshotMethod_Zfs;
+	}
 	return snapshot_method;
 }
 
@@ -777,6 +810,11 @@ void BackupServer::testSnapshotAvailability(IDatabase *db)
 		{
 			Server->Log("Backup destination does handle subvolumes and snapshots. Snapshots enabled for image backups.", LL_INFO);
 		}
+
+		if (reflink_is_copy)
+		{
+			Server->Log("Emulating reflinks via copying", LL_INFO);
+		}
 	}
 	else
 	{
@@ -834,6 +872,110 @@ void BackupServer::testFilesystemTransactionAvailabiliy( IDatabase *db )
 	}
 }
 
+void BackupServer::testFilesystemLinkAvailability(IDatabase *db, bool reflink)
+{
+	std::string test_type = reflink ? "reflink" : "hardlink";
+	std::string test_type_c = reflink ? "Reflink" : "Hardlink";
+	Server->Log("Testing for "+ test_type+"s in backup destination...", LL_DEBUG);
+
+	ServerSettings settings(db);
+
+	const std::string testfilename = "FGHTR654kgfdfg5764578kldsfsdfgre66juzfo";
+	const std::string testfilename_reflinked = testfilename + "_2";
+
+	std::string backupfolder = settings.getSettings()->backupfolder;
+
+	writestring("test", backupfolder + os_file_sep() + testfilename);
+
+	ScopedDeleteFn delete_fn(backupfolder + os_file_sep() + testfilename);
+
+	bool b = os_create_hardlink(backupfolder + os_file_sep() + testfilename_reflinked,
+		backupfolder + os_file_sep() + testfilename, reflink, NULL);
+
+	if (!b)
+	{
+		Server->Log("Could not create "+test_type+" at backup destination. "+ test_type_c+"s disabled. "+os_last_error_str(), LL_DEBUG);
+		return;
+	}
+
+	ScopedDeleteFn delete_fn_2(backupfolder + os_file_sep() + testfilename_reflinked);
+
+	if (getFile(backupfolder + os_file_sep() + testfilename_reflinked)
+		== "test")
+	{
+		Server->Log("Could create "+test_type+" at backup destination. "+ test_type_c +"s enabled.", LL_DEBUG);
+		if (reflink)
+			can_reflink = true;
+		else
+			can_hardlink = true;
+	}
+	else
+	{
+		Server->Log(test_type_c+" test file content wrong. "+ test_type_c +"s disabled.", LL_DEBUG);
+	}
+}
+
+void BackupServer::testFilesystemLinkAvailability(IDatabase * db)
+{
+	testFilesystemLinkAvailability(db, false);
+	testFilesystemLinkAvailability(db, true);
+}
+
+void BackupServer::testFilesystemSyncFs(IDatabase * db)
+{
+	ServerSettings settings(db);
+
+	std::string backupfolder = settings.getSettings()->backupfolder;
+
+	std::string testfolder = backupfolder + os_file_sep() + "05b89714-7c4d-40fb-a908-e45f326c0624";
+
+	if (!os_directory_exists(testfolder)
+		&& !os_create_dir(testfolder))
+	{
+		Server->Log("Could not create folder at " + testfolder + ". " + os_last_error_str(), LL_DEBUG);
+		return;
+	}
+
+	if (!os_sync(testfolder))
+	{
+		Server->Log("Syncing " + testfolder + " failed. This may cause problems with backup consistency on disk. " + os_last_error_str(), LL_WARNING);
+		can_syncfs = false;
+	}
+	else
+	{
+		can_syncfs = true;
+	}
+
+	os_remove_dir(testfolder);
+}
+
+void BackupServer::testFilesystem(IDatabase * db)
+{
+	IScopedLock lock(fs_test_mutex);
+
+	testSnapshotAvailability(db);
+	testFilesystemTransactionAvailabiliy(db);
+	testFilesystemLinkAvailability(db);
+	testFilesystemSyncFs(db);
+}
+
+namespace
+{
+	class TestFilesystemThread : public IThread
+	{
+	public:
+		void operator()() {
+			BackupServer::testFilesystem(Server->getDatabase(Server->getThreadID(), URBACKUPDB_SERVER));
+			delete this;
+		}
+	};
+}
+
+void BackupServer::testFilesystemThread()
+{
+	Server->createThread(new TestFilesystemThread, "test fs");
+}
+
 bool BackupServer::isFilesystemTransactionEnabled()
 {
 	return filesystem_transactions_enabled;
@@ -844,6 +986,21 @@ bool BackupServer::canMountImages()
 	return can_mount_images;
 }
 
+bool BackupServer::canReflink()
+{
+	return can_reflink;
+}
+
+bool BackupServer::canHardlink()
+{
+	return can_hardlink;
+}
+
+bool BackupServer::canSyncFs()
+{
+	return can_syncfs;
+}
+
 void BackupServer::updateDeletePending()
 {
 	IScopedLock lock(throttle_mutex);
@@ -922,11 +1079,45 @@ void BackupServer::enableSnapshots(int method)
 	if (snapshot_method == ESnapshotMethod_Btrfs)
 	{
 		file_snapshots_enabled = true;
-	}	
+		can_reflink = true;
+		can_hardlink = false;
+	}
+	else if (snapshot_method == ESnapshotMethod_ZfsFile)
+	{
+		file_snapshots_enabled = true;
+		reflink_is_copy = true;
+		can_hardlink = false;
+	}
+}
+
+namespace
+{
+	class FileCleanups : public IThread
+	{
+	public:
+		void operator()()
+		{
+			logid_t logid = ServerLogger::getLogId(LOG_CATEGORY_CLEANUP);
+			ScopedProcess recovery_process(std::string(), sa_startup_recovery, std::string(), logid, false, LOG_CATEGORY_CLEANUP);
+
+			for (size_t i = 0; i < cleanups.size(); ++i)
+			{
+				std::string backupinfo = "[id=" + convert(cleanups[i]->getCleanupAction().backupid)+ ", clientid=" + convert(cleanups[i]->getCleanupAction().clientid) + "]";
+				ServerLogger::Log(logid, "Deleting file backup " + backupinfo + "...", LL_WARNING);
+
+				Server->getThreadPool()->executeWait(cleanups[i], "delete fbackup");
+			}
+			delete this;
+		}
+
+		std::vector<ServerCleanupThread*> cleanups;
+	};
 }
 
 void BackupServer::runServerRecovery(IDatabase * db)
 {
+	return;
+
 	logid_t logid = ServerLogger::getLogId(LOG_CATEGORY_CLEANUP);
 	ScopedProcess recovery_process(std::string(), sa_startup_recovery, std::string(), logid, false, LOG_CATEGORY_CLEANUP);
 
@@ -974,6 +1165,14 @@ void BackupServer::runServerRecovery(IDatabase * db)
 		}
 	}
 
+	if (!delete_missing)
+	{
+		ServerLogger::Log(logid, "Didn't find any backup on backup storage. Not removing any backups from database during startup cleanup...", LL_ERROR);
+	}
+
+	ServerBackupDao backupdao(db);
+	std::vector<std::string> old_backupfolders = backupdao.getOldBackupfolders();
+
 	std::vector<db_single_result> to_delete;
 	bool has_delete = false;
 
@@ -988,6 +1187,25 @@ void BackupServer::runServerRecovery(IDatabase * db)
 			std::string backuppath = backupfolder + os_file_sep() + res["name"] + os_file_sep() + res["path"];
 			std::string backupinfo = "[id=" + res["backupid"] + ", path=" + res["path"] + ", backuptime=" + res["backuptime"] + ", clientid=" + res["clientid"] + ", client=" + res["name"] + "]";
 
+			if(!os_directory_exists(os_file_prefix(backuppath))
+				&& os_directory_exists(os_file_prefix(backuppath + ".startup-del")) )
+			{
+				backuppath += ".startup-del";
+			}
+
+			if (!os_directory_exists(os_file_prefix(backuppath)))
+			{
+				for (size_t i = 0; i < old_backupfolders.size(); ++i)
+				{
+					std::string old_backuppath = old_backupfolders[i] + os_file_sep() + res["name"] + os_file_sep() + res["path"];
+					if (os_directory_exists(os_file_prefix(old_backuppath)))
+					{
+						backuppath = old_backuppath;
+						break;
+					}
+				}
+			}
+
 			bool delete_backup = false;
 
 			if (res["done"] == "0")
@@ -1029,23 +1247,48 @@ void BackupServer::runServerRecovery(IDatabase * db)
 		}
 	}
 
+	IQuery* q_set_done = db->Prepare("UPDATE backups SET done=0, complete=0 WHERE id=?");
+	std::auto_ptr<FileCleanups> file_cleanups(new FileCleanups);
 	for (size_t i = 0; i < to_delete.size(); ++i)
 	{
 		has_delete = true;
 
 		db_single_result res = to_delete[i];
+
+		q_set_done->Bind(res["backupid"]);
+		q_set_done->Write();
+		q_set_done->Reset();
+
 		std::string backupinfo = "[id=" + res["backupid"] + ", path=" + res["path"] + ", backuptime=" + res["backuptime"] + ", clientid=" + res["clientid"] + ", client=" + res["name"] + "]";
 		ServerLogger::Log(logid, "Deleting file backup "+backupinfo+"...", LL_WARNING);
-		Server->getThreadPool()->executeWait(
-			new ServerCleanupThread(CleanupAction(backupfolder, watoi(res["clientid"]), watoi(res["backupid"]), true, NULL)),
-			"delete fbackup");
+		std::string backuppath = backupfolder + os_file_sep() + res["name"] + os_file_sep() + res["path"];
+
+		ServerCleanupThread* cleanup = new ServerCleanupThread(CleanupAction(backupfolder, watoi(res["clientid"]), watoi(res["backupid"]), true, NULL));
+		if (!os_directory_exists(os_file_prefix(backuppath))
+			&& os_directory_exists(os_file_prefix(backuppath + ".startup-del")))
+		{
+			file_cleanups->cleanups.push_back(cleanup);
+		}
+		else if (os_rename_file(os_file_prefix(backuppath), os_file_prefix(backuppath + ".startup-del")))
+		{
+			file_cleanups->cleanups.push_back(cleanup);
+		}
+		else
+		{
+			Server->getThreadPool()->executeWait(cleanup, "delete fbackup");
+		}
+	}
+
+	if (!file_cleanups->cleanups.empty())
+	{
+		Server->getThreadPool()->execute(file_cleanups.release());
 	}
 
 	to_delete.clear();
 
 	num_extra_check = 30;
 
-	std::vector<db_single_result> todelete_images;
+	std::vector<std::pair<db_single_result, bool> > todelete_images;
 
 	{
 		IQuery* q_synced = db->Prepare("SELECT backup_images.id AS backupid, name, path, clientid, backuptime FROM (backup_images INNER JOIN clients ON backup_images.clientid=clients.id) WHERE synctime IS NOT NULL ORDER BY synctime DESC");
@@ -1053,11 +1296,33 @@ void BackupServer::runServerRecovery(IDatabase * db)
 		db_single_result res;
 		while (cur.next(res))
 		{
-			std::string backupinfo = "[id=" + res["backupid"] + ", path=" + res["path"] + ", backuptime=" + res["backuptime"] + ", clientid=" + res["clientid"] + ", client=" + res["name"] + "]";
+			std::string path = res["path"];
+			std::string backupinfo = "[id=" + res["backupid"] + ", path=" + path + ", backuptime=" + res["backuptime"] + ", clientid=" + res["clientid"] + ", client=" + res["name"] + "]";
 
-			std::auto_ptr<IFile> image_f(Server->openFile(os_file_prefix(res["path"]), MODE_READ));
+			std::auto_ptr<IFile> image_f(Server->openFile(os_file_prefix(path), MODE_READ));
+
+			if (image_f.get() == NULL)
+			{
+				for (size_t i = 0; i < old_backupfolders.size(); ++i)
+				{
+					std::string old_path = path;
+					if (next(old_path, 0, backupfolder + os_file_sep()))
+					{
+						old_path.erase(0, backupfolder.size() + os_file_sep().size());
+						old_path = old_backupfolders[i] + os_file_sep() + old_path;
+					}
+
+					image_f.reset(Server->openFile(os_file_prefix(old_path), MODE_READ));
+					if (image_f.get() != NULL)
+					{
+						path = old_path;
+						break;
+					}
+				}
+			}
 
 			bool delete_backup = false;
+			bool delete_db_entry = false;
 
 			if (image_f.get() == NULL)
 			{
@@ -1065,6 +1330,7 @@ void BackupServer::runServerRecovery(IDatabase * db)
 				{
 					ServerLogger::Log(logid, "Image backup " + backupinfo + " does not exist on backup storage. Deleting it from database.", LL_WARNING);
 					delete_backup = true;
+					delete_db_entry = true;
 				}
 			}
 			else
@@ -1073,14 +1339,14 @@ void BackupServer::runServerRecovery(IDatabase * db)
 
 				if (sync_f.get() == NULL)
 				{
-					ServerLogger::Log(logid, "Image backup " + backupinfo + " is not synced to backup storage. Deleting it.", LL_WARNING);
+					ServerLogger::Log(logid, "Image backup " + backupinfo + " is not synced to backup storage. Setting it to incomplete.", LL_WARNING);
 					delete_backup = true;
 				}
 			}
 
 			if (delete_backup)
 			{
-				to_delete.push_back(res);
+				todelete_images.push_back( std::make_pair(res, delete_db_entry) );
 			}
 			else
 			{
@@ -1094,22 +1360,36 @@ void BackupServer::runServerRecovery(IDatabase * db)
 	}
 
 	IQuery* q_delete_image = db->Prepare("DELETE FROM backup_images WHERE id=?");
-	for (size_t i = 0; i < to_delete.size(); ++i)
+	IQuery* q_set_complete = db->Prepare("UPDATE backup_images SET complete=0 WHERE id=?");
+	for (size_t i = 0; i < todelete_images.size(); ++i)
 	{
 		has_delete = true;
 
-		db_single_result res = to_delete[i];
+		db_single_result res = todelete_images[i].first;
+		bool delete_db_entry = todelete_images[i].second;
 		std::string backupinfo = "[id=" + res["backupid"] + ", path=" + res["path"] + ", backuptime=" + res["backuptime"] + ", clientid=" + res["clientid"] + ", client=" + res["name"] + "]";
-		ServerLogger::Log(logid, "Deleting image backup " + backupinfo + "...", LL_WARNING);
 
-		ServerCleanupThread::deleteImage(logid, res["name"], res["path"]);
+		if (delete_db_entry)
+		{
+			ServerLogger::Log(logid, "Deleting image backup " + backupinfo + "...", LL_WARNING);
 
-		q_delete_image->Bind(watoi(res["backupid"]));
-		q_delete_image->Write();
-		q_delete_image->Reset();
+			ServerCleanupThread::deleteImage(logid, res["name"], res["path"]);
+
+			q_delete_image->Bind(watoi(res["backupid"]));
+			q_delete_image->Write();
+			q_delete_image->Reset();
+		}
+		else
+		{
+			ServerLogger::Log(logid, "Setting image backup " + backupinfo + " to incomplete...", LL_WARNING);
+
+			q_set_complete->Bind(watoi(res["backupid"]));
+			q_set_complete->Write();
+			q_set_complete->Reset();
+		}
 	}
 
-	to_delete.clear();
+	todelete_images.clear();
 
 	std::string db_backupfolder = backupfolder + os_file_sep() + "urbackup";
 
@@ -1147,6 +1427,11 @@ void BackupServer::runServerRecovery(IDatabase * db)
 	db->destroyAllQueries();
 }
 
+void BackupServer::wakeupNewClient()
+{
+	exitpipe->Write("nc");
+}
+
 std::string BackupServer::findFile(const std::string & path, const std::string & fn)
 {
 	std::vector<SFile> files = getFiles(path);
@@ -1208,7 +1493,7 @@ void BackupServer::setupUseTreeHashing()
 void BackupServer::setVirtualClients( const std::string& clientname, const std::string& new_virtual_clients )
 {
 	std::vector<std::string> toks;
-	TokenizeMail(new_virtual_clients, toks, "|");
+	Tokenize(new_virtual_clients, toks, "|");
 
 	IScopedLock lock(virtual_clients_mutex);
 	virtual_clients[clientname] = toks;
