@@ -35,11 +35,13 @@
 #include <memory.h>
 #include <algorithm>
 #include <limits.h>
+#include <math.h>
 #include "../fileservplugin/IFileServ.h"
 #include "server_log.h"
 #include "restore_client.h"
 #include "serverinterface/backups.h"
 #include "dao/ServerBackupDao.h"
+#include "../urbackupcommon/mbrdata.h"
 
 const unsigned short serviceport=35623;
 extern IFSImageFactory *image_fak;
@@ -152,10 +154,13 @@ bool ServerChannelThread::isOnline()
 }
 
 ServerChannelThread::ServerChannelThread(ClientMain *client_main, const std::string& clientname, int clientid,
-	bool internet_mode, const std::string& identity, std::string server_token, const std::string& virtual_client) :
+	bool internet_mode, bool allow_restore,
+	 std::string server_token, const std::string& virtual_client,
+	ServerChannelThread* parent) :
 	client_main(client_main), clientname(clientname), clientid(clientid), settings(NULL),
-		internet_mode(internet_mode), keepalive_thread(NULL), server_token(server_token),
-	virtual_client(virtual_client)
+		internet_mode(internet_mode), allow_restore(allow_restore), keepalive_thread(NULL), server_token(server_token),
+	virtual_client(virtual_client), allow_shutdown(true),
+	parent(parent), next_reauth_time(0), reauth_tries(0)
 {
 	do_exit=false;
 	mutex=Server->createMutex();
@@ -168,7 +173,17 @@ ServerChannelThread::~ServerChannelThread(void)
 	Server->destroy(mutex);
 }
 
-void ServerChannelThread::operator()(void)
+void ServerChannelThread::operator()()
+{
+	run();
+
+	if (parent != NULL)
+	{
+		delete this;
+	}
+}
+
+void ServerChannelThread::run()
 {
 	int64 lastpingtime=0;
 	lasttime=0;
@@ -178,11 +193,11 @@ void ServerChannelThread::operator()(void)
 
 	std::string curr_ident;
 
-	while(do_exit==false)
+	while(!do_exit)
 	{
 		if(input==NULL)
 		{
-			IPipe *np=client_main->getClientCommandConnection(10000, &client_addr);
+			IPipe *np=client_main->getClientCommandConnection(settings, 10000, &client_addr);
 			if(np==NULL)
 			{
 				Server->Log("Connecting Channel to "+clientname+" failed - CONNECT error -55", LL_DEBUG);
@@ -195,6 +210,8 @@ void ServerChannelThread::operator()(void)
 					input=np;
 				}
 				curr_ident = client_main->getIdentity();
+				tcpstack.reset();
+				tcpstack.setAddChecksum(client_main->isOnInternetConnection());
 				tcpstack.Send(input, curr_ident +"1CHANNEL capa="+convert(constructCapabilities())+"&token="+server_token+"&restore_version=1&virtual_client="+EscapeParamString(virtual_client));
 
 				lasttime=Server->getTimeMS();
@@ -206,19 +223,13 @@ void ServerChannelThread::operator()(void)
 			if(Server->getTimeMS()-lasttime>180000)
 			{
 				Server->Log("Resetting channel to "+clientname+" because of timeout.", LL_DEBUG);
-				IScopedLock lock(mutex);
-				Server->destroy(input);
-				input=NULL;
-				tcpstack.reset();
+				reset();
 			}
 
 			if (curr_ident != client_main->getIdentity())
 			{
 				Server->Log("Resetting channel to " + clientname + " because session identity changed.", LL_DEBUG);
-				IScopedLock lock(mutex);
-				Server->destroy(input);
-				input = NULL;
-				tcpstack.reset();
+				reset();
 			}
 
 			if(input!=NULL)
@@ -246,10 +257,7 @@ void ServerChannelThread::operator()(void)
 					if(input!=NULL && was_updated)
 					{
 						Server->Log("Settings changed. Capabilities may have changed. Reconnecting channel...", LL_DEBUG);
-						IScopedLock lock(mutex);
-						Server->destroy(input);
-						input=NULL;
-						tcpstack.reset();
+						reset();
 					}
 				}
 				else if(rc==0)
@@ -257,10 +265,7 @@ void ServerChannelThread::operator()(void)
 					if(input->hasError())
 					{
 						Server->Log("Lost channel connection to "+clientname+". has_error=true", LL_DEBUG);
-						IScopedLock lock(mutex);
-						Server->destroy(input);
-						input=NULL;
-						tcpstack.reset();
+						reset();
 						Server->wait(1000);
 					}
 					else
@@ -294,7 +299,8 @@ void ServerChannelThread::doExit(void)
 {
 	IScopedLock lock(mutex);
 	do_exit=true;
-	if(input!=NULL)
+	if(input!=NULL
+		&& allow_shutdown)
 	{
 		input->shutdown();
 	}
@@ -302,8 +308,26 @@ void ServerChannelThread::doExit(void)
 
 std::string ServerChannelThread::processMsg(const std::string &msg)
 {
+	if (msg != "ERR")
+	{
+		reauth_tries = 0;
+		next_reauth_time = 0;
+	}
+
 	if(msg=="ERR")
 	{
+		if (next_reauth_time == 0
+			|| Server->getTimeMS() > next_reauth_time)
+		{
+			client_main->forceReauthenticate();
+			client_main->sendToPipe("WAKEUP");
+
+			++reauth_tries;
+
+			unsigned int waittime = (std::min)(static_cast<unsigned int>(1000.*pow(2., static_cast<double>(reauth_tries))), (unsigned int)30 * 60 * 1000);
+			waittime = (std::max)((unsigned int)60000, waittime);
+			next_reauth_time = Server->getTimeMS()+waittime;
+		}
 		return std::string();
 	}
 	else if(msg=="START BACKUP INCR")
@@ -342,28 +366,35 @@ std::string ServerChannelThread::processMsg(const std::string &msg)
 		ParseParamStrHttp(s_params, &params);
 		client_main->setConnectionMetered(params["metered"] == "1");
 	}
-	else if(next(msg, 0, "LOGIN ") && !internet_mode)
+	else if(next(msg, 0, "LOGIN ") && allow_restore)
 	{
 		std::string s_params=msg.substr(6);
 		str_map params;
 		ParseParamStrHttp(s_params, &params);
 		LOGIN(params);
 	}
-	else if(next(msg, 0, "SALT ") && !internet_mode)
+	else if(next(msg, 0, "SALT ") && allow_restore)
 	{
 		std::string s_params=msg.substr(5);
 		str_map params;
 		ParseParamStrHttp(s_params, &params);
 		SALT(params);
 	}
-	else if(msg=="GET BACKUPCLIENTS" && !internet_mode && hasDownloadImageRights() )
+	else if(msg=="GET BACKUPCLIENTS" && allow_restore && hasDownloadImageRights())
 	{
 		GET_BACKUPCLIENTS();
 	}
-	else if(next(msg, 0, "GET BACKUPIMAGES ") && !internet_mode && hasDownloadImageRights())
+	else if(next(msg, 0, "GET BACKUPIMAGES ") && allow_restore)
 	{
-		std::string name=(msg.substr(17));
-		GET_BACKUPIMAGES(name);
+		if (!hasDownloadImageRights())
+		{
+			tcpstack.Send(input, "0|0|0|NO RIGHTS");
+		}
+		else
+		{
+			std::string name = (msg.substr(17));
+			GET_BACKUPIMAGES(name);
+		}
 	}
     else if(next(msg, 0, "GET FILE BACKUPS TOKENS"))
     {
@@ -396,13 +427,26 @@ std::string ServerChannelThread::processMsg(const std::string &msg)
 		str_map params;
 		ParseParamStrHttp(s_params, &params);
 	}
-	else if(next(msg, 0, "DOWNLOAD IMAGE ") && !internet_mode && hasDownloadImageRights())
+	else if(next(msg, 0, "DOWNLOAD IMAGE ") && allow_restore && hasDownloadImageRights())
 	{
 		std::string s_params=msg.substr(15);
 		str_map params;
 		ParseParamStrHttp(s_params, &params);
 
+		{
+			IScopedLock lock(mutex);
+			allow_shutdown = false;
+		}
+
+		add_extra_channel();
 		DOWNLOAD_IMAGE(params);
+		remove_extra_channel();
+		Server->getDatabase(Server->getThreadID(), URBACKUPDB_SERVER)->destroyAllQueries();
+
+		{
+			IScopedLock lock(mutex);
+			allow_shutdown = true;
+		}
 	}
 	else if(next(msg, 0, "DOWNLOAD FILES TOKENS "))
 	{
@@ -468,10 +512,12 @@ std::string ServerChannelThread::processMsg(const std::string &msg)
 	}
 	else
 	{
-		IScopedLock lock(mutex);
-		Server->destroy(input);
-		input=NULL;
-		tcpstack.reset();
+		{
+			IScopedLock lock(mutex);
+			Server->destroy(input);
+			input = NULL;
+			tcpstack.reset();
+		}
 		Server->wait(60000);
 	}
 	return "";
@@ -514,7 +560,7 @@ int ServerChannelThread::constructCapabilities(void)
 		capa |= DONT_ALLOW_FILE_RESTORE;
 	if (!cs->allow_component_restore)
 		capa |= DONT_ALLOW_COMPONENT_RESTORE;
-	if (!cs->allow_component_restore)
+	if (!cs->allow_component_config)
 		capa |= DONT_ALLOW_COMPONENT_CONFIG;
 
 	return capa;
@@ -555,7 +601,7 @@ void ServerChannelThread::LOGIN(str_map& params)
 		{
 			helper.getSession()->id=user_id;
 			PARAMS["REMOTE_ADDR"]=client_addr;
-			logSuccessfullLogin(helper, PARAMS, params["username"], LoginMethod_RestoreCD);
+			logSuccessfulLogin(helper, PARAMS, params["username"], LoginMethod_RestoreCD);
 			tcpstack.Send(input, "ok");
 		}
 		else
@@ -567,7 +613,7 @@ void ServerChannelThread::LOGIN(str_map& params)
 	}
 	else
 	{
-		logSuccessfullLogin(helper, PARAMS, "anonymous", LoginMethod_RestoreCD);
+		logSuccessfulLogin(helper, PARAMS, "anonymous", LoginMethod_RestoreCD);
 		tcpstack.Send(input, "ok");
 	}
 }
@@ -704,14 +750,13 @@ void ServerChannelThread::GET_BACKUPIMAGES(const std::string& clientname)
 {
 	IDatabase *db=Server->getDatabase(Server->getThreadID(), URBACKUPDB_SERVER);
 	//TODO language
-	IQuery *q=db->Prepare("SELECT backupid AS id, strftime('%s', backuptime) AS timestamp, strftime('%Y-%m-%d %H:%M',backuptime) AS backuptime, letter, clientid FROM ((SELECT id AS backupid, clientid, backuptime, letter, complete FROM backup_images) c INNER JOIN (SELECT id FROM clients WHERE name=?) b ON c.clientid=b.id) a WHERE a.complete=1 AND length(a.letter)<=2 ORDER BY backuptime DESC");
+	IQuery *q=db->Prepare("SELECT backupid AS id, strftime('%s', backuptime) AS timestamp, strftime('%Y-%m-%d %H:%M',backuptime,'localtime') AS backuptime, letter, clientid FROM ((SELECT id AS backupid, clientid, backuptime, letter, complete FROM backup_images) c INNER JOIN (SELECT id FROM clients WHERE name=?) b ON c.clientid=b.id) a WHERE a.complete=1 AND a.letter!='SYSVOL' AND a.letter!='ESP' ORDER BY backuptime DESC");
 	q->Bind(clientname);
 	db_results res=q->Read();
 
 	for(size_t i=0;i<res.size();++i)
 	{
-		if(!all_client_rights &&
-			std::find(client_right_ids.begin(), client_right_ids.end(), watoi(res[i]["clientid"]))==client_right_ids.end())
+		if(!has_restore_permission(clientname, watoi(res[i]["clientid"])))
 		{
 			tcpstack.Send(input, "0|0|0|NO RIGHTS");
 			db->destroyAllQueries();
@@ -720,7 +765,7 @@ void ServerChannelThread::GET_BACKUPIMAGES(const std::string& clientname)
 	}
 
 	std::string r;
-	q=db->Prepare("SELECT backupid AS id,strftime('%s', backuptime) AS timestamp, strftime('%Y-%m-%d %H:%M',backuptime) AS backuptime FROM ((SELECT id AS backupid, clientid, backuptime, letter, complete FROM backup_images) c INNER JOIN (SELECT assoc_id, img_id FROM assoc_images WHERE img_id=?) b ON backupid=b.assoc_id) a WHERE a.complete=1 ORDER BY backuptime DESC");
+	q=db->Prepare("SELECT backupid AS id,strftime('%s', backuptime) AS timestamp, strftime('%Y-%m-%d %H:%M',backuptime,'localtime') AS backuptime FROM ((SELECT id AS backupid, clientid, backuptime, letter, complete FROM backup_images) c INNER JOIN (SELECT assoc_id, img_id FROM assoc_images WHERE img_id=?) b ON backupid=b.assoc_id) a WHERE a.complete=1 ORDER BY backuptime DESC");
 	for(size_t i=0;i<res.size();++i)
 	{
 		r+=convert(img_id_offset+watoi(res[i]["id"]))+"|"+(res[i]["timestamp"])+"|"+(res[i]["backuptime"])+"|"+(res[i]["letter"])+"\n";
@@ -743,7 +788,7 @@ void ServerChannelThread::GET_BACKUPIMAGES(const std::string& clientname)
 void ServerChannelThread::GET_FILE_BACKUPS( const std::string& clientname )
 {
 	IDatabase *db=Server->getDatabase(Server->getThreadID(), URBACKUPDB_SERVER);
-	IQuery *q=db->Prepare("SELECT backupid AS id, strftime('%s', backuptime) AS timestamp, strftime('%Y-%m-%d %H:%M',backuptime) AS backuptime, clientid, tgroup FROM "
+	IQuery *q=db->Prepare("SELECT backupid AS id, strftime('%s', backuptime) AS timestamp, strftime('%Y-%m-%d %H:%M',backuptime,'localtime') AS backuptime, clientid, tgroup FROM "
 		"((SELECT id AS backupid, clientid, backuptime, complete, tgroup FROM backups) c INNER JOIN (SELECT id FROM clients WHERE name=?) b ON c.clientid=b.id) a "
 		"WHERE a.complete=1 ORDER BY backuptime DESC");
 	q->Bind(clientname);
@@ -751,8 +796,7 @@ void ServerChannelThread::GET_FILE_BACKUPS( const std::string& clientname )
 
 	for(size_t i=0;i<res.size();++i)
 	{
-		if(!all_client_rights &&
-			std::find(client_right_ids.begin(), client_right_ids.end(), watoi(res[i]["clientid"]))==client_right_ids.end())
+		if(!has_restore_permission(clientname, watoi(res[i]["clientid"])))
 		{
 			tcpstack.Send(input, "0|0|0|NO RIGHTS");
 			db->destroyAllQueries();
@@ -875,10 +919,45 @@ void ServerChannelThread::GET_FILE_LIST_TOKENS(str_map& params)
 	ServerStatus::updateActive();
 }
 
+namespace
+{
+	class ScopedDestroyVhdfile
+	{
+		IVHDFile* vhdfile;
+	public:
+		ScopedDestroyVhdfile(IVHDFile* vhdfile)
+			: vhdfile(vhdfile) {}
+		~ScopedDestroyVhdfile() {
+			image_fak->destroyVHDFile(vhdfile);
+		}
+	};
+
+	class ScopedSetRestoreDone
+	{
+		bool ok;
+		ServerBackupDao& backup_dao;
+		int64 restore_id;
+	public:
+		ScopedSetRestoreDone(ServerBackupDao& backup_dao, int64 restore_id)
+			: backup_dao(backup_dao), ok(false), restore_id(restore_id)
+		{}
+
+		~ScopedSetRestoreDone() {
+			backup_dao.setRestoreDone(ok ? 1 : 0, restore_id);
+		}
+
+		void set_ok() {
+			ok = true;
+		}
+	};
+}
+
 void ServerChannelThread::DOWNLOAD_IMAGE(str_map& params)
 {
 	IDatabase *db = Server->getDatabase(Server->getThreadID(), URBACKUPDB_SERVER);
 
+	const _u32 img_send_timeout = 30000;
+
 	ServerBackupDao backup_dao(db);
 
 	int img_id=watoi(params["img_id"])-img_id_offset;
@@ -889,32 +968,51 @@ void ServerChannelThread::DOWNLOAD_IMAGE(str_map& params)
 	db_results res=q->Read();
 	if(res.empty())
 	{
+		Server->Log("Image to download not found (img_id=" + convert(img_id) + " time=" + params["time"] + ")", LL_DEBUG);
 		_i64 r=-1;
-		input->Write((char*)&r, sizeof(_i64));
+		if (!input->Write((char*)&r, sizeof(_i64), img_send_timeout))
+		{
+			reset();
+			return;
+		}
 	}
 	else
 	{
-		if( !all_client_rights &&
-			std::find(client_right_ids.begin(), client_right_ids.end(), watoi(res[0]["clientid"]))==client_right_ids.end())
+		std::string clientname = get_clientname(db, watoi(res[0]["clientid"]));
+
+		if(clientname.empty()
+			|| !has_restore_permission(clientname, watoi(res[0]["clientid"])))
 		{
+			Server->Log("No permission to download image of client with id " + res[0]["clientid"], LL_DEBUG);
 			_i64 r=-1;
-			input->Write((char*)&r, sizeof(_i64));
+			if (!input->Write((char*)&r, sizeof(_i64), img_send_timeout))
+			{
+				reset();
+			}
 			return;
 		}
 
 		int img_version=watoi(res[0]["version"]);
 		if(params["mbr"]=="true")
 		{
-			IFile *f=Server->openFile(os_file_prefix(res[0]["path"]+".mbr"), MODE_READ);
-			if(f==NULL)
+			std::auto_ptr<IFile> f(Server->openFile(os_file_prefix(res[0]["path"]+".mbr"), MODE_READ));
+			if(f.get()==NULL)
 			{
 				_i64 r=little_endian(-1);
-				input->Write((char*)&r, sizeof(_i64));
+				if (!input->Write((char*)&r, sizeof(_i64), img_send_timeout))
+				{
+					reset();
+				}
+				return;
 			}
 			else
 			{
 				_i64 r=little_endian(f->Size());
-				input->Write((char*)&r, sizeof(_i64));
+				if (!input->Write((char*)&r, sizeof(_i64), img_send_timeout))
+				{
+					reset();
+					return;
+				}
 				char buf[4096];
 				_u32 rc;
 				do
@@ -922,22 +1020,17 @@ void ServerChannelThread::DOWNLOAD_IMAGE(str_map& params)
 					rc=f->Read(buf, 4096);
 					if(rc>0)
 					{
-						bool b=input->Write(buf, rc);
+						bool b=input->Write(buf, rc, img_send_timeout);
 						if(!b)
 						{
 							Server->Log("Writing to output pipe failed processMsg-2", LL_ERROR);
-							Server->destroy(f);
-							db->destroyAllQueries();
-							Server->destroy(input);
-							input=NULL;
+							reset();
 							return;
 						}
 						lasttime=Server->getTimeMS();
 					}
 				}
 				while(rc>0);
-				Server->destroy(f);
-				db->destroyAllQueries();
 				return;
 			}
 		}
@@ -963,18 +1056,23 @@ void ServerChannelThread::DOWNLOAD_IMAGE(str_map& params)
 			vhdfile = image_fak->createVHDFile(res[0]["path"], true, 0);
 		}
 
+		ScopedDestroyVhdfile destroy_vhdfile(vhdfile);
+
 		if(vhdfile==NULL 
 			|| !vhdfile->isOpen())
 		{
 			_i64 r=-1;
-			input->Write((char*)&r, sizeof(_i64));
+			if (!input->Write((char*)&r, sizeof(_i64), img_send_timeout))
+			{
+				reset();
+			}
 		}
 		else
 		{
 			std::string clientname = backup_dao.getClientnameByImageid(img_id).value;
 			ScopedProcess restore_process(clientname, sa_restore_image, res[0]["letter"], logid_t(), false, watoi(res[0]["clientid"]));
 			backup_dao.addRestore(watoi(res[0]["clientid"]), std::string(), std::string(), 1, res[0]["letter"]);
-			int64 restore_id = db->getLastInsertID();
+			ScopedSetRestoreDone restore_done(backup_dao, db->getLastInsertID());
 
 			int skip=1024*512;
 
@@ -982,9 +1080,17 @@ void ServerChannelThread::DOWNLOAD_IMAGE(str_map& params)
 				&& file_extension!="raw")
 				skip=512*512;
 
+			if (is_disk_mbr(res[0]["path"] + ".mbr"))
+				skip = 0;
+
 			_i64 imgsize = (_i64)vhdfile->getSize() - skip;
 			_i64 r=little_endian(imgsize);
-			input->Write((char*)&r, sizeof(_i64));
+			if (!input->Write((char*)&r, sizeof(_i64), img_send_timeout))
+			{
+				Server->Log("Sending image size failed", LL_DEBUG);
+				reset();
+				return;
+			}
 			unsigned int blocksize=vhdfile->getBlocksize();
 			char buffer[4096];
 			size_t read;
@@ -1025,10 +1131,15 @@ void ServerChannelThread::DOWNLOAD_IMAGE(str_map& params)
 			if (params["with_used_bytes"] == "1")
 			{
 				_i64 r = little_endian(used_bytes - skip);
-				input->Write((char*)&r, sizeof(_i64));
+				if (!input->Write((char*)&r, sizeof(_i64)))
+				{
+					Server->Log("Sending used bytes failed", LL_DEBUG);
+					reset();
+					return;
+				}
 			}
 
-			vhdfile->Seek(skip);
+			vhdfile->Seek(skip + currpos);
 
 			bool is_ok=true;
 			do
@@ -1048,16 +1159,15 @@ void ServerChannelThread::DOWNLOAD_IMAGE(str_map& params)
 					}
 						
 					uint64 currpos_endian = little_endian(currpos);
-					input->Write((char*)&currpos_endian, sizeof(uint64));
-					bool b=input->Write(buffer, (_u32)read);
+					bool b = input->Write((char*)&currpos_endian, sizeof(uint64), img_send_timeout, false);
+					if (b)
+					{
+						b = input->Write(buffer, (_u32)read, img_send_timeout, false);
+					}
 					if(!b)
 					{
 						Server->Log("Writing to output pipe failed processMsg-1", LL_ERROR);
-						image_fak->destroyVHDFile(vhdfile);
-						db->destroyAllQueries();
-						Server->destroy(input);
-						input=NULL;
-						backup_dao.setRestoreDone(0, restore_id);
+						reset();
 						return;
 					}
 					used_transferred_bytes += read;
@@ -1068,13 +1178,22 @@ void ServerChannelThread::DOWNLOAD_IMAGE(str_map& params)
 					if(Server->getTimeMS()-lasttime>30000)
 					{
 						uint64 currpos_endian = little_endian(currpos);
-						input->Write((char*)&currpos_endian, sizeof(uint64));
+						bool b = input->Write((char*)&currpos_endian, sizeof(uint64), img_send_timeout, false);
 						memset(buffer, 0, 4096);
-						input->Write(buffer, (_u32)4096);
+						if (b)
+						{
+							 b = input->Write(buffer, (_u32)4096, img_send_timeout, true);
+						}
+						if (!b)
+						{
+							Server->Log("Sending keep-alive block failed", LL_DEBUG);
+							reset();
+							return;
+						}
 						lasttime=Server->getTimeMS();
 					}
 					read=4096;
-					vhdfile->Seek(skip+currpos+4096);
+					vhdfile->Seek(skip+currpos+read);
 				}					
 				currpos+=read;
 
@@ -1098,14 +1217,28 @@ void ServerChannelThread::DOWNLOAD_IMAGE(str_map& params)
 			if((_i64)currpos>=imgsize)
 			{
 				r = little_endian(0x7fffffffffffffffLL);
-				input->Write((char*)&r, sizeof(int64));
+				if (!input->Write((char*)&r, sizeof(int64), img_send_timeout))
+				{
+					Server->Log("Sending image end failed", LL_WARNING);
+					reset();
+					return;
+				}
+			}
+			else
+			{
+				if (!input->Flush())
+				{
+					reset();
+					return;
+				}
 			}
 
-			backup_dao.setRestoreDone(is_ok ? 1 : 0, restore_id);
+			if (is_ok)
+			{
+				restore_done.set_ok();
+			}
 		}
-		image_fak->destroyVHDFile(vhdfile);
 	}
-	db->destroyAllQueries();
 }
 
 void ServerChannelThread::DOWNLOAD_FILES( str_map& params )
@@ -1125,8 +1258,10 @@ void ServerChannelThread::DOWNLOAD_FILES( str_map& params )
 	}
 	else
 	{
-		if( !all_client_rights &&
-			std::find(client_right_ids.begin(), client_right_ids.end(), watoi(res[0]["clientid"]))==client_right_ids.end())
+		int clientid = watoi(res[0]["clientid"]);
+		std::string clientname = get_clientname(db, clientid);
+		if(clientname.empty()
+			|| !has_restore_permission(clientname, clientid) )
 		{
 			JSON::Object ret;
 			ret.set("err", 4);
@@ -1154,8 +1289,10 @@ void ServerChannelThread::DOWNLOAD_FILES( str_map& params )
 			restore_flags = watoi64(restore_flags_it->second);
 		}
 
+		std::vector<std::string> tokens;
+
 		if(create_clientdl_thread(backupid, clientname, clientid, restore_id, status_id, log_id,
-			params["restore_token"], map_paths, clean_other, ignore_other_fs, follow_symlinks, restore_flags))
+			params["restore_token"], map_paths, clean_other, ignore_other_fs, follow_symlinks, restore_flags, tokens))
 		{
 			JSON::Object ret;
 			ret.set("ok", true);
@@ -1251,7 +1388,10 @@ void ServerChannelThread::DOWNLOAD_FILES_TOKENS(str_map& params)
 		logid_t log_id;
 
 		std::string filename;
-		if(path_info.is_file)
+		std::string p_filter = params["filter"];
+		if(path_info.is_file
+			|| (p_filter.empty() && !u_path.empty()
+				&& !path_info.rel_path.empty()) )
 		{
 			filename = ExtractFileName(path_info.full_path, os_file_sep());
 			if(!filename.empty())
@@ -1269,7 +1409,7 @@ void ServerChannelThread::DOWNLOAD_FILES_TOKENS(str_map& params)
 
 		if(filename.empty())
 		{
-			filename = params["filter"];
+			filename = p_filter;
 		}
 
 		std::vector<std::pair<std::string, std::string> > map_paths;
@@ -1291,7 +1431,8 @@ void ServerChannelThread::DOWNLOAD_FILES_TOKENS(str_map& params)
 
 		if(!create_clientdl_thread(clientname, clientid, clientid, path_info.full_path, path_info.full_metadata_path, filename, 
 			path_info.rel_path.empty(), path_info.rel_path, restore_id, status_id, log_id, params["restore_token"],
-			map_paths, clean_other, ignore_other_fs, greplace(os_file_sep(), "/", path_info.rel_path), follow_symlinks, restore_flags, ticket))
+			map_paths, clean_other, ignore_other_fs, greplace(os_file_sep(), "/", path_info.rel_path), follow_symlinks, restore_flags,
+			ticket, tokens, path_info.backup_tokens))
 		{
 			ret.set("err", 5);
 			break;
@@ -1388,4 +1529,78 @@ void ServerChannelThread::RESTORE_DONE( str_map params )
 	
 }
 
+void ServerChannelThread::reset()
+{
+	IScopedLock lock(mutex);
+	Server->destroy(input);
+	input = NULL;
+	tcpstack.reset();
+}
+
+bool ServerChannelThread::has_restore_permission(const std::string& clientname, int clientid)
+{
+	if (!all_client_rights
+		&& std::find(client_right_ids.begin(), client_right_ids.end(), clientid) == client_right_ids.end())
+	{
+		return false;
+	}
+
+	std::vector<std::string> allow_restore_clients = client_main->getAllowRestoreClients();
+	if (!allow_restore_clients.empty()
+		&& std::find(allow_restore_clients.begin(), allow_restore_clients.end(), clientname)
+		== allow_restore_clients.end())
+	{
+		return false;
+	}
+
+	return true;
+}
+
+std::string ServerChannelThread::get_clientname(IDatabase* db, int clientid)
+{
+	IQuery* q_name = db->Prepare("SELECT name FROM clients WHERE id=?");
+	q_name->Bind(clientid);
+	db_results res_name = q_name->Read();
+	q_name->Reset();
+
+	if (!res_name.empty())
+		return res_name[0]["name"];
+
+	return std::string();
+}
+
+void ServerChannelThread::add_extra_channel()
+{
+	if (parent != NULL)
+	{
+		parent->add_extra_channel();
+	}
+
+	IScopedLock lock(mutex);
+
+	ServerChannelThread* extra = new ServerChannelThread(client_main,
+		clientname, clientid, internet_mode, allow_restore,
+		server_token, virtual_client, this);
+
+	extra_channel_threads.push_back(extra);
+
+	Server->getThreadPool()->execute(extra, "channel extra");
+}
+
+void ServerChannelThread::remove_extra_channel()
+{
+	if (parent != NULL)
+	{
+		parent->remove_extra_channel();
+	}
+
+	IScopedLock lock(mutex);
+
+	if (!extra_channel_threads.empty())
+	{
+		extra_channel_threads[extra_channel_threads.size() - 1]->doExit();
+		extra_channel_threads.erase(extra_channel_threads.begin() + extra_channel_threads.size() - 1);
+	}
+}
+
 
