@@ -181,6 +181,16 @@ void ServerCleanupThread::operator()(void)
 			}
 		}
 	}
+	else
+	{
+		ServerSettings settings(db);
+		if (os_directory_exists(settings.getSettings()->backupfolder + os_file_sep() + "inode_db"))
+		{
+			Server->deleteFile(settings.getSettings()->backupfolder + os_file_sep() + "inode_db" + os_file_sep() + "inode_db.lmdb");
+			Server->deleteFile(settings.getSettings()->backupfolder + os_file_sep() + "inode_db" + os_file_sep() + "inode_db.lmdb-lock");
+			os_remove_dir(settings.getSettings()->backupfolder + os_file_sep() + "inode_db");
+		}
+	}
 
 	{
 		ScopedActiveThread sat;
@@ -299,7 +309,8 @@ void ServerCleanupThread::operator()(void)
 			ServerSettings settings(db);
 			std::vector<STimeSpan> tw=settings.getCleanupWindow();
 			if( ( (!tw.empty() && ServerSettings::isInTimeSpan(tw)) || ( tw.empty() && (chour==3 || chour==4) ) )
-				&& Server->getTimeSeconds()-last_cleanup>min_cleanup_interval)
+				&& Server->getTimeSeconds()-last_cleanup>min_cleanup_interval
+				&& os_directory_exists(settings.getSettings()->backupfolder) )
 			{
 				IScopedLock lock(a_mutex);
 
@@ -580,9 +591,8 @@ void ServerCleanupThread::do_remove_unknown(void)
 				continue;
 
 			if(cf.isdir
-				|| ( cf.issym 
-						&& cf.name.find(".")==std::string::npos
-					    && cf.name.find("Image")!=std::string::npos ) )
+				|| ( !cf.isdir 
+						&& cf.name.find(".")==std::string::npos ) )
 			{
 				if (cf.name.find("Image") == std::string::npos)
 				{
@@ -592,9 +602,15 @@ void ServerCleanupThread::do_remove_unknown(void)
 					{
 						Server->Log("File backup \"" + cf.name + "\" of client \"" + clientname + "\" not found in database. Deleting it.", LL_WARNING);
 						bool remove_folder = false;
-						if (BackupServer::isFileSnapshotsEnabled())
+						if (!cf.isdir)
 						{
-							if (!SnapshotHelper::removeFilesystem(clientname, cf.name))
+							SnapshotHelper::removeFilesystem(false, clientname, cf.name);
+							Server->deleteFile(os_file_prefix(backupfolder
+								+ os_file_sep() + clientname + os_file_sep() + cf.name));
+						}
+						else if (BackupServer::isFileSnapshotsEnabled())
+						{
+							if (!SnapshotHelper::removeFilesystem(false, clientname, cf.name))
 							{
 								remove_folder = true;
 							}
@@ -645,7 +661,7 @@ void ServerCleanupThread::do_remove_unknown(void)
 							
 							if (extension == "raw")
 							{
-								SnapshotHelper::removeFilesystem(clientname, cf.name);
+								SnapshotHelper::removeFilesystem(true, clientname, cf.name);
 							}
 							else
 							{
@@ -656,9 +672,9 @@ void ServerCleanupThread::do_remove_unknown(void)
 
 					if (!found_image)
 					{
-						if (cf.issym)
+						if (!cf.isdir)
 						{
-							SnapshotHelper::removeFilesystem(clientname, cf.name);
+							SnapshotHelper::removeFilesystem(true, clientname, cf.name);
 							Server->deleteFile(os_file_prefix(backupfolder + os_file_sep() + clientname + os_file_sep() + cf.name));
 						}
 						else
@@ -668,7 +684,7 @@ void ServerCleanupThread::do_remove_unknown(void)
 							if(os_directory_exists(os_file_prefix(backupfolder + os_file_sep() + clientname + os_file_sep() + cf.name))
 								&& BackupServer::isImageSnapshotsEnabled() )
 							{
-								SnapshotHelper::removeFilesystem(clientname, cf.name);
+								SnapshotHelper::removeFilesystem(true, clientname, cf.name);
 							}
 						}
 					}
@@ -714,7 +730,7 @@ void ServerCleanupThread::do_remove_unknown(void)
 				}
 			}
 		}
-		check_symlinks(res_clients[i], backupfolder);
+		check_symlinks(res_clients[i], backupfolder, false);
 	}
 
 	Server->Log("Removing dangling file entries...", LL_INFO);
@@ -798,7 +814,8 @@ bool ServerCleanupThread::deleteImage(logid_t logid, std::string clientname, std
 {
 	std::string image_extension = findextension(path);
 
-	if (image_extension != "raw")
+	if (image_extension != "raw"
+		|| !BackupServer::isImageSnapshotsEnabled())
 	{
 		bool b = true;
 		if (!deleteAndTruncateFile(logid, path))
@@ -818,6 +835,7 @@ bool ServerCleanupThread::deleteImage(logid_t logid, std::string clientname, std
 		}
 		deleteAndTruncateFile(logid, path + ".cbitmap");
 		deleteAndTruncateFile(logid, path + ".sync");
+		deleteAndTruncateFile(logid, path + ".bitmap");
 
 		if (b && ExtractFileName(ExtractFilePath(path)) != clientname)
 		{
@@ -828,13 +846,25 @@ bool ServerCleanupThread::deleteImage(logid_t logid, std::string clientname, std
 	}
 	else
 	{
-		return SnapshotHelper::removeFilesystem(clientname, ExtractFileName(ExtractFilePath(path)));
+		bool b = SnapshotHelper::removeFilesystem(true, clientname, ExtractFileName(ExtractFilePath(path)));
+
+		if (b
+			&& BackupServer::getSnapshotMethod(true) == BackupServer::ESnapshotMethod_Zfs)
+		{
+			Server->deleteFile(ExtractFilePath(path));
+		}
+
+		return b;
 	}
 }
 
-int ServerCleanupThread::max_removable_incr_images(ServerSettings& settings, int backupid)
+int ServerCleanupThread::max_removable_incr_images(ServerSettings& settings, int backupid,
+	int del_in_stack)
 {
 	int incr_image_num=cleanupdao->getIncrNumImagesForBackup(backupid);
+	incr_image_num -= del_in_stack;
+	if (incr_image_num < 0)
+		incr_image_num = 0;
 	int min_incr_image_num = static_cast<int>(settings.getSettings()->min_image_incr);
 
 	int max_allowed_del_refs = 0;
@@ -847,7 +877,7 @@ int ServerCleanupThread::max_removable_incr_images(ServerSettings& settings, int
 	return max_allowed_del_refs;
 }
 
-bool ServerCleanupThread::cleanup_images_client(int clientid, int64 minspace, std::vector<int> &imageids, bool cleanup_only_one)
+bool ServerCleanupThread::cleanup_one_imagebackup_client(int clientid, int64 minspace, int& imagebid)
 {
 	ServerSettings settings(db, clientid);
 
@@ -900,18 +930,10 @@ bool ServerCleanupThread::cleanup_images_client(int clientid, int64 minspace, st
 			}
 			else
 			{
-				imageids.push_back(backupid);
-
-				if (cleanup_only_one)
-				{
-					return true;
-				}
+				imagebid = backupid;
+				return true;
 			}
 		}
-
-		int r=hasEnoughFreeSpace(minspace, &settings);
-		if( r==-1 || r==1 )
-			return true;
 				
 		full_image_num=(int)getImagesFullNum(clientid, backupid, notit);
 	}
@@ -956,24 +978,16 @@ bool ServerCleanupThread::cleanup_images_client(int clientid, int64 minspace, st
 		}
 		else
 		{
-			if(!removeImage(backupid, &settings, true, false, true, true))
+			if(!removeImage(backupid, &settings, true, false, true, true, 1))
 			{
 				notit.push_back(backupid);
 			}
 			else
 			{
-				imageids.push_back(backupid);
-
-				if (cleanup_only_one)
-				{
-					return true;
-				}
+				imagebid = backupid;
+				return true;
 			}
 		}
-
-		int r=hasEnoughFreeSpace(minspace, &settings);
-		if( r==-1 || r==1 )
-			return true;
 				
 		incr_image_num=(int)getImagesIncrNum(clientid, backupid, notit);
 	}
@@ -986,7 +1000,8 @@ void ServerCleanupThread::cleanup_images(int64 minspace)
 	std::vector<ServerCleanupDao::SIncompleteImages> incomplete_images=cleanupdao->getIncompleteImages();
 	for(size_t i=0;i<incomplete_images.size();++i)
 	{
-		if (!isImageLockedFromCleanup(incomplete_images[i].id))
+		if (!isImageLockedFromCleanup(incomplete_images[i].id)
+			&& cleanupdao->getIncompleteImage(incomplete_images[i].id).exists )
 		{
 			ServerLogger::Log(logid, "Deleting incomplete image file \"" + incomplete_images[i].path + "\"...", LL_INFO);
 			if (!deleteImage(logid, incomplete_images[i].clientname, incomplete_images[i].path))
@@ -1020,34 +1035,43 @@ void ServerCleanupThread::cleanup_images(int64 minspace)
 		}
 	}
 
+	ServerSettings settings(db);
+	cleanup_all_system_images(settings);
+
+	int r=hasEnoughFreeSpace(minspace, &settings);
+	if( r==-1 || r==1)
+		return;
+
+	bool deleted_something = true;
+	while (deleted_something)
 	{
-		ServerSettings settings(db);
+		deleted_something = false;
 
-		cleanup_all_system_images(settings);
-
-		int r=hasEnoughFreeSpace(minspace, &settings);
-		if( r==-1 || r==1)
-			return;
-	}
-
-	std::vector<int> res=cleanupdao->getClientsSortImagebackups();
-	for(size_t i=0;i<res.size();++i)
-	{
-		int clientid=res[i];
-		
-		std::vector<int> imageids;
-		if(cleanup_images_client(clientid, minspace, imageids, false))
+		std::vector<int> res = cleanupdao->getClientsSortImagebackups();
+		for (size_t i = 0; i<res.size(); ++i)
 		{
-			if(minspace!=-1)
+			int clientid = res[i];
+
+			int imagebid;
+			if (cleanup_one_imagebackup_client(clientid, minspace, imagebid))
 			{
-				return;
+				int r = hasEnoughFreeSpace(minspace, &settings);
+				if (r == -1 || r == 1)
+					return;
+
+				deleted_something = true;
 			}
 		}
-	}
+
+		int r = hasEnoughFreeSpace(minspace, &settings);
+		if (r == -1 || r == 1)
+			return;
+	}	
 }
 
 bool ServerCleanupThread::removeImage(int backupid, ServerSettings* settings, 
-	bool update_stat, bool force_remove, bool remove_associated, bool remove_references)
+	bool update_stat, bool force_remove, bool remove_associated, bool remove_references,
+	int del_incr_in_stack)
 {
 	int64 deleted_size_bytes = 0;
 
@@ -1071,16 +1095,19 @@ bool ServerCleanupThread::removeImage(int backupid, ServerSettings* settings,
 
 		for(size_t i=0;i<refs.size();++i)
 		{
-			if(max_removable_incr_images(*settings, refs[i].id)<=0)
+			if(max_removable_incr_images(*settings, refs[i].id, del_incr_in_stack)<=0)
 			{
 				ServerLogger::Log(logid, "Cannot delete image because incremental image backups referencing this image are not allowed to be deleted", LL_INFO);
 				return false;
 			}
 
-			bool b=removeImage(refs[i].id, settings, true, force_remove, remove_associated, remove_references);
+			bool b=removeImage(refs[i].id, settings, true, force_remove, 
+				remove_associated, remove_references,
+				del_incr_in_stack+1);
 			if(!b)
 			{
-				ret=false;
+				ServerLogger::Log(logid, "Cannot delete image because incremental image backup referencing this image cannot be deleted", LL_ERROR);
+				return false;
 			}
 		}
 	}
@@ -1102,7 +1129,8 @@ bool ServerCleanupThread::removeImage(int backupid, ServerSettings* settings,
 		{
 			int64 is=getImageSize(assoc[i]);
 			if(is>0) deleted_size_bytes+=is;
-			removeImage(assoc[i], settings, false, force_remove, remove_associated, remove_references);
+			removeImage(assoc[i], settings, false, force_remove, remove_associated,
+				remove_references, del_incr_in_stack);
 		}
 	}
 
@@ -1267,7 +1295,11 @@ bool ServerCleanupThread::cleanup_one_filebackup_client(int clientid, int64 mins
 	int full_file_num=(int)getFilesFullNum(clientid, backupid);
 	ServerLogger::Log(logid, "Client with id="+convert(clientid)+" has "+convert(full_file_num)+" full file backups max="+convert(max_file_full), LL_DEBUG);
 	while(full_file_num>max_file_full
-		&& full_file_num>0)
+		&& full_file_num>0
+		&& !(full_file_num==1
+			&& settings.getSettings()->max_file_incr>0
+			&& settings.getUpdateFreqFileIncr()>=0
+			&& getFilesIncrNum(clientid, backupid)==0) )
 	{
 		ServerCleanupDao::SFileBackupInfo res_info=cleanupdao->getFileBackupInfo(backupid);
 		ServerCleanupDao::CondString clientname=cleanupdao->getClientName(clientid);
@@ -1451,10 +1483,18 @@ bool ServerCleanupThread::deleteFileBackup(const std::string &backupfolder, int
 	}
 
 	std::string path=backupfolder+os_file_sep()+clientname+os_file_sep()+backuppath;
+
+	if (!os_directory_exists(os_file_prefix(path))
+		&& os_directory_exists(os_file_prefix(path + ".startup-del")))
+	{
+		backuppath += ".startup-del";
+		path += ".startup-del";
+	}
+
 	bool b=false;
 	if( BackupServer::isFileSnapshotsEnabled())
 	{
-		b=SnapshotHelper::removeFilesystem(clientname, backuppath);
+		b=SnapshotHelper::removeFilesystem(false, clientname, backuppath);
 
 		if(!b)
 		{
@@ -1462,7 +1502,7 @@ bool ServerCleanupThread::deleteFileBackup(const std::string &backupfolder, int
 
 			b=remove_directory_link_dir(path, link_dao, clientid);
 
-			if(!b && SnapshotHelper::isSubvolume(clientname, backuppath) )
+			if(!b && SnapshotHelper::isSubvolume(false, clientname, backuppath) )
 			{
 				ServerLogger::Log(logid, "Deleting directory failed. Trying to truncate all files in subvolume to zero...", LL_ERROR);
 				b=truncate_files_recurisve(os_file_prefix(path));
@@ -1473,6 +1513,10 @@ bool ServerCleanupThread::deleteFileBackup(const std::string &backupfolder, int
 				}
 			}
 		}
+		else if (BackupServer::getSnapshotMethod(false) == BackupServer::ESnapshotMethod_ZfsFile)
+		{
+			Server->deleteFile(path);
+		}
 	}
 	else
 	{
@@ -1551,7 +1595,7 @@ void ServerCleanupThread::removeClient(int clientid)
 			ServerLogger::Log(logid, "Removing file backup with id \""+convert(backupid)+"\"", LL_INFO);
 			bool b=deleteFileBackup(settings.getSettings()->backupfolder, clientid, backupid, true);
 			if(b)
-				ServerLogger::Log(logid, "Removing file backup with id \""+convert(backupid)+"\" successfull.", LL_INFO);
+				ServerLogger::Log(logid, "Removing file backup with id \""+convert(backupid)+"\" successful.", LL_INFO);
 			else
 				ServerLogger::Log(logid, "Removing file backup with id \""+convert(backupid)+"\" failed.", LL_ERROR);
 		}
@@ -1566,47 +1610,7 @@ void ServerCleanupThread::removeClient(int clientid)
 
 	ServerLogger::Log(logid, "Deleting database table entries of client..", LL_INFO);
 
-	//Remove logentries
-	IQuery *q=db->Prepare("DELETE FROM logs WHERE clientid=?", false);
-	q->Bind(clientid); q->Write(); q->Reset();
-	db->destroyQuery(q);
-
-	//history data
-	/*q=db->Prepare("SELECT hist_id FROM clients_hist WHERE id=?", false);
-	q->Bind(clientid);
-	res=q->Read(); q->Reset();
-	db->destroyQuery(q);
-	q=db->Prepare("DELETE FROM clients_hist_id WHERE id=?", false);
-	for(size_t i=0;i<res.size();++i)
-	{
-		q->Bind(res[i][L"hist_id"]);
-		q->Write();
-		q->Reset();
-	}
-	db->destroyQuery(q);*/
-	q=db->Prepare("DELETE FROM clients_hist WHERE id=?", false);
-	q->Bind(clientid); q->Write(); q->Reset();
-	db->destroyQuery(q);
-
-	//settings
-	q=db->Prepare("DELETE FROM settings_db.settings WHERE clientid=?", false);
-	q->Bind(clientid); q->Write(); q->Reset();
-	db->destroyQuery(q);
-
-	//stats
-	q=db->Prepare("DELETE FROM del_stats WHERE clientid=?", false);
-	q->Bind(clientid); q->Write(); q->Reset();
-	db->destroyQuery(q);
-
-	//client
-	q=db->Prepare("DELETE FROM clients WHERE id=?", false);
-	q->Bind(clientid); q->Write(); q->Reset();
-	db->destroyQuery(q);
-
-	q=db->Prepare("DELETE FROM settings_db.extra_clients WHERE id=?", false);
-	q->Bind(clientid); q->Write(); q->Reset();
-	db->destroyQuery(q);
-
+	deleteClientSQL(db, clientid);
 	//delete dirs
 	os_remove_nonempty_dir(settings.getSettings()->backupfolder+os_file_sep()+clientname);
 	Server->deleteFile(settings.getSettings()->backupfolder+os_file_sep()+"clients"+os_file_sep()+clientname);
@@ -1815,7 +1819,8 @@ bool ServerCleanupThread::backup_database(void)
 
 					db_results res = copy_db->Read("PRAGMA quick_check");
 
-					integrity_ok = !res.empty() && res[0]["quick_check"] == "ok";
+					integrity_ok = !res.empty() && (res[0]["quick_check"] == "ok"
+													|| res[0]["integrity_check"] == "ok");
 
 					if (!integrity_ok)
 					{
@@ -2079,21 +2084,33 @@ bool ServerCleanupThread::enforce_quota(int clientid, std::ostringstream& log)
 
 			if(state==0)
 			{
-				std::vector<int> imageids;
-				cleanup_images_client(clientid, target_minspace, imageids, true);
-				if(!imageids.empty())
+				int imagebid;
+				bool enough_space = false;
+				while (true)
 				{
-					log << "Removed " << imageids.size() << " image with id ";
-					for(size_t i=0;i<imageids.size();++i) log << imageids[i] << " ";
-					log << std::endl;
+					if (cleanup_one_imagebackup_client(clientid, target_minspace, imagebid))
+					{
+						log << "Removed image backupd with id " << imagebid << std::endl;
+						did_remove_something = true;
+						//TODO: wait here for btrfs subvol remove to finish
+						if (hasEnoughFreeSpace(target_minspace, &client_settings))
+						{
+							enough_space = true;
+							break;
+						}
+					}
+					else
+					{
+						++nopc;
+						break;
+					}
+				}
 
-					did_remove_something=true;
+				if (enough_space)
+				{
 					break;
 				}
-				else
-				{
-					++nopc;
-				}
+				
 			}
 			else
 			{
@@ -2104,6 +2121,7 @@ bool ServerCleanupThread::enforce_quota(int clientid, std::ostringstream& log)
 
 					did_remove_something=true;
 					nopc=0;
+					//TODO: wait here for btrfs subvol remove to finish
 					if(hasEnoughFreeSpace(target_minspace, &client_settings))
 					{
 						break;
@@ -2172,7 +2190,8 @@ bool ServerCleanupThread::correct_poolname( const std::string& backupfolder, con
 	return false;
 }
 
-void ServerCleanupThread::check_symlinks( const ServerCleanupDao::SClientInfo& client_info, const std::string& backupfolder )
+void ServerCleanupThread::check_symlinks( const ServerCleanupDao::SClientInfo& client_info, const std::string& backupfolder,
+	bool deep_check)
 {
 	const int clientid=client_info.id;
 	const std::string& clientname=client_info.name;
@@ -2188,6 +2207,7 @@ void ServerCleanupThread::check_symlinks( const ServerCleanupDao::SClientInfo& c
 
 	IDatabaseCursor* cursor = q_directory_links->Cursor();
 
+	std::vector<ServerLinkDao::DirectoryLinkEntry> deep_check_links;
 	db_single_result res;
 	while(cursor->next(res))
 	{
@@ -2237,6 +2257,14 @@ void ServerCleanupThread::check_symlinks( const ServerCleanupDao::SClientInfo& c
 						Server->Log("Error deleting symlink \""+new_target+"\"", LL_ERROR);
 					}
 				}
+
+				if (deep_check)
+				{
+					ServerLinkDao::DirectoryLinkEntry e;
+					e.target = symlink_pool_path;
+					e.name = pool_name;
+					deep_check_links.push_back(e);
+				}
 			}
 		}
 	}
@@ -2252,6 +2280,12 @@ void ServerCleanupThread::check_symlinks( const ServerCleanupDao::SClientInfo& c
 		link_dao.updateLinkReferenceTarget(target_adjustments[i].second, target_adjustments[i].first);
 	}
 
+	for (size_t i = 0; i < deep_check_links.size(); ++i)
+	{
+		reference_contained_directory_links(link_dao, clientid, deep_check_links[i].name,
+			std::string(), deep_check_links[i].target);
+	}
+
 	if(os_directory_exists(pool_root))
 	{
 		std::vector<SFile> first_files = getFiles(pool_root, NULL);
@@ -2265,23 +2299,46 @@ void ServerCleanupThread::check_symlinks( const ServerCleanupDao::SClientInfo& c
 			std::string curr_path = pool_root + os_file_sep() + first_files[i].name;
 			std::vector<SFile> pool_files = getFiles(curr_path, NULL);
 
-			for(size_t j=0;i<pool_files.size();++i)
+			if (pool_files.empty())
 			{
-				if(!pool_files[i].isdir)
+				os_remove_dir(curr_path);
+				continue;
+			}
+			
+			bool has_dir = false;
+			for(size_t j=0;j<pool_files.size();++j)
+			{
+				if(!pool_files[j].isdir)
 					continue;
-				if(pool_files[i].issym)
+				if(pool_files[j].issym)
 					continue;
 
-				std::string pool_path = curr_path + os_file_sep() + pool_files[i].name;
+				std::string pool_path = curr_path + os_file_sep() + pool_files[j].name;
 
-				if(link_dao.getDirectoryRefcount(clientid, pool_files[i].name)==0)
+				if(link_dao.getDirectoryRefcount(clientid, pool_files[j].name)==0)
 				{
-					Server->Log("Refcount of \""+pool_path+"\" is zero. Deleting pool folder.");
+					Server->Log("Refcount of \""+pool_path+"\" is zero. Deleting pool folder.", LL_WARNING);
+
+					if (!deep_check)
+					{
+						Server->Log("Restarting symlink check. Starting deep directory link check...", LL_WARNING);
+						check_symlinks(client_info, backupfolder, true);
+						return;
+					}
 					if(!remove_directory_link_dir(pool_path, link_dao, clientid))
 					{
 						Server->Log("Could not remove pool folder \""+pool_path+"\"", LL_ERROR);
 					}
 				}
+				else
+				{
+					has_dir = true;
+				}
+			}
+
+			if (!has_dir)
+			{
+				os_remove_dir(curr_path);
 			}
 		}
 	}	
@@ -2469,6 +2526,14 @@ void ServerCleanupThread::cleanup_other()
 	ServerLogger::Log(logid, "Cleaning up client lists...", LL_INFO);
 	cleanup_clientlists();
 	ServerLogger::Log(logid, "Done cleaning up client lists.", LL_INFO);
+	ServerLogger::Log(logid, "Cleaning up old client access tokens...", LL_INFO);
+	cleanupdao->getDatabase()->Write("DELETE FROM user_tokens WHERE id IN ("
+		"SELECT a.id FROM user_tokens a WHERE EXISTS ("
+		"SELECT id FROM user_tokens b WHERE a.username=b.username "
+		"AND a.clientid=b.clientid AND a.tgroup=b.tgroup AND b.created>a.created "
+		"AND (strftime('%s', b.created)-strftime('%s', a.created)) > 15552000));");
+	ServerLogger::Log(logid, "Done cleaning up old client access tokens.", LL_INFO);
+
 }
 
 void ServerCleanupThread::removeFileBackupSql( int backupid )
@@ -2716,6 +2781,34 @@ bool ServerCleanupThread::cleanup_clientlists()
 	return !has_error;
 }
 
+void ServerCleanupThread::deleteClientSQL(IDatabase * db, int clientid)
+{
+	//Remove logentries
+	IQuery *q = db->Prepare("DELETE FROM logs WHERE clientid=?", false);
+	q->Bind(clientid); q->Write(); q->Reset();
+	db->destroyQuery(q);
+
+	//history data
+	q = db->Prepare("DELETE FROM clients_hist WHERE id=?", false);
+	q->Bind(clientid); q->Write(); q->Reset();
+	db->destroyQuery(q);
+
+	//settings
+	q = db->Prepare("DELETE FROM settings_db.settings WHERE clientid=?", false);
+	q->Bind(clientid); q->Write(); q->Reset();
+	db->destroyQuery(q);
+
+	//stats
+	q = db->Prepare("DELETE FROM del_stats WHERE clientid=?", false);
+	q->Bind(clientid); q->Write(); q->Reset();
+	db->destroyQuery(q);
+
+	//client
+	q = db->Prepare("DELETE FROM clients WHERE id=?", false);
+	q->Bind(clientid); q->Write(); q->Reset();
+	db->destroyQuery(q);
+}
+
 
 
 
