@@ -20,6 +20,7 @@
 #include "dao/ServerFilesDao.h"
 #include "FileIndex.h"
 #include "create_files_index.h"
+#include "FileBackup.h"
 
 extern std::string server_identity;
 extern std::string server_token;
@@ -77,7 +78,7 @@ public:
 
 		logid = ServerLogger::getLogId(clientid);
 
-		local_hash.reset(new BackupServerHash(NULL, clientid, use_snapshots, use_reflink, use_tmpfiles, logid, use_snapshots));
+		local_hash.reset(new BackupServerHash(NULL, clientid, use_snapshots, use_reflink, use_tmpfiles, logid, use_snapshots, max_file_id));
 	}
 
 	~BackupServerContinuous()
@@ -556,7 +557,7 @@ private:
 	std::string getOsFp(const std::string& fn)
 	{
 		std::vector<std::string> tokens;
-		TokenizeMail(fn, tokens, "/");
+		Tokenize(fn, tokens, "/");
 
 		std::string fp;
 		for(size_t i=0;i<tokens.size();++i)
@@ -718,7 +719,7 @@ private:
 			continuous_hash_path, continuous_path, std::string(), hashed_transfer_full,
 			false, clientid, clientname, std::string(), use_tmpfiles, tmpfile_path, server_token,
 			use_reflink, backupid, true, hashpipe_prepare, client_main, client_main->getProtocolVersions().file_protocol_version,
-			0, logid, true, shares_without_snapshot, true, NULL, false));
+			0, logid, true, shares_without_snapshot, true, NULL, false, filepath_corrections, max_file_id));
 
 		server_download_ticket = Server->getThreadPool()->execute(server_download.get(), "backup download");
 	}
@@ -745,7 +746,8 @@ private:
 
 		if(!fileclient_chunked.get())
 		{
-			if(!client_main->getClientChunkedFilesrvConnection(fileclient_chunked, server_settings.get()))
+			if(!client_main->getClientChunkedFilesrvConnection(fileclient_chunked, server_settings.get(),
+				NULL))
 			{
 				ServerLogger::Log(logid, "Connect error during continuous backup (fileclient_chunked-1)", LL_ERROR);
 				return false;
@@ -767,7 +769,7 @@ private:
 		int64 created;
 		int64 modified;
 		int64 accessed=0; //TODO
-		_u32 rc = fileclient_metadata->GetFileHashAndMetadata(change.fn1, hash, permissions, filesize, created, modified);
+		_u32 rc = ERR_ERROR; // fileclient_metadata->GetFileHashAndMetadata(change.fn1, hash, permissions, filesize, created, modified);
 
 		if(rc!=ERR_SUCCESS)
 		{
@@ -952,5 +954,9 @@ private:
 	std::auto_ptr<ServerFilesDao> filesdao;
 	std::auto_ptr<FileIndex> fileindex;
 
+	FilePathCorrections filepath_corrections;
+
+	MaxFileId max_file_id;
+
 	logid_t logid;
 };
\ No newline at end of file
