@@ -15,7 +15,7 @@ class IScopedLock;
 void dir_link_lock_client_mutex(int clientid, IScopedLock& lock);
 
 bool link_directory_pool(int clientid, const std::string& target_dir, const std::string& src_dir, const std::string& pooldir, bool with_transaction,
-	ServerLinkDao*& link_dao, ServerLinkJournalDao*& link_journal_dao);
+	ServerLinkDao*& link_dao, ServerLinkJournalDao*& link_journal_dao, int depth);
 
 bool replay_directory_link_journal();
 
@@ -24,4 +24,7 @@ bool remove_directory_link(const std::string &path, ServerLinkDao& link_dao, int
 
 bool remove_directory_link_dir(const std::string &path, ServerLinkDao& link_dao, int clientid, bool delete_root=true, bool with_transaction=true);
 
+bool reference_contained_directory_links(ServerLinkDao& link_dao, int clientid,
+	const std::string& pool_name, const std::string &path, const std::string& link_path);
+
 bool is_directory_link(const std::string& path);
\ No newline at end of file
