@@ -18,6 +18,7 @@
 #include "server_hash.h"
 #include "server_prepare_hash.h"
 #include "server_settings.h"
+#include "server.h"
 #include "../urbackupcommon/fileclient/FileClientChunked.h"
 #include "../Interface/Server.h"
 #include "../Interface/File.h"
@@ -32,6 +33,7 @@
 #include <algorithm>
 #include <memory.h>
 #include "../urbackupcommon/file_metadata.h"
+#include "FileBackup.h"
 #include <assert.h>
 #ifdef _WIN32
 #include <Windows.h>
@@ -53,9 +55,9 @@ void destroy_mutex1(void)
 }
 
 BackupServerHash::BackupServerHash(IPipe *pPipe, int pClientid, bool use_snapshots, bool use_reflink, bool use_tmpfiles, logid_t logid,
-	bool snapshot_file_inplace)
+	bool snapshot_file_inplace, MaxFileId& max_file_id)
 	: use_snapshots(use_snapshots), use_reflink(use_reflink), use_tmpfiles(use_tmpfiles), filesdao(NULL), old_backupfolders_loaded(false),
-	  logid(logid), snapshot_file_inplace(snapshot_file_inplace)
+	  logid(logid), snapshot_file_inplace(snapshot_file_inplace), max_file_id(max_file_id)
 {
 	pipe=pPipe;
 	clientid=pClientid;
@@ -140,6 +142,9 @@ void BackupServerHash::operator()(void)
 
 			if(action==EAction_LinkOrCopy)
 			{
+				int64 fileid;
+				rd.getVarInt(&fileid);
+
 				std::string temp_fn;
 				rd.getStr(&temp_fn);
 
@@ -202,18 +207,24 @@ void BackupServerHash::operator()(void)
 					}
 
 					addFile(backupid, incremental, tf, tfn, hashpath, sha2,
-						old_file_fn, hashoutput_fn, t_filesize, metadata, with_hashes!=0, extent_iterator.get());
+						old_file_fn, hashoutput_fn, t_filesize, metadata, with_hashes!=0, extent_iterator.get(), fileid);
 				}
 
 				if(!hashoutput_fn.empty())
 				{
 					Server->deleteFile(hashoutput_fn);
 				}
+
+				max_file_id.setMaxDownloaded(fileid);
 			}
 			else if(action==EAction_Copy)
 			{
+				int64 fileid;
+				bool b = rd.getVarInt(&fileid);
+				assert(b);
+
 				std::string source;
-				bool b = rd.getStr(&source);
+				b = rd.getStr(&source);
 				assert(b);
 				
 				std::string dest;
@@ -618,7 +629,8 @@ bool BackupServerHash::findFileAndLink(const std::string &tfn, IFile *tf, std::s
 		{
 			if (tf!=NULL
 				&& tf->Size() <= 8
-				&& os_get_file_type(os_file_prefix(tfn)) != 0)
+				&& os_get_file_type(os_file_prefix(tfn)) != 0
+				&& os_file_truncate(os_file_prefix(tfn), t_filesize) )
 			{
 				//zero patch
 				b = true;
@@ -630,7 +642,7 @@ bool BackupServerHash::findFileAndLink(const std::string &tfn, IFile *tf, std::s
 		}
 		if (!b)
 		{
-			b = os_create_hardlink(os_file_prefix(tfn), os_file_prefix(existing_file.fullpath), use_snapshots, &too_many_hardlinks);
+			b = FileBackup::create_hardlink(os_file_prefix(tfn), os_file_prefix(existing_file.fullpath), use_snapshots || (use_reflink && !BackupServer::canHardlink()), &too_many_hardlinks, &copy);
 		}
 		if(!b)
 		{
@@ -775,7 +787,7 @@ bool BackupServerHash::findFileAndLink(const std::string &tfn, IFile *tf, std::s
 					{
 						if(!os_file_truncate(os_file_prefix(hash_fn), get_hashdata_size(t_filesize)))
 						{
-							ServerLogger::Log(logid, "Error truncating hashdata file -1", LL_ERROR);
+							ServerLogger::Log(logid, "Error truncating hashdata file -1. "+os_last_error_str(), LL_ERROR);
 						}
 					}
 				}
@@ -837,7 +849,7 @@ bool BackupServerHash::findFileAndLink(const std::string &tfn, IFile *tf, std::s
 
 void BackupServerHash::addFile(int backupid, int incremental, IFile *tf, const std::string &tfn,
 	std::string hash_fn, const std::string &sha2, const std::string &orig_fn, const std::string &hashoutput_fn, int64 t_filesize,
-	FileMetadata& metadata, bool with_hashes, ExtentIterator* extent_iterator)
+	FileMetadata& metadata, bool with_hashes, ExtentIterator* extent_iterator, int64 fileid)
 {
 	bool copy=true;
 	bool tries_once = false;
@@ -854,7 +866,7 @@ void BackupServerHash::addFile(int backupid, int incremental, IFile *tf, const s
 		false, tries_once, ff_last, hardlink_limit, copied_file, entryid, entryclientid, rsize, next_entryid,
 		metadata, false, extent_iterator))
 	{
-		ServerLogger::Log(logid, "HT: Linked file: \""+tfn+"\"", LL_DEBUG);
+		ServerLogger::Log(logid, "HT: Linked file: \""+tfn+"\" (id="+convert(fileid)+")", LL_DEBUG);
 		copy=false;
 		std::string temp_fn=tf->getFilename();
 		Server->destroy(tf);
@@ -882,7 +894,7 @@ void BackupServerHash::addFile(int backupid, int incremental, IFile *tf, const s
 	
 	if(copy)
 	{
-		ServerLogger::Log(logid, "HT: Copying file: \""+tfn+"\"", LL_DEBUG);
+		ServerLogger::Log(logid, "HT: Copying file: \""+tfn+"\" (id=" + convert(fileid) + ")", LL_DEBUG);
 		int64 fs=tf->Size();
 		if(!use_reflink)
 		{
@@ -918,15 +930,32 @@ void BackupServerHash::addFile(int backupid, int incremental, IFile *tf, const s
 
 			if( fs>0 && available_space<=fs )
 			{
-				if(space_logcnt==0)
+				if (extent_iterator != NULL)
 				{
-					ServerLogger::Log(logid, "HT: No free space available deleting backups...", LL_WARNING);
+					extent_iterator->reset();
+
+					IFsFile::SSparseExtent ext;
+					while ((ext = extent_iterator->nextExtent()).offset!=-1)
+					{
+						if(ext.size>0)
+							fs -= ext.size;
+					}
+
+					extent_iterator->reset();
 				}
-				else
+
+				if (fs > 0 && available_space <= fs)
 				{
-					ServerLogger::Log(logid, "HT: No free space available deleting backups...", LL_WARNING);
+					if (space_logcnt == 0)
+					{
+						ServerLogger::Log(logid, "HT: No free space available deleting backups...", LL_WARNING);
+					}
+					else
+					{
+						ServerLogger::Log(logid, "HT: No free space available deleting backups...", LL_WARNING);
+					}
+					free_ok = freeSpace(fs, os_file_prefix(tfn));
 				}
-				free_ok=freeSpace(fs, os_file_prefix(tfn));
 			}
 
 			if(!free_ok)
@@ -1125,7 +1154,7 @@ ServerFilesDao::SFindFileEntry BackupServerHash::findFileHash(const std::string
 			entryid=state.prev.next_entry;
 		}
 	}
-	else if(state.state=4)
+	else if(state.state==4)
 	{
 		if(state.prev.prev_entry==0)
 		{
@@ -1228,6 +1257,8 @@ IFsFile* BackupServerHash::openFileRetry(const std::string &dest, int mode, std:
 
 bool BackupServerHash::copyFile(IFile *tf, const std::string &dest, ExtentIterator* extent_iterator)
 {
+	ServerLogger::Log(logid, "HT: Copying file to \""+dest+"\"", LL_DEBUG);
+
 	std::string errstr;
 	std::auto_ptr<IFsFile> dst(openFileRetry(dest, MODE_WRITE, errstr));
 	if (dst.get() == NULL)
@@ -1326,6 +1357,8 @@ bool BackupServerHash::copyFile(IFile *tf, const std::string &dest, ExtentIterat
 
 bool BackupServerHash::copyFileWithHashoutput(IFile *tf, const std::string &dest, const std::string hash_dest, ExtentIterator* extent_iterator)
 {
+	ServerLogger::Log(logid, "HT: Copying file with hash output to \""+dest+"\"", LL_DEBUG);
+
 	std::string errstr;
 	IFsFile *dst=openFileRetry(dest, MODE_WRITE, errstr);
 	if(dst==NULL) return false;
@@ -1447,19 +1480,24 @@ bool BackupServerHash::patchFile(IFile *patch, const std::string &source, const
 	_i64 dstfsize;
 	{
 		has_reflink=false;
+		bool copy = false;
 		if( use_reflink )
 		{
 			if( (!snapshot_file_inplace || (os_get_file_type(os_file_prefix(dest)) & EFileType_File)==0)
-				&& !os_create_hardlink(os_file_prefix(dest), os_file_prefix(source), true, NULL) )
+				&& !FileBackup::create_hardlink(os_file_prefix(dest), os_file_prefix(source), true, NULL, &copy) )
 			{
 				ServerLogger::Log(logid, "Reflinking file \""+dest+"\" failed", LL_WARNING);
 			}
 			else
 			{
-				has_reflink=true;
+				has_reflink=!copy;
 			}
 		}
 
+		ServerLogger::Log(logid, "HT: Patching file \""+dest+"\" with source \""+source+"\" patch size "+PrettyPrintBytes(patch->Size())+" reflink="+convert(has_reflink), LL_DEBUG);
+
+
+
 		std::string errstr;
 		chunk_output_fn=openFileRetry(dest, has_reflink?MODE_RW:MODE_WRITE, errstr);
 		if (chunk_output_fn == NULL)
@@ -1543,7 +1581,7 @@ bool BackupServerHash::patchFile(IFile *patch, const std::string &source, const
 	{
 		if (!os_file_truncate(dest, chunk_patcher.getFilesize()))
 		{
-			ServerLogger::Log(logid, "Error truncating \""+dest+"\" to "+convert(chunk_patcher.getFilesize()), LL_ERROR);
+			ServerLogger::Log(logid, "Error truncating \""+dest+"\" to "+convert(chunk_patcher.getFilesize())+". "+os_last_error_str(), LL_ERROR);
 			return false;
 		}
 	}
@@ -1577,8 +1615,9 @@ const size_t RP_COPY_BLOCKSIZE=1024;
 
 bool BackupServerHash::replaceFile(IFile *tf, const std::string &dest, const std::string &orig_fn, ExtentIterator* extent_iterator)
 {
+	bool copy = false;
 	if( (!snapshot_file_inplace || (os_get_file_type(os_file_prefix(dest)) & EFileType_File) == 0)
-		&& ! os_create_hardlink(os_file_prefix(dest), os_file_prefix(orig_fn), true, NULL) )
+		&& ! FileBackup::create_hardlink(os_file_prefix(dest), os_file_prefix(orig_fn), true, NULL, &copy) )
 	{
 		ServerLogger::Log(logid, "Reflinking file \""+dest+"\" failed -2", LL_ERROR);
 
@@ -1665,6 +1704,10 @@ bool BackupServerHash::replaceFile(IFile *tf, const std::string &dest, const std
 				return false;
 			}
 		}
+		else if (copy)
+		{
+			cow_filesize += read1;
+		}
 
 		dst_pos+=read1;
 	}
@@ -1678,7 +1721,7 @@ bool BackupServerHash::replaceFile(IFile *tf, const std::string &dest, const std
 	{
 		if( !os_file_truncate(os_file_prefix(dest), tf->Size()) )
 		{
-			ServerLogger::Log(logid, "Error truncating file \""+dest+"\" -2", LL_ERROR);
+			ServerLogger::Log(logid, "Error truncating file \""+dest+"\" -2. "+os_last_error_str(), LL_ERROR);
 			return false;
 		}
 	}
@@ -1689,8 +1732,9 @@ bool BackupServerHash::replaceFile(IFile *tf, const std::string &dest, const std
 bool BackupServerHash::replaceFileWithHashoutput(IFile *tf, const std::string &dest,
 	const std::string hash_dest, const std::string &orig_fn, ExtentIterator* extent_iterator)
 {
+	bool copy = false;
 	if( (!snapshot_file_inplace || (os_get_file_type(os_file_prefix(dest)) & EFileType_File) == 0)
-		&& !os_create_hardlink(os_file_prefix(dest), os_file_prefix(orig_fn), true, NULL) )
+		&& !FileBackup::create_hardlink(os_file_prefix(dest), os_file_prefix(orig_fn), true, NULL, &copy) )
 	{
 		ServerLogger::Log(logid, "Reflinking file \""+dest+"\" failed -3", LL_ERROR);
 
@@ -1726,13 +1770,18 @@ bool BackupServerHash::replaceFileWithHashoutput(IFile *tf, const std::string &d
 
 		_i64 dst_size=dst->Size();
 
+		if (copy)
+		{
+			cow_filesize = dst_size;
+		}
+
 		dst_s.clear();
 
 		if( dst_size!=tf->Size() )
 		{
 			if( !os_file_truncate(os_file_prefix(dest), tf->Size()) )
 			{
-				ServerLogger::Log(logid, "Error truncating file \""+dest+"\" -2", LL_ERROR);
+				ServerLogger::Log(logid, "Error truncating file \""+dest+"\" -2. "+os_last_error_str(), LL_ERROR);
 				return false;
 			}
 		}
@@ -1743,6 +1792,8 @@ bool BackupServerHash::replaceFileWithHashoutput(IFile *tf, const std::string &d
 
 bool BackupServerHash::renameFileWithHashoutput(IFile *tf, const std::string &dest, const std::string hash_dest, ExtentIterator* extent_iterator)
 {
+	ServerLogger::Log(logid, "HT: Renaming file to \""+dest+"\" with hash output", LL_DEBUG);
+
 	if(tf->Size()>0)
 	{
 		std::string errstr;
@@ -1761,11 +1812,16 @@ bool BackupServerHash::renameFileWithHashoutput(IFile *tf, const std::string &de
 	}
 
 
-	return renameFile(tf, dest);
+	return renameFile(tf, dest, false);
 }
 
-bool BackupServerHash::renameFile(IFile *tf, const std::string &dest)
+bool BackupServerHash::renameFile(IFile *tf, const std::string &dest, bool log_info)
 {
+	if(log_info)
+	{
+		ServerLogger::Log(logid, "HT: Renaming file to \""+dest+"\"", LL_DEBUG);
+	}
+
 	std::string tf_fn=tf->getFilename();
 	Server->destroy(tf);
 
@@ -1773,7 +1829,7 @@ bool BackupServerHash::renameFile(IFile *tf, const std::string &dest)
 	{
 		std::string err = os_last_error_str();
 		if (use_reflink &&
-			os_create_hardlink(os_file_prefix(dest), os_file_prefix(tf_fn), true, NULL))
+			FileBackup::create_hardlink(os_file_prefix(dest), os_file_prefix(tf_fn), true, NULL, NULL))
 		{
 			Server->deleteFile(os_file_prefix(tf_fn));
 		}
@@ -1802,28 +1858,89 @@ bool BackupServerHash::correctPath( std::string& ff, std::string& f_hashpath )
 		backupfolder = settings.getSettings()->backupfolder;
 	}
 
+	if (correctClientName(backupfolder, ff, f_hashpath))
+	{
+		return true;
+	}
+
 	for(size_t i=0;i<old_backupfolders.size();++i)
 	{
 		size_t erase_size = old_backupfolders[i].size() + os_file_sep().size();
 		if(ff.size()>erase_size &&
-			next(ff, 0, old_backupfolders[i]))
+			next(ff, 0, old_backupfolders[i] + os_file_sep()))
 		{
 			std::string tmp_ff = backupfolder + os_file_sep() + ff.substr(erase_size);
+
+			std::string tmp_hashff;
+			if (f_hashpath.size() > erase_size)
+			{
+				tmp_hashff = backupfolder + os_file_sep() + f_hashpath.substr(erase_size);
+			}
 						
-			IFile* f = Server->openFile(tmp_ff, MODE_READ);
+			IFile* f = Server->openFile(os_file_prefix(tmp_ff), MODE_READ);
 			if(f!=NULL)
 			{
 				Server->destroy(f);
 
-				if(f_hashpath.size()>erase_size)
-				{
-					f_hashpath = backupfolder + os_file_sep() + f_hashpath.substr(erase_size);
-				}
-
+				f_hashpath = tmp_hashff;
 				ff = tmp_ff;
 
 				return true;
 			}
+			else if (correctClientName(backupfolder, tmp_ff, tmp_hashff))
+			{
+				return true;
+			}
+		}
+	}
+
+	return false;
+}
+
+bool BackupServerHash::correctClientName(const std::string& backupfolder, std::string & ff, std::string & f_hashpath)
+{
+	if (ff.size() <= backupfolder.size()
+		|| f_hashpath.size() <= backupfolder.size())
+		return false;
+
+	
+	size_t clientname_off = ff.find(os_file_sep(), backupfolder.size() + os_file_sep().size());
+	
+	if (clientname_off == std::string::npos)
+		return false;
+
+	std::string clientname = ff.substr(backupfolder.size() + 1, clientname_off - backupfolder.size() - 1);
+
+	ServerBackupDao backupdao(Server->getDatabase(Server->getThreadID(), URBACKUPDB_SERVER));
+
+	std::map<std::string, std::vector<std::string> >::iterator it = client_moved_to.find(clientname);
+
+	if (it == client_moved_to.end())
+	{
+		client_moved_to[clientname] = backupdao.getClientMovedFrom(clientname);
+		it = client_moved_to.find(clientname);
+	}
+
+	if (it->second.empty())
+		return false;
+
+	for (size_t i = 0; i < it->second.size(); ++i)
+	{
+		std::string tmp_ff = backupfolder + os_file_sep() + it->second[i] + ff.substr(clientname_off);
+
+		IFile* f = Server->openFile(os_file_prefix(tmp_ff), MODE_READ);
+		if (f != NULL)
+		{
+			Server->destroy(f);
+
+			if (f_hashpath.size()>clientname_off)
+			{
+				f_hashpath = backupfolder + os_file_sep() + it->second[i] + f_hashpath.substr(clientname_off);
+			}
+
+			ff = tmp_ff;
+
+			return true;
 		}
 	}
 
