@@ -31,6 +31,14 @@
 IMutex *ServerSettings::g_mutex=NULL;
 std::map<int, SSettings*> ServerSettings::g_settings_cache;
 
+#ifndef DEFAULT_BACKUP_FOLDER
+#ifdef _WIN32
+#define DEFAULT_BACKUP_FOLDER "C:\\urbackup"
+#else
+#define DEFAULT_BACKUP_FOLDER "/mnt/backups/urbackup"
+#endif
+#endif
+
 //#define CLEAR_SETTINGS_CACHE
 
 void ServerSettings::init_mutex(void)
@@ -56,9 +64,10 @@ void ServerSettings::clear_cache()
 	{
 		if(it->second->refcount==0)
 		{
-			std::map<int, SSettings*>::iterator delit=it++;
-			delete delit->second;
-			g_settings_cache.erase(delit);
+			std::map<int, SSettings*>::iterator it_curr = it;
+			++it;
+			delete it_curr->second;
+			g_settings_cache.erase(it_curr);
 		}
 		else
 		{
@@ -71,6 +80,9 @@ void ServerSettings::clear_cache()
 ServerSettings::ServerSettings(IDatabase *db, int pClientid)
 	: local_settings(NULL), clientid(pClientid), db(db)
 {
+#ifndef NDEBUG
+	init_thread_id = Server->getThreadID();
+#endif
 	IScopedLock lock(g_mutex);
 		
 	std::map<int, SSettings*>::iterator iter=g_settings_cache.find(clientid);
@@ -176,9 +188,10 @@ void ServerSettings::updateAll(void)
 	{
 		if(it->second->refcount==0)
 		{
-			std::map<int, SSettings*>::iterator delit=it++;
-			delete delit->second;
-			g_settings_cache.erase(delit);
+			std::map<int, SSettings*>::iterator it_curr = it;
+			++it;
+			delete it_curr->second;
+			g_settings_cache.erase(it_curr);
 		}
 		else
 		{
@@ -290,6 +303,7 @@ void ServerSettings::update(bool force_update)
 
 void ServerSettings::updateInternal(bool* was_updated)
 {
+	assert(init_thread_id == Server->getThreadID());
 	if(local_settings->needs_update)
 	{
 		if(was_updated!=NULL)
@@ -339,7 +353,7 @@ void ServerSettings::readSettingsDefault(ISettingsReader* settings_default,
 	settings->no_file_backups=(settings_global->getValue("no_file_backups", "false")=="true");
 	settings->overwrite=false;
 	settings->allow_overwrite=(settings_default->getValue("allow_overwrite", "true")=="true");
-	settings->backupfolder=trim(settings_global->getValue("backupfolder", "C:\\urbackup"));
+	settings->backupfolder=trim(settings_global->getValue("backupfolder", DEFAULT_BACKUP_FOLDER));
 	settings->backupfolder_uncompr=trim(settings_global->getValue("backupfolder_uncompr", settings->backupfolder));
 	settings->autoshutdown=(settings_global->getValue("autoshutdown", "false")=="true");;
 	settings->startup_backup_delay=settings_default->getValue("startup_backup_delay", 0);
@@ -354,6 +368,7 @@ void ServerSettings::readSettingsDefault(ISettingsReader* settings_default,
 	settings->exclude_files=settings_default->getValue("exclude_files", "");
 	settings->include_files=settings_default->getValue("include_files", "");
 	settings->default_dirs=settings_default->getValue("default_dirs", "");
+	settings->backup_dirs_optional = settings_default->getValue("backup_dirs_optional", "false") == "true";
 	settings->cleanup_window= settings_global->getValue("cleanup_window", "1-7/3-4");
 	settings->allow_config_paths=(settings_default->getValue("allow_config_paths", "true")=="true");
 	settings->allow_starting_full_file_backups=(settings_default->getValue("allow_starting_full_file_backups", "true")=="true");
@@ -366,6 +381,7 @@ void ServerSettings::readSettingsDefault(ISettingsReader* settings_default,
 	settings->image_letters=settings_default->getValue("image_letters", "C");
 	settings->backup_database=(settings_global->getValue("backup_database", "true")=="true");
 	settings->internet_server_port=(unsigned short)(atoi(settings_global->getValue("internet_server_port", "55415").c_str()));
+	settings->internet_server_proxy = settings_global->getValue("internet_server_proxy", "");
 	settings->client_set_settings=false;
 	settings->internet_server= settings_global->getValue("internet_server", "");
 	settings->internet_image_backups=(settings_default->getValue("internet_image_backups", "false")=="true");
@@ -393,6 +409,7 @@ void ServerSettings::readSettingsDefault(ISettingsReader* settings_default,
 	settings->client_quota=settings_default->getValue("client_quota", "");
 	settings->end_to_end_file_backup_verification=(settings_default->getValue("end_to_end_file_backup_verification", "false")=="true");
 	settings->internet_calculate_filehashes_on_client=(settings_default->getValue("internet_calculate_filehashes_on_client", "true")=="true");
+	settings->internet_parallel_file_hashing = (settings_default->getValue("internet_parallel_file_hashing", "false") == "true");	
 	settings->use_incremental_symlinks=(settings_global->getValue("use_incremental_symlinks", "true")=="true");
 	settings->internet_connect_always=(settings_default->getValue("internet_connect_always", "false")=="true");
 	settings->show_server_updates=(settings_global->getValue("show_server_updates", "true")=="true");
@@ -419,6 +436,8 @@ void ServerSettings::readSettingsDefault(ISettingsReader* settings_default,
 	settings->file_snapshot_groups = settings_default->getValue("file_snapshot_groups", "");
 	settings->internet_file_dataplan_limit = settings_default->getValue("internet_file_dataplan_limit", 5LL*1000*1024*1024);
 	settings->internet_image_dataplan_limit = settings_default->getValue("internet_image_dataplan_limit", 20LL * 1000* 1024 * 1024);
+	settings->alert_script = settings_default->getValue("alert_script", 1);
+	settings->alert_params = settings_default->getValue("alert_params", "");
 }
 
 void ServerSettings::readSettingsClient(ISettingsReader* settings_client)
@@ -507,6 +526,7 @@ void ServerSettings::readSettingsClient(ISettingsReader* settings_client)
 	if(!stmp.empty())
 		settings->local_speed=stmp;
 
+	readBoolClientSetting(settings_client, "backup_dirs_optional", &settings->backup_dirs_optional);
 	readBoolClientSetting(settings_client, "client_set_settings", &settings->client_set_settings);
 	readBoolClientSetting(settings_client, "internet_mode_enabled", &settings->internet_mode_enabled);
 	readBoolClientSetting(settings_client, "internet_full_file_backups", &settings->internet_full_file_backups);
@@ -517,24 +537,35 @@ void ServerSettings::readSettingsClient(ISettingsReader* settings_client)
 
 	readStringClientSetting(settings_client, "vss_select_components", &settings->vss_select_components);
 
+	if (!settings->overwrite
+		&& settings->image_snapshot_groups.empty())
+	{
+		readStringClientSetting(settings_client, "image_snapshot_groups_def", &settings->image_snapshot_groups);
+	}
+
+	if (settings_client->getValue("virtual_clients_add", &stmp))
+	{
+		settings->virtual_clients_add = stmp;
+	}
+
+	stmp = settings_client->getValue("backup_window_incr_file", "");
+	if (!stmp.empty())
+		settings->backup_window_incr_file = stmp;
+	stmp = settings_client->getValue("backup_window_full_file", "");
+	if (!stmp.empty())
+		settings->backup_window_full_file = stmp;
+	stmp = settings_client->getValue("backup_window_incr_image", "");
+	if (!stmp.empty())
+		settings->backup_window_incr_image = stmp;
+	stmp = settings_client->getValue("backup_window_full_image", "");
+	if (!stmp.empty())
+		settings->backup_window_full_image = stmp;
+
 	if(!settings->overwrite)
 		return;
 
 	//Following settings are not configurable by the client
 
-	stmp=settings_client->getValue("backup_window_incr_file", "");
-	if(!stmp.empty())
-		settings->backup_window_incr_file=stmp;
-	stmp=settings_client->getValue("backup_window_full_file", "");
-	if(!stmp.empty())
-		settings->backup_window_full_file=stmp;
-	stmp=settings_client->getValue("backup_window_incr_image", "");
-	if(!stmp.empty())
-		settings->backup_window_incr_image=stmp;
-	stmp=settings_client->getValue("backup_window_full_image", "");
-	if(!stmp.empty())
-		settings->backup_window_full_image=stmp;
-
 	stmp=settings_client->getValue("client_quota", "");
 	if(!stmp.empty())
 		settings->client_quota=stmp;
@@ -551,7 +582,8 @@ void ServerSettings::readSettingsClient(ISettingsReader* settings_client)
 	readStringClientSetting(settings_client, "internet_image_transfer_mode", &settings->internet_image_transfer_mode);
 
 	readBoolClientSetting(settings_client, "end_to_end_file_backup_verification", &settings->end_to_end_file_backup_verification);
-	readBoolClientSetting(settings_client, "internet_calculate_filehashes_on_client", &settings->internet_calculate_filehashes_on_client);	
+	readBoolClientSetting(settings_client, "internet_calculate_filehashes_on_client", &settings->internet_calculate_filehashes_on_client);
+	readBoolClientSetting(settings_client, "internet_parallel_file_hashing", &settings->internet_parallel_file_hashing);	
 	readBoolClientSetting(settings_client, "silent_update", &settings->silent_update);
 
 	readBoolClientSetting(settings_client, "allow_config_paths", &settings->allow_config_paths);
@@ -587,6 +619,9 @@ void ServerSettings::readSettingsClient(ISettingsReader* settings_client)
 
 	readInt64ClientSetting(settings_client, "internet_file_dataplan_limit", &settings->internet_file_dataplan_limit);
 	readInt64ClientSetting(settings_client, "internet_image_dataplan_limit", &settings->internet_image_dataplan_limit);
+
+	readIntClientSetting(settings_client, "alert_script", &settings->alert_script);
+	readStringClientSetting(settings_client, "alert_params", &settings->alert_params);
 }
 
 void ServerSettings::readBoolClientSetting(ISettingsReader* settings_client, const std::string &name, bool *output)
@@ -809,7 +844,8 @@ std::string ServerSettings::getImageFileFormatInt( const std::string& image_file
 {
 	if(image_file_format == image_file_format_default)
 	{
-		if(BackupServer::isImageSnapshotsEnabled())
+		if(BackupServer::isImageSnapshotsEnabled()
+			|| BackupServer::canReflink())
 		{
 			return image_file_format_cowraw;
 		}
@@ -955,6 +991,18 @@ int ServerSettings::getGlobalInternetSpeed()
 	return static_cast<int>(round(currentTimeSpanValue(getSettings()->global_internet_speed)));
 }
 
+std::string ServerSettings::getVirtualClients()
+{
+	std::string ret = getSettings()->virtual_clients;
+	std::string add = getSettings()->virtual_clients_add;
+	if (!add.empty())
+	{
+		if (!ret.empty())ret += "|";
+		ret += add;
+	}
+	return ret;
+}
+
 double ServerSettings::currentTimeSpanValue(std::string time_span_value)
 {
 	std::vector<std::pair<double, STimeSpan > > time_span_values = parseTimeSpanValue(time_span_value);
