@@ -442,6 +442,16 @@ namespace backupaccess
 
 		Helper helper(Server->getThreadID(), NULL, NULL);
 
+		int last_filebackup = 0;
+		IQuery* q_last = db->Prepare("SELECT id FROM backups WHERE clientid=? AND done=1 ORDER BY backuptime DESC LIMIT 1");
+		q_last->Bind(t_clientid);
+		db_results res_last = q_last->Read();
+		q_last->Reset();		
+		if (!res_last.empty())
+		{
+			last_filebackup = watoi(res_last[0]["id"]);
+		}
+
 		IQuery *q=db->Prepare("SELECT id, strftime('"+helper.getTimeFormatString()+"', backuptime) AS t_backuptime, incremental, size_bytes, archived, archive_timeout, path, delete_pending FROM backups WHERE complete=1 AND done=1 AND clientid=? ORDER BY backuptime DESC");
 		q->Bind(t_clientid);
 		db_results res=q->Read();
@@ -464,7 +474,8 @@ namespace backupaccess
 			has_access = true;
 
 			JSON::Object obj;
-			obj.set("id", watoi(res[i]["id"])+backupid_offset);
+			int backupid = watoi(res[i]["id"]);
+			obj.set("id", backupid+backupid_offset);
 			obj.set("backuptime", watoi64(res[i]["t_backuptime"]));
 			obj.set("incremental", watoi(res[i]["incremental"]));
             obj.set("size_bytes", watoi64(res[i]["size_bytes"]));
@@ -487,6 +498,10 @@ namespace backupaccess
             {
                 obj.set("archive_timeout", archive_timeout);
             }
+			if (backupid == last_filebackup)
+			{
+				obj.set("disable_delete", true);
+			}
 			backups.add(obj);
 		}
 
@@ -905,7 +920,8 @@ namespace backupaccess
 		return backups;
 	}
 
-	JSON::Object get_image_info(IDatabase* db, int backupid, int t_clientid, int backupid_offset, std::string& path)
+	JSON::Object get_image_info(IDatabase* db, int backupid, int t_clientid, int backupid_offset, std::string& path,
+		std::vector<IFSImageFactory::SPartition>& partitions)
 	{
 		Helper helper(Server->getThreadID(), NULL, NULL);
 
@@ -943,65 +959,137 @@ namespace backupaccess
 			std::string filename = ExtractFileName(path);
 			std::string extension = findextension(filename);
 
-			if (extension == "vhd" || extension == "vhdz")
+			bool disk_image = false;
+			std::auto_ptr<IFile> mbrfile(Server->openFile(os_file_prefix(path + ".mbr"), MODE_READ));
+			if (mbrfile.get() && mbrfile->Size()<10*1024*1024)
 			{
-				std::auto_ptr<IVHDFile> vhdfile(image_fak->createVHDFile(path, true, 0));
-				if (vhdfile.get() != NULL)
+				std::string mbrdata_header = mbrfile->Read(0LL, 2);
+				CRData mbrdata_header_view(mbrdata_header.data(), mbrdata_header.size());
+				char version = 0;
+				if (mbrdata_header_view.getChar(&version)
+					&& mbrdata_header_view.getChar(&version)
+					&& version != 100)
 				{
-					ret.set("volume_size", vhdfile->getSize());
+					std::string mbrdata = mbrfile->Read(0LL, static_cast<_u32>(mbrfile->Size()));
+					if (mbrdata.size() == mbrfile->Size())
+					{
+						CRData mbrdata_view(mbrdata.data(), mbrdata.size());
+						SMBRData mbr(mbrdata_view);
+						if (!mbr.hasError())
+						{
+							disk_image = false;
+							ret.set("part_table", mbr.gpt_style ? "GPT" : "MBR");
+							ret.set("disk_number", mbr.device_number);
+							ret.set("partition_number", mbr.partition_number);
+							ret.set("volume_name", mbr.volume_name);
+							ret.set("fs_type", mbr.fsn);
+							ret.set("serial_number", mbr.serial_number);
+						}
+					}
+				}
+				else if (version == 100)
+				{
+					disk_image = true;
 				}
 			}
-			else if (extension == "raw")
+
+			std::auto_ptr<IVHDFile> vhdfile;
+			if (extension == "vhd"
+				|| extension == "vhdz")
 			{
-				std::auto_ptr<IFile> rawfile(Server->openFile(os_file_prefix(path), MODE_READ));
-				if (rawfile.get() != NULL)
-				{
-					ret.set("volume_size", rawfile->Size());
-				}
+				vhdfile.reset(image_fak->createVHDFile(path, true, 0));
+			}
+			else if(extension=="raw")
+			{
+				vhdfile.reset(image_fak->createVHDFile(path, true, 0, 2*1024*1024, false, IFSImageFactory::ImageFormat_RawCowFile));
 			}
 			else
 			{
 				assert(false);
 			}
 
-			std::auto_ptr<IFile> mbrfile(Server->openFile(os_file_prefix(path + ".mbr"), MODE_READ));
-			if (mbrfile.get() && mbrfile->Size()<10*1024*1024)
+			if (vhdfile.get() != NULL
+				&& vhdfile->isOpen())
 			{
-				std::string mbrdata = mbrfile->Read(static_cast<_u32>(mbrfile->Size()));
-				if (mbrdata.size() == mbrfile->Size())
+				ret.set("volume_size", vhdfile->getSize());
+
+				if (disk_image)
 				{
-					CRData mbrdata_view(mbrdata.data(), mbrdata.size());
-					SMBRData mbr(mbrdata_view);
-					if (!mbr.hasError())
-					{
-						ret.set("part_table", mbr.gpt_style ? "GPT" : "MBR");
-						ret.set("disk_number", mbr.device_number);
-						ret.set("partition_number", mbr.partition_number);
-						ret.set("volume_name", mbr.volume_name);
-						ret.set("fs_type", mbr.fsn);
-						ret.set("serial_number", mbr.serial_number);
-					}
+					bool gpt_style;
+					partitions = image_fak->readPartitions(vhdfile.get(), 0, gpt_style);
+					ret.set("part_table", gpt_style ? "GPT" : "MBR");
 				}
 			}
+			
 		}
 		return ret;
 	}
 
 	bool get_image_files(IDatabase* db, int backupid, int t_clientid, std::string clientname,
-			const std::string& u_path, int backupid_offset, bool do_mount, JSON::Object& ret)
+			std::string u_path, int backupid_offset, bool do_mount, JSON::Object& ret)
 	{
 		std::string path;
-		JSON::Object image_backup_info = get_image_info(db, backupid, t_clientid, backupid_offset, path);
+		std::vector<IFSImageFactory::SPartition> partitions;
+		JSON::Object image_backup_info = get_image_info(db, backupid, t_clientid, 
+			backupid_offset, path, partitions);
 		ret.set("image_backup_info", image_backup_info);
 		ret.set("single_item", false);
+
+		int partition = -1;
+		if (partitions.size()>1)
+		{
+			std::string m_u_path = u_path;
+			while (!m_u_path.empty()
+				&& m_u_path[0] == '/')
+			{
+				m_u_path.erase(0, 1);
+			}
+
+			if (next(m_u_path, 0, "partition "))
+			{
+				std::string part = getbetween("partition ", "/", m_u_path);
+				if (part.empty())
+				{
+					part = getafter("partition ", m_u_path);
+				}
+				partition = watoi(part)-1;
+				u_path = m_u_path.substr(10+part.size());
+			}
+			else
+			{
+				JSON::Array files;
+				for (size_t i = 0; i < partitions.size(); ++i)
+				{
+					JSON::Object obj;
+					obj.set("name", "partition "+convert(i+1));
+					obj.set("dir", true);
+					obj.set("size", partitions[i].length);
+					obj.set("mod", 0);
+					obj.set("creat", 0);
+					obj.set("access", 0);
+					files.add(obj);
+				}
+				ret.set("files", files);
+				ret.set("no_zip", true);
+
+				ret.set("backuptime", image_backup_info.get("backuptime"));
+				ret.set("backupid", backupid + backupid_offset);
+				return true;
+			}
+		}
+		else
+		{
+			partition = 0;
+		}
 		
 		if (!path.empty())
 		{
 			ScopedMountedImage mounted_image;
 			bool has_mount_timeout;
-			std::string content_path = ImageMount::get_mount_path(backupid, 
+			std::string mount_errmsg;
+			std::string content_path = ImageMount::get_mount_path(backupid, t_clientid, partition,
 				!u_path.empty() && u_path!="/", mounted_image,
-				10000, has_mount_timeout);
+				10000, has_mount_timeout, mount_errmsg);
 
 			if (content_path.empty()
 				|| !os_directory_exists(content_path) )
@@ -1030,8 +1118,8 @@ namespace backupaccess
 				}
 				else
 				{
-					content_path = ImageMount::get_mount_path(backupid, true,
-						mounted_image, 10000, has_mount_timeout);
+					content_path = ImageMount::get_mount_path(backupid, t_clientid, partition, true,
+						mounted_image, 10000, has_mount_timeout, mount_errmsg);
 					if (content_path.empty())
 					{
 						if (has_mount_timeout)
@@ -1041,6 +1129,10 @@ namespace backupaccess
 						else
 						{
 							ret.set("mount_failed", true);
+							if (!mount_errmsg.empty())
+							{
+								ret.set("mount_errmsg", mount_errmsg);
+							}
 						}
 					}
 				}
@@ -1285,6 +1377,7 @@ ACTION_IMPL(backups)
 	JSON::Object ret;
 	SUser *session=helper.getSession();
 
+	std::string rights = helper.getRights("browse_backups");
 	bool has_tokens = CURRP.find("tokens0")!=CURRP.end();
 
 	bool token_authentication=false;
@@ -1315,7 +1408,8 @@ ACTION_IMPL(backups)
 			}
 		}
 	}
-	else if(session!=NULL && session->id==SESSION_ID_TOKEN_AUTH)
+	else if(session!=NULL && session->id==SESSION_ID_TOKEN_AUTH
+		&& rights!=RIGHT_ALL)
 	{
 		fileaccesstokens = session->mStr["fileaccesstokens"];
 		if(!fileaccesstokens.empty())
@@ -1338,7 +1432,6 @@ ACTION_IMPL(backups)
 	}
 
 	std::string sa=CURRP["sa"];
-	std::string rights=helper.getRights("browse_backups");
 	std::string archive_rights=helper.getRights("manual_archive");
 	std::string delete_rights = helper.getRights("delete_backups");
 	std::vector<int> clientid;
@@ -1638,11 +1731,13 @@ ACTION_IMPL(backups)
 						else
 						{
 							bool has_mount_timeout;
-							backuppath = ImageMount::get_mount_path(-1*backupid, true, mounted_image, -1, has_mount_timeout);
+							std::string mount_errmsg;
+							backuppath = ImageMount::get_mount_path(-1*backupid, t_clientid, 0, true, mounted_image, -1, has_mount_timeout, mount_errmsg);
 							path_info = backupaccess::get_image_path_info(u_path, clientname, backupfolder, backupid, backuppath);
 						}
 
-						if(token_authentication && !path_info.can_access_path)
+						if( (token_authentication && !path_info.can_access_path)
+							|| backuppath.empty())
 						{
 							JSON::Object err_ret;
 							err_ret.set("err", "access_denied");
@@ -1686,11 +1781,16 @@ ACTION_IMPL(backups)
 							size_t status_id;
 							logid_t log_id;
 							THREADPOOL_TICKET ticket;
+							int64 restore_flags = 0;
+							if (!token_authentication)
+							{
+								restore_flags |= restore_flag_ignore_permissions;
+							}
 
 							if(!create_clientdl_thread(clientname, t_clientid, t_clientid, path_info.full_path, path_info.full_metadata_path, CURRP["filter"],
 								path_info.rel_path.empty(), path_info.rel_path, restore_id, status_id, log_id, std::string(),
-								std::vector< std::pair<std::string, std::string> >(), true, true, greplace(os_file_sep(), "/", path_info.rel_path), 0,
-								true, ticket))
+								std::vector< std::pair<std::string, std::string> >(), true, true, greplace(os_file_sep(), "/", path_info.rel_path), true,
+								restore_flags, ticket, tokens, path_info.backup_tokens))
 							{
 								ret.set("err", "internal_error");
                                 helper.Write(ret.stringify(false));
