@@ -88,6 +88,41 @@ std::vector<std::string> getMailSettingsList(void)
 	return tmp;
 }
 
+JSON::Array getAlertScripts(IDatabase* db)
+{
+	db_results res_scripts = db->Read("SELECT id, name FROM alert_scripts");
+
+	IQuery* q_get_params = db->Prepare("SELECT name, label, default_value, has_translation, type FROM alert_script_params WHERE script_id=? ORDER BY idx ASC");
+	JSON::Array ret;
+	for (size_t i = 0; i < res_scripts.size(); ++i)
+	{
+		q_get_params->Bind(res_scripts[i]["id"]);
+		db_results res_params = q_get_params->Read();
+		q_get_params->Reset();
+
+		JSON::Object script;
+		script.set("id", res_scripts[i]["id"]);
+		script.set("name", res_scripts[i]["name"]);
+		
+		JSON::Array params;
+		for (size_t j = 0; j < res_params.size(); ++j)
+		{
+			JSON::Object obj;
+			obj.set("name", res_params[j]["name"]);
+			obj.set("label", res_params[j]["label"]);
+			obj.set("default_value", res_params[j]["default_value"]);
+			obj.set("has_translation", watoi(res_params[j]["has_translation"]));
+			obj.set("type", res_params[j]["type"]);
+			params.add(obj);
+		}
+		script.set("params", params);
+
+		ret.add(script);
+	}
+
+	return ret;
+}
+
 JSON::Object getJSONClientSettings(ServerSettings &settings)
 {
 	JSON::Object ret;
@@ -115,6 +150,7 @@ JSON::Object getJSONClientSettings(ServerSettings &settings)
 	SET_SETTING(exclude_files);
 	SET_SETTING(include_files);
 	SET_SETTING(default_dirs);
+	SET_SETTING(backup_dirs_optional);
 	SET_SETTING(allow_config_paths);
 	SET_SETTING(allow_starting_full_file_backups);
 	SET_SETTING(allow_starting_incr_file_backups);
@@ -143,6 +179,7 @@ JSON::Object getJSONClientSettings(ServerSettings &settings)
 	SET_SETTING(internet_image_transfer_mode);
 	SET_SETTING(end_to_end_file_backup_verification);
 	SET_SETTING(internet_calculate_filehashes_on_client);
+	SET_SETTING(internet_parallel_file_hashing);
 	ret.set("image_file_format", settings.getImageFileFormat());
 	SET_SETTING(internet_connect_always);
 	SET_SETTING(verify_using_client_hashes);
@@ -165,6 +202,8 @@ JSON::Object getJSONClientSettings(ServerSettings &settings)
 	SET_SETTING(image_snapshot_groups);
 	SET_SETTING(internet_file_dataplan_limit);
 	SET_SETTING(internet_image_dataplan_limit);
+	SET_SETTING(alert_script);
+	SET_SETTING(alert_params);
 #undef SET_SETTING
 	return ret;
 }
@@ -195,6 +234,7 @@ void getGeneralSettings(JSON::Object& obj, IDatabase *db, ServerSettings &settin
 	SET_SETTING(backup_database);
 	SET_SETTING(internet_server);
 	SET_SETTING(internet_server_port);
+	SET_SETTING(internet_server_proxy);
 	SET_SETTING(global_local_speed);
 	SET_SETTING(global_soft_fs_quota);
 	SET_SETTING(global_internet_speed);
@@ -206,6 +246,8 @@ void getGeneralSettings(JSON::Object& obj, IDatabase *db, ServerSettings &settin
 	SET_SETTING(show_server_updates);
 	SET_SETTING(server_url);
 	SET_SETTING_DB_BOOL(update_dataplan_db, true);
+	SET_SETTING_DB(restore_authkey, std::string());
+	SET_SETTING_DB(internet_expect_endpoint, std::string());
 
 #undef SET_SETTING
 #undef SET_SETTING_DB
@@ -316,14 +358,16 @@ namespace
 		}
 	}
 
-	std::string fixupBackupfolder(const std::string val, ServerBackupDao& backupdao, ServerSettings &server_settings)
+	std::string fixupBackupfolder(const std::string val, ServerBackupDao& backupdao, ServerSettings &server_settings, bool& changed_backupfolder)
 	{
 		if(val!=server_settings.getSettings()->backupfolder)
 		{
 			backupdao.addToOldBackupfolders(server_settings.getSettings()->backupfolder);
+			changed_backupfolder = true;
 		}
 
-		if(val.find(os_file_sep())==val.size()-os_file_sep().size())
+		if(val.find_last_of(os_file_sep())==val.size()-os_file_sep().size()
+			&& val.size()>1)
 		{
 			return val.substr(0, val.size()-os_file_sep().size());
 		}
@@ -334,7 +378,7 @@ namespace
 	}
 }
 
-void saveGeneralSettings(str_map &POST, IDatabase *db, ServerBackupDao& backupdao, ServerSettings &server_settings)
+void saveGeneralSettings(str_map &POST, IDatabase *db, ServerBackupDao& backupdao, ServerSettings &server_settings, bool& changed_backupfolder)
 {
 	IQuery *q_get=db->Prepare("SELECT value FROM settings_db.settings WHERE clientid=0 AND key=?");
 	IQuery *q_update=db->Prepare("UPDATE settings_db.settings SET value=? WHERE key=? AND clientid=0");
@@ -349,7 +393,7 @@ void saveGeneralSettings(str_map &POST, IDatabase *db, ServerBackupDao& backupda
 			std::string val = UnescapeSQLString(it->second);
 			if(settings[i]=="backupfolder")
 			{
-				val = fixupBackupfolder(val, backupdao, server_settings);
+				val = fixupBackupfolder(val, backupdao, server_settings, changed_backupfolder);
 
 #ifndef _WIN32
 				os_create_dir("/etc/urbackup");
@@ -717,7 +761,7 @@ ACTION_IMPL(settings)
 				&& helper.getRights("groupmod") == RIGHT_ALL)
 			{
 				std::vector<std::string> changes;
-				TokenizeMail(group_mem_changes, changes, ";");
+				Tokenize(group_mem_changes, changes, ";");
 				for (size_t i = 0; i < changes.size(); ++i)
 				{
 					int mod_clientid = watoi(getuntil("-", changes[i]));
@@ -846,7 +890,7 @@ ACTION_IMPL(settings)
 				{
 					JSON::Object u;
 
-					int lclientid = watoi(it->second["id"]);
+					int lclientid = watoi(it->second["clientid"]);
 					ServerSettings client_settings(db, lclientid);
 					u.set("id", lclientid);
 					u.set("name", it->second["name"]);
@@ -1006,6 +1050,11 @@ ACTION_IMPL(settings)
 				JSON::Object obj=getJSONClientSettings(settings);
 				s=getClientSettings(db, t_clientid);				
 				obj.set("clientid", t_clientid);
+				obj.set("alert_scripts", getAlertScripts(db));
+				if (helper.getRights(RIGHT_ALERT_SCRIPTS) == RIGHT_ALL)
+				{
+					obj.set("can_edit_scripts", true);
+				}
 				if (t_clientid <= 0)
 				{
 					obj.set("groupid", t_clientid*-1);
@@ -1021,7 +1070,8 @@ ACTION_IMPL(settings)
 					obj.set("overwrite", s.overwrite);
 				}
 
-				ret.set("cowraw_available", BackupServer::isImageSnapshotsEnabled());
+				ret.set("cowraw_available", BackupServer::isImageSnapshotsEnabled()
+					|| BackupServer::canReflink() );
 				ret.set("settings",  obj);
 
 				getArchiveSettings(ret, db, t_clientid);
@@ -1188,13 +1238,19 @@ ACTION_IMPL(settings)
 				propagateGlobalClientSettings(backupdao, db, POST);
 				updateClientSettings(0, POST, db);
 				updateArchiveSettings(0, POST, db);
-				saveGeneralSettings(POST, db, backupdao, serv_settings);
+				bool changed_backupfolder = false;
+				saveGeneralSettings(POST, db, backupdao, serv_settings, changed_backupfolder);
 				db->EndTransaction();
 
 				ServerSettings::updateAll();
 
 				updateAllOnlineClientSettings(db);
 
+				if (changed_backupfolder)
+				{
+					BackupServer::testFilesystemThread();
+				}
+
 				ret.set("saved_ok", true);
 				sa="general";
 			}
@@ -1205,6 +1261,11 @@ ACTION_IMPL(settings)
 
 				ServerSettings serv_settings(db);
 				JSON::Object obj=getJSONClientSettings(serv_settings);
+				obj.set("alert_scripts", getAlertScripts(db));
+				if (helper.getRights(RIGHT_ALERT_SCRIPTS) == RIGHT_ALL)
+				{
+					obj.set("can_edit_scripts", true);
+				}
 				getGeneralSettings(obj, db, serv_settings);
 				#ifdef _WIN32
 				obj.set("ONLY_WIN32_BEGIN","");
@@ -1214,7 +1275,8 @@ ACTION_IMPL(settings)
 				obj.set("ONLY_WIN32_END","-->");
 				#endif //_WIN32
 
-				ret.set("cowraw_available", BackupServer::isImageSnapshotsEnabled());
+				ret.set("cowraw_available", BackupServer::isImageSnapshotsEnabled()
+					|| BackupServer::canReflink() );
 				ret.set("settings", obj);
 
 				getArchiveSettings(ret, db, t_clientid);
