@@ -25,6 +25,7 @@
 #include "../../cryptoplugin/ICryptoFactory.h"
 #include "../server.h"
 #include "../ClientMain.h"
+#include "../dao/ServerBackupDao.h"
 
 #include <algorithm>
 #include <memory>
@@ -110,84 +111,13 @@ void set_server_version_info(IDatabase* db, JSON::Object& ret)
 	}
 }
 
-void access_dir_details(std::string folder, std::string& ret)
-{
-	bool has_error = false;
-	getFiles(folder, &has_error);
-	if (has_error)
-	{
-		ret += "Cannot access " + folder + ". " + os_last_error_str() + "\n";
-	}
-	else
-	{
-		ret += "Can access " + folder + "\n";
-	}
-}
-
-std::string access_err_details(std::string folder)
-{
-	std::vector<std::string> toks;
-	TokenizeMail(folder, toks, os_file_sep());
-
-	std::string ret;
-
-	std::string cdir = os_file_sep();
-	access_dir_details(cdir, ret);
-
-	for (size_t i = 0; i < toks.size(); ++i)
-	{
-		if (toks[i].empty()) continue;
-
-		if (cdir!=os_file_sep())
-		{
-			cdir += os_file_sep();
-		}
-		cdir += toks[i];
-
-		access_dir_details(cdir, ret);
-	}
-
-	return ret;
-}
-
-std::string access_dir_hint(std::string folder)
-{
-	if (folder.size() > 1 && folder[1] == ':')
-	{
-		bool has_error = false;
-		getFiles(folder.substr(0, 2) + os_file_sep(), &has_error);
-
-		if (has_error)
-		{
-			return "volume_not_accessible";
-		}
-	}
-
-	if ( (folder.size() > 2 && folder[0] == '\\'
-		&& folder[1] == '\\'
-		&& folder[2] != '?' )
-		|| next(folder, 0, "\\\\?\\UNC") )
-	{
-		bool has_error = false;
-		getFiles(folder, &has_error);
-
-		if (has_error && os_last_error() == 5
-			|| os_last_error()== 1326 )
-		{
-			return "folder_unc_access_denied";
-		}
-	}
-
-	return std::string();
-}
-
 void add_remove_stop_show(IDatabase* db, std::string stop_show, bool add)
 {
 	db_results res = db->Read("SELECT tvalue FROM misc WHERE tkey='stop_show'");
 	if (!res.empty())
 	{
 		std::vector<std::string> toks;
-		TokenizeMail(res[0]["tvalue"], toks, ",");
+		Tokenize(res[0]["tvalue"], toks, ",");
 		std::vector<std::string>::iterator it = std::find(toks.begin(), toks.end(), stop_show);
 		if (add)
 		{
@@ -225,18 +155,6 @@ void add_remove_stop_show(IDatabase* db, std::string stop_show, bool add)
 	}
 }
 
-bool is_stop_show(IDatabase* db, std::string stop_key)
-{
-	db_results res = db->Read("SELECT tvalue FROM misc WHERE tkey='stop_show'");
-	if (!res.empty())
-	{
-		std::vector<std::string> toks;
-		TokenizeMail(res[0]["tvalue"], toks, ",");
-		return std::find(toks.begin(), toks.end(), stop_key) != toks.end();
-	}
-	return false;
-}
-
 void set_stop_show_version(IDatabase* db, std::string ver)
 {
 	db_results res = db->Read("SELECT tvalue FROM misc WHERE tkey='stop_show_version'");
@@ -256,238 +174,36 @@ void set_stop_show_version(IDatabase* db, std::string ver)
 	}
 }
 
-void add_stop_show(IDatabase* db, JSON::Object& ret, std::string dir_error_stop_show_key)
+bool is_stop_show(IDatabase* db, std::string stop_key)
 {
-	ret.set("dir_error_stop_show_key", dir_error_stop_show_key);
-	if (is_stop_show(db, dir_error_stop_show_key))
+	db_results res = db->Read("SELECT tvalue FROM misc WHERE tkey='stop_show'");
+	if (!res.empty())
 	{
-		ret.set("dir_error_show", false);
+		std::vector<std::string> toks;
+		Tokenize(res[0]["tvalue"], toks, ",");
+		return std::find(toks.begin(), toks.end(), stop_key) != toks.end();
 	}
+	return false;
 }
 
-void access_dir_checks(IDatabase* db, ServerSettings& settings, std::string backupfolder, std::string backupfolder_uncompr,
-	JSON::Object& ret)
+std::string ipAddrToStr(const std::string& ip_addr_binary)
 {
-#ifdef _WIN32
-	if (backupfolder.size() == 2 && backupfolder[1] == ':')
+	FileClient::SAddrHint addr_hint;
+	if (ip_addr_binary.size() == sizeof(addr_hint.addr_ipv6))
 	{
-		backupfolder += os_file_sep();
+		addr_hint.is_ipv6 = true;
+		memcpy(addr_hint.addr_ipv6, ip_addr_binary.data(), ip_addr_binary.size());
 	}
-	if (backupfolder_uncompr.size() == 2 && backupfolder_uncompr[1] == ':')
+	else if (ip_addr_binary.size() == sizeof(addr_hint.addr_ipv4))
 	{
-		backupfolder_uncompr += os_file_sep();
-	}
-#endif
-
-	if (backupfolder.empty() || !os_directory_exists(os_file_prefix(backupfolder)) || !os_directory_exists(os_file_prefix(backupfolder_uncompr)))
-	{
-		if (!backupfolder.empty())
-		{
-			ret.set("system_err", os_last_error_str());
-		}
-
-		ret.set("dir_error", true);
-
-		if (settings.getSettings()->backupfolder.empty())
-		{
-			ret.set("dir_error_ext", "err_name_empty");
-		}
-		else if (!os_directory_exists(os_file_prefix(settings.getSettings()->backupfolder)))
-		{
-			ret.set("dir_error_ext", "err_folder_not_found");
-			add_stop_show(db, ret, "dir_error_not_found");
-		}
-		else
-		{
-			add_stop_show(db, ret, "dir_error_misc");
-		}
-
-#ifdef _WIN32
-		std::string hint = access_dir_hint(settings.getSettings()->backupfolder);
-		if (!hint.empty())
-		{
-			ret.set("dir_error_hint", hint);
-		}
-#endif
-
-#ifndef _WIN32
-		ret.set("detail_err_str", access_err_details(settings.getSettings()->backupfolder));
-#endif
-	}
-	else if (!os_directory_exists(os_file_prefix(backupfolder + os_file_sep() + "clients")) && !os_create_dir(os_file_prefix(backupfolder + os_file_sep() + "clients")))
-	{
-		ret.set("system_err", os_last_error_str());
-		ret.set("dir_error", true);
-		ret.set("dir_error_ext", "err_cannot_create_subdir");
-
-		add_stop_show(db, ret, "dir_error_cannot_create_subdir");
-
-#ifdef _WIN32
-		std::string hint = access_dir_hint(backupfolder);
-		if (!hint.empty())
-		{
-			ret.set("dir_error_hint", hint);
-		}
-#endif
+		addr_hint.is_ipv6 = false;
+		memcpy(&addr_hint.addr_ipv4, ip_addr_binary.data(), ip_addr_binary.size());
 	}
 	else
 	{
-		bool has_access_error = false;
-		std::string testfoldername = "testfolderHvfgh---dFFoeRRRf";
-		std::string testfolderpath = backupfolder + os_file_sep() + testfoldername;
-		if (os_directory_exists(os_file_prefix(testfolderpath)))
-		{
-			if (!os_remove_dir(os_file_prefix(testfolderpath)))
-			{
-				ret.set("system_err", os_last_error_str());
-				ret.set("dir_error", true);
-				ret.set("dir_error_ext", "err_cannot_create_subdir");
-				add_stop_show(db, ret, "dir_error_cannot_create_subdir2");
-				has_access_error = true;
-#ifdef _WIN32
-				std::string hint = access_dir_hint(backupfolder);
-				if (!hint.empty())
-				{
-					ret.set("dir_error_hint", hint);
-				}
-#endif
-			}
-		}
-
-		SSettings* server_settings = settings.getSettings();
-
-		if (!has_access_error
-			&& !os_create_dir(os_file_prefix(testfolderpath)))
-		{
-			ret.set("system_err", os_last_error_str());
-			ret.set("dir_error", true);
-			ret.set("dir_error_ext", "err_cannot_create_subdir");
-			add_stop_show(db, ret, "dir_error_cannot_create_subdir3");
-
-#ifdef _WIN32
-			std::string hint = access_dir_hint(backupfolder);
-			if (!hint.empty())
-			{
-				ret.set("dir_error_hint", hint);
-			}
-#endif
-			has_access_error = true;
-		}
-#ifdef _WIN32
-		else if (!server_settings->no_file_backups
-			&& os_directory_exists(os_file_prefix(backupfolder + os_file_sep() + "testfo~1")))
-		{
-			ret.set("dir_error", true);
-			ret.set("dir_error_ext", "dos_names_created");
-			add_stop_show(db, ret, "dos_names_created");
-			ret.set("dir_error_hint", "dos_names_created");
-			if (backupfolder.size() > 2
-				&& backupfolder[1]==':')
-			{
-				ret.set("dir_error_volume", backupfolder.substr(0, 2));
-			}
-			else
-			{
-				ret.set("dir_error_volume", "<VOLUME>");
-			}
-		}
-#endif
-
-		if (!server_settings->no_file_backups)
-		{
-			std::string linkfolderpath = testfolderpath + "_link";
-			os_remove_symlink_dir(os_file_prefix(linkfolderpath));
-			Server->deleteFile(os_file_prefix(linkfolderpath));
-
-			if (!has_access_error
-				&& !os_link_symbolic(os_file_prefix(testfolderpath), os_file_prefix(linkfolderpath)))
-			{
-				ret.set("system_err", os_last_error_str());
-				ret.set("dir_error", true);
-				ret.set("dir_error_ext", "err_cannot_create_symbolic_links");
-				ret.set("dir_error_hint", "UrBackup cannot create symbolic links on the backup storage. "
-					"Your backup storage must support symbolic links in order for UrBackup to work correctly. "
-					"The UrBackup Server must run as administrative user on Windows (If not you get error code 1314). "
-					"Note: As of 2016-05-07 samba (which is used by many Linux based NAS operating systems for Windows file sharing) has not "
-					"implemented the necessary functionality to support symbolic link creation from Windows (With this you get error code 4390).");
-				add_stop_show(db, ret, "dir_error_cannot_create_symbolic_links");
-			}
-
-			os_remove_symlink_dir(os_file_prefix(linkfolderpath));
-		}
-
-		if (!server_settings->no_file_backups)
-		{
-			bool use_tmpfiles = server_settings->use_tmpfiles;
-			std::string tmpfile_path;
-			if (!use_tmpfiles)
-			{
-				tmpfile_path = server_settings->backupfolder + os_file_sep() + "urbackup_tmp_files";
-			}
-			
-			std::auto_ptr<IFile> tmp_f(ClientMain::getTemporaryFileRetry(use_tmpfiles, tmpfile_path, logid_t()));
-
-			if (tmp_f.get() == NULL)
-			{
-				ret.set("tmpdir_error", true);
-				ret.set("tmpdir_error_stop_show_key", "tmpdir_error");
-				if (is_stop_show(db, "tmpdir_error"))
-				{
-					ret.set("tmpdir_error_show", false);
-				}
-			}
-			else
-			{
-				std::string teststring = base64_decode("WDVPIVAlQEFQWzRcUFpYNTQoUF4pN0NDKTd9JEVJQ0FSLVNUQU5EQVJELUFOVElWSVJVUy1URVNULUZJTEUhJEgrSCo=");
-				tmp_f->Write(teststring);
-				std::string tmp_fn = tmp_f->getFilename();
-				tmp_f.reset();
-				tmp_f.reset(Server->openFile(tmp_fn, MODE_RW));
-
-				std::string readstring;
-				if (tmp_f.get() != NULL)
-				{
-					readstring = tmp_f->Read(static_cast<_u32>(teststring.size()));
-				}
-
-				tmp_f.reset();
-				Server->deleteFile(tmp_fn);
-
-				if (teststring != readstring)
-				{
-					ret.set("virus_error", true);
-					ret.set("virus_error_path", ExtractFilePath(tmp_fn));
-					ret.set("virus_error_stop_show_key", "virus_error");
-					if (is_stop_show(db, "virus_error"))
-					{
-						ret.set("virus_error_show", false);
-					}
-				}
-			}
-		}
-
-		if (!has_access_error
-			&& !os_remove_dir(os_file_prefix(testfolderpath)))
-		{
-			ret.set("system_err", os_last_error_str());
-			ret.set("dir_error", true);
-			ret.set("dir_error_ext", "err_cannot_create_subdir");
-			add_stop_show(db, ret, "dir_error_cannot_create_subdir4");
-			has_access_error = true;
-		}
-	}
-
-	IFile *tmp = Server->openTemporaryFile();
-	ScopedDeleteFile delete_tmp_file(tmp);
-	if (tmp == NULL)
-	{
-		ret.set("tmpdir_error", true);
-		ret.set("tmpdir_error_stop_show_key", "tmpdir_error");
-		if (is_stop_show(db, "tmpdir_error"))
-		{
-			ret.set("tmpdir_error_show", false);
-		}
+		assert(false);
 	}
+	return addr_hint.toString();
 }
 
 }
@@ -545,10 +261,7 @@ ACTION_IMPL(status)
 
 		if(rights=="all")
 		{
-			ServerSettings settings(db);
-
-			access_dir_checks(db, settings, settings.getSettings()->backupfolder,
-				settings.getSettings()->backupfolder_uncompr, ret);
+			ret.set("has_status_check", true);
 
 			if(ServerStatus::getServerNospcStalled()>0)
 			{
@@ -630,23 +343,9 @@ ACTION_IMPL(status)
 			}
 		}
 		db_results res=db->Read("SELECT c.id AS id, delete_pending, c.name AS name, strftime('"+helper.getTimeFormatString()+"', lastbackup) AS lastbackup, strftime('"+helper.getTimeFormatString()+"', lastseen) AS lastseen,"
-			"strftime('"+helper.getTimeFormatString()+"', lastbackup_image) AS lastbackup_image, last_filebackup_issues, os_simple, os_version_str, client_version_str, cg.name AS groupname FROM "
+			"strftime('"+helper.getTimeFormatString()+"', lastbackup_image) AS lastbackup_image, last_filebackup_issues, os_simple, os_version_str, client_version_str, cg.name AS groupname, file_ok, image_ok FROM "
 			" clients c LEFT OUTER JOIN settings_db.si_client_groups cg ON c.groupid = cg.id "+filter+" ORDER BY name");
 
-		double backup_ok_mod_file=3.;
-		db_results res_t=db->Read("SELECT value FROM settings_db.settings WHERE key='backup_ok_mod_file' AND clientid=0");
-		if(res_t.size()>0)
-		{
-			backup_ok_mod_file=atof((res_t[0]["value"]).c_str());
-		}
-
-		double backup_ok_mod_image=3.;
-		res_t=db->Read("SELECT value FROM settings_db.settings WHERE key='backup_ok_mod_image' AND clientid=0");
-		if(res_t.size()>0)
-		{
-			backup_ok_mod_image=atof((res_t[0]["value"]).c_str());
-		}
-
 		std::vector<SStatus> client_status=ServerStatus::getStatus();
 
 		for(size_t i=0;i<res.size();++i)
@@ -656,11 +355,34 @@ ACTION_IMPL(status)
 			std::string clientname=res[i]["name"];
 			stat.set("id", clientid);
 			stat.set("name", clientname);
-			stat.set("lastbackup", watoi64(res[i]["lastbackup"]));
-			stat.set("lastbackup_image", watoi64(res[i]["lastbackup_image"]));
+			int64 lastbackup = watoi64(res[i]["lastbackup"]);
+			int64 lastbackup_image = watoi64(res[i]["lastbackup_image"]);
+			stat.set("lastbackup", lastbackup);
+			stat.set("lastbackup_image", lastbackup_image);
 			stat.set("delete_pending", res[i]["delete_pending"] );
-			stat.set("last_filebackup_issues", watoi(res[i]["last_filebackup_issues"]));
+			int issues = watoi(res[i]["last_filebackup_issues"]);
+			if (issues == ServerBackupDao::num_issues_no_backuppaths)
+			{
+				stat.set("last_filebackup_issues", 0);
+				stat.set("no_backup_paths", true);
+			}
+			else
+			{
+				stat.set("last_filebackup_issues", issues);
+			}
 			stat.set("groupname", res[i]["groupname"]);
+			stat.set("file_ok", res[i]["file_ok"] == "1" && lastbackup!=0);
+			stat.set("image_ok", res[i]["image_ok"] == "1" && lastbackup_image!=0);
+
+			if (res[i]["file_ok"] == "-1")
+			{
+				stat.set("file_disabled", true);
+			}
+
+			if (res[i]["image_ok"] == "-1")
+			{
+				stat.set("image_disabled", true);
+			}
 
 			std::string ip="-";
 			std::string client_version_string = res[i]["client_version_str"];
@@ -682,8 +404,7 @@ ACTION_IMPL(status)
 						online=true;
 					}
 
-					unsigned char *ips=(unsigned char*)&client_status[j].ip_addr;
-					ip=convert(ips[0])+"."+convert(ips[1])+"."+convert(ips[2])+"."+convert(ips[3]);
+					ip = ipAddrToStr(client_status[j].ip_addr_binary);
 
 					client_version_string=client_status[j].client_version_string;
 					os_version_string=client_status[j].os_version_string;
@@ -728,36 +449,6 @@ ACTION_IMPL(status)
 			stat.set("processes", processes);
 			stat.set("lastseen", lastseen);
 
-			ServerSettings settings(db, clientid);
-
-			int time_filebackup=settings.getUpdateFreqFileIncr();
-			int time_filebackup_full=settings.getUpdateFreqFileFull();
-			if( time_filebackup_full>=0 &&
-				(time_filebackup<0 || time_filebackup_full<time_filebackup) )
-			{
-				time_filebackup=time_filebackup_full;
-			}
-			
-			IQuery *q=db->Prepare("SELECT id FROM clients WHERE lastbackup IS NOT NULL AND datetime('now','-"+convert((int)(time_filebackup*backup_ok_mod_file+0.5))+" seconds')<lastbackup AND id=?");
-			q->Bind(clientid);
-			db_results res_file_ok=q->Read();
-			q->Reset();
-			stat.set("file_ok", !res_file_ok.empty());
-
-			int time_imagebackup=settings.getUpdateFreqImageIncr();
-			int time_imagebackup_full=settings.getUpdateFreqImageFull();
-			if( time_imagebackup_full>=0 &&
-				(time_imagebackup<0 || time_imagebackup_full<time_imagebackup) )
-			{
-				time_imagebackup=time_imagebackup_full;
-			}
-
-			q=db->Prepare("SELECT id FROM clients WHERE lastbackup_image IS NOT NULL AND datetime('now','-"+convert((int)(time_imagebackup*backup_ok_mod_image+0.5))+" seconds')<lastbackup_image AND id=?");
-			q->Bind(clientid);
-			res_file_ok=q->Read();
-			q->Reset();
-			stat.set("image_ok", !res_file_ok.empty());
-
 			status.add(stat);
 		}
 
@@ -786,10 +477,7 @@ ACTION_IMPL(status)
 				stat.set("lastbackup_image", (std::string)"-");
 				stat.set("online", client_status[i].r_online);
 				stat.set("delete_pending", 0);
-				std::string ip;
-				unsigned char *ips=(unsigned char*)&client_status[i].ip_addr;
-				ip=convert(ips[0])+"."+convert(ips[1])+"."+convert(ips[2])+"."+convert(ips[3]);
-				stat.set("ip", ip);
+				stat.set("ip", ipAddrToStr(client_status[i].ip_addr_binary));
 
 				switch(client_status[i].status_error)
 				{
@@ -833,13 +521,13 @@ ACTION_IMPL(status)
 
 				extra_client.set("hostname", res[i]["hostname"]);
 
-				_i64 i_ip=os_atoi64(res[i]["lastip"]);
+				std::string i_ip=res[i]["lastip"];
 
 				bool online=false;
 
 				for(size_t j=0;j<client_status.size();++j)
 				{
-					if(i_ip==(_i64)client_status[j].ip_addr)
+					if(i_ip==client_status[j].ip_addr_binary)
 					{
 						online=true;
 					}
@@ -894,6 +582,19 @@ ACTION_IMPL(status)
 			ret.set("big_endian", true);
 		}
 	}
+	else if (session != NULL)
+	{
+		ret.set("status", JSON::Array());
+		ret.set("extra_clients", JSON::Array());
+		ret.set("server_identity", helper.getStrippedServerIdentity());
+		ServerSettings settings(db);
+		ret.set("no_images", settings.getSettings()->no_images);
+		ret.set("no_file_backups", settings.getSettings()->no_file_backups);
+		if (is_big_endian())
+		{
+			ret.set("big_endian", true);
+		}
+	}
 	else
 	{
 		ret.set("error", 1);
